!!* Contains routines to calculate the coulombic interaction in non periodic
!!* and periodic systems.
module coulomb
# include "assert.h"
# include "allocate.h"  
  use accuracy
  use message
#ifdef EXTERNALERFC
  use external_erfc
#endif
  use constants, only : pi
  implicit none

  private

  public :: invR, sumInvR, addInvRPrime, getOptimalAlphaEwald, getMaxGEwald, &
      & getMaxREwald, ewald

  interface sumInvR
    module procedure sumInvR_cluster_asymm
    module procedure sumInvR_periodic_asymm
  end interface

  interface invR
    module procedure invR_cluster
    module procedure invR_periodic
  end interface

  interface addInvRPrime
    module procedure addInvRPrime_cluster
    module procedure addInvRPrime_cluster_asymm
    module procedure addInvRPrime_periodic
    module procedure addInvRPrime_periodic_asymm
  end interface
  
  character(len=100) :: error_string     !* Used to return runtime diagnostics
  character(len=*), parameter :: ftTooClose = &
      &"('The objects with the following indexes are too close to each other',&
      &I5,I5)"

  !! Maximal argument value of erf, after which it is constant
  real(dp), parameter :: erfArgLimit = 10.0_dp

  
contains


  !!* Calculates the 1/R Matrix for all atoms for the non-periodic case.
  !!* @param invRMat Matrix of 1/R values for each atom pair.
  !!* @param coord   List of atomic coordinates.
  !!* @note Only the lower triangle is constructed.
  subroutine invR_cluster(invRMat, nAtom, coord)
    real(dp), intent(out) :: invRMat(:,:)
    integer, intent(in) :: nAtom
    real(dp), intent(in)  :: coord(:,:)

    integer :: ii, jj
    real(dp) :: dist, vect(3)

    ASSERT(all(shape(invRMat) == (/ nAtom, nAtom /)))
    ASSERT(size(coord, dim=1) == 3)
    ASSERT(size(coord, dim=2) >= nAtom)

    invRMat(:,:) = 0.0_dp
    do ii = 1, nAtom
      do jj = ii + 1, nAtom
        vect(:) = coord(:,ii) - coord(:,jj)
        dist = sqrt(sum(vect(:)**2))
        invRMat(jj,ii) = 1.0_dp/dist
      end do
    end do

  end subroutine invR_cluster



  !!* Calculates the summed 1/R vector for all atoms for the non-periodic case
  !!* asymmmetric case (like interaction of atoms with point charges).
  !!* @param invRVec Vector of sum_i q_i/|R_atom - R_i] values for each atom
  !!* @param nAtom0 Number of atoms in the first group
  !!* @param nAtom1 Number of atoms in the second group
  !!* @param coord0 Coordinates of the first group of objects (atoms)
  !!* @param coord1 Coordinates of the 2nd group of objects (point charges)
  !!* @param charges1 Charges of the 2nd group of objects
  !!* @param blurWidths1 Gaussian blur widht of the charges in the 2nd group
  subroutine sumInvR_cluster_asymm(invRVec, nAtom0, nAtom1, coord0, &
      &coord1, charges1, blurWidths1)
    real(dp), intent(out) :: invRVec(:)
    integer, intent(in) :: nAtom0
    integer, intent(in) :: nAtom1
    real(dp), intent(in)  :: coord0(:,:)
    real(dp), intent(in)  :: coord1(:,:)
    real(dp), intent(in)  :: charges1(:)
    real(dp), intent(in), optional :: blurWidths1(:)

    integer :: iAt0, iAt1
    real(dp) :: dist, vect(3), fTmp

    ASSERT(size(invRVec) == nAtom0)
    ASSERT(size(coord0, dim=2) >= nAtom0)
    ASSERT(size(coord0, dim=1) == 3)
    ASSERT(size(coord1, dim=2) >= nAtom1)
    ASSERT(size(coord1, dim=1) == 3)
    ASSERT(size(charges1) == nAtom1)
    ASSERT_ENV(if (present(blurWidths1)) then)
    ASSERT_ENV(  ASSERT(size(blurWidths1) == nAtom1))
    ASSERT_ENV(end if)

    !! Doing blurring and non blurring case separately in order to avoid
    !! the if branch deep in the loop
    invRVec(:) = 0.0_dp
    if (present(blurWidths1)) then
      do iAt0 = 1, nAtom0
        do iAt1 = 1, nAtom1
          vect(:) = coord0(:,iAt0) - coord1(:,iAt1)
          dist = sqrt(sum(vect(:)**2))
          if (dist > epsilon(0.0_dp)) then
            fTmp = charges1(iAt1) / dist
            if (dist < erfArgLimit * blurWidths1(iAt1)) then
#ifdef EXTERNALERFC
              fTmp = fTmp * extErf(dist/blurWidths1(iAt1))
#else              
              fTmp = fTmp * erf(dist/blurWidths1(iAt1))
#endif              
            end if
            invRVec(iAt0) = invRVec(iAt0) + fTmp
          else
            write (error_string, ftTooClose) iAt0, iAt1
            call error(trim(error_string))
          end if
        end do
      end do
    else
      do iAt0 = 1, nAtom0
        do iAt1 = 1, nAtom1
          vect(:) = coord0(:,iAt0) - coord1(:,iAt1)
          dist = sqrt(sum(vect(:)**2))
          if (dist > epsilon(0.0_dp)) then
            invRVec(iAt0) = invRVec(iAt0) + charges1(iAt1) / dist
          else
            write (error_string, ftTooClose) iAt0, iAt1
            call error(trim(error_string))
          end if
        end do
      end do
    end if
    
  end subroutine sumInvR_cluster_asymm
  


  !!* Calculates the 1/R Matrix for all atoms for the periodic case.
  !!* @param invRMat      Matrix of 1/R values for each atom pair.
  !!* @param nAtom  Number of atoms.
  !!* @param coord        List of atomic coordinates (all atoms).
  !!* @param nNeighborEwald Nr. of neighbors for each atom for real part of
  !!* Ewald.
  !!* @param iNeighbor    List of neighbors for the real space part of Ewald.
  !!* @param img2CentCell Image of each atom in the central cell.
  !!* @param recPoint     Contains the points included in the reciprocal sum.
  !!*                     The set should not include the origin or
  !!*                     inversion related points.
  !!* @param alpha        Parameter for Ewald summation.
  !!* @param volume       Volume of the real space unit cell.
  !!* @note Only the lower triangle is constructed.
  subroutine invR_periodic(invRMat, nAtom, coord, nNeighborEwald, iNeighbor, &
      &img2CentCell, recPoint, alpha, volume)
    real(dp), intent(out) :: invRMat(:,:)
    integer, intent(in) :: nAtom
    real(dp), intent(in)  :: coord(:,:)
    integer,  intent(in)  :: nNeighborEwald(:)
    integer,  intent(in)  :: iNeighbor(0:,:)
    integer,  intent(in)  :: img2CentCell(:)
    real(dp), intent(in)  :: recPoint(:,:)
    real(dp), intent(in)  :: alpha
    real(dp), intent(in)  :: volume

    integer :: iAtom1, iAtom2, iAtom2f, iNeigh

    ASSERT(all(shape(invRMat) == (/ nAtom, nAtom /)))
    ASSERT(size(coord, dim=1) == 3)
    ASSERT(size(coord, dim=2) >= nAtom)
    ASSERT(size(nNeighborEwald) == nAtom)
    ASSERT(size(iNeighbor, dim=2) == nAtom)
    ASSERT(volume > 0.0_dp)

    !! Real space part of the Ewald sum.
    invRMat(:,:) = 0.0_dp
    do iAtom1 = 1, nAtom
      do iNeigh = 1, nNeighborEwald(iAtom1)
        iAtom2 = iNeighbor(iNeigh, iAtom1)
        iAtom2f = img2CentCell(iAtom2)
        invRMat(iAtom2f, iAtom1) = invRMat(iAtom2f, iAtom1) &
            &+rTerm(sqrt(sum((coord(:,iAtom1)-coord(:,iAtom2))**2)), alpha)
      end do
    end do

    !! Reciprocal space part of the Ewald sum.
    do iAtom1 = 1, nAtom
      do iAtom2 = iAtom1, nAtom
        invRMat(iAtom2, iAtom1) = invRMat(iAtom2, iAtom1) &
            &+ ewaldReciprocal(coord(:,iAtom1)-coord(:,iAtom2), recPoint, &
            &alpha, volume) - pi / (volume * alpha**2)
      end do
    end do

    !! Extra contribution for self interaction.
    do iAtom1 = 1, nAtom
      invRMat(iAtom1, iAtom1) = invRMat(iAtom1, iAtom1) &
          &- 2.0_dp * alpha / sqrt(pi)
    end do

  end subroutine invR_periodic



  !!* Calculates summed 1/R vector for two groups of objects for the
  !!* periodic case.
  !!* @param invRVec Vector of sum_i q_i/|R_atom - R_i] values for each atom
  !!* @param nAtom0 Number of atoms in the first group
  !!* @param nAtom1 Number of atoms in the second group
  !!* @param coord0 Coordinates of the first group of objects (atoms)
  !!* @param coord1 Coordinates of the 2nd group of objects (point charges)
  !!* @param charges1 Charges of the 2nd group of objects
  !!* @param rLat Lattice vectors to be used for the real Ewald summation
  !!* @param gLat Lattice vectors to be used for the reciprocal Ewald summation.
  !!* @param alpha Parameter of the Ewald summation
  !!* @param vol Volume of the supercell.
  subroutine sumInvR_periodic_asymm(invRVec, nAtom0, nAtom1, coord0, &
      &coord1, charges1, rLat, gLat, alpha, volume)
    real(dp), intent(out) :: invRVec(:)
    integer, intent(in) :: nAtom0
    integer, intent(in) :: nAtom1    
    real(dp), intent(in)  :: coord0(:,:)
    real(dp), intent(in)  :: coord1(:,:)
    real(dp), intent(in)  :: charges1(:)
    real(dp), intent(in)  :: rLat(:,:)
    real(dp), intent(in)  :: gLat(:,:)
    real(dp), intent(in)  :: alpha
    real(dp), intent(in)  :: volume

    integer :: iAt0, iAt1
    real(dp) :: rTmp, rr(3)
    
    ASSERT(size(invRVec) == nAtom0)
    ASSERT(size(coord0, dim=2) >= nAtom0)
    ASSERT(size(coord0, dim=1) == 3)
    ASSERT(size(coord1, dim=2) >= nAtom1)
    ASSERT(size(coord1, dim=1) == 3)
    ASSERT(size(charges1) == nAtom1)
    ASSERT(size(rLat, dim=1) == 3)
    ASSERT(size(gLat, dim=1) == 3)    
    ASSERT(volume > 0.0_dp)

    invRVec(:) = 0.0_dp
    do iAt0 = 1, nAtom0
      do iAt1 = 1, nAtom1
        rr = coord0(:,iAt0) - coord1(:,iAt1)
        rTmp = ewald(rr, rLat, gLat, alpha, volume)
        invRVec(iAt0) = invRVec(iAt0) + rTmp * charges1(iAt1)
      end do
    end do

  end subroutine sumInvR_periodic_asymm


  
  !!* Calculates the -1/R**2 deriv contribution for all atoms for the 
  !!* non-periodic case, without storing anything.
  !!* @param deriv Contains the derivative on exit.
  !!* @param nAtom Number of atoms
  !!* @param coord List of atomic coordinates.
  !!* @param deltaQAtom List of charges on each atom
  subroutine addInvRPrime_cluster(deriv, nAtom, coord, deltaQAtom)
    real(dp), intent(inout) :: deriv(:,:)
    integer, intent(in) :: nAtom
    real(dp), intent(in)  :: coord(:,:)
    real(dp), intent(in)  :: deltaQAtom(:)
    
    integer :: ii, jj
    real(dp) :: dist, vect(3), fTmp

    ASSERT(size(deriv, dim=1) == 3)
    ASSERT(size(deriv, dim=2) >= nAtom)
    ASSERT(size(coord, dim=1) == 3)
    ASSERT(size(coord, dim=2) >= nAtom)
    ASSERT(size(deltaQAtom) == nAtom)
    
    do ii = 1, nAtom
      do jj = ii + 1, nAtom
        vect(:) = coord(:,ii) - coord(:,jj)
        dist = sqrt(sum(vect(:)**2))
        fTmp = -deltaQAtom(ii) * deltaQAtom(jj) / (dist**3)
        deriv(:,ii) = deriv(:,ii) + vect(:)*fTmp
        !! Skew-symmetric 1/r2 interaction, so the other triangle is
        !! calculated :
        deriv(:,jj) = deriv(:,jj) - vect(:)*fTmp
      end do
    end do
    
  end subroutine addInvRPrime_cluster



  !!* Calculates the -1/R**2 deriv contribution for charged atoms interacting 
  !!* with a group of charged objects (like point charges) for the non-periodic
  !!* case, without storing anything.
  !!* @param deriv0 Contains the derivative for the first group
  !!* @param deriv1 Contains the derivative for the second group
  !!* @param nAtom0 Number of atoms in the first group
  !!* @param nAtom1 Number of atoms in the second group
  !!* @param coord0 List of atomic coordinates.
  !!* @param coord1 List of the point charge coordinates
  !!* @param charge0 Charge of the atoms.
  !!* @param charge1 Charge of the point charges.
  subroutine addInvRPrime_cluster_asymm(deriv0, deriv1, nAtom0, nAtom1, &
      &coord0, coord1, charge0, charge1, blurWidths1)
    real(dp), intent(inout) :: deriv0(:,:)
    real(dp), intent(inout) :: deriv1(:,:)
    integer, intent(in) :: nAtom0
    integer, intent(in) :: nAtom1    
    real(dp), intent(in)  :: coord0(:,:)
    real(dp), intent(in)  :: coord1(:,:)
    real(dp), intent(in)  :: charge0(:)
    real(dp), intent(in)  :: charge1(:)
    real(dp), intent(in), optional  :: blurWidths1(:)
    
    integer :: iAt0, iAt1
    real(dp) :: dist, vect(3), fTmp(3), sigma, rs

    ASSERT(size(deriv0, dim=1) == 3)
    ASSERT(size(deriv0, dim=2) >= nAtom0)
    ASSERT(size(deriv1, dim=1) == 3)
    ASSERT(size(deriv1, dim=2) >= nAtom1)
    ASSERT(size(coord0, dim=1) == 3)
    ASSERT(size(coord0, dim=2) >= nAtom0)
    ASSERT(size(coord1, dim=1) == 3)
    ASSERT(size(coord1, dim=2) >= nAtom1)
    ASSERT(size(charge0) == nAtom0)
    ASSERT(size(charge1) == nAtom1)
    ASSERT_ENV(if (present(blurWidths1)) then)
    ASSERT_ENV(  ASSERT(size(blurWidths1) == nAtom1))
    ASSERT_ENV(end if)

    !! Doing blured and unblured cases separately to avoid ifs in the loop
    if (present(blurWidths1)) then
      do iAt0 = 1, nAtom0
        do iAt1 = 1, nAtom1
          vect(:) = coord0(:,iAt0) - coord1(:,iAt1)
          dist = sqrt(sum(vect(:)**2))
          if (dist > epsilon(0.0_dp)) then
            fTmp = -vect(:) / (dist**3)
            if (dist < erfArgLimit * blurWidths1(iAt1)) then
              sigma = blurWidths1(iAt1)
              rs = dist / sigma
#ifdef EXTERNALERFC
              fTmp = fTmp * (extErf(rs) &
                  &- 2.0_dp/(sqrt(pi)*sigma) * dist * exp(-(rs**2)))
#else              
              fTmp = fTmp * (erf(rs) &
                  &- 2.0_dp/(sqrt(pi)*sigma) * dist * exp(-(rs**2)))
#endif              
            end if
            fTmp = charge0(iAt0) * charge1(iAt1) * fTmp
            deriv0(:,iAt0) = deriv0(:,iAt0) + fTmp(:)
            deriv1(:,iAt1) = deriv1(:,iAt1) - fTmp(:)
          else
            write (error_string, ftTooClose) iAt0, iAt1
            call error(trim(error_string))
          end if
        end do
      end do
    else
      do iAt0 = 1, nAtom0
        do iAt1 = 1, nAtom1
          vect(:) = coord0(:,iAt0) - coord1(:,iAt1)
          dist = sqrt(sum(vect(:)**2))
          if (dist > epsilon(0.0_dp)) then
            fTmp = -charge0(iAt0) * charge1(iAt1) / (dist**3) * vect(:)
            deriv0(:,iAt0) = deriv0(:,iAt0) + fTmp(:)
            deriv1(:,iAt1) = deriv1(:,iAt1) - fTmp(:)
          else
            write (error_string, ftTooClose) iAt0, iAt1
            call error(trim(error_string))
          end if
        end do
      end do
    end if
    
  end subroutine addInvRPrime_cluster_asymm


  
  !!* Calculates the -1/R**2 deriv contribution for the periodic case, without
  !!* storing anything.
  !!* @param deriv          Derivative on exit
  !!* @param nAtom          Number of atoms
  !!* @param coord          List of atomic coordinates (all atoms).
  !!* @param nNeighborEwald Nr. of neighbors for each atom for real part
  !!*   ofEwald.
  !!* @param iNeighbor      List of neighbors for the real space part of Ewald.
  !!* @param img2CentCell   Image of each atom in the central cell.
  !!* @param recPoint       Contains the points included in the reciprocal sum.
  !!*                       The set should not include the origin or
  !!*                       inversion related points.
  !!* @param alpha          Parameter for Ewald summation.
  !!* @param volume         Volume of the real space unit cell.
  !!* @param deltaQAtom     List of charges on each atom
  subroutine addInvRPrime_periodic(deriv, nAtom, coord, nNeighborEwald, &
      &iNeighbor, img2CentCell, recPoint, alpha, volume,deltaQAtom)
    real(dp), intent(inout) :: deriv(:,:)
    integer, intent(in) :: nAtom
    real(dp), intent(in)    :: coord(:,:)
    integer,  intent(in)    :: nNeighborEwald(:)
    integer,  intent(in)    :: iNeighbor(0:,:)
    integer,  intent(in)    :: img2CentCell(:)
    real(dp), intent(in)    :: recPoint(:,:)
    real(dp), intent(in)    :: alpha
    real(dp), intent(in)    :: volume
    real(dp), intent(in)    :: deltaQAtom(:)
    
    integer :: iAtom1, iAtom2, iAtom2f, iNeigh
    real(dp) :: r(3)
    
    ASSERT(size(deriv, dim=1) == 3)
    ASSERT(size(deriv, dim=2) >= nAtom)
    ASSERT(size(coord, dim=1) == 3)
    ASSERT(size(coord, dim=2) >= nAtom)
    ASSERT(size(nNeighborEwald) == nAtom)
    ASSERT(size(iNeighbor, dim=2) == nAtom)
    ASSERT(volume > 0.0_dp)
    ASSERT(size(deltaQAtom) == nAtom)

    do iAtom1 = 1, nAtom
      
      ! d(1/R)/dr real space      
      do iNeigh = 1, nNeighborEwald(iAtom1)
        iAtom2 = iNeighbor(iNeigh, iAtom1)
        iAtom2f = img2CentCell(iAtom2)
        if (iAtom2f /= iAtom1) then
          r(:) = coord(:,iAtom1)-coord(:,iAtom2)
          deriv(:,iAtom1) = deriv(:,iAtom1) &
            & + derivRTerm(r,alpha) * &
            & deltaQAtom(iAtom1) * deltaQAtom(iAtom2f)
          deriv(:,iAtom2f) = deriv(:,iAtom2f) &
            & - derivRTerm(r,alpha) * &
            & deltaQAtom(iAtom1) * deltaQAtom(iAtom2f)
        end if
      end do
      
      ! d(1/R)/dr reciprocal space
      do iAtom2 = iAtom1+1, nAtom
        r(:) = coord(:,iAtom1)-coord(:,iAtom2)
        deriv(:,iAtom1) = deriv(:,iAtom1) &
            & + derivEwaldReciprocal(r, recPoint, &
            & alpha, volume) * deltaQAtom(iAtom1) * deltaQAtom(iAtom2)
        deriv(:,iAtom2) = deriv(:,iAtom2) &
            & - derivEwaldReciprocal(r, recPoint, &
            & alpha, volume) * deltaQAtom(iAtom1) * deltaQAtom(iAtom2)
      end do
      
    end do

  end subroutine addInvRPrime_periodic

  

  !!* Calculates the -1/R**2 deriv contribution for charged atoms interacting 
  !!* with a group of charged objects (like point charges) for the periodic
  !!* case, without storing anything.
  !!* @param deriv0 Contains the derivative for the first group on exit
  !!* @param deriv1 Contains the derivative for the second group on exit
  !!* @param nAtom0 Number of atoms in the first group
  !!* @param nAtom1 Number of atoms in the second group
  !!* @param coord0 List of atomic coordinates.
  !!* @param coord1 List of the point charge coordinates
  !!* @param charge0 Charge of the atoms.
  !!* @param charge1 Charge of the point charges.
  !!* @param deltaQAtom     List of charges on each atom
  !!* @param rVec Lattice vectors to be used for the real Ewald summation
  !!* @param gVec Lattice vectors to be used for the reciprocal Ewald summation.
  !!* @param alpha Parameter of the Ewald summation
  !!* @param vol Volume of the supercell.
  subroutine addInvRPrime_periodic_asymm(deriv0, deriv1, nAtom0, nAtom1, &
      &coord0, coord1, charge0, charge1, rVec, gVec, alpha, vol)
    real(dp), intent(inout) :: deriv0(:,:)
    real(dp), intent(inout) :: deriv1(:,:)
    integer, intent(in) :: nAtom0
    integer, intent(in) :: nAtom1
    real(dp), intent(in)    :: coord0(:,:)
    real(dp), intent(in)    :: coord1(:,:)
    real(dp), intent(in)    :: charge0(:)
    real(dp), intent(in)    :: charge1(:)
    real(dp), intent(in)    :: rVec(:,:)
    real(dp), intent(in)    :: gVec(:,:)
    real(dp), intent(in)    :: alpha
    real(dp), intent(in)    :: vol
    
    integer :: iAt0, iAt1
    real(dp) :: dist, vect(3), fTmp(3)
    
    ASSERT(size(deriv0, dim=1) == 3)
    ASSERT(size(deriv0, dim=2) == nAtom0)
    ASSERT(size(deriv1, dim=1) == 3)
    ASSERT(size(deriv1, dim=2) >= nAtom1)
    ASSERT(size(coord0, dim=1) == 3)
    ASSERT(size(coord0, dim=2) >= nAtom0)
    ASSERT(size(coord1, dim=1) == 3)
    ASSERT(size(coord1, dim=2) >= nAtom1)
    ASSERT(size(charge0) == nAtom0)
    ASSERT(size(charge1) == nAtom1)
    ASSERT(size(rVec, dim=1) == 3)
    ASSERT(size(gVec, dim=1) == 3)
    ASSERT(vol > 0.0_dp)

    do iAt0 = 1, nAtom0
      do iAt1 = 1, nAtom1
        vect(:) = coord0(:,iAt0) - coord1(:,iAt1)
        dist = sqrt(sum(vect(:)**2))
        if (dist > tolSameDist) then
          fTmp(:) = (derivEwaldReal(vect, rVec, alpha) &
              &+ derivEwaldReciprocal(vect, gVec, alpha, vol)) &
              &* charge0(iAt0) * charge1(iAt1)
          deriv0(:,iAt0) = deriv0(:,iAt0) + fTmp(:)
          deriv1(:,iAt1) = deriv1(:,iAt1) - fTmp(:)
        else
          write (error_string, ftTooClose) iAt0, iAt1
          call error(trim(error_string))
        end if
      end do
    end do
    
  end subroutine addInvRPrime_periodic_asymm
  


  !!* Get optimal alpha-parameter for the Ewald summation by finding alpha,
  !!* where decline of real and reciprocal part of Ewald are equal.
  !!* @param latVec    Lattice vectors.
  !!* @param recVec    Reciprocal vectors.
  !!* @param volume    Volume of the unit cell.
  !!* @param tolerance Tolerance for difference in real and rec. part.
  !!* @return Optimal alpha.
  !!* @note The function stops, if the optimal alpha cannot be found.
  function getOptimalAlphaEwald(latVec, recVec, volume, tolerance) result(alpha)
    real(dp), intent(in) :: latVec(:,:)
    real(dp), intent(in) :: recVec(:,:)
    real(dp), intent(in) :: volume
    real(dp), intent(in) :: tolerance
    real(dp) :: alpha

    real(dp) :: alphaLeft, alphaRight
    real(dp), parameter :: alphaInit = 1.0e-8_dp

    real(dp) :: minG, minR, diff
    integer  :: iIter
    integer  :: iError

    ASSERT(all(shape(latVec) == (/3, 3/)))
    ASSERT(all(shape(recVec) == (/3, 3/)))
    ASSERT(volume > 0.0_dp)
    ASSERT(tolerance > 0.0_dp)

    minG = sqrt(minval(sum(recVec(:,:)**2, dim=1)))
    minR = sqrt(minval(sum(latVec(:,:)**2, dim=1)))

    iError = 0
    alpha = alphaInit
    diff = diffRecReal(alpha, minG, minR, volume)
    do while (diff < -tolerance .and. alpha <= huge(1.0_dp))
      alpha = 2.0_dp * alpha
      diff = diffRecReal(alpha, minG, minR, volume)
    end do
    if (alpha > huge(1.0_dp)) then
      iError = 1
    elseif (alpha == alphaInit) then
      iError = 2
    end if

    if (iError == 0) then
      alphaLeft = 0.5_dp * alpha
      do while (diff < tolerance .and. alpha <= huge(1.0_dp))
        alpha = 2.0_dp * alpha
        diff = diffRecReal(alpha, minG, minR, volume)
      end do
      if (alpha > huge(1.0_dp)) then
        iError = 3
      end if
    end if

    if (iError == 0) then
      alphaRight = alpha
      alpha = (alphaLeft + alphaRight) / 2.0
      iIter = 0
      diff = diffRecReal(alpha, minG, minR, volume)
      do while (abs(diff) > tolerance .and. iIter <= nSearchIter)
        if (diff < 0) then
          alphaLeft = alpha
        else
          alphaRight = alpha
        end if
        alpha = (alphaLeft + alphaRight) / 2.0
        diff = diffRecReal(alpha, minG, minR, volume)
        iIter = iIter + 1
      end do
      if (iIter > nSearchIter) then
        iError = 4
      end if
    end if

    if (iError /= 0) then
      !!alpha = exp(-0.310104 * log(volume) + 0.786382) / 2.0
99000 format ('Failure in determining optimal alpha for Ewaldsum.', &
          & ' Error code: ',I3)
      write(error_string, 99000) iError
      call error(error_string)
    end if

  end function getOptimalAlphaEwald



  !!* Returns the longest reciprocal vector which gives a bigger contribution
  !!* to the Ewald sum as a certain tolerance.
  !!* @param alpha     Parameter of the ewald summation.
  !!* @param volume    Volume of the unit cell.
  !!* @param minValue  Tolerance value.
  function getMaxGEwald(alpha, volume, minValue) result(xx)
    real(dp), intent(in) :: alpha
    real(dp), intent(in) :: volume
    real(dp), intent(in) :: minValue
    real(dp) :: xx

    real(dp), parameter :: gInit = 1.0e-8_dp
    real(dp) :: xLeft, xRight, yLeft, yRight, yy
    integer :: iError, iIter

    iError = 0
    xx = gInit
    yy = gTerm(xx, alpha, volume)
    do while (yy > minValue .and. xx <= huge(1.0_dp))
      xx = 2.0_dp * xx
      yy = gTerm(xx, alpha, volume)
    end do
    if (xx > huge(1.0_dp)) then
      iError = 1
    elseif (xx == gInit) then
      iError = 2
    end if

    if (iError == 0) then
      xLeft = 0.5_dp * xx
      yLeft = gTerm(xLeft, alpha, volume)
      xRight = xx
      yRight = yy

      iIter = 1
      do while (yLeft - yRight > minValue .and. iIter <= nSearchIter)
        xx = 0.5_dp * (xLeft + xRight)
        yy = gTerm(xx, alpha, volume)
        if (yy >= minValue) then
          xLeft = xx
          yLeft = yy
        else
          xRight = xx
          yRight = yy
        end if
        iIter = iIter + 1
      end do
      if (iIter > nSearchIter) then
        iError = 3
      end if
    end if

    if (iError /= 0) then
99010 format ('Failure in getMaxGEwald.', ' Error nr: ',I3)
      write(error_string, 99010) iError
      call error(error_string)
    end if

  end function getMaxGEwald



  !!* Returns the longest real space vector which gives a bigger contribution
  !!* to the Ewald sum as a certain tolerance.
  !!* @param alpha     Parameter of the ewald summation.
  !!* @param minValue  Tolerance value.
  function getMaxREwald(alpha, minValue) result(xx)
    real(dp), intent(in) :: alpha
    real(dp), intent(in) :: minValue
    real(dp) :: xx

    real(dp), parameter :: rInit = 1.0e-8_dp
    real(dp) :: xLeft, xRight, yLeft, yRight, yy
    integer  :: iError, iIter

    iError = 0
    xx = rInit
    yy = rTerm(xx, alpha)
    do while (yy > minValue .and. xx <= huge(1.0_dp))
      xx = 2.0_dp * xx
      yy = rTerm(xx, alpha)
    end do
    if (xx > huge(1.0_dp)) then
      iError = 1
    elseif (xx == rInit) then
      iError = 2
    end if

    if (iError == 0) then
      xLeft = 0.5_dp * xx
      yLeft = rTerm(xLeft, alpha)
      xRight = xx
      yRight = yy

      iIter = 1
      do while (yLeft - yRight > minValue .and. iIter <= nSearchIter)
        xx = 0.5_dp * (xLeft + xRight)
        yy = rTerm(xx, alpha)
        if (yy >= minValue) then
          xLeft = xx
          yLeft = yy
        else
          xRight = xx
          yRight = yy
        end if
        iIter = iIter + 1
      end do
      if (iIter > nSearchIter) then
        iError = 3
      end if
    end if

    if (iError /= 0) then
99020 format ('Failure in getMaxREwald.', ' Error nr: ',I3)
      write(error_string, 99020) iError
      call error(error_string)
    end if

  end function getMaxREwald



  !!* Returns the Ewald sum for a given lattice in a given point.
  !!* @param rr    Vector where to calculate the Ewald sum.
  !!* @param rVec  Real space vectors to sum over. (Should contain origin).
  !!* @param gVec  Reciprocal space vectors to sum over (Should not contain
  !!*   either origin nor inversion related points).
  !!* @param alpha Parameter for the Ewald summation.
  !!* @param vol   Volume of the real space unit cell.
  function ewald(rr, rVec, gVec, alpha, vol)
    real(dp), intent(in) :: rr(:)
    real(dp), intent(in) :: rVec(:,:)
    real(dp), intent(in) :: gVec(:,:)
    real(dp), intent(in) :: alpha
    real(dp), intent(in) :: vol
    real(dp) :: ewald

    ewald = ewaldReciprocal(rr, gVec, alpha, vol) &
        &+ ewaldReal(rr, rVec, alpha) - pi / (vol*alpha**2)
    if (sum(rr(:)**2) < tolSameDist2) then
      ewald = ewald - 2.0_dp * alpha / sqrt(pi)
    end if

  end function ewald


  
  !!* Returns the reciprocal part of the Ewald sum.
  !!* @param rr    Vector where to calculate the Ewald sum.
  !!* @param gVec  Reciprocal space vectors to sum over (Should not contain
  !!*   either origin nor inversion related points).
  !!* @param alpha Parameter for the Ewald summation.
  !!* @param vol   Volume of the real space unit cell.
  function ewaldReciprocal(rr, gVec, alpha, vol) result(recSum)
    real(dp), intent(in) :: rr(:)
    real(dp), intent(in) :: gVec(:,:)
    real(dp), intent(in) :: alpha
    real(dp), intent(in) :: vol
    real(dp) :: recSum

    real(dp) :: gg(3), g2
    integer  :: iG

    ASSERT(size(gVec, dim=1) == 3)
    ASSERT(size(rr) == 3)
    ASSERT(vol > 0.0_dp)

    recSum = 0.0_dp
    do iG = 1, size(gVec, dim=2)
      gg = gVec(:,iG)
      g2 = sum(gg(:)**2)
      recSum = recSum + exp(-g2/(4.0_dp*alpha**2))/g2 * cos(dot_product(gg,rr))
    end do
    ! note factor of 2 as only summing half of reciprocal space
    recSum = 2.0_dp * recSum * 4.0_dp * pi / vol

  end function ewaldReciprocal


  
  !!* Returns the derivative of the reciprocal part of the Ewald sum.
  !!* @param rr    Vector where to calculate the Ewald sum.
  !!* @param gVec  Reciprocal space vectors to sum over (Should not contain
  !!*   either origin nor inversion related points).
  !!* @param alpha Parameter for the Ewald summation.
  !!* @param vol   Volume of the real space unit cell.
  function derivEwaldReciprocal(rr, gVec, alpha, vol) result(recSum)
    real(dp), intent(in) :: rr(:)
    real(dp), intent(in) :: gVec(:,:)
    real(dp), intent(in) :: alpha
    real(dp), intent(in) :: vol
    real(dp) :: recSum(3)

    real(dp) :: gg(3), g2
    integer  :: iG
    
    ASSERT(size(gVec, dim=1) == 3)
    ASSERT(size(rr) == 3)
    ASSERT(vol > 0.0_dp)
    
    recSum(:) = 0.0_dp
    do iG = 1, size(gVec, dim=2)
      gg(:) = gVec(:,iG)
      g2 = sum(gg(:)**2)
      recSum(:) = recSum(:) - &
          & gg(:)*sin(dot_product(gg,rr))*exp(-g2/(4.0_dp*alpha**2))/g2
    end do
    ! note factor of 2 as only summing over half of reciprocal space
    recSum(:) = 2.0_dp * recSum(:) * 4.0_dp * pi / vol

  end function derivEwaldReciprocal


  
  !!* Returns the real space part of the Ewald sum.
  !!* @param rr    Vector where to calculate the Ewald sum.
  !!* @param rVec  Real space vectors to sum over. (Should contain origin).
  !!* @param alpha Parameter for the Ewald summation.
  function ewaldReal(rr, rVec, alpha) result(realSum)
    real(dp), intent(in) :: rVec(:,:)
    real(dp), intent(in) :: alpha
    real(dp), intent(in) :: rr(:)
    real(dp) :: realSum

    real(dp) :: absRR
    integer  :: iR


    ASSERT(size(rVec, dim=1) == 3)
    ASSERT(size(rr) == 3)

    realSum = 0.0_dp;
    do iR = 1, size(rVec, dim=2)
      absRR = sqrt(sum((rr(:) + rVec(:,iR))**2))
      if (absRR < tolSameDist) then
        cycle
      end if
#ifdef EXTERNALERFC
      realSum = realSum +  extErfc(alpha*absRR)/absRR
#else
      realSum = realSum + erfc(alpha*absRR)/absRR
#endif
    end do

  end function ewaldReal

  

  !!* Returns the derivative of the real space part of the Ewald sum.
  !!* @param rr    Vector where to calculate the Ewald sum.
  !!* @param rVec  Real space vectors to sum over. (Should contain origin).
  !!* @param alpha Parameter for the Ewald summation.
  function derivEwaldReal(rdiff, rVec, alpha) result(dewr)
    real(dp), intent(in) :: rdiff(:)
    real(dp), intent(in) :: rVec(:,:)
    real(dp), intent(in) :: alpha
    real(dp) :: dewr(3)

    real(dp) :: rNew(3)
    real(dp) :: rr
    integer  :: iR

    ASSERT(size(rVec, dim=1) == 3)
    ASSERT(size(rdiff) == 3)

    dewr = 0.0_dp;
    do iR = 1, size(rVec, dim=2)
      rNew(:) = rdiff(:) + rVec(:,iR)
      rr = sqrt(sum(rNew**2))
      if (rr < tolSameDist2) then
        cycle
      end if
#ifdef EXTERNALERFC
      dewr(:) = dewr &
          &+ rNew(:) * (-2.0_dp/sqrt(pi)*exp(-(alpha*rr)**2)*alpha*rr &
          &- extErfc(alpha*rr))/(rr**3)
#else
      dewr(:) = dewr &
          &+ rNew(:) * (-2.0_dp/sqrt(pi)*exp(-alpha*alpha*rr*rr)* alpha*rr &
          &- erfc(alpha*rr))/(rr*rr*rr)
#endif
    end do

  end function derivEwaldReal
  

  
  !!* Returns the difference in the decrease of the real and reciprocal parts
  !!* of the Ewald sum.
  !!* @param alpha  Parameter for the Ewald summation.
  !!* @param minG   Length of the shortest reciprocal space vector in the sum.
  !!* @param minR   Length of the shortest real space vector in the sum.
  !!* @param volume Volume of the real space unit cell.
  !!* @note In order to make the real space part shorter as the reciprocal
  !!*   space part, the inclinations are taken at different points for the
  !!*   the real space and the reciprocal space part.
  function diffRecReal(alpha, minG, minR, volume) result(diff)
    real(dp), intent(in) :: alpha
    real(dp), intent(in) :: minG
    real(dp), intent(in) :: minR
    real(dp), intent(in) :: volume
    real(dp) :: diff

    ASSERT(volume > 0.0_dp)

    diff = ((gTerm(4.0_dp*minG, alpha, volume) &
        &- gTerm(5.0_dp*minG, alpha, volume))) &
        &- (rTerm(2.0_dp*minR, alpha) - rTerm(3.0_dp*minR, alpha))

  end function diffRecReal



  !!* Returns the max. value of a term in the reciprocal space part of the
  !!* Ewald summation for a given vector length.
  !!* @param gg    Length of the reciprocal space vector.
  !!* @param alpha Parameter of the Ewald summation.
  !!* @param vol   Volume of the real space unit cell.
  function gTerm(gg, alpha, vol)
    real(dp), intent(in) :: gg
    real(dp), intent(in) :: alpha
    real(dp), intent(in) :: vol
    real(dp) :: gTerm

    gTerm = 4.0_dp*pi*(exp(-0.25_dp*gg**2/(alpha**2))/(vol*gg**2))

  end function gTerm



  !!* Returns the max. value of a term in the real space part of the
  !!* Ewald summation for a given vector length.
  !!* @param rr    Length of the real space vector.
  !!* @param alpha Parameter of the Ewald summation.
  function rTerm(rr, alpha)
    real(dp), intent(in) :: rr
    real(dp), intent(in) :: alpha
    real(dp) :: rTerm

    ASSERT(rr >= epsilon(1.0_dp))

#ifdef EXTERNALERFC
    rTerm = extErfc(alpha*rr)/rr
#else
    rTerm = erfc(alpha*rr)/rr
#endif

  end function rTerm


  
  !!* Returns the derivative of a term in the real space part of the
  !!* Ewald summation for a given vector length.
  !!* @param rr    Length of the real space vector.
  !!* @param alpha Parameter of the Ewald summation.
  function derivRTerm(r, alpha)
    real(dp), intent(in) :: r(3)
    real(dp), intent(in) :: alpha
    real(dp) :: derivRTerm(3)

    real(dp) :: rr
    rr = sqrt(sum(r(:)**2))
    
    ASSERT(rr >= epsilon(1.0_dp))
 
#ifdef EXTERNALERFC
    derivRTerm (:) = r(:) * &
        & (-2.0_dp/sqrt(pi)*exp(-(alpha*rr)**2)*alpha*rr &
        & - extErfc(alpha*rr))/(rr**3)
#else
    derivRTerm (:) = r(:)*(-2.0_dp/sqrt(pi)*exp(-alpha*alpha*rr*rr)* &
        & alpha*rr - erfc(alpha*rr))/(rr*rr*rr)
#endif

  end function derivRTerm

end module coulomb
