!!* Module containing various routines for DFTB+U calculations
!!* Intended to be used with SCC switched on !
module DFTBplsU
# include "assert.h"
  use accuracy
  use message
  use FileId
  use CommonTypes
  use spin, only : spinIndex
  implicit none
  private

  public :: add_VplsU, E_DFTBplsU, writeRhoToFile, readRhoFromFile

  interface add_VplsU
    module procedure V_Orbital
  end interface

  interface E_DFTBplsU
    module procedure E_Orbital
    module procedure E_Orbital_atom
  end interface

contains

  !!* Construct the Orbital contribution to the Hamiltonian
  !!* @param H the sparse Hamiltonian to add the +U potential to
  !!* @param rho the density matrix
  !!* @param nNeigh number of surrounding neighbours for each atom
  !!* @param iNeigh list of surrounding neighbours for each atom
  !!* @param specie list of the specie for each atom
  !!* @param orb  Angular momentum information about the orbitals.
  !!* @param iPair indexing array for the Hamiltonian/density matrix
  !!* @param img2CentCell indexing array to fold from periodic image atom
  !!* numbers back to the central cell number
  !!* @param UJ list of U-J values for each species
  !!* @param nUBlocks number of +U blocks to calculate for each species
  !!* @param nLinBlock number of l values contained in each block for each
  !!* species
  !!* @param iLVals list of l values in each block for each species
  !!* @param functional choice of functional, so far FLL, pSIC (1,2)
  !!* @ref Petukhov, Mazin, Chioncel, and Lichtenstein PHYSICAL REVIEW B
  !!* 67 (15): 153106 APR 15 2003
  !!* @author B. Hourahine
  !!* @todo add correct handling of ghost atoms, where the block used
  !!* ranges accross two atoms, also add other +U potentials
  subroutine V_Orbital(ham, rho, nNeigh, iNeigh, specie, orb, iPair, &
      &img2CentCell, UJ, nUBlocks, nLinBlock, iLVals, functional)
    real(dp), intent(inout) :: ham(:,:)
    real(dp), intent(in)    :: rho(:,:)
    integer,  intent(in)    :: nNeigh(:)
    integer,  intent(in)    :: iNeigh(0:,:)
    integer,  intent(in)    :: specie(:)
    type(TOrbitals), intent(in) :: orb
    integer,  intent(in)    :: iPair(0:,:)
    integer,  intent(in)    :: img2CentCell(:)
    real(dp), intent(in)    :: UJ(:,:)
    integer,  intent(in)    :: nUBlocks(:)
    integer,  intent(in)    :: nLinBlock(:,:)
    integer,  intent(in)    :: iLVals(:,:,:)
    integer,  intent(in), optional :: functional

    real(dp)    :: Htmp(orb%mOrb,orb%mOrb), rhoTmp(orb%mOrb,orb%mOrb)
    complex(dp) :: rhoTmp_nc(orb%mOrb,orb%mOrb), HTmp_nc(orb%mOrb,orb%mOrb)
    integer     :: nAtom, iAt1, iNeigh1, jj, ijj, kk, ll, la, lb
    integer     :: iOrig, iSpin, nSpin, nOrb1
    integer     :: iAlpha, iBeta
    integer     :: iFunctional
   
    ASSERT(size(rho(:,:),dim=2) == size(ham(:,:),dim=2))
    ASSERT_ENV(if (present(functional)) then)
    ASSERT_ENV(  ASSERT(functional == 1 .or. functional == 2))
    ASSERT_ENV(end if)

    nAtom = size(orb%nOrbAtom)
    nSpin = size(rho,dim=2) ! check for spin-polarisation

    if (present(functional)) then
      iFunctional = functional
    else
      iFunctional = 1
    end if

    if (nSpin < 3) then
      !! Collinear case
      do iSpin = 1, nSpin
        do iAt1 = 1, nAtom
          nOrb1 = orb%nOrbAtom(iAt1)

          !! On diagonal contribution for matrices.
          !! Shifts the trace of the matrix hence only R=0 needs to be adjusted
          if (iFunctional == 1) then
            iOrig = iPair(0,iAt1) + 1  ! counts from 0
            HTmp(:,:) = 0.0_dp
            do jj = 1, nUBlocks(specie(iAt1))
              do kk = 1, nLinBlock(jj,specie(iAt1))
                la = iLVals(kk,jj,specie(iAt1))
                do ll = la**2+1, (la+1)**2
                  HTmp(ll,ll) = HTmp(ll,ll) + 0.5_dp * UJ(jj,specie(iAt1))
                end do
              end do
            end do
            ham(iOrig:iOrig+nOrb1**2-1,iSpin) = &
                & ham(iOrig:iOrig+nOrb1**2-1,iSpin) + &
                & reshape(HTmp(1:nOrb1,1:nOrb1), (/nOrb1**2/))
          end if

          !! on atomic block contributions from the density matrix. Image atoms
          !! also have to be processed to give the k-dependance of the potential
          do iNeigh1 = 0, nNeigh(iAt1)
            !! ugly hack list of periodic images would be more elegant
            if (img2CentCell(iNeigh(iNeigh1,iAt1)) == iAt1) then
              iOrig = iPair(iNeigh1,iAt1) + 1  ! counts from 0
              HTmp(:,:) = 0.0_dp
              rhoTmp(:,:) = 0.0_dp
              rhoTmp(1:nOrb1,1:nOrb1) = &
                  & reshape(rho(iOrig:iOrig+nOrb1**2-1,iSpin), (/nOrb1,nOrb1/))
              do ijj = 1, nUBlocks(specie(iAt1))
                do kk = 1, nLinBlock(ijj,specie(iAt1))
                  la = iLVals(kk,ijj,specie(iAt1))
                  do ll = 1, nLinBlock(ijj,specie(iAt1))
                    lb = iLVals(ll,ijj,specie(iAt1))
                    HTmp(la**2+1:(la+1)**2,lb**2+1:(lb+1)**2)= &
                        & + UJ(ijj,specie(iAt1)) * &
                        & rhoTmp(la**2+1:(la+1)**2,lb**2+1:(lb+1)**2)
                  end do
                end do
              end do
              !! allow for spin average case where eigenvalues of rho are 2 and
              !! spin polarised where they are 1, so divide by 2 for s-a and
              !! 1 for s-p
              HTmp(:,:) = HTmp(:,:) / real(3-nSpin,dp)
              ham(iOrig:iOrig+nOrb1**2-1,iSpin) = &
                  & ham(iOrig:iOrig+nOrb1**2-1,iSpin) - &
                  & reshape(HTmp(1:nOrb1,1:nOrb1),(/nOrb1**2/))
            end if
          end do
        end do
      end do
    else
      !! Non collinear case
      do iAlpha = 1,2
        iSpin = spinIndex(iAlpha,iAlpha)
        do iAt1 = 1, nAtom
          nOrb1 = orb%nOrbAtom(iAt1)
          ! On diagonal contribution for matrices.
          ! Shifts the trace of the matrix hence only R=0 needs to be adjusted
          if (iFunctional == 1) then
            iOrig = iPair(0,iAt1) + 1  ! counts from 0
            HTmp_nc(:,:) = cmplx(0.0_dp,0.0_dp,dp)
            do jj = 1, nUBlocks(specie(iAt1))
              do kk = 1, nLinBlock(jj,specie(iAt1))
                la = iLVals(kk,jj,specie(iAt1))
                do ll = la**2+1, (la+1)**2
                  HTmp(ll,ll) = HTmp(ll,ll) + 0.5_dp * UJ(jj,specie(iAt1))
                end do
              end do
            end do
            ham(iOrig:iOrig+nOrb1**2-1,iSpin) = &
                & ham(iOrig:iOrig+nOrb1**2-1,iSpin) + &
                & reshape(HTmp(1:nOrb1,1:nOrb1),(/nOrb1**2/))
          end if
          do iBeta = 1,2
            iSpin = spinIndex(iAlpha,iBeta)
            !! on atomic block contributions from the density matrix. Image 
            !! atoms also have to be processed to give the k-dependance of the
            !! potential
            do iNeigh1 = 0, nNeigh(iAt1)
              !! ugly hack - a list of periodic images would be more elegant
              if (img2CentCell(iNeigh(iNeigh1,iAt1)) == iAt1) then
                iOrig = iPair(iNeigh1,iAt1)+1  ! counts from 0
                HTmp_nc(:,:) = cmplx(0.0_dp,0.0_dp, dp)
                rhoTmp_nc(:,:) = cmplx(0.0_dp,0.0_dp, dp)
                rhoTmp_nc(1:nOrb1,1:nOrb1) = &
                    & reshape(rho(iOrig:iOrig+nOrb1**2-1,iSpin), &
                    &(/nOrb1,nOrb1/))
                do ijj = 1, nUBlocks(specie(iAt1))
                  do kk = 1, nLinBlock(ijj,specie(iAt1))
                    la = iLVals(kk,ijj,specie(iAt1))
                    do ll = 1, nLinBlock(ijj,specie(iAt1))
                      lb = iLVals(ll,ijj,specie(iAt1))
                      HTmp_nc(la**2+1:(la+1)**2,lb**2+1:(lb+1)**2)= &
                          & + cmplx(UJ(ijj,specie(iAt1)),0.0_dp,dp) * &
                          & rhoTmp_nc(la**2+1:(la+1)**2,lb**2+1:(lb+1)**2)
                    end do
                  end do
                end do
                ham(iOrig:iOrig+nOrb1**2-1,iSpin) = &
                    & ham(iOrig:iOrig+nOrb1**2-1,iSpin) - &
                    & real(reshape(HTmp_nc(1:nOrb1,1:nOrb1),(/nOrb1**2/)), dp)
              end if
            end do
          end do
        end do
      end do
    end if
    
  end subroutine V_Orbital


  
  !!* Calculate the Orbital contribution to the total energy
  !!* @param rslt energy per atom
  !!* @param rho the density matrix
  !!* @param nAtom number of atoms in the central cell
  !!* @param nNeigh number of surrounding neighbours for each atom
  !!* @param iNeigh list of surrounding neighbours for each atom
  !!* @param specie list of the specie for each atom
  !!* @param orb  Angular moment information about orbitals.
  !!* @param iPair indexing array for the density matrix
  !!* @param img2CentCell indexing array to fold from periodic image atom
  !!* numbers back to the central cell number
  !!* @param UJ list of U-J values for each species
  !!* @param nUBlocks number of +U blocks to calculate for each species
  !!* @param nLinBlock number of l values contained in each block for each
  !!* species
  !!* @param iLVals list of l values in each block for each species
  !!* @param functional choice of functional, so far FLL, pSIC (1,2)
  !!* @ref Petukhov, Mazin, Chioncel, and Lichtenstein PHYSICAL REVIEW B
  !!* 67 (15): 153106 APR 15 2003
  !!* @author B. Hourahine
  !!* @todo add correct handling of ghost atoms, where the block used
  !!* ranges accross two atoms, also add other +U potentials
  subroutine E_Orbital(rslt, rho, nNeigh, iNeigh, specie, orb, iPair, &
      &img2CentCell, UJ, nUBlocks, nLinBlock, iLVals, functional)
    real(dp), intent(out) :: rslt
    real(dp), intent(in)  :: rho(:,:)
    integer,  intent(in)  :: nNeigh(:)
    integer,  intent(in)  :: iNeigh(0:,:)
    integer,  intent(in)  :: specie(:)
    type(TOrbitals), intent(in) :: orb
    integer,  intent(in)  :: iPair(0:,:)
    integer,  intent(in)  :: img2CentCell(:)
    real(dp), intent(in)  :: UJ(:,:)
    integer,   intent(in) :: nUBlocks(:)
    integer,   intent(in) :: nLinBlock(:,:)
    integer,   intent(in) :: iLVals(:,:,:)
    integer,  intent(in), optional :: functional

    real(dp)              :: atom_rslt(size(orb%nOrbAtom))

    rslt = 0.0_dp
    atom_rslt(:) = 0.0_dp

    if (present(functional)) then
      call E_Orbital_atom(atom_rslt, rho, nNeigh, iNeigh, specie, &
          &orb, iPair, img2CentCell, UJ, nUBlocks, nLinBlock, iLVals, &
          &functional)
    else
      call E_Orbital_atom(atom_rslt, rho, nNeigh, iNeigh, specie, &
          &orb, iPair, img2CentCell, UJ, nUBlocks, nLinBlock, iLVals)
    end if
    rslt = sum(atom_rslt(:))

  end subroutine E_Orbital

  

  !!* Calculate the Orbital contribution to the total energy for each atom
  !!* @param rslt energy per atom
  !!* @param rho the density matrix
  !!* @param nAtom number of atoms in the central cell
  !!* @param nNeigh number of surrounding neighbours for each atom
  !!* @param iNeigh list of surrounding neighbours for each atom
  !!* @param specie list of the specie for each atom
  !!* @param orb Angular information about orbitals.
  !!* @param iPair indexing array for the density matrix
  !!* @param img2CentCell indexing array to fold from periodic image atom
  !!* numbers back to the central cell number
  !!* @param UJ list of U-J values for each species
  !!* @param nUBlocks number of +U blocks to calculate for each species
  !!* @param nLinBlock number of l values contained in each block for each
  !!* species
  !!* @param iLVals list of l values in each block for each species
  !!* @param functional choice of functional, so far FLL, pSIC and relaxation
  !!* corected pSIC (1,2)
  !!* @ref Petukhov, Mazin, Chioncel, and Lichtenstein PHYSICAL REVIEW B
  !!* 67 (15): 153106 APR 15 2003
  !!* @author B. Hourahine
  !!* @todo add correct handling of ghost atoms, where the block used
  !!* ranges accross two atoms, also add other +U potentials
  subroutine E_Orbital_atom(rslt, rho, nNeigh, iNeigh, specie, orb, &
      &iPair, img2CentCell, UJ, nUBlocks, nLinBlock, iLVals, functional)
    real(dp), intent(out) :: rslt(:)
    real(dp), intent(in)  :: rho(:,:)
    integer,  intent(in)  :: nNeigh(:)
    integer,  intent(in)  :: iNeigh(0:,:)
    integer,  intent(in)  :: specie(:)
    type(TOrbitals),  intent(in) :: orb
    integer,  intent(in)  :: iPair(0:,:)
    integer,  intent(in)  :: img2CentCell(:)
    real(dp), intent(in)  :: UJ(:,:)
    integer,   intent(in) :: nUBlocks(:)
    integer,   intent(in) :: nLinBlock(:,:)
    integer,   intent(in) :: iLVals(:,:,:)
    integer,  intent(in), optional :: functional

    real(dp)    :: rhoTmp(orb%mOrb,orb%mOrb), HTmp(orb%mOrb,orb%mOrb)
    complex(dp) :: rhoTmp_nc(orb%mOrb,orb%mOrb), HTmp_nc(orb%mOrb,orb%mOrb)
    integer     :: nAtom, iAt1, jj, ijj, kk, ll, la, lb
    integer     :: iOrig, iSpin, nSpin, nOrb1
    integer     :: iAlpha, iBeta
    integer     :: iFunctional

    ASSERT_ENV(if (present(functional)) then)
    ASSERT_ENV(  ASSERT(functional == 1 .or. functional == 2))
    ASSERT_ENV(end if)

    nAtom = size(orb%nOrbAtom)
    
    if (present(functional)) then
      iFunctional = functional
    else
      iFunctional = 1
    end if
    
    rslt(:) = 0.0_dp

    nSpin = size(rho(:,:),dim=2) ! check for spin-polarisation

    if (nSpin < 3) then
      do iSpin = 1,nSpin
        do iAt1 = 1, nAtom
          nOrb1 = orb%nOrbAtom(iAt1)
          !! On diagonal contribution from rho. Only R=0 needs to be included
          iOrig = iPair(0,iAt1) + 1  ! counts from 0
          HTmp(:,:) = 0.0_dp
          rhoTmp(:,:) = 0.0_dp
          if (iFunctional == 1) then
            rhoTmp(1:nOrb1,1:nOrb1) = &
                & reshape(rho(iOrig:iOrig+nOrb1**2-1,iSpin), (/nOrb1,nOrb1/))
            do jj = 1, nUBlocks(specie(iAt1))
              do kk = 1, nLinBlock(jj,specie(iAt1))
                la = iLVals(kk,jj,specie(iAt1))
                do ll = la**2+1, (la+1)**2
                  HTmp(ll,ll) = HTmp(ll,ll) + 0.5_dp * UJ(jj,specie(iAt1)) * &
                      &rhoTmp(ll,ll)
                end do
              end do
            end do
            rslt(iAt1) = rslt(iAt1) + sum(HTmp)
          end if
          ! on atomic block contributions from the density matrix.
          ! Image atoms also have to be processed to give the k-dependance
          ! of the potential
          do jj = 0, nNeigh(iAt1)
            !! ugly hack - a list of periodic images would be more elegant
            if (img2CentCell(iNeigh(jj,iAt1)) == iAt1) then
              iOrig = iPair(jj,iAt1)+1  ! counts from 0
              HTmp(:,:) = 0.0_dp
              rhoTmp(:,:) = 0.0_dp
              rhoTmp(1:nOrb1,1:nOrb1) = &
                  & reshape(rho(iOrig:iOrig+nOrb1**2-1,iSpin), (/nOrb1,nOrb1/))
              do ijj = 1, nUBlocks(specie(iAt1))
                do kk = 1, nLinBlock(ijj,specie(iAt1))
                  la = iLVals(kk,ijj,specie(iAt1))
                  do ll = 1, nLinBlock(ijj,specie(iAt1))
                    lb = iLVals(ll,ijj,specie(iAt1))
                    HTmp(la**2+1:(la+1)**2,lb**2+1:(lb+1)**2) = &
                        & + UJ(ijj,specie(iAt1)) * 0.5_dp * &
                        & rhoTmp(la**2+1:(la+1)**2,lb**2+1:(lb+1)**2)**2
                  end do
                end do
              end do
              !! allow for spin average case where eigenvalues of rho are 2 or 0
              !! and spin polarised where they are 1 or 0, so divide by 2
              !! for spin average and 1 for polarised
              HTmp(:,:) = HTmp(:,:) / real(3-nSpin,dp) 
              rslt(iAt1) = rslt(iAt1) - sum(HTmp)
            end if
          end do
        end do
      end do
    else
      do iAlpha = 1,2
        iSpin = spinIndex(iAlpha,iAlpha)
        do iAt1 = 1, nAtom
          nOrb1 = orb%nOrbAtom(iAt1)
          ! On diagonal contribution from rho. Only R=0 needs to be
          ! included
          iOrig = iPair(0,iAt1) + 1  ! counts from 0
          HTmp_nc(:,:) = cmplx(0.0_dp,0.0_dp,dp)
          rhoTmp_nc(:,:) = cmplx(0.0_dp,0.0_dp,dp)
          if (iFunctional == 1) then
            rhoTmp_nc(1:nOrb1,1:nOrb1) = &
                & reshape(rho(iOrig:iOrig+nOrb1**2-1,iSpin), (/nOrb1,nOrb1/))
            do jj = 1, nUBlocks(specie(iAt1))
              do kk = 1, nLinBlock(jj,specie(iAt1))
                la = iLVals(kk,jj,specie(iAt1))
                do ll = la**2+1, (la+1)**2
                  HTmp_nc(ll,ll) = HTmp_nc(ll,ll) + &
                      & cmplx(0.5_dp*UJ(jj,specie(iAt1)),0.0_dp,dp) * &
                      & rhoTmp_nc(ll,ll)
                end do
              end do
            end do
            rslt(iAt1) = rslt(iAt1) + sum(real(HTmp_nc))
          end if
        end do
        do iBeta = 1,2
          iSpin = spinIndex(iAlpha,iBeta)
          do iAt1 = 1, nAtom
            nOrb1 = orb%nOrbAtom(iAt1)
            ! on atomic block contributions from the density matrix.
            ! Image atoms also have to be processed to give the k-dependance
            ! of the potential
            do jj = 0, nNeigh(iAt1)
              ! ugly hack - a list of periodic images would be more elegant
              if (img2CentCell(iNeigh(jj,iAt1)) == iAt1) then
                iOrig = iPair(jj,iAt1)+1  ! counts from 0
                HTmp(:,:) = 0.0_dp
                rhoTmp_nc(:,:) = cmplx(0.0_dp,0.0_dp,dp)
                rhoTmp_nc(1:nOrb1,1:nOrb1) = &
                    & reshape(rho(iOrig:iOrig+nOrb1**2-1,iSpin), &
                    & (/nOrb1,nOrb1/))
                do ijj = 1, nUBlocks(specie(iAt1))
                  do kk = 1, nLinBlock(ijj,specie(iAt1))
                    la = iLVals(kk,ijj,specie(iAt1))
                    do ll = 1, nLinBlock(ijj,specie(iAt1))
                      lb = iLVals(ll,ijj,specie(iAt1))
                      HTmp(la**2+1:(la+1)**2,lb**2+1:(lb+1)**2)= &
                          & +UJ(ijj,specie(iAt1))*0.5_dp*real( &
                          & rhoTmp_nc(la**2+1:(la+1)**2,lb**2+1:(lb+1)**2)* &
                          & conjg( &
                          & rhoTmp_nc(la**2+1:(la+1)**2,lb**2+1:(lb+1)**2)))
                    end do
                  end do
                end do
                rslt(iAt1) = rslt(iAt1) - sum(HTmp)
              end if
            end do
          end do
        end do
      end do
    end if

  end subroutine E_Orbital_atom

  

  !!* Writes parts of the density matrix to disc, necessary to reconstruct
  !!* the LDA+U potential.
  !!* @param rho          Density matrix
  !!* @param fileName     Name of the file to contain the density
  !!* @param nNeighbor    Number of neighors for each atom
  !!* @param iNeighbor    Neighbor map
  !!* @param img2CentCell Indexing array to fold from periodic image atom
  !!*   numbers back to the central cell
  !!* @param iPair        Indexing array for the density matrix
  !!* @param specie       Specie list (shape = (nAtom,))
  !!* @param orb Orbital information
  subroutine writeRhoToFile(rho, fileName, nNeighbor, iNeighbor, &
      &img2CentCell, iPair, specie, orb)
    real(dp), intent(in) :: rho(:,:)
    character(*), intent(in) :: fileName
    integer, intent(in) :: nNeighbor(:)
    integer, intent(in) :: iNeighbor(0:,:)
    integer, intent(in) :: img2CentCell(:)
    integer, intent(in) :: iPair(0:,:)
    integer, intent(in) :: specie(:)
    type(TOrbitals), intent(in) :: orb

    integer :: nAtom, nSpin
    integer :: iAtom, iSpin, iNeigh, iOrig
    integer :: nOrb1, nOrb2
    integer :: iErr
    integer, save :: fileId = -1

    nAtom = size(orb%nOrbAtom)
    nSpin = size(rho, dim=2)

    ASSERT(size(nNeighbor) >= nAtom)
    ASSERT(size(img2CentCell) >= maxval(iNeighbor))
    ASSERT(all(shape(iNeighbor) >= (/ maxval(nNeighbor), nAtom /)))
    ASSERT(all(shape(iPair) >= (/ maxval(nNeighbor), nAtom /)))
    ASSERT(size(specie) >= nAtom)

    if (fileId == -1) then
      fileId = getFileId()
    end if

    open(fileId, file=fileName, position="rewind", status="replace", &
        &form="unformatted", iostat=iErr)
    if (iErr /= 0) then
      call error("Can't create external file for density matrix")
    end if

    do iSpin = 1, nSpin
      do iAtom = 1, nAtom
        nOrb1 = orb%nOrbAtom(iAtom)
        do iNeigh = 0, nNeighbor(iAtom)
          nOrb2 = orb%nOrbAtom(img2CentCell(iNeighbor(iNeigh, iAtom)))
          !! We need only on site blocks and on-site images
          !if (img2CentCell(iNeighbor(iNeigh, iAtom)) /= iAtom) then
          !  continue
          !end if
          iOrig = iPair(iNeigh, iAtom)
          write (fileId, iostat=iErr) rho(iOrig+1:iOrig+nOrb1*nOrb2, iSpin)
          if (iErr /= 0) then
            call error("Can't write external file for density matrix")
          end if
        end do
      end do
    end do
    close(fileId)

  end subroutine writeRhoToFile



  !!* Reads parts of the density matrix from disc to reconstruct the LDA+U
  !!* potential.
  !!* @param rho          Density matrix
  !!* @param fileName     Name of the file to contain the density
  !!* @param nNeighbor    Number of neighors for each atom
  !!* @param iNeighbor    Neighbor map
  !!* @param img2CentCell Indexing array to fold from periodic image atom
  !!*   numbers back to the central cell
  !!* @param iPair        Indexing array for the density matrix
  !!* @param specie       Specie list
  !!* @param orb          Orbital information
  subroutine readRhoFromFile(rho, fileName, nNeighbor, iNeighbor, &
      &img2CentCell, iPair, orb)
    real(dp), intent(out) :: rho(:,:)
    character(*), intent(in) :: fileName
    integer, intent(in) :: nNeighbor(:)
    integer, intent(in) :: iNeighbor(0:,:)
    integer, intent(in) :: img2CentCell(:)
    integer, intent(in) :: iPair(0:,:)
    type(TOrbitals), intent(in) :: orb

    integer :: nAtom, nSpin
    integer :: iAt1, iSp1, iNeigh1, iOrig
    integer :: nOrb1, nOrb2
    integer :: iErr
    integer, save :: fileId = -1

    nAtom = size(orb%nOrbAtom)
    nSpin = size(rho, dim=2)

    ASSERT(size(nNeighbor) >= nAtom)
    ASSERT(size(img2CentCell) >= maxval(iNeighbor))
    ASSERT(all(shape(iNeighbor) >= (/ maxval(nNeighbor), nAtom /)))
    ASSERT(all(shape(iPair) >= (/ maxval(nNeighbor), nAtom /)))

    if (fileId == -1) then
      fileId = getFileId()
    end if

    rho(:,:) = 0.0_dp
    open(fileId, file=fileName, action="read", status="old", &
        &form="unformatted", iostat=iErr)
    if (iErr /= 0) then
      call error("Can't open external file for density matrix")
    end if
    do iSp1 = 1, nSpin
      do iAt1 = 1, nAtom
        nOrb1 = orb%nOrbAtom(iAt1)
        do iNeigh1 = 0, nNeighbor(iAt1)
          nOrb2 = orb%nOrbAtom(img2CentCell(iNeighbor(iNeigh1, iAt1)))
          !! We need only blocks for atoms with themselves and their images.
          !! if (img2CentCell(iNeighbor(iNeigh, iAtom)) /= iAtom) then
          !!  continue
          !! end if
          iOrig = iPair(iNeigh1, iAt1)
          read (fileId, iostat=iErr) rho(iOrig+1:iOrig+nOrb1*nOrb2, iSp1)
          if (iErr /= 0) then
            call error("Failure to read external file for density matrix")
          end if
        end do
      end do
    end do
    close(fileId)

  end subroutine readRhoFromFile

end module DFTBplsU
