!!* Code to calculate forces for several different types of calculation 
!!* (non-scc, scc, sDFTB etc)
module forces
#include "assert.h"
#include "allocate.h"
  use accuracy
  use nonscc, only: H0Sprime
  use SCC
  use CommonTypes
  use SlakoCont
  implicit none

  private
  
  public :: derivative_nonSCC, derivative_SCC, derivative_spin

contains
  
  !!* The non-SCC electronic force contribution for all atoms, calculated from
  !!* $F_\alpha = \sum_{\mu\nu} \rho_{\mu\nu}\frac{H^0_{\mu\nu}}{\partial 
  !!* R_\alpha} - \rho^E_{\mu\nu}\frac{S_{\mu\nu}}{\partial R_\alpha}$
  !!* with $\rho$ and $\rho^E$ being the density and energy-density
  !!* matrices
  !!* @param deriv x,y,z derivatives for each real atom in the system
  !!* @param DM density matrix in packed format
  !!* @param EDM energy-weighted density matrix in packed format 
  !!* @param skHamCont Container for SK Hamiltonian integrals
  !!* @param skOverCont Container for SK overlap integrals
  !!* @param coords list of all atomic coordinates
  !!* @param species list of all atomic species
  !!* @param iNeighbor neighbor list for atoms
  !!* @param nNeighbor number of neighbors of each atom
  !!* @param img2CentCell indexing array for periodic image atoms
  !!* @param iPair indexing array for the Hamiltonian
  !!* @param orb Information about the shells and orbitals in the system.
  subroutine derivative_nonSCC(deriv, DM, EDM, skHamCont, skOverCont, coords, &
      &species, iNeighbor, nNeighbor, img2CentCell, iPair, orb)
    real(dp), intent(out) :: deriv(:,:)
    real(dp), intent(in) :: DM(:)
    real(dp), intent(in) :: EDM(:)
    type(OSlakoCont), intent(in) :: skHamCont, skOverCont
    real(dp), intent(in) :: coords(:,:)
    integer, intent(in) :: species(:)
    integer, intent(in) :: iNeighbor(0:,:)
    integer, intent(in) :: nNeighbor(:)
    integer, intent(in) :: img2CentCell(:)
    integer, intent(in) :: iPair(0:,:)
    type(TOrbitals), intent(in) :: orb
    
    integer   :: iOrig, ii
    integer   :: nAtom, iNeigh, iAtom1, iAtom2, iAtom2f
    integer   :: nOrb1, nOrb2
    real(dp)  :: sqrDMTmp(orb%mOrb,orb%mOrb), sqrEDMTmp(orb%mOrb,orb%mOrb)
    real(dp)  :: hPrimeTmp(orb%mOrb,orb%mOrb,3), sPrimeTmp(orb%mOrb,orb%mOrb,3)
    
    ASSERT(size(deriv,dim=1) == 3)

    nAtom = size(orb%nOrbAtom)
    deriv(:,:) = 0.0_dp

    do iAtom1 = 1, nAtom
      nOrb1 = orb%nOrbAtom(iAtom1)
      !! loop from 1 as no contribution from the atom itself
      do iNeigh = 1, nNeighbor(iAtom1) 
        iAtom2 = iNeighbor(iNeigh, iAtom1)
        iAtom2f = img2CentCell(iAtom2)
        if (iAtom1 /= iAtom2f) then
          nOrb2 = orb%nOrbAtom(iAtom2f)
          iOrig = iPair(iNeigh,iAtom1)
          sqrDMTmp(:,:) = 0.0_dp
          sqrEDMTmp(:,:) = 0.0_dp
          hPrimeTmp(:,:,:) = 0.0_dp
          sPrimeTmp(:,:,:) = 0.0_dp          
          sqrDMTmp(1:nOrb2,1:nOrb1) = &
              & reshape(DM(iOrig+1:iOrig+nOrb1*nOrb2), (/nOrb2,nOrb1/))
          sqrEDMTmp(1:nOrb2,1:nOrb1) = &
              & reshape(EDM(iOrig+1:iOrig+nOrb1*nOrb2), (/nOrb2,nOrb1/))
          call H0Sprime(hPrimeTmp, sPrimeTmp, skHamCont, skOverCont, coords, &
              &species, iAtom1, iAtom2, orb)
          do ii = 1, 3
            deriv(ii,iAtom1) = deriv(ii,iAtom1) &
                &+ sum(sqrDMTmp(1:nOrb2,1:nOrb1)&
                &* hPrimeTmp(1:nOrb2,1:nOrb1,ii)) &
                &- sum(sqrEDMTmp(1:nOrb2,1:nOrb1)&
                &* sPrimeTmp(1:nOrb2,1:nOrb1,ii))
          end do
          !! Add contribution to the other triangle sum, using the symmetry
          !! in the blocks, and also the skew symmetry in the derivatives
          do ii = 1, 3
            deriv(ii,iAtom2f) = deriv(ii,iAtom2f) &
                &- sum(sqrDMTmp(1:nOrb2,1:nOrb1) &
                &* hPrimeTmp(1:nOrb2,1:nOrb1,ii)) &
                &+ sum(sqrEDMTmp(1:nOrb2,1:nOrb1)&
                &* sPrimeTmp(1:nOrb2,1:nOrb1,ii))
          end do
        end if
      end do
    end do
    
  end subroutine derivative_nonSCC



  !!* The SCC electronic force contribution for all atoms, calculated from
  !!* $F_\alpha = \sum_{\mu\nu} \rho_{\mu\nu}\frac{H^0_{\mu\nu}}{\partial 
  !!* R_\alpha} - (\rho^E_{\mu\nu}-\frac{H^1_{\mu\nu}}{S_{\mu\nu}}
  !!* \rho_{\mu\nu})\frac{S_{\mu\nu}}{\partial R_\alpha}$
  !!* with $\rho$ and $\rho^E$ being the density and energy-density
  !!* matrices
  !!* @param deriv x,y,z derivatives for each real atom in the system
  !!* @param H1 the SCC Hamiltonian elements
  !!* @param over the overlap matrix
  !!* @param DM density matrix in packed format
  !!* @param EDM energy-weighted density matrix in packed format 
  !!* @param skHamCont Container for SK Hamiltonian integrals
  !!* @param skOverCont Container for SK overlap integrals
  !!* @param coords list of all atomic coordinates
  !!* @param species list of all atomic species
  !!* @param iNeighbor neighbor list for atoms
  !!* @param nNeighbor number of neighbors of each atom
  !!* @param img2CentCell indexing array for periodic image atoms
  !!* @param iPair indexing array for the Hamiltonian
  !!* @param orb Information about the shells and orbitals in the system.
  subroutine derivative_SCC(deriv, DM, EDM, skHamCont, skOverCont, coords, &
      &species, iNeighbor, nNeighbor, img2CentCell, iPair, orb)
    real(dp), intent(out) :: deriv(:,:)
    real(dp), intent(in) :: DM(:)
    real(dp), intent(in) :: EDM(:)
    type(OSlakoCont), intent(in) :: skHamCont, skOverCont
    real(dp), intent(in) :: coords(:,:)
    integer, intent(in) :: species(:)
    integer, intent(in) :: iNeighbor(0:,:)
    integer, intent(in) :: nNeighbor(:)
    integer, intent(in) :: img2CentCell(:)
    integer, intent(in) :: iPair(0:,:)
    type(TOrbitals), intent(in) :: orb

    integer   :: iOrig, ii
    integer   :: nAtom, iNeigh, iAtom1, iAtom2, iAtom2f, iSp1, iSp2, iSh1, iSh2
    integer   :: nOrb1, nOrb2, iSh1l, iSh1r, iSh2l, iSh2r
    real(dp)  :: sqrDMTmp(orb%mOrb,orb%mOrb), sqrEDMTmp(orb%mOrb,orb%mOrb)
    real(dp)  :: sqrGammaDMTmp(orb%mOrb,orb%mOrb)
    real(dp)  :: hPrimeTmp(orb%mOrb,orb%mOrb,3), sPrimeTmp(orb%mOrb,orb%mOrb,3)
    real(dp), allocatable  :: shiftPerAtom(:), shiftPerL(:,:)

    
    ASSERT(size(deriv,dim=1) == 3)
    nAtom = size(orb%nOrbAtom)
    ALLOCATE_(shiftPerAtom, (nAtom))
    ALLOCATE_(shiftPerL, (orb%mShell, nAtom))

    call getShiftPerAtom(shiftPerAtom)
    call getShiftPerL(shiftPerL)
    
    deriv(:,:) = 0.0_dp

    do iAtom1 = 1, nAtom
      iSp1 = species(iAtom1)
      nOrb1 = orb%nOrbSpecie(iSp1)
      do iNeigh = 1, nNeighbor(iAtom1)
        iAtom2 = iNeighbor(iNeigh, iAtom1)
        iAtom2f = img2CentCell(iAtom2)
        iSp2 = species(iAtom2f)
        if (iAtom1 /= iAtom2f) then
          nOrb2 = orb%nOrbSpecie(iSp2)
          iOrig = iPair(iNeigh,iAtom1)+1
          sqrDMTmp(:,:)=0.0_dp
          sqrEDMTmp(:,:)=0.0_dp
          hPrimeTmp(:,:,:)=0.0_dp
          sPrimeTmp(:,:,:)=0.0_dp          
          sqrDMTmp(1:nOrb2,1:nOrb1) = &
              & reshape(DM(iOrig:iOrig+nOrb1*nOrb2-1),(/nOrb2,nOrb1/))
          sqrGammaDMTmp(:,:) = sqrDMTmp(:,:)
          sqrEDMTmp(1:nOrb2,1:nOrb1) = &
              & reshape(EDM(iOrig:iOrig+nOrb1*nOrb2-1),(/nOrb2,nOrb1/))
          ! 1/R and per-l parts of the SCC shift
          do iSh1 = 1, orb%nShell(iSp1)
            iSh1l = orb%posShell(iSh1, iSp1)
            iSh1r = orb%posShell(iSh1+1, iSp1) - 1
            do iSh2 = 1, orb%nShell(iSp2)
              iSh2l = orb%posShell(iSh2, iSp2)
              iSh2r = orb%posShell(iSh2+1, iSp2) - 1
              sqrGammaDMTmp(iSh2l:iSh2r,iSh1l:iSh1r) = &
                  & sqrGammaDMTmp(iSh2l:iSh2r,iSh1l:iSh1r) * 0.5_dp &
                  &* ((shiftPerL(iSh1,iAtom1) + shiftPerL(iSh2,iAtom2f)) &
                  &+ (shiftPerAtom(iAtom1) + shiftPerAtom(iAtom2f)))
            end do
          end do
          call H0Sprime(hPrimeTmp, sPrimeTmp, skHamCont, skOverCont, coords, &
              &species, iAtom1, iAtom2, orb)
          do ii = 1, 3
            deriv(ii,iAtom1) = deriv(ii,iAtom1) &
                & + sum(sqrDMTmp(:,:) * hPrimeTmp(:,:,ii)) &
                & - sum(sqrEDMTmp(:,:) * sPrimeTmp(:,:,ii)) &
                & + sum(sqrGammaDMTmp(:,:) * sPrimeTmp(:,:,ii))
          end do
          ! Add contribution to the other triangle sum, using the symmetry
          ! in the blocks, and also the skew symmetry in the derivatives
          do ii = 1, 3
            deriv(ii,iAtom2f) = deriv(ii,iAtom2f) &
                & - sum(sqrDMTmp(:,:) * hPrimeTmp(:,:,ii)) &
                & + sum(sqrEDMTmp(:,:) * sPrimeTmp(:,:,ii)) &
                & - sum(sqrGammaDMTmp(:,:) * sPrimeTmp(:,:,ii))
          end do
        end if
      enddo
    enddo

    DEALLOCATE_(shiftPerAtom)
    DEALLOCATE_(shiftPerL)

  end subroutine derivative_SCC


  
  !!* The SCC and spin electronic force contribution for all atoms, calculated
  !!* from
  !!* $F_\alpha = \sum_{\mu\nu} \rho_{\mu\nu}\frac{H^0_{\mu\nu}}{\partial 
  !!* R_\alpha} - (\rho^E_{\mu\nu}-\frac{H^1_{\mu\nu}}{S_{\mu\nu}}
  !!* \rho_{\mu\nu})\frac{S_{\mu\nu}}{\partial R_\alpha}$
  !!* with $\rho$ and $\rho^E$ being the density and energy-density
  !!* matrices
  !!* @param deriv x,y,z derivatives for each real atom in the system
  !!* @param H1 the SCC Hamiltonian elements
  !!* @param over the overlap matrix
  !!* @param DM density matrix in packed format
  !!* @param EDM energy-weighted density matrix in packed format 
  !!* @param skHamCont Container for SK Hamiltonian integrals
  !!* @param skOverCont Container for SK overlap integrals
  !!* @param coords list of all atomic coordinates
  !!* @param species list of all atomic species
  !!* @param iNeighbor neighbor list for atoms
  !!* @param nNeighbor number of neighbors of each atom
  !!* @param nAtom number of real atoms
  !!* @param img2CentCell indexing array for periodic image atoms
  !!* @param iPair indexing array for the Hamiltonian
  !!* @param ww array with W for all angular momentum combinations on all
  !!* species
  !!* @param qOrbital Mulliken charges, both spin components are used !
  !!* @param orb Information about the orbitals
  subroutine derivative_spin(deriv, DM, EDM, skHamCont, skOverCont, coords, &
      &species, iNeighbor, nNeighbor, img2CentCell, iPair, ww, &
      &qOrbital, orb)
    real(dp), intent(out) :: deriv(:,:)
    real(dp), intent(in) :: DM(:,:)
    real(dp), intent(in) :: EDM(:,:)
    type(OSlakoCont) :: skHamCont, skOverCont
    real(dp), intent(in) :: coords(:,:)
    integer, intent(in) :: species(:)
    integer, intent(in) :: iNeighbor(0:,:)
    integer, intent(in) :: nNeighbor(:)
    integer, intent(in) :: img2CentCell(:)
    integer, intent(in) :: iPair(0:,:)
    real(dp),  intent(in) :: ww(:,:,:)
    real(dp), intent(in)  :: qOrbital(:,:,:)
    type(TOrbitals), intent(in) :: orb

    integer  :: iOrig, ii
    integer  :: nAtom, iNeigh, iAtom1, iAtom2, iAtom2f, iSh1, iSh2, iSp1, iSp2
    integer  :: nOrb1, nOrb2, iSh1l, iSh1r, iSh2l, iSh2r
    real(dp) :: sqrDMTmp(orb%mOrb,orb%mOrb), sqrEDMTmp(orb%mOrb,orb%mOrb)
    real(dp) :: sqrGammaDMTmp(orb%mOrb,orb%mOrb)
    real(dp) :: sqrSpinDMTmp(orb%mOrb,orb%mOrb,2)
    real(dp) :: hPrimeTmp(orb%mOrb,orb%mOrb,3), sPrimeTmp(orb%mOrb,orb%mOrb,3)
    real(dp) :: deltaQ(size(qOrbital,dim=1),size(qOrbital,dim=2))
    real(dp) :: spinShiftPerL(orb%mShell,size(qOrbital,dim=2))
    real(dp) :: deltaQPerLShell(orb%mShell,size(qOrbital,dim=2))
    real(dp), allocatable  :: shiftPerAtom(:), shiftPerL(:,:)
    
    ASSERT(size(deriv,dim=1) == 3)

    nAtom = size(orb%nOrbAtom)
    ALLOCATE_(shiftPerAtom, (nAtom))
    ALLOCATE_(shiftPerL, (orb%mShell, nAtom))

    call getShiftPerAtom(shiftPerAtom)
    call getShiftPerL(shiftPerL)

    deltaQ(:,:) = (qOrbital(:,:,1) - qOrbital(:,:,2))
    deltaQPerLShell(:,:) = 0.0_dp    
    do iAtom1 = 1, nAtom
      iSp1 = species(iAtom1)
      do iSh1 = 1, orb%nShell(iSp1)
        deltaQPerLShell(iSh1,iAtom1) = sum(deltaQ(orb%posShell(iSh1,iSp1)&
            &:orb%posShell(iSh1+1,iSp1)-1,iAtom1))
      end do
    end do
    spinShiftPerL(:,:) = 0.0_dp
    do iAtom1 = 1, nAtom
      iSp1 = species(iAtom1)
      do iSh1 = 1, orb%nShell(iSp1)
        do iSh2 = 1, orb%nShell(iSp1)
          spinShiftPerL(iSh1,iAtom1) =  spinShiftPerL(iSh1,iAtom1) &
              &+ deltaQPerLShell(iSh2,iAtom1) * ww(iSh2,iSh1,iSp1)
        end do
      end do
    end do
    
    deriv(:,:) = 0.0_dp

    do iAtom1 = 1, nAtom
      iSp1 = species(iAtom1)
      nOrb1 = orb%nOrbSpecie(iSp1)
      do iNeigh = 1, nNeighbor(iAtom1)
        iAtom2 = iNeighbor(iNeigh, iAtom1)
        iAtom2f = img2CentCell(iAtom2)
        iSp2 = species(iAtom2f)
        if (iAtom1 /= iAtom2f) then
          nOrb2 = orb%nOrbSpecie(iSp2)
          iOrig = iPair(iNeigh,iAtom1) + 1
          
          sqrDMTmp(:,:) = 0.0_dp
          sqrEDMTmp(:,:) = 0.0_dp
          hPrimeTmp(:,:,:) = 0.0_dp
          sPrimeTmp(:,:,:) = 0.0_dp
          sqrSpinDMTmp(:,:,:) = 0.0_dp
          do ii = 1, 2
            sqrSpinDMTmp(1:nOrb2,1:nOrb1,ii) = &
                & reshape(DM(iOrig:iOrig+nOrb1*nOrb2-1,ii), (/nOrb2,nOrb1/))
            sqrDMTmp(1:nOrb2,1:nOrb1) = sqrDMTmp(1:nOrb2,1:nOrb1) &
                &+ sqrSpinDMTmp(1:nOrb2,1:nOrb1,ii)
            sqrEDMTmp(1:nOrb2,1:nOrb1) = sqrEDMTmp(1:nOrb2,1:nOrb1)&
                &+ reshape(EDM(iOrig:iOrig+nOrb1*nOrb2-1,ii), (/nOrb2,nOrb1/))
          end do
          sqrGammaDMTmp(:,:) = sqrDMTmp(:,:)
          ! 1/R and per-l parts of the SCC shift, and the spin-shift
          do iSh1 = 1, orb%nShell(iSp1)
            iSh1l = orb%posShell(iSh1, iSp1)
            iSh1r = orb%posShell(iSh1+1, iSp1) - 1
            do iSh2 = 1, orb%nShell(iSp2)
              iSh2l = orb%posShell(iSh2, iSp2)
              iSh2r = orb%posShell(iSh2+1, iSp2) - 1
              sqrGammaDMTmp(iSh2l:iSh2r,iSh1l:iSh1r) = &
                  &sqrGammaDMTmp(iSh2l:iSh2r,iSh1l:iSh1r) * 0.5_dp &
                  &* ((shiftPerL(iSh1,iAtom1) + shiftPerL(iSh2,iAtom2f)) &
                  &+ (shiftPerAtom(iAtom1) + shiftPerAtom(iAtom2f)))
              sqrSpinDMTmp(iSh2l:iSh2r,iSh1l:iSh1r,:) = &
                  &sqrSpinDMTmp(iSh2l:iSh2r,iSh1l:iSh1r,:) * 0.5_dp &
                  &* (spinShiftPerL(iSh1,iAtom1) + spinShiftPerL(iSh2,iAtom2f))
            end do
          end do
          
          call H0Sprime(hPrimeTmp, sPrimeTmp, skHamCont, skOverCont, coords, &
              &species, iAtom1, iAtom2, orb)
          do ii = 1, 3
            deriv(ii,iAtom1) = deriv(ii,iAtom1) &
                & + sum(sqrDMTmp(:,:) * hPrimeTmp(:,:,ii)) &
                & - sum(sqrEDMTmp(:,:) * sPrimeTmp(:,:,ii)) &
                & + sum(sqrGammaDMTmp(:,:) * sPrimeTmp(:,:,ii)) &
                & + sum(sqrSpinDMTmp(:,:,1) * sPrimeTmp(:,:,ii)) &
                & - sum(sqrSpinDMTmp(:,:,2) * sPrimeTmp(:,:,ii))
          end do
          ! Add contribution to the other triangle sum, using the symmetry
          ! in the blocks, and also the skew symmetry in the derivatives
          do ii = 1, 3
            deriv(ii,iAtom2f) = deriv(ii,iAtom2f) &
                & - sum(sqrDMTmp(:,:) * hPrimeTmp(:,:,ii)) &
                & + sum(sqrEDMTmp(:,:) * sPrimeTmp(:,:,ii)) &
                & - sum(sqrGammaDMTmp(:,:) * sPrimeTmp(:,:,ii)) &
                & - sum(sqrSpinDMTmp(:,:,1) * sPrimeTmp(:,:,ii)) &
                & + sum(sqrSpinDMTmp(:,:,2) * sPrimeTmp(:,:,ii))
          end do
        end if
      enddo
    enddo
    
    DEALLOCATE_(shiftPerAtom)
    DEALLOCATE_(shiftPerL)
    
  end subroutine derivative_spin
  
end module forces
