!!* Contains subroutines for the periodic calculations and unpacking
!!* Hamiltonian between the square and 1-dimensional representations
module periodic
# include "assert.h"
# include "allocate.h"
  use accuracy
  use constants, only : pi
  use message
  use sorting
  use bisect
  use LinkedList
  use SimpleAlgebra, only : determinant33
  use LapackRoutines, only : matinv
  use spin, only : spinIndex
  use CommonTypes
  use memman
  implicit none

  private

  public :: unpackHS, packHS, blockSymmetrizeHS
  public :: getCellTranslations, getLatticePoints, foldCoordToUnitCell
  public :: reallocateHS, buildSquaredAtomIndex
  public :: TNeighborList, init, destroy
  public :: updateNeighborList, updateNeighborListAndSpecies
  public :: getNrOfNeighbors, getNrOfNeighborsForAll
  public :: getImgRange, getSuperSampling, printMatrix
  
  !!* Unpack sparse matrix (Hamiltonian, overlap, etc.) to square form
  interface unpackHS
    module procedure unpackHS_real
    module procedure unpackHS_real_kpts
    module procedure unpackHS_cmplx
    !module procedure unpackHS_real_banded
    module procedure unpack2compHS
    module procedure unpack2compHS_kpts
  end interface

  interface reallocateHS
    module procedure reallocateHS_1
    module procedure reallocateHS_2
    module procedure reallocateHS_Single
    module procedure reallocateHS_nc
  end interface

  !!* Pack square matrix to sparse form.
  interface packHS
    module procedure packHS_real
    module procedure packHS_cmplx
  end interface

  !!* Symmetrize the square matrix except the on-site blocks
  interface blockSymmetrizeHS
    module procedure blockSymmetrizeHS_real
    module procedure blockSymmetrizeHS_cmplx
  end interface

  !!* Initializes ADTs defined in this module
  interface init
    module procedure init_TNeighborList
  end interface

  !!* Destroys ADTs defined in this module
  interface destroy
    module procedure destroy_TNeighborList
  end interface

  ! routines to print sparse matrices to file
  interface printMatrix
    module procedure sparse_realprint
    module procedure sparse_cmplxprint
  end interface
  
  !!* Contains essential data for the neighborlist
  type TNeighborList
    integer,  pointer :: iNeighbor(:,:) => null()  !* index of neighbor atoms
    integer,  pointer :: nNeighbor(:) => null()    !* nr. of neighbors
    real(dp), pointer :: neighDist2(:,:) => null() !* temporary array for
                                                   !* neighbor distances
    real(dp)            :: cutoff          !* cutoff it was generated for
    logical             :: initialized = .false.
  end type TNeighborList

contains

  !!* Initializes a neighborlist instance.
  !!* @param neibhorlist   Neighborlist data.
  !!* @param nAtom         Nr. of atoms in the system.
  !!* @param nInitNeighbor Expected nr. of neighbors per atom.
  subroutine init_TNeighborList(neighborList, nAtom, nInitNeighbor)
    type(TNeighborList), intent(out) :: neighborList
    integer,             intent(in)  :: nAtom
    integer,             intent(in)  :: nInitNeighbor

    ASSERT(.not. neighborList%initialized)
    ASSERT(nAtom > 0)
    ASSERT(nInitNeighbor > 0)

    INITALLOCATE_PARR(neighborList%nNeighbor, (nAtom))
    INITALLOCATE_PARR(neighborList%iNeighbor, (0:nInitNeighbor, nAtom))
    INITALLOCATE_PARR(neighborList%neighDist2, (0:nInitNeighbor, nAtom))

    neighborList%cutoff = -1.0_dp
    neighborList%initialized = .true.

  end subroutine init_TNeighborList



  !!* Destroys the neighborlist
  !!* @param neighborlist Neighborlist data.
  subroutine destroy_TNeighborList(neighborList)
    type(TNeighborList), intent(inout) :: neighborList

    ASSERT(neighborlist%initialized)

    DEALLOCATE_PARR(neighborList%nNeighbor)
    DEALLOCATE_PARR(neighborList%iNeighbor)
    neighborList%cutoff = -1.0_dp

    neighborList%initialized = .false.

  end subroutine destroy_TNeighborList

  

  !!* Calculates the translation vectors for cells, which could contain atoms
  !!* interacting with any of the atoms in the central cell.
  !!* @param cellVec  Returns cell translation vectors in relative coordinates.
  !!* @param rCellVec Returns cell translation vectors in absolute units.
  !!* @param latVec   Lattice vectors
  !!* @param recVec2p Reciprocal lattice vectors in 2*pi units.
  !!* @param cutoff   Global cutoff for the diatomic interactions
  !!* @note This subroutine uses a simple guess to get the necessary translation
  !!*   vectors. This results in a set of vectors wich could for very asymmetric
  !!*   cells a large amount bigger than the real necessary one.
  subroutine getCellTranslations(cellVec, rCellVec, latVec, recVec2p, cutoff)
    real(dp), pointer    :: cellVec(:, :)
    real(dp), pointer    :: rCellVec(:,:)
    real(dp), intent(in) :: latVec(:,:)
    real(dp), intent(in) :: recVec2p(:,:)
    real(dp), intent(in) :: cutoff

    integer  :: ii
    
    ASSERT(all(shape(latVec) == (/3, 3/)))
    ASSERT(all(shape(recVec2p) == (/3, 3/)))
    ASSERT(cutoff >= 0.0_dp)
    
    call getLatticePoints(cellVec, latVec, recVec2p, cutoff, posExtension=1, &
        &negExtension=1)
    INITALLOCATE_PARR(rCellVec, (3, size(cellVec, dim=2)))
    do ii = 1, size(rCellVec, dim=2)
      rCellVec(:,ii) = matmul(latVec(:,:), cellVec(:,ii))
    end do

  end subroutine getCellTranslations



  subroutine getImgRange(imgRange, dist, recVec2p, posExt, negExt)
    integer, intent(out) :: imgRange(:,:)
    real(dp), intent(in) :: dist
    real(dp), intent(in) :: recVec2p(:,:)
    integer, intent(in) :: posExt
    integer, intent(in) :: negExt

    integer :: ii, iTmp

    ASSERT(all(shape(imgRange) == (/ 2, 3 /)))
    ASSERT(dist >= 0.0_dp)
    ASSERT(all(shape(recVec2p) == (/ 3, 3 /)))
    
    do ii = 1, 3
      iTmp = floor(dist * sqrt(sum(recVec2p(:, ii)**2)))
      imgRange(1, ii) = -(iTmp + negExt)
      imgRange(2, ii) = iTmp + posExt
    end do

  end subroutine getImgRange


  
  !!* Returns a set which definitely contains all the points of a 3D grid
  !!* which are nearer to the origin as a given distance.
  !!* @param latPoint     Returns grid points in relative coords.
  !!* @param latVec       Lattice vectors.
  !!* @param recVec2p     Reciprocal lattice vectors in 2*pi units.
  !!* @param dist         Global cutoff for the diatomic interactions.
  !!* @param posExtension Extend the set along the positive lattice vectors with
  !!*    that many additional lattice vectors.
  !!* @param negExtension Same as posExtension for negative lattice vectors
  !!* @param onlyInside   Return only those lattice points which are really
  !!*                     not outside the given distance.
  !!* @note Without the onlyInside parameter, the returned set of lattice points
  !!*   shape a parallelepipedon. With the onlyInside parameter its not
  !!*   necessarily the case.
  !!* @todo Refine the algorithm with the help of a new routine which can
  !!*   calculate the minimal distance between two arbitary cells.
  subroutine getLatticePoints(latPoint, latVec, recVec2p, dist, posExtension, &
      &negExtension, onlyInside, reduceByInversion, withoutOrigin)
    real(dp), pointer    :: latPoint(:, :)
    real(dp), intent(in) :: latVec(:,:)
    real(dp), intent(in) :: recVec2p(:,:)
    real(dp), intent(in) :: dist
    integer,  intent(in), optional :: posExtension
    integer,  intent(in), optional :: negExtension
    logical,  intent(in), optional :: onlyInside
    logical,  intent(in), optional :: reduceByInversion
    logical,  intent(in), optional :: withoutOrigin

    integer  :: imgRange(2, 3)
    integer  :: negExt, posExt
    integer  :: ii, jj, kk, ind, ii0, jj0, kk0
    real(dp) :: ri, rj, rk, rTmp, dist2
    logical  :: tAll, tOrig, tNoInv
    real(dp), pointer :: tmpLatPoint(:,:)

    ASSERT(all(shape(latVec) == (/3, 3/)))
    ASSERT(all(shape(recVec2p) == (/3, 3/)))
    ASSERT(dist >= 0.0_dp)

    if (present(negExtension)) then
      negExt = negExtension
    else
      negExt = 0
    end if
    if (present(posExtension)) then
      posExt = posExtension
    else
      posExt = 0
    end if
    if (present(onlyInside)) then
      tAll = .not. onlyInside
    else
      tAll = .true.
    end if
    if (present(withoutOrigin)) then
      tOrig = .not. withoutOrigin
    else
      tOrig = .true.
    end if
    if (present(reduceByInversion)) then
      tNoInv = reduceByInversion
    else
      tNoInv = .false.
    end if

    call getImgRange(imgRange, dist, recVec2p, posExt, negExt)

    INITALLOCATE_PARR(tmpLatPoint, \
        (3, product(sum(abs(imgRange(:,:)), dim=1)+1)))
    
    dist2 = dist**2
    if (tOrig) then
      tmpLatPoint(:, 1) = (/0.0_dp, 0.0_dp, 0.0_dp/)
      ind = 1
    else
      ind = 0
    end if

    if (tNoInv) then
      ii0 = 0
    else
      ii0 = imgRange(1,1)
    end if
    iiLoop: do ii = ii0, imgRange(2, 1)
      ri = real(ii)

      if (tNoInv .and. ii == 0) then
        jj0 = 0
      else
        jj0 = imgRange(1, 2)
      end if
      jjLoop: do jj = jj0, imgRange(2, 2)
        rj = real(jj)

        if (tNoInv .and. ii == 0 .and. jj == 0) then
          kk0 = 0
        else
          kk0 = imgRange(1, 3)
        end if
        kkLoop: do kk = kk0, imgRange(2, 3)
          if (kk == 0 .and. jj == 0 .and. ii == 0) then
            cycle
          end if
          rk = real(kk)
          if (tAll) then
            ind = ind + 1
            tmpLatPoint(:, ind) = (/ri, rj, rk/)
          else
            rTmp = sum(matmul(latVec, reshape((/ri, rj, rk/), (/3, 1/)))**2)
            if (rTmp <= dist2) then
              ind = ind + 1
              tmpLatPoint(:, ind) = (/ri, rj, rk/)
            end if
          end if
        end do kkLoop
      end do jjLoop
    end do iiLoop
    
    if (ind == size(tmpLatPoint, dim=2)) then
      latPoint => tmpLatPoint
    else
      INITALLOCATE_PARR(latPoint, (3, ind))
      latPoint(:,:) = tmpLatPoint(:,:ind)
      DEALLOCATE_PARR(tmpLatPoint)
    end if

  end subroutine getLatticePoints



  !!* Fold coordinates back in the central cell
  !!* @param coord       Contains the original coordinates on call and
  !!*   the folded ones on return.
  !!* @param latVec      Lattice vectors (column format).
  !!* @param recVec2p    Reciprocal vectors in units of 2pi (column format).
  !!* @param invShift    Contains difference vectors old_coords - new_coords.
  !!* @desc
  !!*   Throw away the integer part of the relative coordinates of every
  !!*   atom. If the resulting coordinate is very near to 1.0 (closer
  !!*   than tolSameDist in absolute length), fold it to 0.0 to make the
  !!*   algorithm more predictable and independent of numeric noises.
  subroutine foldCoordToUnitCell(coord, latVec, recVec2p, invShift)
    real(dp), intent(inout) :: coord(:,:)
    real(dp), intent(in)    :: latVec(:,:)
    real(dp), intent(in)    :: recVec2p(:,:)
    real(dp), intent(out), optional :: invShift(:,:)

    integer  :: nAtom             ! Nr. of atoms in the cell.
    integer  :: ii, jj
    real(dp) :: frac(3), frac2(3), tmp3(3), vecLen(3)

    nAtom = size(coord, dim=2)

    ASSERT(size(coord, dim=1) == 3)
    ASSERT(all(shape(latVec) == (/3, 3/)))
    ASSERT(all(shape(recVec2p) == (/3, 3/)))
    ASSERT_ENV(if (present(invShift)) then)
    ASSERT_ENV(  ASSERT(all(shape(invShift) == shape(coord))))
    ASSERT_ENV(end if)

    vecLen(:) = sqrt(sum(latVec(:,:)**2, dim=1))
    do ii = 1, nAtom
      do jj = 1, 3
        frac(jj) = dot_product(recVec2p(:,jj), coord(:,ii))
      end do
      tmp3(:) = coord(:,ii)
      frac2(:) = frac(:) - real(floor(frac(:)), dp)
      where (abs(vecLen*(1.0 - frac2)) < tolSameDist) frac2 = 0.0_dp
      coord(:, ii) = matmul(latVec, frac2)
      if (present(invShift)) then
        invShift(:,ii) = tmp3(:) - coord(:,ii)
      end if
    end do

  end subroutine foldCoordToUnitCell



  !!* Updates the neighbor list and the species arrays.
  !!* @param coord Coordinates of all interacting atoms on exit
  !!* @param specie Specie of all interacting atoms on exit.
  !!* @param img2CentCell Mapping on atoms in the central cell
  !!* @param iCellVec Shift vector index for every interacting atom
  !!* @param neig Updated neighbor list.
  !!* @param nAllAtom Number of all interacting atoms
  !!* @param coord0 Coordinates of the atoms in the central cell
  !!* @param specie0 Species of the atoms in the central cell
  !!* @param cutoff Cutoff until neighborlist should be created
  !!* @param rCellVec Cell vector for the translated cells to consider.
  subroutine updateNeighborListAndSpecies(coord, specie, img2CentCell, &
      &iCellVec, neigh, nAllAtom, coord0, specie0, cutoff, rCellVec)
    real(dp), pointer                  :: coord(:,:)
    integer,  pointer                  :: specie(:)
    integer,  pointer                  :: img2CentCell(:)
    integer,  pointer                  :: iCellVec(:)
    type(TNeighborList), intent(inout) :: neigh
    integer,  intent(out)              :: nAllAtom
    real(dp), intent(in)               :: coord0(:,:)
    integer,  intent(in)               :: specie0(:)
    real(dp), intent(in)               :: cutoff
    real(dp), intent(in)               :: rCellVec(:,:)
    

    call updateNeighborList(coord, img2CentCell, iCellVec, neigh, nAllAtom, &
        &coord0, cutoff, rCellVec)
    if (size(specie) < nAllAtom) then
      DEALLOCATE_PARR(specie)
      ALLOCATE_PARR(specie, (nAllAtom))
    end if
    specie(1:nAllAtom) = specie0(img2CentCell(1:nAllAtom))

  end subroutine updateNeighborListAndSpecies
  


  !!* Updates the neighbor list according a given geometry.
  !!* @param coord1       Coordinates of the objects interacting with the
  !!*   objects in the central cell (on exit).
  !!* @param img2CentCell Returns for all objects the index of an object in
  !!*   the central cell which the object is mapped on.
  !!* @param iCellVec     Returns the index of the translating
  !!*   superlattice vector for each object.
  !!* @param neigh        Neighborlist.
  !!* @param nAllAtom     Returns the nr. of all objects (including those
  !!*   in the translated cells.)
  !!* @param coord0In     Coordinates of the objects in the central cell.
  !!* @param coord1In     Coorindates of the periodical images interacting
  !!*   with the objects in the centrall cell.
  !!* @param cutoff       Cutoff radius for the interactions.
  !!* @param rCellVec     Absolute coordinates of the shifted supercells which
  !!*   could have interacting atoms with the central cell.
  !!* @desc The neighborlist for the given cutoff is calculated. Arrays are
  !!*   resized if necessary. The neighbor list determination is a simple
  !!*   N^2 algorithm, calculating the distance between the possible atom pairs.
  subroutine updateNeighborList(coord, img2CentCell, iCellVec, neigh, &
      &nAllAtom, coord0, cutoff, rCellVec)
    real(dp), pointer                  :: coord(:,:)
    integer,  pointer                  :: img2CentCell(:)
    integer,  pointer                  :: iCellVec(:)
    type(TNeighborList), intent(inout) :: neigh
    integer,  intent(out)              :: nAllAtom
    real(dp), intent(in)               :: coord0(:,:)
    real(dp), intent(in)               :: cutoff
    real(dp), intent(in)               :: rCellVec(:,:)

    integer  :: nAtom             ! Nr. of atoms in the system
    integer  :: mAtom             ! Max. nr. of atom without reallocation
    integer  :: maxNeighbor       ! Max. nr. of neighbors without reallocation
    integer  :: nCellVec          ! Nr. of cell translation vectors

    !! Square of the diatomic interaction cutoffs
    real(dp) :: cutoff2

    real(dp) :: dist2
    real(dp) :: rCell(3), rr(3)
    integer  :: ii, iAtom1, oldIAtom1, iAtom2
    integer  :: nn1

    integer,  allocatable :: indx(:)
    character(len=100)    :: strError

    nAtom = size(neigh%nNeighbor, dim=1)
    mAtom = size(coord, dim=2)
    maxNeighbor = ubound(neigh%iNeighbor, dim=1)
    nCellVec = size(rCellVec, dim=2)

    ASSERT(nAtom <= mAtom)
    ASSERT(associated(coord))
    ASSERT(size(coord, dim=1) == 3)
    ASSERT(associated(img2CentCell))
    ASSERT(size(img2CentCell) == mAtom)
    ASSERT(associated(iCellVec))
    ASSERT(size(iCellVec) == mAtom)
    ASSERT(size(neigh%iNeighbor, dim=2) == nAtom)
    ASSERT((size(coord0, dim=1) == 3) .and. size(coord0, dim=2) >= nAtom)
    ASSERT((size(rCellVec, dim=1) == 3))
    ASSERT(cutoff >= 0.0_dp)

    neigh%cutoff = cutoff
    cutoff2 = cutoff**2
    nAllAtom = 0

    !! Clean arrays.
    !! (Every atom is the 0th neigbhor of itself with zero distance square.)
    neigh%nNeighbor(:) = 0
    neigh%iNeighbor(:,:) = 0
    do ii = 1, nAtom
      neigh%iNeighbor(0, ii) = ii
    end do
    neigh%neighDist2(:,:) = 0.0_dp


    !! Loop over all possible neighbors for all atoms in the central cell.
    !! Only those neighbors are considered which map on atom with a higher
    !! or equal index in the central cell.
    !! Outer two loops: all atoms in all cells.
    !! Inner loop: all atoms in the central cell.
    lpCellVec: do ii = 1, nCellVec
      rCell(:) = rCellVec(:, ii)
      oldIAtom1 = 0
      lpIAtom1: do iAtom1 = 1, nAtom
        rr(:) = coord0(:, iAtom1) + rCell(:)
        lpIAtom2: do iAtom2 = 1, iAtom1

          !! If distance greater than cutoff -> skip
          dist2 = sum((coord0(:, iAtom2) - rr(:))**2)
          if (dist2 > cutoff2) then
            cycle
          end if

          !! New interacting atom -> append
          !! We need that before checking for interaction with dummy atom or
          !! with itself to make sure that atoms in the central cell are
          !! appended  exactly in the same order as found in the coord0 array.
          if (iAtom1 /= oldIAtom1) then
            nAllAtom = nAllAtom + 1
            if (nAllAtom > mAtom) then
              mAtom = incrmntOfArray(mAtom)
              call reallocateArrays1(img2CentCell, iCellVec, coord, mAtom)
            end if
            coord(:, nAllAtom) = rr(:)
            img2CentCell(nAllAtom) = iAtom1
            iCellVec(nAllAtom) = ii
            oldIAtom1 = iAtom1
          end if

          !! Check if atoms are not too close to each other
          if (dist2 < minNeighDist2) then
            if (ii == 1 .and. iAtom1 == iAtom2) then
              !! We calculated the distance between the same atom in the
              !! unit cell
              cycle
            else
99000         format ('Atoms ',I5,' and ',I5,' too close to each other!', &
                  & ' (dist=',E13.6,')')
              write (strError, 99000) iAtom2, nAllAtom, sqrt(dist2)
              call warning(strError)
            end if
          end if

          neigh%nNeighbor(iAtom2) = neigh%nNeighbor(iAtom2) + 1
          if (neigh%nNeighbor(iAtom2) > maxNeighbor) then
            maxNeighbor = incrmntOfArray(maxNeighbor)
            call reallocateArrays3(neigh%iNeighbor, neigh%neighDist2, &
                & maxNeighbor)

          end if
          neigh%iNeighbor(neigh%nNeighbor(iAtom2), iAtom2) = nAllAtom
          neigh%neighDist2(neigh%nNeighbor(iAtom2), iAtom2) = dist2

        end do lpIAtom2
      end do lpIAtom1
    end do lpCellVec

    !! Sort neighbors for all atom by distance
    ALLOCATE_(indx, (maxNeighbor))
    do iAtom1 = 1, nAtom
      nn1 = neigh%nNeighbor(iAtom1)
      call index_heap_sort(indx(1:nn1), neigh%neighDist2(1:nn1, iAtom1),&
          &tolSameDist2)
      neigh%iNeighbor(1:nn1, iAtom1) = neigh%iNeighbor(indx(:nn1), iAtom1)
      neigh%neighDist2(1:nn1, iAtom1) = neigh%neighDist2(indx(:nn1), iAtom1)
    end do
    DEALLOCATE_(indx)
    coord(:,nAllAtom+1:size(coord, dim=2)) = 0.0_dp
    
  end subroutine updateNeighborList

  

  !!* Returns the nr. of neighbors for a given cutoff for all atoms.
  !!* @param nNeighbor Contains the nr. of neighbors for each atom on exit.
  !!* @param neigh     Initialized neighborlist
  !!* @param cutoff    Maximal neighbor distance to consider.
  subroutine getNrOfNeighborsForAll(nNeighbor, neigh, cutoff)
    integer,             intent(out) :: nNeighbor(:)
    type(TNeighborList), intent(in)  :: neigh
    real(dp),            intent(in)  :: cutoff

    integer  :: nAtom, iAtom

    nAtom = size(nNeighbor)

    ASSERT(size(neigh%iNeighbor, dim=2) == nAtom)
    ASSERT(size(neigh%nNeighbor) == nAtom)
    ASSERT(maxval(neigh%nNeighbor) <= size(neigh%iNeighbor, dim=1))
    ASSERT(all(shape(neigh%neighDist2) == shape(neigh%iNeighbor)))
    ASSERT(cutoff >= 0.0_dp)

    !! Get last interacting neighbor for given cutoff
    do iAtom = 1, nAtom
      nNeighbor(iAtom) = getNrOfNeighbors(neigh, cutoff, iAtom)
    end do
     
  end subroutine getNrOfNeighborsForAll



  !!* Returns the nr. of neighbors for a given atom.
  !!* @param neigh   Intialized neihgborlist.
  !!* @param cutoff  Maximal neighbor distance to consider.
  !!* @param iAtom   Index of the atom to get the nr. of neighbors for.
  !!* @return        Nr. of neighbors for the specified atom.
  function getNrOfNeighbors(neigh, cutoff, iAtom) result(nNeighbor)
    type(TNeighborList), intent(in)  :: neigh
    real(dp),            intent(in)  :: cutoff
    integer,             intent(in)  :: iAtom
    integer :: nNeighbor

    character(len=100) :: strError

    ASSERT(cutoff >= 0.0_dp)
    ASSERT(iAtom <= size(neigh%nNeighbor))

    !! Issue warning, if cutoff is bigger as used for the neighborlist.
    if (cutoff > neigh%cutoff) then
99010 format ('Cutoff (', E16.6, ') greater then last cutoff ', &
          & '(', E13.6, ') passed to updateNeighborList!')
      write (strError, 99010) cutoff, neigh%cutoff
      call warning(strError)
    end if

    !! Get last interacting neighbor for given cutoff
    call bisection(nNeighbor, &
        &neigh%neighDist2(1:neigh%nNeighbor(iAtom), iAtom), cutoff**2, &
        &tolSameDist2)
    
  end function getNrOfNeighbors





  !!* Reallocate arrays which depends on the maximal nr. of all atoms.
  subroutine reallocateArrays1(img2CentCell, iCellVec, coord, mNewAtom)
    integer,  pointer    :: img2CentCell(:)
    integer,  pointer    :: iCellVec(:)
    real(dp), pointer    :: coord(:, :)
    integer,  intent(in) :: mNewAtom

    integer               :: mAtom
    integer,  allocatable :: tmpIntR1(:)
    real(dp), allocatable :: tmpRealR2(:, :)

    mAtom = size(img2CentCell)

    ASSERT(size(iCellVec) == mAtom)
    ASSERT(all(shape(coord) == (/ 3, mAtom /)))
    ASSERT((mNewAtom > 0) .and. (mNewAtom > mAtom))

    ALLOCATE_(tmpIntR1, (mAtom))
    tmpIntR1(:) = img2CentCell(:)
    DEALLOCATE_PARR(img2CentCell)
    ALLOCATE_PARR(img2CentCell, (mNewAtom))
    img2CentCell(:) = 0
    img2CentCell(:mAtom) = tmpIntR1(:mAtom)    
    
    tmpIntR1(:) = iCellVec(:)
    DEALLOCATE_PARR(iCellVec)
    ALLOCATE_PARR(iCellVec, (mNewAtom))
    iCellVec(:mAtom) = tmpIntR1(:mAtom)

    ALLOCATE_(tmpRealR2, (3, mNewAtom))
    tmpRealR2(:, 1:size(coord,dim=2)) = coord(:, 1:size(coord,dim=2))
    ! fails on alphas for some reason :
    ! tmpRealR2(:, :) = coord(:, :)
    DEALLOCATE_PARR(coord)
    ALLOCATE_PARR(coord, (3, mNewAtom))
    coord(:, :mAtom) = tmpRealR2(:, :mAtom)
    DEALLOCATE_(tmpRealR2)

  end subroutine reallocateArrays1


  
  !!* Reallocate array which depends on the maximal nr. of neighbors.
  subroutine reallocateArrays3(iNeighbor, neighDist2, mNewNeighbor)
    integer,  pointer    :: iNeighbor(:, :)
    real(dp), pointer    :: neighDist2(:,:)
    integer,  intent(in) :: mNewNeighbor

    integer               :: mNeighbor, mAtom
    integer,  allocatable :: tmpIntR2(:,:)
    real(dp), allocatable :: tmpRealR2(:,:)

    mNeighbor = ubound(iNeighbor, dim=1)
    mAtom = size(iNeighbor, dim=2)

    ASSERT(mNewNeighbor > 0 .and. mNewNeighbor > mNeighbor)
    ASSERT(all(shape(neighDist2) == shape(iNeighbor)))

    ALLOCATE_(tmpIntR2, (0:mNeighbor, mAtom))
    tmpIntR2(:, :) = iNeighbor(:, :)
    DEALLOCATE_PARR(iNeighbor)
    ALLOCATE_PARR(iNeighbor, (0:mNewNeighbor, mAtom))
    iNeighbor(:,:) = 0
    iNeighbor(:mNeighbor, :mAtom) = tmpIntR2(:mNeighbor, :mAtom)
    DEALLOCATE_(tmpIntR2)

    ALLOCATE_(tmpRealR2, (0:mNeighbor, mAtom))
    tmpRealR2(:,:) = neighDist2(:,:)
    DEALLOCATE_PARR(neighDist2)
    ALLOCATE_PARR(neighDist2, (0:mNewNeighbor, mAtom))
    neighDist2 = 0.0_dp
    neighDist2(:mNeighbor, :mAtom) = tmpRealR2(:mNeighbor, :mAtom)
    DEALLOCATE_(tmpRealR2)

  end subroutine reallocateArrays3



  !!* Allocate (reallocate) space for the sparse hamiltonian and overlap
  !!* matrix.
  !!* @param ham       Hamiltonian.
  !!* @param over      Overlap matrix.
  !!* @param iPair     Pair indexing array (specifying the offset for the
  !!*   interaction between atoms in the central cell and their neigbhors).
  !!* @param iNeighbor List of neighbors for each atom in the central cell.
  !!*   (Note: first index runs from 0!)
  !!* @param nNeighbor Nr. of neighbors for each atom in the central cell.
  !!* @param orb Orbitals in the system.
  subroutine reallocateHS_1(ham, over, iPair, iNeighbor, nNeighbor, orb, &
      &img2Centcell)
    real(dp), pointer    :: ham(:)
    real(dp), pointer    :: over(:)
    integer,  pointer    :: iPair(:,:)
    integer,  intent(in) :: iNeighbor(0:,:)
    integer,  intent(in) :: nNeighbor(:)
    type(TOrbitals), intent(in) :: orb
    integer, intent(in) :: img2CentCell(:)

    integer :: nAtom           ! nr. atoms in the central cell
    integer :: nElem, nOldElem ! nr. of elements in the spare H/S
    integer :: mNeighbor       ! nr. of max. possible neighbors (incl. itself)
    integer :: ind
    integer :: iAt1, iNeigh1, nOrb1

    nAtom = size(iNeighbor, dim=2)
    mNeighbor = size(iNeighbor, dim=1)
    nOldElem = size(ham, dim=1)

    ASSERT(associated(ham))
    ASSERT(associated(over))
    ASSERT(size(over) == nOldElem)
    ASSERT(associated(iPair))
    ASSERT(size(iPair, dim=2) == nAtom)

    if (mNeighbor > size(iPair, dim=1)) then
      DEALLOCATE_PARR(iPair)
      ALLOCATE_PARR(iPair, (0:mNeighbor-1, nAtom))
      iPair(:,:) = 0
    end if
    nElem = 0
    ind = 0
    do iAt1 = 1, nAtom
      nOrb1 = orb%nOrbAtom(iAt1)
      do iNeigh1 = 0, nNeighbor(iAt1)
        iPair(iNeigh1, iAt1) = ind
        ind = ind + nOrb1 * orb%nOrbAtom(img2CentCell(iNeighbor(iNeigh1, iAt1)))
      end do
    end do
    nElem = ind
    if (nElem > nOldElem) then
      DEALLOCATE_PARR(ham)
      DEALLOCATE_PARR(over)
      ALLOCATE_PARR(ham, (nElem))
      ALLOCATE_PARR(over, (nElem))
      ham(:) = 0.0_dp
      over(:) = 0.0_dp
    end if

  end subroutine reallocateHS_1


  
  !!* Allocate (reallocate) space for the sparse hamiltonian and overlap
  !!* matrix.
  !!* @param ham       Hamiltonian.
  !!* @param over      Overlap matrix.
  !!* @param iPair     Pair indexing array (specifying the offset for the
  !!*   interaction between atoms in the central cell and their neigbhors).
  !!* @param iNeighbor List of neighbors for each atom in the central cell.
  !!*   (Note: first index runs from 0!)
  !!* @param nNeighbor Nr. of neighbors for each atom in the central cell.
  !!* @param orb Orbitals in the system.
  !!* @param img2CentCell Mapping on atoms in the central cell
  subroutine reallocateHS_2(ham, over, iPair, iNeighbor, nNeighbor, orb, &
      &img2CentCell)
    real(dp), pointer    :: ham(:,:)
    real(dp), pointer    :: over(:)
    integer,  pointer    :: iPair(:,:)
    integer,  intent(in) :: iNeighbor(0:,:)
    integer,  intent(in) :: nNeighbor(:)
    type(TOrbitals), intent(in) :: orb
    integer, intent(in) :: img2CentCell(:)

    integer :: nSpinBlock      ! nr. of spin blocks in the Hamiltonian
    integer :: nAtom           ! nr. atoms in the central cell
    integer :: nElem, nOldElem ! nr. of elements in the spare H/S
    integer :: mNeighbor       ! nr. of max. possible neighbors (incl. itself)
    integer :: ind
    integer :: iAt1, iNeigh1, nOrb1

    nAtom = size(iNeighbor, dim=2)
    mNeighbor = size(iNeighbor, dim=1)
    nSpinBlock = size(ham, dim=2)
    nOldElem = size(ham, dim=1)

    ASSERT(associated(ham))
    ASSERT(associated(over))
    ASSERT(size(over) == nOldElem)
    ASSERT(associated(iPair))
    ASSERT(size(iPair, dim=2) == nAtom)

    if (mNeighbor > size(iPair, dim=1)) then
      DEALLOCATE_PARR(iPair)
      ALLOCATE_PARR(iPair, (0:mNeighbor-1, nAtom))
      iPair(:,:) = 0
    end if
    nElem = 0
    ind = 0
    do iAt1 = 1, nAtom
      nOrb1 = orb%nOrbAtom(iAt1)
      do iNeigh1 = 0, nNeighbor(iAt1)
        iPair(iNeigh1, iAt1) = ind
        ind = ind +  nOrb1 * orb%nOrbAtom(img2CentCell(iNeighbor(iNeigh1,iAt1)))
      end do
    end do
    nElem = ind
    if (nElem > nOldElem) then
      DEALLOCATE_PARR(ham)
      DEALLOCATE_PARR(over)
      ALLOCATE_PARR(ham, (nElem, nSpinBlock))
      ALLOCATE_PARR(over, (nElem))
      ham(:,:) = 0.0_dp
      over(:) = 0.0_dp
    end if

  end subroutine reallocateHS_2



  !!* Allocate (reallocate) space for the sparse non-collinear hamiltonian and 
  !!* overlap matrix.
  !!* @param ham       Hamiltonian - Pauli block matrix form.
  !!* @param over      Overlap matrix - Pauli block matrix form.
  !!* @param iPair     Pair indexing array (specifying the offset for the
  !!*   interaction between atoms in the central cell and their neigbhors).
  !!* @param iNeighbor List of neighbors for each atom in the central cell.
  !!*   (Note: first index runs from 0!)
  !!* @param nNeighbor Nr. of neighbors for each atom in the central cell.
  !!* @param orb Information about the orbitals in the system.
  !!* @param img2CentCell Mapping on atoms in the central cell.
  subroutine reallocateHS_nc(ham, over, iPair, iNeighbor, nNeighbor, orb, &
      &img2CentCell)
    complex(dp), pointer :: ham(:,:)
    complex(dp), pointer :: over(:,:)
    integer,  pointer    :: iPair(:,:)
    integer,  intent(in) :: iNeighbor(0:,:)
    integer,  intent(in) :: nNeighbor(:)
    type(TOrbitals), intent(in) :: orb
    integer, intent(in) :: img2CentCell(:)
    
    integer :: nSpinBlock      ! nr. of spin blocks in the Hamiltonian
    integer :: nAtom           ! nr. atoms in the central cell
    integer :: nElem, nOldElem ! nr. of elements in the spare H/S
    integer :: mNeighbor       ! nr. of max. possible neighbors (incl. itself)
    integer :: ind
    integer :: iAt1, iNeigh1, nOrb1
    
    nAtom = size(iNeighbor, dim=2)
    mNeighbor = size(iNeighbor, dim=1)
    nSpinBlock = size(ham, dim=2)
    nOldElem = size(ham, dim=1)
    
    ASSERT(associated(ham))
    ASSERT(associated(over))
    ASSERT(size(over) == nOldElem)
    ASSERT(associated(iPair))
    ASSERT(size(iPair, dim=2) == nAtom)
    
    if (mNeighbor > size(iPair, dim=1)) then
      DEALLOCATE_PARR(iPair)
      ALLOCATE_PARR(iPair, (0:mNeighbor-1, nAtom))
      iPair(:,:) = 0
    end if
    nElem = 0
    ind = 0
    do iAt1 = 1, nAtom
      nOrb1 = orb%nOrbAtom(iAt1)
      do iNeigh1 = 0, nNeighbor(iAt1)
        iPair(iNeigh1, iAt1) = ind
        ind = ind +  nOrb1 * orb%nOrbAtom(img2CentCell(iNeighbor(iNeigh1,iAt1)))
      end do
    end do
    nElem = ind
    if (nElem > nOldElem) then
      DEALLOCATE_PARR(ham)
      DEALLOCATE_PARR(over)
      ALLOCATE_PARR(ham, (nElem, nSpinBlock))
      ALLOCATE_PARR(over, (nElem, nSpinBlock))
      ham(:,:) = cmplx(0.0_dp,0.0_dp,dp)
      over(:,:) = cmplx(0.0_dp,0.0_dp,dp)
    end if
    
  end subroutine reallocateHS_nc


  
  !!* Allocate (reallocate) space for the sparse hamiltonian and overlap
  !!* matrix.
  !!* @param ham       Hamiltonian.
  !!* @param iPair     Pair indexing array (specifying the offset for the
  !!*   interaction between atoms in the central cell and their neigbhors).
  !!* @param iNeighbor List of neighbors for each atom in the central cell.
  !!*   (Note: first index runs from 0!)
  !!* @param nNeighbor Nr. of neighbors for each atom in the central cell.
  !!* @param orb Information about the orbitals in the system.
  !!* @param img2CentCell Mapping on atoms in the central cell.
  subroutine reallocateHS_Single(ham, iPair, iNeighbor, nNeighbor, orb, &
      &img2CentCell)
    real(dp), pointer    :: ham(:)
    integer,  pointer    :: iPair(:,:)
    integer,  intent(in) :: iNeighbor(0:,:)
    integer,  intent(in) :: nNeighbor(:)
    type(TOrbitals), intent(in) :: orb
    integer, intent(in) :: img2CentCell(:)

    integer :: nAtom           ! nr. atoms in the central cell
    integer :: nElem, nOldElem ! nr. of elements in the spare H/S
    integer :: mNeighbor       ! nr. of max. possible neighbors (incl. itself)
    integer :: ind
    integer :: iAt1, iNeigh1, nOrb1

    nAtom = size(iNeighbor, dim=2)
    mNeighbor = size(iNeighbor, dim=1)
    nOldElem = size(ham, dim=1)

    ASSERT(associated(ham))
    ASSERT(associated(iPair))
    ASSERT(size(iPair, dim=2) == nAtom)

    if (mNeighbor > size(iPair, dim=1)) then
      DEALLOCATE_PARR(iPair)
      ALLOCATE_PARR(iPair, (0:mNeighbor-1, nAtom))
      iPair(:,:) = 0
    end if
    nElem = 0
    ind = 0
    do iAt1 = 1, nAtom
      nOrb1 = orb%nOrbAtom(iAt1)
      do iNeigh1 = 0, nNeighbor(iAt1)
        iPair(iNeigh1, iAt1) = ind
        ind = ind +  nOrb1 * orb%nOrbAtom(img2CentCell(iNeighbor(iNeigh1,iAt1)))
      end do
    end do
    nElem = ind
    if (nElem > nOldElem) then
      DEALLOCATE_PARR(ham)
      ALLOCATE_PARR(ham, (nElem))
      ham(:) = 0.0_dp
    end if

  end subroutine reallocateHS_Single


  
  !!* Builds an atom offset array for the squared hamiltonain/overlap.
  !!* @param iSquare  Returns the offset array for each atom.
  !!* @param orb Information about the orbitals in the system.
  subroutine buildSquaredAtomIndex(iSquare, orb)
    integer, intent(out) :: iSquare(:)
    type(TOrbitals), intent(in) :: orb

    integer :: ind, iAt1
    integer :: nAtom

    nAtom = size(orb%nOrbAtom)

    ASSERT(all(shape(iSquare) == (/ nAtom + 1 /)))

    ind = 1
    do iAt1 = 1, nAtom
      iSquare(iAt1) = ind
      ind = ind + orb%nOrbAtom(iAt1)
    end do
    iSquare(nAtom+1) = ind

  end subroutine buildSquaredAtomIndex



  !!* Unpacks sparse matrix to square form (complex version)
  !!* @param square       Square form matrix on exit.
  !!* @param orig         Sparse matrix
  !!* @param kPoint       <b>Relative</b> coordinates of the K-point where the
  !!*   sparse  matrix    should be unfolded.
  !!* @param iNeighbor    Neighbor list for each atom (First index from 0!)
  !!* @param nNeighbor    Nr. of neighbors for each atom (incl. itself).
  !!* @param iCellVec     Index of the cell translation vector for each atom.
  !!* @param cellVec      Relative coordinates of the cell translation vectors.
  !!* @param iSquare      Atom offset for the squared Hamiltonian
  !!* @param iPair indexing array for the sparse Hamiltonian
  !!* @param img2CentCell Atomic mapping indexes.
  !!* @note The non on-site blocks are only filled in the lower triangle part
  !!*   of the matrix. To fill the matrix completely, apply the
  !!*   blockSymmetrizeHS subroutine.
  subroutine unpackHS_cmplx(square, orig, kPoint, iNeighbor, nNeighbor, &
      &iCellVec, cellVec, iSquare, iPair, img2CentCell)
    complex(dp), intent(out) :: square(:,:)
    real(dp),    intent(in)  :: orig(:)
    real(dp),    intent(in)  :: kPoint(:)
    integer,     intent(in)  :: iNeighbor(0:,:)
    integer,     intent(in)  :: nNeighbor(:)
    integer,     intent(in)  :: iCellVec(:)
    real(dp),    intent(in)  :: cellVec(:,:)
    integer,     intent(in)  :: iSquare(:)
    integer,     intent(in)  :: iPair(0:,:)
    integer,     intent(in)  :: img2CentCell(:)

    complex(dp) :: phase
    integer     :: nAtom
    integer     :: iOrig, ii, jj
    integer     :: iNeigh
    integer     :: iOldVec, iVec
    integer     :: iAtom1, iAtom2, iAtom2f
    integer     :: nOrb1, nOrb2
    real(dp)    :: kPoint2p(3)

    nAtom = size(iNeighbor, dim=2)

    ASSERT(nAtom > 0)
    ASSERT(size(square, dim=1) == size(square, dim=2))
    ASSERT(size(square, dim=1) == iSquare(nAtom+1) - 1)
    ASSERT(all(shape(kPoint) == (/ 3 /)))
    ASSERT(all(shape(nNeighbor) == (/ nAtom /)))
    ASSERT(size(iSquare) == nAtom + 1)

    square(:,:) = cmplx(0, 0, dp)
    kPoint2p(:) = 2.0_dp * pi * kPoint(:)
    iOldVec = 0
    phase = 1.0_dp
    do iAtom1 = 1, nAtom
      ii = iSquare(iAtom1)
      nOrb1 = iSquare(iAtom1+1) - ii
      do iNeigh = 0, nNeighbor(iAtom1)
        iOrig = iPair(iNeigh,iAtom1) + 1
        iAtom2 = iNeighbor(iNeigh, iAtom1)
        iAtom2f = img2CentCell(iAtom2)
        jj = iSquare(iAtom2f)
        ASSERT(jj >= ii)
        nOrb2 = iSquare(iAtom2f+1) - jj
        iVec = iCellVec(iAtom2)
        if (iVec /= iOldVec) then
          phase = exp((0.0_dp, 1.0_dp) &
              &* dot_product(kPoint2p, cellVec(:, iVec)))
          iOldVec = iVec
        end if
        square(jj:jj+nOrb2-1, ii:ii+nOrb1-1) = &
            &square(jj:jj+nOrb2-1, ii:ii+nOrb1-1) &
            &+ phase&
            &*reshape(orig(iOrig:iOrig+nOrb1*nOrb2-1), (/nOrb2, nOrb1/))
      end do
    end do

  end subroutine unpackHS_cmplx



  !!* Unpacks sparse matrix to square form (only real part of the phase factor
  !!* is considered)
  !!* @param square       Square form matrix on exit.
  !!* @param orig         Sparse matrix
  !!* @param kPoint       <b>Relative</b> coordinates of the K-point where the
  !!*   sparse  matrix    should be unfolded.
  !!* @param iNeighbor    Neighbor list for each atom (First index from 0!)
  !!* @param nNeighbor    Nr. of neighbors for each atom (incl. itself).
  !!* @param iCellVec     Index of the cell translation vector for each atom.
  !!* @param cellVec      Relative coordinates of the cell translation vectors.
  !!* @param iSquare      Atom offset for the squared Hamiltonian
  !!* @param iPair indexing array for the sparse Hamiltonian
  !!* @param img2CentCell Atomic mapping indexes.
  !!* @note The non on-site blocks are only filled in the lower triangle part
  !!*   of the matrix. To fill the matrix completely, apply the
  !!*   blockSymmetrizeHS subroutine.
  subroutine unpackHS_real_kpts(square, orig, kPoint, iNeighbor, nNeighbor, &
      &iCellVec, cellVec, iSquare, iPair, img2CentCell)
    real(dp), intent(out) :: square(:,:)
    real(dp), intent(in)  :: orig(:)
    real(dp), intent(in)  :: kPoint(:)
    integer,  intent(in)  :: iNeighbor(0:,:)
    integer,  intent(in)  :: nNeighbor(:)
    integer,  intent(in)  :: iCellVec(:)
    real(dp), intent(in)  :: cellVec(:,:)
    integer,  intent(in)  :: iSquare(:)
    integer,     intent(in)  :: iPair(0:,:)
    integer,  intent(in)  :: img2CentCell(:)

    real(dp) :: phase
    integer  :: nAtom
    integer  :: iOrig, ii, jj
    integer  :: iNeigh
    integer  :: iOldVec, iVec
    integer  :: iAtom1, iAtom2, iAtom2f
    integer  :: nOrb1, nOrb2
    real(dp) :: kPoint2p(3)


    nAtom = size(iNeighbor, dim=2)

    ASSERT(nAtom > 0)
    ASSERT(size(square, dim=1) == size(square, dim=2))
    ASSERT(size(square, dim=1) == iSquare(nAtom+1) -1)
    ASSERT(all(shape(kPoint) == (/ 3 /)))
    ASSERT(all(shape(nNeighbor) == (/ nAtom /)))
    ASSERT(size(iSquare) == nAtom + 1)

    square(:,:) = 0.0_dp
    kPoint2p(:) = 2.0_dp * pi * kPoint(:)
    iOldVec = 0
    phase = 1.0_dp
    do iAtom1 = 1, nAtom
      ii = iSquare(iAtom1)
      nOrb1 = iSquare(iAtom1+1) - ii
      do iNeigh = 0, nNeighbor(iAtom1)
        iOrig = iPair(iNeigh,iAtom1) + 1
        iAtom2 = iNeighbor(iNeigh, iAtom1)
        iAtom2f = img2CentCell(iAtom2)
        jj = iSquare(iAtom2f)
        ASSERT(jj >= ii)
        nOrb2 = iSquare(iAtom2f+1) - jj
        iVec = iCellVec(iAtom2)
        if (iVec /= iOldVec) then
          phase = cos(dot_product(kPoint2p, cellVec(:, iVec)))
          iOldVec = iVec
        end if
        square(jj:jj+nOrb2-1,ii:ii+nOrb1-1) = &
            &square(jj:jj+nOrb2-1,ii:ii+nOrb1-1) &
            &+ phase &
            &* reshape(orig(iOrig:iOrig+nOrb1*nOrb2-1),(/nOrb2,nOrb1/))
      end do
    end do
  end subroutine unpackHS_real_kpts



  !!* Unpacks sparse matrix to square form (real version for Gamma point)
  !!* @param square       Square form matrix on exit.
  !!* @param orig         Sparse matrix
  !!* @param iNeighbor    Neighbor list for each atom (First index from 0!)
  !!* @param nNeighbor    Nr. of neighbors for each atom (incl. itself).
  !!* @param iSquare      Atom offset for the squared Hamiltonian
  !!* @param iPair indexing array for the sparse Hamiltonian
  !!* @param img2CentCell Atomic mapping indexes.
  !!* @note The non on-site blocks are only filled in the lower triangle part
  !!*   of the matrix. To fill the matrix completely, apply the
  !!*   blockSymmetrizeHS subroutine.
  subroutine unpackHS_real(square, orig, iNeighbor, nNeighbor, iSquare, &
      &iPair, img2CentCell)
    real(dp), intent(out) :: square(:,:)
    real(dp), intent(in)  :: orig(:)
    integer,  intent(in)  :: iNeighbor(0:,:)
    integer,  intent(in)  :: nNeighbor(:)
    integer,  intent(in)  :: iSquare(:)
    integer,     intent(in)  :: iPair(0:,:)
    integer,  intent(in)  :: img2CentCell(:)

    integer     :: nAtom
    integer     :: iOrig, ii, jj
    integer     :: iNeigh
    integer     :: iAtom1, iAtom2, iAtom2f
    integer     :: nOrb1, nOrb2

    nAtom = size(iNeighbor, dim=2)

    ASSERT(nAtom > 0)
    ASSERT(size(square, dim=1) == size(square, dim=2))
    ASSERT(size(square, dim=1) == iSquare(nAtom+1) - 1)
    ASSERT(all(shape(nNeighbor) == (/ nAtom /)))
    ASSERT(size(iSquare) == nAtom + 1)

    square(:,:) = 0.0_dp

    do iAtom1 = 1, nAtom
      ii = iSquare(iAtom1)
      nOrb1 = iSquare(iAtom1+1) - ii
      do iNeigh = 0, nNeighbor(iAtom1)
        iOrig = iPair(iNeigh,iAtom1) + 1
        iAtom2 = iNeighbor(iNeigh, iAtom1)
        iAtom2f = img2CentCell(iAtom2)
        jj = iSquare(iAtom2f)
        ASSERT(jj >= ii)
        nOrb2 = iSquare(iAtom2f+1) - jj
        square(jj:jj+nOrb2-1, ii:ii+nOrb1-1) = &
            &square(jj:jj+nOrb2-1, ii:ii+nOrb1-1) &
            &+ reshape(orig(iOrig:iOrig+nOrb1*nOrb2-1), (/nOrb2,nOrb1/))
      end do
    end do

  end subroutine unpackHS_real


  
  !!!* Unpacks sparse matrix to a banded form (real version for Gamma point)
  !!!* @param band         Lower triangle of banded form of matrix on exit.
  !!!* @param orig         Sparse matrix
  !!!* @param iNeighbor    Neighbor list for each atom (First index from 0!)
  !!!* @param nNeighbor    Nr. of neighbors for each atom (incl. itself).
  !!!* @param mmAng        Maximum total angular momentum
  !!!* @param iSquare      Atom offset for the squared Hamiltonian
  !!!* @param iPair indexing array for the sparse Hamiltonian
  !!!* @param img2CentCell Atomic mapping indexes.
  !!!* @param bandwidth    The bandwidth of the matrix band(bandwidth,:)
  !!!* @note Only the lower triangle of the matrix is present, unlike the
  !!!* square matrices, where the other triangle is present but 0
  !subroutine unpackHS_real_banded(band, orig, iNeighbor, nNeighbor, mmAng, &
  !    &iSquare, iPair, img2CentCell, bandwidth)
  !  real(dp), intent(out) :: band(:,:)
  !  real(dp), intent(in)  :: orig(:)
  !  integer,  intent(in)  :: iNeighbor(0:,:)
  !  integer,  intent(in)  :: nNeighbor(:)
  !  integer,  intent(in)  :: mmAng
  !  integer,  intent(in)  :: iSquare(:)
  !  integer,  intent(in)  :: iPair(0:,:)
  !  integer,  intent(in)  :: img2CentCell(:)
  !  integer,  intent(in)  :: bandwidth
  !
  !  integer     :: nAtom
  !  integer     :: iOrig, ii, jj, kk, ll
  !  integer     :: iNeigh, iAng1, iAng2
  !  integer     :: iAtom1, iAtom2, iAtom2f
  !  integer     :: nOrb1, nOrb2
  !  real(dp)    :: tmpSqr((mmAng+1)**2,(mmAng+1)**2)
  !
  !  nAtom = size(iNeighbor, dim=2)
  !
  !  ASSERT(nAtom > 0)
  !  ASSERT(size(band, dim=2) == iSquare(nAtom+1) - 1)
  !  ASSERT(all(shape(nNeighbor) == (/ nAtom /)))
  !  ASSERT(size(iSquare) == nAtom + 1)
  !  ASSERT(size(band,dim=1)==bandwidth)
  !
  !  band(:,:) = 0.0_dp
  !
  !  do iAtom1 = 1, nAtom
  !    ii = iSquare(iAtom1)
  !    nOrb1 = iSquare(iAtom1+1) - ii
  !    do iNeigh = 0, nNeighbor(iAtom1)
  !      iOrig = iPair(iNeigh,iAtom1) + 1
  !      iAtom2 = iNeighbor(iNeigh, iAtom1)
  !      iAtom2f = img2CentCell(iAtom2)
  !      jj = iSquare(iAtom2f)
  !      ASSERT(jj >= ii)
  !      nOrb2 = iSquare(iAtom2f+1) - jj
  !      if (iAtom2f == iAtom1) then ! on-diagonal cases          
  !        ! need to use a tempory square block :
  !        tmpSqr(1:nOrb2,1:nOrb1) = &
  !            & reshape(orig(iOrig:iOrig+nOrb1*nOrb2-1), (/nOrb2,nOrb1/))
  !        do kk = 1, nOrb1
  !          do ll = kk, nOrb2
  !            band(ll-kk+1,ii+kk-1) = band(ll-kk+1,ii+kk-1) &
  !                & + tmpSqr(ll,kk)
  !          end do
  !        end do
  !      else        
  !        do iAng1 = 1, nOrb1
  !          do iAng2 = 1, nOrb2
  !            !! if UPLO = 'L', AB(1+i-j,j) = A(i,j) for j<=i<=min(n,j+ka).
  !            band(1+jj+iAng2-1-(ii+iAng1-1), ii+iAng1-1) = &
  !                &band(1+jj+iAng2-1-(ii+iAng1-1), ii+iAng1-1) &
  !                & + orig(iOrig+iAng2-1)
  !          end do
  !        end do
  !      end if
  !    end do
  !  end do
  !
  !end subroutine unpackHS_real_banded

  

  !!* Unpacks 2-component sparse matrix to square form 
  !!* @param square       Square form matrix on exit, 2*2 block structure.
  !!* @param orig         Sparse Pauli matrix, final index of 4 blocks
  !!* @param iNeighbor    Neighbor list for each atom (First index from 0!)
  !!* @param nNeighbor    Nr. of neighbors for each atom (incl. itself).
  !!* @param iSquare      Atom offset for the squared Hamiltonian
  !!* @param iPair indexing array for the sparse Hamiltonian
  !!* @param img2CentCell Atomic mapping indexes.
  !!* @note The non on-site blocks are only filled in the lower triangle part
  !!*   of the matrix. To fill the matrix completely, apply the
  !!*   blockSymmetrizeHS subroutine.
  subroutine unpack2compHS(square, orig, iNeighbor, nNeighbor, iSquare, iPair, &
      &img2CentCell)
    complex(dp), intent(out) :: square(:,:)
    complex(dp), intent(in)  :: orig(:,:)
    integer,     intent(in)  :: iNeighbor(0:,:)
    integer,     intent(in)  :: nNeighbor(:)
    integer,     intent(in)  :: iSquare(:)
    integer,     intent(in)  :: iPair(0:,:)
    integer,     intent(in)  :: img2CentCell(:)
    
    integer     :: nAtom
    integer     :: iOrig, ii, jj
    integer     :: iNeigh
    integer     :: iAtom1, iAtom2, iAtom2f
    integer     :: nOrb1, nOrb2
    integer     :: mOffset, iAlpha, iBeta, iOffA, iOffB, iIndex
    
    nAtom = size(iNeighbor, dim=2)
    
    ASSERT(nAtom > 0)
    ASSERT(size(square, dim=1) == size(square, dim=2))
    ! of Pauli block size?:
    ASSERT(size(square, dim=1) == 2*(iSquare(nAtom+1) - 1))
    ASSERT(size(orig,dim=2) == 4)
    ASSERT(all(shape(nNeighbor) == (/ nAtom /)))
    ASSERT(size(iSquare) == nAtom + 1)
    
    mOffset = size(square, dim=1)/2    
    square(:,:) = cmplx(0.0_dp, 0.0_dp, dp)
    
    do iBeta = 1,2
      iOffB = (iBeta-1)*mOffset      
      do iAlpha = 1,2
        iOffA = (iAlpha-1)*mOffset
        iIndex = spinIndex(iAlpha,iBeta)
        do iAtom1 = 1, nAtom
          ii = iSquare(iAtom1)
          nOrb1 = iSquare(iAtom1+1) - 1
          if (iAlpha == 1 .and. iBeta == 2) then
            do iNeigh = 1, nNeighbor(iAtom1)
              iOrig = iPair(iNeigh,iAtom1) + 1
              iAtom2 = iNeighbor(iNeigh, iAtom1)
              iAtom2f = img2CentCell(iAtom2)
              jj = iSquare(iAtom2f)
              ASSERT(jj >= ii)
              nOrb2 = iSquare(iAtom2f+1) - 1
              square(ii+iOffB:ii+nOrb1-1+iOffB,jj+iOffA:jj+nOrb2-1+iOffA)= &
                  & square(ii+iOffB:ii+nOrb1-1+iOffB, &
                  & jj+iOffA:jj+nOrb2-1+iOffA) &
                  & + conjg(transpose( &
                  & reshape(orig(iOrig:iOrig+nOrb1*nOrb2-1,iIndex) &
                  & ,(/nOrb2,nOrb1/))))
            end do
          else
            do iNeigh = 0, nNeighbor(iAtom1)
              iOrig = iPair(iNeigh,iAtom1) + 1
              iAtom2 = iNeighbor(iNeigh, iAtom1)
              iAtom2f = img2CentCell(iAtom2)
              jj = iSquare(iAtom2f)
              ASSERT(jj >= ii)
              nOrb2 = iSquare(iAtom2f+1) - jj
              square(jj+iOffA:jj+nOrb2-1+iOffA,ii+iOffB:ii+nOrb1-1+iOffB)= &
                  & square(jj+iOffA:jj+nOrb2-1+iOffA, &
                  & ii+iOffB:ii+nOrb1-1+iOffB) &
                  & +reshape(orig(iOrig:iOrig+nOrb1*nOrb2-1,iIndex),&
                  & (/nOrb2,nOrb1/))
            end do
          end if
        end do
      end do
    end do
    
  end subroutine unpack2compHS


  
  !!* Unpacks 2-component sparse matrix to square form with k-points
  !!* @param square       Square form matrix on exit, 2*2 block structure.
  !!* @param orig         Sparse Pauli matrix, final index of 4 blocks
  !!* @param kPoint       <b>Relative</b> coordinates of the K-point where the
  !!*   sparse  matrix    should be unfolded.
  !!* @param iNeighbor    Neighbor list for each atom (First index from 0!)
  !!* @param nNeighbor    Nr. of neighbors for each atom (incl. itself).
  !!* @param iCellVec     Index of the cell translation vector for each atom.
  !!* @param cellVec      Relative coordinates of the cell translation vectors.
  !!* @param iSquare      Atom offset for the squared Hamiltonian
  !!* @param iPair indexing array for the sparse Hamiltonian
  !!* @param img2CentCell Atomic mapping indexes.
  !!* @note The non on-site blocks are only filled in the lower triangle part
  !!*   of the matrix. To fill the matrix completely, apply the
  !!*   blockSymmetrizeHS subroutine.
  subroutine unpack2compHS_kpts(square, orig, kPoint, iNeighbor, nNeighbor, &
      &iCellVec, cellVec, iSquare, iPair, img2CentCell)
    complex(dp), intent(out) :: square(:,:)
    complex(dp), intent(in)  :: orig(:,:)
    real(dp),    intent(in)  :: kPoint(:)
    integer,     intent(in)  :: iNeighbor(0:,:)
    integer,     intent(in)  :: nNeighbor(:)
    integer,     intent(in)  :: iCellVec(:)
    real(dp),    intent(in)  :: cellVec(:,:)
    integer,     intent(in)  :: iSquare(:)
    integer,     intent(in)  :: iPair(0:,:)
    integer,     intent(in)  :: img2CentCell(:)
    
    complex(dp) :: phase
    integer     :: nAtom
    integer     :: iOrig, ii, jj
    integer     :: iNeigh
    integer     :: iOldVec, iVec
    integer     :: iAtom1, iAtom2, iAtom2f
    integer     :: nOrb1, nOrb2
    real(dp)    :: kPoint2p(3)
    integer     :: mOffset, iAlpha, iBeta, iOffA, iOffB, iIndex
    
    nAtom = size(iNeighbor, dim=2)
    
    ASSERT(nAtom > 0)
    ASSERT(size(square, dim=1) == size(square, dim=2))
    ! of Pauli block size?:
    ASSERT(size(square, dim=1) == 2*(iSquare(nAtom+1) - 1))
    ASSERT(size(orig,dim=2) == 4)
    ASSERT(all(shape(kPoint) == (/ 3 /)))
    ASSERT(all(shape(nNeighbor) == (/ nAtom /)))
    ASSERT(size(iSquare) == nAtom + 1)
    
    mOffset = size(square, dim=1)/2    
    square(:,:) = cmplx(0.0_dp, 0.0_dp, dp)
    kPoint2p(:) = 2.0_dp * pi * kPoint(:)
    
    do iBeta = 1,2
      iOffB = (iBeta-1)*mOffset      
      do iAlpha = 1,2
        iOffA = (iAlpha-1)*mOffset
        iIndex = spinIndex(iAlpha,iBeta)
        iOldVec = 0
        phase = 1.0_dp
        do iAtom1 = 1, nAtom
          ii = iSquare(iAtom1)
          nOrb1 = iSquare(iAtom1+1) - ii
          if (iAlpha == 1 .and. iBeta == 2) then
            do iNeigh = 1, nNeighbor(iAtom1)
              iOrig = iPair(iNeigh,iAtom1) + 1
              iAtom2 = iNeighbor(iNeigh, iAtom1)
              iAtom2f = img2CentCell(iAtom2)
              jj = iSquare(iAtom2f)
              ASSERT(jj >= ii)
              nOrb2 = iSquare(iAtom2f+1) - jj
              iVec = iCellVec(iAtom2)
              if (iVec /= iOldVec) then
                phase = exp((0.0_dp, 1.0_dp) &
                    &* dot_product(kPoint2p, cellVec(:, iVec)))
                iOldVec = iVec
              end if
              square(ii+iOffB:ii+nOrb1-1+iOffB,jj+iOffA:jj+nOrb2-1+iOffA)= &
                  & square(ii+iOffB:ii+nOrb1-1+iOffB, &
                  & jj+iOffA:jj+nOrb2-1+iOffA) &
                  & + conjg(phase*transpose( &
                  & reshape(orig(iOrig:iOrig+nOrb1*nOrb2-1,iIndex) &
                  & ,(/nOrb2,nOrb1/))))
            end do
          else
            do iNeigh = 0, nNeighbor(iAtom1)
              iOrig = iPair(iNeigh,iAtom1) + 1
              iAtom2 = iNeighbor(iNeigh, iAtom1)
              iAtom2f = img2CentCell(iAtom2)
              jj = iSquare(iAtom2f)
              ASSERT(jj >= ii)
              nOrb2 = iSquare(iAtom2f+1) - jj
              iVec = iCellVec(iAtom2)
              if (iVec /= iOldVec) then
                phase = exp((0.0_dp, 1.0_dp) &
                    &* dot_product(kPoint2p, cellVec(:, iVec)))
                iOldVec = iVec
              end if
              square(jj+iOffA:jj+nOrb2-1+iOffA,ii+iOffB:ii+nOrb1-1+iOffB)= &
                  & square(jj+iOffA:jj+nOrb2-1+iOffA, &
                  & ii+iOffB:ii+nOrb1-1+iOffB) &
                  & +phase*reshape(orig(iOrig:iOrig+nOrb1*nOrb2-1,iIndex),&
                  & (/nOrb2,nOrb1/))
            end do
          end if
        end do
      end do
    end do

  end subroutine unpack2compHS_kpts


  
  !!* Pack squared matrix in the sparse form (complex version).
  !!* @param primitive    Sparse matrix
  !!* @param square       Squared form matrix
  !!* @param kPoint       Relative coordinates of the K-point
  !!* @param kWeight      Weight of the K-point
  !!* @param iNeighbor    Neighbor list for the atoms (First index from 0!)
  !!* @param nNeighbor    Nr. of neighbors for the atoms.
  !!* @param mOrb         Maximal number of orbitals on an atom.
  !!* @param iCellVec     Index of the cell translation vector for each atom.
  !!* @param cellVec      Relative coordinates of the cell translation vectors.
  !!* @param iSquare      Atom offset for the squared matrix
  !!* @param iPair indexing array for the sparse Hamiltonian
  !!* @param img2CentCell Mapping between image atoms and correspondent atom in
  !!*   the central cell.
  subroutine packHS_cmplx(primitive, square, kPoint, kWeight, iNeighbor, &
      & nNeighbor, mOrb, iCellVec, cellVec, iSquare, iPair, img2CentCell)
    real(dp),    intent(inout) :: primitive(:)
    complex(dp), intent(in)    :: square(:,:)
    real(dp),    intent(in)    :: kPoint(:)
    real(dp),    intent(in)    :: kweight
    integer,     intent(in)    :: iNeighbor(0:,:)
    integer,     intent(in)    :: nNeighbor(:)
    integer,     intent(in)    :: mOrb
    integer,     intent(in)    :: iCellVec(:)
    real(dp),    intent(in)    :: cellVec(:,:)
    integer,     intent(in)    :: iSquare(:)
    integer,     intent(in)    :: iPair(0:,:)
    integer,     intent(in)    :: img2CentCell(:)

    complex(dp) :: phase
    integer     :: nAtom
    integer     :: iOrig, ii, jj, kk
    integer     :: iNeigh
    integer     :: iOldVec, iVec
    integer     :: iAtom1, iAtom2, iAtom2f
    integer     :: nOrb1, nOrb2
    real(dp)    :: kPoint2p(3)
    complex(dp) :: tmpSqr(mOrb,mOrb)
    ASSERT_ENV(integer :: sizePrim)

    nAtom = size(iNeighbor, dim=2)
    ASSERT_ENV(sizePrim = size(primitive))

    ASSERT(nAtom > 0)
    ASSERT(size(square, dim=1) == size(square, dim=2))
    ASSERT(size(square, dim=1) == iSquare(nAtom+1) - 1)
    ASSERT(all(shape(kPoint) == (/ 3 /)))
    ASSERT(all(shape(nNeighbor) == (/ nAtom /)))
    ASSERT(kWeight > 0.0_dp)
    ASSERT(size(iSquare) == nAtom + 1)

    kPoint2p(:) = 2.0_dp * pi * kPoint(:)
    iOldVec = 0
    phase = 1.0_dp
    do iAtom1 = 1, nAtom
      ii = iSquare(iAtom1)
      nOrb1 = iSquare(iAtom1+1) - ii
      do iNeigh = 0, nNeighbor(iAtom1)
        iOrig = iPair(iNeigh,iAtom1) + 1
        iAtom2 = iNeighbor(iNeigh, iAtom1)
        iAtom2f = img2CentCell(iAtom2)
        jj = iSquare(iAtom2f)
        ASSERT(jj >= ii)
        nOrb2 = iSquare(iAtom2f+1) - jj
        iVec = iCellVec(iAtom2)
        if (iVec /= iOldVec) then
          phase = exp(cmplx(0,-1,dp) &
              &* dot_product(kPoint2p(:), cellVec(:, iVec)))
          iOldVec = iVec
        end if
        tmpSqr(1:nOrb2, 1:nOrb1) = square(jj:jj+nOrb2-1, ii:ii+nOrb1-1)

        !! Symmetrize the on-site block before packing, just in case
        if (iAtom1 == iAtom2f) then
          do kk = 1, nOrb2
            tmpSqr(kk, kk+1:nOrb1) = conjg(tmpSqr(kk+1:nOrb1, kk))
          end do
        end if

        ASSERT(sizePrim >= iOrig + nOrb1*nOrb2 - 1)
        primitive(iOrig : iOrig + nOrb1*nOrb2 - 1) = &
            & primitive(iOrig : iOrig + nOrb1*nOrb2 - 1) &
            &+ kWeight &
            &* real(phase &
            &* reshape(tmpSqr(1:nOrb2, 1:nOrb1), (/nOrb1*nOrb2/)), dp)
      end do
    end do
  end subroutine packHS_cmplx



  !!* Pack squared matrix in the sparse form (real version).
  !!* @param primitive    Sparse matrix
  !!* @param square       Squared form matrix
  !!* @param iNeighbor    Neighbor list for the atoms (First index from 0!)
  !!* @param nNeighbor    Nr. of neighbors for the atoms.
  !!* @param mOrb         Maximal number of orbitals on an atom.
  !!* @param iCellVec     Index of the cell translation vector for each atom.
  !!* @param cellVec      Relative coordinates of the cell translation vectors.
  !!* @param iSquare      Atom offset for the squared matrix
  !!* @param iPair indexing array for the sparse Hamiltonian
  !!* @param img2CentCell Mapping between image atoms and correspondent atom in
  !!*   the central cell.
  subroutine packHS_real(primitive, square, iNeighbor, nNeighbor, mOrb, &
      &iSquare, iPair, img2CentCell)
    real(dp), intent(inout) :: primitive(:)
    real(dp), intent(in)    :: square(:,:)
    integer,  intent(in)    :: iNeighbor(0:,:)
    integer,  intent(in)    :: nNeighbor(:)
    integer,  intent(in)    :: mOrb
    integer,  intent(in)    :: iSquare(:)
    integer,  intent(in)    :: iPair(0:,:)
    integer,  intent(in)    :: img2CentCell(:)

    integer     :: nAtom
    integer     :: iOrig, ii, jj, kk
    integer     :: iNeigh
    integer     :: iAtom1, iAtom2, iAtom2f
    integer     :: nOrb1, nOrb2
    real(dp)    :: tmpSqr(mOrb,mOrb)
    ASSERT_ENV(integer :: sizePrim)

    nAtom = size(iNeighbor, dim=2)
    ASSERT_ENV(sizePrim = size(primitive))

    ASSERT(nAtom > 0)
    ASSERT(size(square, dim=1) == size(square, dim=2))
    ASSERT(size(square, dim=1) == iSquare(nAtom+1) - 1)
    ASSERT(all(shape(nNeighbor) == (/ nAtom /)))

    do iAtom1 = 1, nAtom
      ii = iSquare(iAtom1)
      nOrb1 = iSquare(iAtom1+1) - ii
      do iNeigh = 0, nNeighbor(iAtom1)
        iOrig = iPair(iNeigh,iAtom1) + 1
        iAtom2 = iNeighbor(iNeigh, iAtom1)
        iAtom2f = img2CentCell(iAtom2)
        jj = iSquare(iAtom2f)
        ASSERT(jj >= ii)
        nOrb2 = iSquare(iAtom2f+1) - jj
        tmpSqr(1:nOrb2, 1:nOrb1) = square(jj:jj+nOrb2-1, ii:ii+nOrb1-1)

        ! Symmetrize the on-site block before packing, just in case
        if (iAtom1 == iAtom2f) then
          do kk = 1, nOrb2
            tmpSqr(kk, kk+1:nOrb1) = tmpSqr(kk+1:nOrb1, kk)
          end do
        end if

        ASSERT(sizePrim >= iOrig + nOrb1*nOrb2 - 1)
        primitive(iOrig : iOrig + nOrb1*nOrb2 - 1) = &
            &primitive(iOrig : iOrig + nOrb1*nOrb2 - 1) &
            &+ reshape(tmpSqr(1:nOrb2,1:nOrb1), (/nOrb1*nOrb2/))
      end do
    end do
  end subroutine packHS_real



  !!* Symmetrize a squared matrix leaving the on-site atomic blocks alone.
  !!* (Complex version)
  !!* @param square   Square form matrix.
  !!* @param iSquare  Returns the offset array for each atom.
  subroutine blockSymmetrizeHS_cmplx(square, iSquare)
    complex(dp), intent(inout) :: square(:,:)
    integer, intent(in) :: iSquare(:)

    integer     :: nAtom
    integer     :: iAtom1, iOrb1
    integer     :: ind
    integer     :: nOrb1

    nAtom = size(iSquare, dim=1) - 1

    ASSERT(nAtom > 0)
    ASSERT(size(square, dim=1) == size(square, dim=2))
    ASSERT(size(square, dim=1) == iSquare(nAtom+1) - 1)

    !! Fill upper triangle part (except on-site blocks)
    do iAtom1 = 1, nAtom
      ind = iSquare(iAtom1)
      nOrb1 = iSquare(iAtom1+1) - ind
      do iOrb1 = 0, nOrb1 - 1
        square(ind+iOrb1, ind+nOrb1:) = square(ind+nOrb1:, ind+iOrb1)
      end do
    end do

  end subroutine blockSymmetrizeHS_cmplx



  !!* Symmetrize a squared matrix leaving the on-site atomic blocks alone.
  !!* (Real version)
  !!* @param square   Square form matrix.
  !!* @param iSquare  Returns the offset array for each atom.
  subroutine blockSymmetrizeHS_real(square, iSquare)
    real(dp), intent(inout) :: square(:,:)
    integer,  intent(in)    :: iSquare(:)

    integer     :: nAtom
    integer     :: iAtom1, iOrb1
    integer     :: ind
    integer     :: nOrb1

    nAtom = size(iSquare) - 1

    ASSERT(nAtom > 0)
    ASSERT(size(square, dim=1) == size(square, dim=2))
    ASSERT(size(square, dim=1) == iSquare(nAtom+1) - 1)

    !! Fill upper triangle part (except on-site blocks)
    do iAtom1 = 1, nAtom
      ind = iSquare(iAtom1)
      nOrb1 = iSquare(iAtom1+1) - ind
      do iOrb1 = 0, nOrb1 - 1
        square(ind+iOrb1, ind+nOrb1:) = square(ind+nOrb1:, ind+iOrb1)
      end do
    end do

  end subroutine blockSymmetrizeHS_real


  
  !!* Creates a K-points sampling, equivalent to folding of a reciprocal point
  !!*   of a super lattice.
  !!* @param coeffs Coefficients of the lattice vectors in the linear 
  !!*   combination for the super lattice vectors (should be integer values)
  !!* @param shifts Shift of the grid along the three small reciprocal lattice
  !!*   vectors (between 0.0 and 1.0)
  !!* @param latVecs Lattice vector of the original grid
  !!* @param recVecs2p Reciprocal lattice vectors in 2p units (inverse(latVecs))
  !!* @param kPoints Contains the kPoints on exit.
  !!* @param kWeights Contains the weights of the kPoints on exit.
  !!* @param reduceByInversion If points equivalent by inversion should be
  !!*   reduced.
  !!* @desc The routine calculates those reciprocal lattice points of the
  !!*   super lattice, which are inside the Brillouin zone of the original
  !!*   lattice. The resulting points are then all shifted by sum(shift(i)*B(i))
  !!*   where B(i) are the reciprocal lattice vectors of the super lattice.
  !!*   Finally, points equivalent by inversion are reduced, unless specified
  !!*   otherwise.
  subroutine getSuperSampling(coeffs, shifts, latVecs, recVecs2p, kPoints, &
      &kWeights, reduceByInversion)
    real(dp), intent(in) :: coeffs(:,:)
    real(dp), intent(in) :: shifts(:)
    real(dp), intent(in) :: latVecs(:,:)
    real(dp), intent(in) :: recVecs2p(:,:)
    real(dp), pointer :: kPoints(:,:)
    real(dp), pointer :: kWeights(:)
    logical, intent(in), optional :: reduceByInversion

    real(dp), allocatable :: allKPoints(:,:), allKWeights(:)
    logical, allocatable :: irreducible(:)
    logical :: tReduce
    real(dp) :: superLatVecs(3,3), superRecVecs(3,3)
    real(dp) :: invCoeffs(3,3), rr(3)
    real(dp) :: mDist2, dist2
    integer :: imgRange(2,3)
    integer :: nAllKPoint, nKPoint
    integer :: i1, i2, i3
    type(listRealR1) :: lr1

    real(dp), parameter :: tol = 1e-4_dp
    real(dp), parameter :: minLim = -tol, maxLim = 1.0_dp - tol

    ASSERT(all(shape(coeffs) == (/ 3, 3 /)))
    ASSERT(size(shifts) == 3)
    ASSERT(all(shape(latVecs) == (/ 3, 3 /)))
    ASSERT(all(shape(recVecs2p) == (/ 3, 3 /)))

    if (present(reduceByInversion)) then
      tReduce = reduceByInversion
    else
      tReduce = .true.
    end if

    superLatVecs(:,:) = matmul(latVecs, coeffs)
    superRecVecs(:,:) = superLatVecs(:,:)
    call matinv(superRecVecs)
    superRecVecs(:,:) = reshape(superRecVecs, (/3, 3/), order=(/2, 1/))  ! B
    invCoeffs(:,:) = coeffs(:,:)
    call matinv(invCoeffs)
    invCoeffs = reshape(invCoeffs, (/3, 3 /), order=(/2,1/)) !n^-1^t
    mDist2 = 0.0_dp
    do i1 = 0, 1
      do i2 = 0, 1
        do i3 = 0, 1
          dist2 = sum(matmul(recVecs2p, real((/ i1, i2, i3 /), dp))**2)
          if (dist2 > mDist2) then
            mDist2 = dist2
          end if
        end do
      end do
    end do
    !! To make sure, the shifts do not cause problems, we take one shell
    !! more then really necessary
    call getImgRange(imgRange, sqrt(mDist2), superLatVecs, 1, 1)
    call init(lr1)
    do i1 = imgRange(1, 1), imgRange(2, 1)
      do i2 = imgRange(1, 2), imgRange(2, 2)
        do i3 = imgRange(1, 3), imgRange(2, 3)
          rr(:) = matmul(invCoeffs, real((/ i1, i2, i3 /), dp) + shifts)
          if (all(rr >= minLim) .and. all(rr < maxLim)) then
            call append(lr1, rr)
          end if
        end do
      end do
    end do
    nAllKPoint = len(lr1)
    if (abs(real(nAllKPoint,dp) - determinant33(coeffs)) > tol) then
      call error("Monkhorst-Pack routine failed to find all K-points.")
    end if
    ALLOCATE_(allKPoints, (3, nAllKPoint))
    ALLOCATE_(allKWeights, (nAllKPoint))
    call asArray(lr1, allKPoints)
    allKWeights = 1.0_dp / real(nAllKPoint, dp)
    call destroy(lr1)
    if (tReduce) then
      ALLOCATE_(irreducible, (nAllKPoint))
      irreducible(:) = .true.
      do i1 = 1, nAllKPoint
        if (.not. irreducible(i1)) then
          cycle
        end if
        rr(:) = modulo(-1.0_dp * allKPoints(:,i1), 1.0_dp)
        do i2 = i1 + 1, nAllKPoint
          if (.not. irreducible(i2)) then
            cycle
          end if
          if (all(abs(allKPoints(:,i2) - rr(:)) < tol)) then
            irreducible(i2) = .false.
            allKWeights(i1) = allKWeights(i1) + allKWeights(i2)
          end if
        end do
      end do
      nKPoint = count(irreducible)
      INITALLOCATE_PARR(kPoints, (3, nKPoint))
      INITALLOCATE_PARR(kWeights, (nKPoint))
      i1 = 1
      i2 = 1
      do while (i2 <= nKpoint)
        if (irreducible(i1)) then
          kPoints(:,i2) = allKPoints(:,i1)
          kWeights(i2) = allKWeights(i1)
          i2 = i2 + 1
        end if
        i1 = i1 + 1
      end do
      DEALLOCATE_(irreducible)
    else
      INITALLOCATE_PARR(kPoints, (3, nAllKPoint))
      INITALLOCATE_PARR(kWeights, (nAllKPoint))
      kPoints(:,:) = allKPoints
      kWeights(:) = allKWeights
    end if
    DEALLOCATE_(allKPoints)
    DEALLOCATE_(allkWeights)
    
  end subroutine getSuperSampling


  !!* Prints sparse matrix to file
  !!* @param fdOverlap    File ID to print to
  !!* @param orig         Sparse matrix
  !!* @param kPoint       <b>Relative</b> coordinates of the K-point where the
  !!*   sparse  matrix    should be unfolded.
  !!* @param iNeighbor    Neighbor list for each atom (First index from 0!)
  !!* @param nNeighbor    Nr. of neighbors for each atom (incl. itself).
  !!* @param iCellVec     Index of the cell translation vector for each atom.
  !!* @param cellVec      Relative coordinates of the cell translation vectors.
  !!* @param iSquare      Atom offset for the squared Hamiltonian
  !!* @param iPair indexing array for the sparse Hamiltonian
  !!* @param img2CentCell Atomic mapping indexes.
  !!* @param orb Information about the orbitals in the system
  subroutine sparse_cmplxprint(fdOverlap, orig, kPoint, iNeighbor, &
      & nNeighbor, iCellVec, cellVec, iSquare, iPair, img2CentCell, orb)
    integer,      intent(in) :: fdOverlap
    real(dp),    intent(in)  :: orig(:)
    real(dp),    intent(in)  :: kPoint(:)
    integer,     intent(in)  :: iNeighbor(0:,:)
    integer,     intent(in)  :: nNeighbor(:)
    integer,     intent(in)  :: iCellVec(:)
    real(dp),    intent(in)  :: cellVec(:,:)
    integer,     intent(in)  :: iSquare(:)
    integer,     intent(in)  :: iPair(0:,:)
    integer,     intent(in)  :: img2CentCell(:)
    type(TOrbitals), intent(in) :: orb
    
    complex(dp) :: phase
    integer     :: nAtom
    integer     :: iOrig, ii, jj, kk, ll
    integer     :: iNeigh
    integer     :: iOldVec, iVec
    integer     :: iAtom1, iAtom2, iAtom2f
    integer     :: nOrb1, nOrb2
    real(dp)    :: kPoint2p(3)
    complex(dp) :: tmp(orb%mOrb,orb%mOrb)
    
    nAtom = size(iNeighbor, dim=2)

    ASSERT(nAtom > 0)
    ASSERT(all(shape(kPoint) == (/ 3 /)))
    ASSERT(all(shape(nNeighbor) == (/ nAtom /)))
    ASSERT(size(iSquare) == nAtom + 1)

    kPoint2p(:) = 2.0_dp * pi * kPoint(:)
    iOldVec = 0
    phase = 1.0_dp
    do iAtom1 = 1, nAtom
      ii = iSquare(iAtom1)
      nOrb1 = iSquare(iAtom1+1) - ii
      do iNeigh = 0, nNeighbor(iAtom1)
        iOrig = iPair(iNeigh,iAtom1) + 1
        iAtom2 = iNeighbor(iNeigh, iAtom1)
        iAtom2f = img2CentCell(iAtom2)
        jj = iSquare(iAtom2f)
        ASSERT(jj >= ii)
        nOrb2 = iSquare(iAtom2f+1) - jj
        iVec = iCellVec(iAtom2)
        if (iVec /= iOldVec) then
          phase = exp((0.0_dp, 1.0_dp) &
              &* dot_product(kPoint2p, cellVec(:, iVec)))
          iOldVec = iVec
        end if
        tmp(1:nOrb2,1:nOrb1) = phase * &
            & reshape(orig(iOrig:iOrig+nOrb1*nOrb2-1), (/nOrb2, nOrb1/))
        if (iNeigh == 0) then
          do ll = 1, nOrb1
            do kk = ll, nOrb2
              write(fdOverlap,*)kk+jj-1,ll+ii-1, tmp(kk,ll)
            end do
          end do
        else
          do ll = 1, nOrb1
            do kk = 1, nOrb2
              write(fdOverlap,*)kk+jj-1,ll+ii-1, tmp(kk,ll)
            end do
          end do
        end if
      end do
    end do

  end subroutine sparse_cmplxprint



  !!* Prints sparse matrix to file
  !!* @param fdOverlap    File ID to print to
  !!* @param orig         Sparse matrix
  !!* @param iNeighbor    Neighbor list for each atom (First index from 0!)
  !!* @param nNeighbor    Nr. of neighbors for each atom (incl. itself).
  !!* @param iSquare      Atom offset for the squared Hamiltonian
  !!* @param iPair indexing array for the sparse Hamiltonian
  !!* @param img2CentCell Atomic mapping indexes.
  !!* @param orb Information about the orbitals in the system
  subroutine sparse_realprint(fdOverlap, orig, iNeighbor, nNeighbor, &
      & iSquare, iPair, img2CentCell,orb)
    integer,  intent(in)  :: fdOverlap
    real(dp), intent(in)  :: orig(:)
    integer,  intent(in)  :: iNeighbor(0:,:)
    integer,  intent(in)  :: nNeighbor(:)
    integer,  intent(in)  :: iSquare(:)
    integer,     intent(in)  :: iPair(0:,:)
    integer,  intent(in)  :: img2CentCell(:)
    type(TOrbitals), intent(in) :: orb
    
    integer     :: nAtom
    integer     :: iOrig, ii, jj, kk, ll
    integer     :: iNeigh
    integer     :: iAtom1, iAtom2, iAtom2f
    integer     :: nOrb1, nOrb2
    real(dp)    :: tmp(orb%mOrb,orb%mOrb)
    
    nAtom = size(iNeighbor, dim=2)

    ASSERT(nAtom > 0)
    ASSERT(all(shape(nNeighbor) == (/ nAtom /)))
    ASSERT(size(iSquare) == nAtom + 1)

    do iAtom1 = 1, nAtom
      ii = iSquare(iAtom1)
      nOrb1 = iSquare(iAtom1+1) - ii
      do iNeigh = 0, nNeighbor(iAtom1)
        iOrig = iPair(iNeigh,iAtom1) + 1
        iAtom2 = iNeighbor(iNeigh, iAtom1)
        iAtom2f = img2CentCell(iAtom2)
        jj = iSquare(iAtom2f)
        ASSERT(jj >= ii)
        nOrb2 = iSquare(iAtom2f+1) - jj
        tmp(1:nOrb2,1:nOrb1) = &
            & reshape(orig(iOrig:iOrig+nOrb1*nOrb2-1), (/nOrb2,nOrb1/))
        if (iNeigh == 0) then
          do ll = 1, nOrb1
            do kk = ll, nOrb2
              write(fdOverlap,*)kk+jj-1,ll+ii-1, tmp(kk,ll)
            end do
          end do
        else
          do ll = 1, nOrb1
            do kk = 1, nOrb2
              write(fdOverlap,*)kk+jj-1,ll+ii-1, tmp(kk,ll)
            end do
          end do
        end if
      end do
    end do
    
  end subroutine sparse_realprint


end module periodic
