!!* Calculates various types of charge populations
!!* @todo extend to other populations than Mulliken
!!* @author Ben Hourahine
module populations
# include "assert.h"
# include "allocate.h"  
  use accuracy
  use constants
  use periodic
  use spin, only : spinIndex
  use CommonTypes
  implicit none
  private

  public :: mulliken

  !!* Provides an interface to calculate the two types of Mulliken, either
  !!* orbitally resolved or atom resolved
  interface mulliken
    module procedure mullikenPerOrbital
    module procedure mullikenPerAtom
    module procedure mullikenPerOrbital_nc
  end interface
  
contains

  !!* Calculate the Mulliken population for each atom in the system, by summing
  !!* the individual orbital contriutions on each atom
  !!* @param q The charge per atom
  !!* @param s Overlap matrix in packed format
  !!* @param rho Density matrix in Packed format
  !!* @param orb Information about the orbitals.
  !!* @param iNeighbor Number of neighbours of each real atom (central cell)
  !!* @param nNeighbor List of neighbours for each atom, starting at 0 for
  !!*    itself
  !!* @param img2CentCell indexing array to convert images of atoms
  !!* back into their number in the central cell
  !!* @param iPair indexing array for the Hamiltonian
  subroutine mullikenPerAtom(qq, over, rho, orb, iNeighbor, nNeighbor, &
      &img2CentCell, iPair)
    real(dp), intent(out) :: qq(:)
    real(dp), intent(in) :: over(:)
    real(dp), intent(in) :: rho(:)
    type(TOrbitals), intent(in) :: orb
    integer, intent(in) :: iNeighbor(0:,:)
    integer, intent(in) :: nNeighbor(:)
    integer, intent(in) :: img2CentCell(:)
    integer, intent(in) :: iPair(0:,:)
    
    real(dp), allocatable :: qPerOrbital(:,:)
    integer :: nAtom

    nAtom = size(orb%nOrbAtom)
    ASSERT(size(qq) == nAtom)
    ASSERT(size(over) == size(rho))
    
    ALLOCATE_(qPerOrbital,(orb%mOrb, nAtom))
    qPerOrbital(:,:) = 0.0_dp
    
    call mullikenPerOrbital(qPerOrbital, over, rho, orb, iNeighbor, nNeighbor, &
        &img2CentCell, iPair)

    qq(:) = sum(qPerOrbital, dim=1)
    DEALLOCATE_(qPerOrbital)

  end subroutine mullikenPerAtom

  

  !!* Calculate the Mulliken population for each orbital in the system
  !!* using purely real-space overlap and density matrix values.
  !!* Currently Mulliken defined as 
  !!* $q_a = \sum_k w_k\sum_{\mu on a}\sum_\nu S_{\nu\mu}(k).\rho_{\mu\nu}(k)$
  !!* but transformed into real space sums over one triangle of real space 
  !!* extended matrices 
  !!* @param qq The charge per orbital
  !!* @param over Overlap matrix in packed format
  !!* @param rho Density matrix in Packed format
  !!* @param orb Information about the orbitals.
  !!* @param iNeighbor Number of neighbours of each real atom (central cell)
  !!* @param nNeighbor List of neighbours for each atom, starting at 0 for
  !!* itself
  !!* @param img2CentCell indexing array to convert images of atoms
  !!* back into their number in the central cell
  !!* @param iPair indexing array for the Hamiltonian
  !!* @todo add description of algorithm to programer manual / documentation.
  subroutine mullikenPerOrbital(qq, over, rho, orb, iNeighbor, nNeighbor, &
      &img2CentCell, iPair)    
    real(dp), intent(out) :: qq(:,:)
    real(dp), intent(in) :: over(:)
    real(dp), intent(in) :: rho(:)
    type(TOrbitals), intent(in) :: orb
    integer, intent(in) :: iNeighbor(0:,:)
    integer, intent(in) :: nNeighbor(:)
    integer, intent(in) :: img2CentCell(:)
    integer, intent(in) :: iPair(0:,:)

    integer   :: iOrig
    integer   :: iNeigh
    integer   :: nAtom, iAtom1, iAtom2, iAtom2f
    integer   :: nOrb1, nOrb2
    real(dp)  :: sqrTmp(orb%mOrb,orb%mOrb)
    real(dp)  :: mulTmp(orb%mOrb**2)

    nAtom = size(orb%nOrbAtom)
    
    ASSERT(all(shape(qq) == (/orb%mOrb, nAtom/)))
    ASSERT(size(over) == size(rho))

    qq(:,:) = 0.0_dp

    do iAtom1 = 1, nAtom
      nOrb1 = orb%nOrbAtom(iAtom1)
      do iNeigh = 0, nNeighbor(iAtom1)
        sqrTmp(:,:) = 0.0_dp
        mulTmp(:) = 0.0_dp
        iAtom2 = iNeighbor(iNeigh, iAtom1)
        iAtom2f = img2CentCell(iAtom2)
        nOrb2 = orb%nOrbAtom(iAtom2f)
        iOrig = iPair(iNeigh,iAtom1) + 1
        mulTmp(1:nOrb1*nOrb2) = over(iOrig:iOrig+nOrb1*nOrb2-1) &
            &* rho(iOrig:iOrig+nOrb1*nOrb2-1)
        sqrTmp(1:nOrb2,1:nOrb1) = &
            & reshape(mulTmp(1:nOrb1*nOrb2), (/nOrb2,nOrb1/))
        qq(1:nOrb1,iAtom1) = qq(1:nOrb1,iAtom1) &
            &+ sum(sqrTmp(1:nOrb2,1:nOrb1), dim=1)
        ! Add contribution to the other triangle sum, using the symmetry
        ! but only when off diagonal
        if (iAtom1 /= iAtom2f) then
          qq(1:nOrb2,iAtom2f) = qq(1:nOrb2,iAtom2f) &
              &+ sum(sqrTmp(1:nOrb2,1:nOrb1), dim=2)
        end if
      end do
    end do

  end subroutine mullikenPerOrbital



  !!* Calculate the non-colinear Mulliken population for each orbital in the 
  !!* system using purely real space overlap and density matrix values.
  !!* @param qq The charge vector per orbital, (q,p$_x$,p$_y$,p$_z$)
  !!* @param over Overlap matrix in packed format
  !!* @param rho 2-component Density matrix in Packed format
  !!* @param orb Orbital information
  !!* @param iNeighbor Number of neighbours of each real atom (central cell)
  !!* @param nNeighbor List of neighbours for each atom, starting at 0 for 
  !!* itself
  !!* @param nAtom Total number of atoms including any periodic images
  !!* @param img2CentCell indexing array to convert images of atoms
  !!* back into their number in the central cell
  !!* @param iPair indexing array for the Hamiltonian
  !!* @todo add description of algorithm to programer manual / documentation.
  subroutine mullikenPerOrbital_nc(qq, over, rho, orb, iNeighbor, nNeighbor, &
      &img2CentCell, iPair)
    real(dp),  intent(out) :: qq(:,:,:)
    real(dp),  intent(in)  :: over(:)
    complex(dp),intent(in) :: rho(:,:)
    type(TOrbitals), intent(in) :: orb
    integer,   intent(in)  :: iNeighbor(0:,:)
    integer,   intent(in)  :: nNeighbor(:)
    integer,   intent(in)  :: img2CentCell(:)
    integer,   intent(in)  :: iPair(0:,:)
    
    integer     :: iOrig
    integer     :: iNeigh
    integer     :: nAtom, iAtom1, iAtom2, iAtom2f
    integer     :: nOrb1, nOrb2
    complex(dp) :: sqrTmp(orb%mOrb,orb%mOrb,2,2)
    complex(dp) :: mulTmp(orb%mOrb**2,4)
    integer     :: iAlpha, iBeta, iIndex

    nAtom = size(orb%nOrbAtom)

    ASSERT(all(shape(qq) == (/orb%mOrb,nAtom,4/)))
    ASSERT(size(over) == size(rho,dim=1))
    ASSERT(size(rho,dim=2) == 4)
    
    qq(:,:,:) = 0.0_dp

    do iAtom1 = 1, nAtom
      nOrb1 = orb%nOrbAtom(iAtom1)
      do iNeigh = 0, nNeighbor(iAtom1)
        sqrTmp(:,:,:,:) = cmplx(0.0,0.0,dp)
        mulTmp(:,:) = 0.0_dp
        iAtom2 = iNeighbor(iNeigh, iAtom1)
        iAtom2f = img2CentCell(iAtom2)
        nOrb2 = orb%nOrbAtom(iAtom2f)
        iOrig = iPair(iNeigh,iAtom1) + 1
        do iAlpha = 1, 2
          do iBeta = 1, 2
            iIndex = spinIndex(iAlpha,iBeta)
            mulTmp(1 : nOrb1*nOrb2,iIndex) = &
                & cmplx(over(iOrig : iOrig + nOrb1*nOrb2 -1),0.0,dp) * &
                & rho(iOrig : iOrig + nOrb1*nOrb2 -1, iIndex)
            sqrTmp(1:nOrb2,1:nOrb1,iAlpha,iBeta) = &
                & reshape(mulTmp(1: nOrb1*nOrb2,iIndex), (/nOrb2,nOrb1/))
          end do
        end do
        qq(1:nOrb1,iAtom1,1) = qq(1:nOrb1,iAtom1,1) &
            &+ real(sum(sqrTmp(1:nOrb2,1:nOrb1,1,1),dim=1)) &
            &+ real(sum(sqrTmp(1:nOrb2,1:nOrb1,2,2),dim=1))
        qq(1:nOrb1,iAtom1,4) = qq(1:nOrb1,iAtom1,4) &
            &+ real(sum(sqrTmp(1:nOrb2,1:nOrb1,1,1),dim=1)) &
            &- real(sum(sqrTmp(1:nOrb2,1:nOrb1,2,2),dim=1))
        qq(1:nOrb1,iAtom1,3) = qq(1:nOrb1,iAtom1,3) &
            &+ real(sum(sqrTmp(1:nOrb2,1:nOrb1,1,2),dim=1)) &
            &+ real(sum(sqrTmp(1:nOrb2,1:nOrb1,2,1),dim=1))
        qq(1:nOrb1,iAtom1,2) = qq(1:nOrb1,iAtom1,2) &
            &- aimag(sum(sqrTmp(1:nOrb2,1:nOrb1,1,2),dim=1)) &
            &+ aimag(sum(sqrTmp(1:nOrb2,1:nOrb1,2,1),dim=1))
        ! Add contribution to the other triangle sum, using the Hermitian 
        ! symmetry but only when off diagonal
        if (iAtom1 /= iAtom2f) then
          qq(1:nOrb2,iAtom2f,1) = qq(1:nOrb2,iAtom2f,1) &
              &+ real(sum(sqrTmp(1:nOrb2,1:nOrb1,1,1),dim=2)) &
              &+ real(sum(sqrTmp(1:nOrb2,1:nOrb1,2,2),dim=2))
          qq(1:nOrb2,iAtom2f,4) = qq(1:nOrb2,iAtom2f,4) &
              &+ real(sum(sqrTmp(1:nOrb2,1:nOrb1,1,1),dim=2)) &
              &- real(sum(sqrTmp(1:nOrb2,1:nOrb1,2,2),dim=2))
          qq(1:nOrb2,iAtom2f,3) = qq(1:nOrb2,iAtom2f,3) &
              &+ real(sum(sqrTmp(1:nOrb2,1:nOrb1,2,1),dim=2)) &
              &+ real(sum(sqrTmp(1:nOrb2,1:nOrb1,1,2),dim=2))
          qq(1:nOrb2,iAtom2f,2) = qq(1:nOrb2,iAtom2f,2) &
              &+ aimag(sum(sqrTmp(1:nOrb2,1:nOrb1,2,1),dim=2)) &
              &- aimag(sum(sqrTmp(1:nOrb2,1:nOrb1,1,2),dim=2))
        end if
      end do
    end do
    
  end subroutine mullikenPerOrbital_nc


  
end module populations
