!!* Module for initializing SCC part of the calculation
module SCC_init
# include "assert.h"
# include "allocate.h"  
  use accuracy
  use message
  use FileId
  use CommonTypes
  implicit none
  private

  public :: initializeCharges, writeQToFile
  
  character(len=120) :: error_string !* Used to return runtime diagnostics  

  interface initializeCharges
     module procedure initQFromAtomChrg
     module procedure initQFromFile
  end interface

contains
  
  !!* Initialise the charge vector from the reference atomic charges
  !!* @param qq The charges per lm,atom,spin
  !!* @param qShell The reference charges per shell
  !!* @param specie List of chemical species for each atom
  !!* @param orb Information about the orbitals.
  !!* @todo Generalise initialization to include other cases than the 
  !!* atomic closed shell
  subroutine initQFromAtomChrg(qq, qShell, specie, orb)
    real(dp), intent(out) :: qq(:,:,:)
    real(dp), intent(in)  :: qShell(:,:)
    integer,   intent(in) :: specie(:)
    type(TOrbitals), intent(in) :: orb
    
    integer :: iAt1, iSp1, iSh1, nOrbL, ind, nAtom

    nAtom = size(orb%nOrbAtom)

    ASSERT(size(qq, dim=1) == orb%mOrb)
    ASSERT(size(qq, dim=2) == nAtom)
    ASSERT(size(qShell, dim=1) == orb%mShell)
    ASSERT(size(qShell, dim=2) == size(orb%angShell, dim=2))
    ASSERT(size(specie) == nAtom)

    qq(:,:,:) = 0.0_dp
    !! fill degenerately over m for each shell l
    do iAt1 = 1, nAtom
      iSp1 = specie(iAt1)
      ind = 1
      do iSh1 = 1, orb%nShell(iSp1)
        nOrbL = 2 * orb%angShell(iSh1, iSp1) + 1
        qq(ind:ind+nOrbL-1, iAt1, 1) = qShell(iSh1, iSp1) / real(nOrbL, dp)
        ind = ind + nOrbL
      end do
    end do
    
    if (size(qq, dim=3) == 2) then
      !! spin polarised case
      qq(:,:,1) = 0.5_dp * qq(:,:,1)
      qq(:,:,2) = qq(:,:,1)
    end if
    
  end subroutine initQFromAtomChrg


  
  !!* Initialise the charge vector from a named external file. Check the total
  !!* charge matches that expected for the calculation.
  !!* @param qq The charges per lm,atom,spin
  !!* @param fileName The external file of charges for the orbitals, currently
  !!*   stored with each line containing the per-orbital charges in order of
  !!*   increasing m and l. Alternating lines give the spin case (if present)
  !!* @param nEl Nr. of electrons for each spin channel
  !!* @param orb Information about the orbitals in the system.
  !!* @todo XML format for external file/option of charges data in input file
  !!*   in XML. Also further test of the input, if the number of orbital charges
  !!*   per atom match the number from the angular momentum. 
  subroutine initQFromFile(qq, fileName, nEl, orb)
    real(dp), intent(out)     :: qq(:,:,:)
    character(*), intent(in)  :: fileName
    real(dp), intent(in)      :: nEl(:)
    type(TOrbitals), intent(in) :: orb

    integer  :: nOrb, nAtom, nSpin ! nr. of orbitals / atoms / spin channels
    integer  :: iErr               ! error returned by the io commands
    integer, save :: file = -1     ! file unit number
    integer  :: iOrb, iAtom, iSpin
    real(dp) :: nEl0
    real(dp), allocatable :: sumChrg(:)

    nAtom = size(qq, dim=2)
    nSpin = size(qq, dim=3)

    ASSERT(size(qq, dim=1) == orb%mOrb)
    ASSERT(size(nEl) == nSpin)

    if (file == -1) then
      file = getFileId()
    end if

    qq(:,:,:) = 0.0_dp
    open(file, file=fileName, status='old', action='READ', form='unformatted', &
        &iostat=iErr)
    if (iErr /= 0) then
      write(error_string, *) "Failure to open external file of charge data"
      call error(error_string)
    end if
    rewind(file)

    do iAtom = 1, nAtom
      nOrb = orb%nOrbAtom(iAtom)
      do iSpin = 1, nSpin
        read(file, iostat=iErr) (qq(iOrb, iAtom, iSpin), iOrb=1,nOrb)
        if (iErr /= 0) then
          call error("Error during reading external file of charge data")
        end if
      end do
    end do
    close(file)

    ALLOCATE_(sumChrg, (nSpin))
    sumChrg(:) = sum(sum(qq, dim=1), dim=1)
    do iSpin = 1, nSpin
      if (nSpin == 1) then
        nEl0 = 2.0_dp * nEl(iSpin)
      else
        nEl0 = nEl(iSpin)
      end if
      if (abs(nEl0 - sumChrg(iSpin)) > elecTolMax) then
        write(error_string, 99000) sumChrg(iSpin), nEl0, iSpin
99000   format ('External file of charges has a total charge:', &
            &E16.8,', instead of ',E16.8,' in spin channel ',I1)
        call warning(error_string)
      end if
    end do
    DEALLOCATE_(sumChrg)

  end subroutine initQFromFile



  !!* Write the current charges to an external file
  !!* @param qq       Array containing the charges
  !!* @param fileName Name of the file to write the charges
  !!* @param orb Information about the orbitals in the system.
  subroutine writeQToFile(qq, fileName, orb)
    real(dp), intent(in)     :: qq(:,:,:)
    character(*), intent(in) :: fileName
    type(TOrbitals), intent(in) :: orb

    integer :: nAtom, nOrb, nSpin
    integer :: iAtom, iOrb, iSpin
    integer, save :: file = -1
    integer :: iErr

    nAtom = size(qq, dim=2)
    nSpin = size(qq, dim=3)

    ASSERT(size(qq, dim=1) >= orb%mOrb)

    if (file == -1) then
      file = getFileId()
    end if
    
    open(file, file=fileName, position="rewind", status="replace", &
        & form="unformatted")
    do iAtom = 1, nAtom
      do iSpin = 1, nSpin
        nOrb = orb%nOrbAtom(iAtom)
        write (file, iostat=iErr) (qq(iOrb, iAtom, iSpin), iOrb = 1,nOrb)
        if (iErr /= 0) then
          write (error_string, *) "Failure to write file for external charges"
          call error(error_string)
        end if
      end do
    end do
    close(file)
    
  end subroutine writeQToFile
    
  
end module SCC_init
