!!* Dispersion a la Slater-Kirkwood as implemented by M. Elstner in old DFTB.
!!* @desc The expression as found in the old DFTB had been reimplemented.
!!*   The periodic case had been completely rewritten using a correct
!!*   Ewald summation (instead of the pure real space summation, which converges
!!*   very poorly).
!!* @note The expression for C6(iAt1,iAt2) is not the same as in the reference 
!!*   paper by M. Elstner, but the one found in the old code (implemented by
!!*   him as well). Furthermore, the expression for the dispersion energy in the
!!*   paper (eq. 9) misses a factor of 1/2.
!!* @ref Elstner et al., J. Chem. Phys., 114, 5149 (2001)
!!* @ref N. Karasawa and W. A. Goddard III, J. Phys. Chem. 93, 7320-7327 (1989)
!!* @ref Zhou-Min Chen, T. Cagin and W. A. Goddard III, J. Comp. Chem. 18, 1365
!!*    (1997)
!!* @todo The generation of the reciprocal lattice vectors should not be done
!!*   locally, but somewhere outside, since the Coulomb module does the same.
module SlaterKirkwood
#include "allocate.h"
#include "assert.h"
  use Accuracy
  use Periodic, only: TNeighborList, getNrOfNeighborsForAll, getLatticePoints
  use Sorting
  use SimpleAlgebra, only : cross3
  use Constants, only : pi
  use Message
#ifdef EXTERNALERFC
  use external_erfc
#endif  
  implicit none
  private

  public :: TSlaterKirkwoodInput, OSlaterKirkwood
  public :: create, destroy, updateCoords, getEnergies, addGradients
  public :: getRCutoff, getGCutoff


  !!* Contains the initialisation data for the Slater-Kirkwood module
  type TSlaterKirkwoodInput
    real(dp), pointer :: polar(:) => null()  !* Atomic polarisabilities (nAtom)
    real(dp), pointer :: rWaals(:) => null()  !* Van der Waals radii (nAtom)
    real(dp), pointer :: charges(:) => null()  !* Effective charges (nAtom)
  end type TSlaterKirkwoodInput


  !!* Data for the Slater-Kirkwood type dispersion
  type OSlaterKirkwood
    private
    real(dp), pointer :: c6(:,:)      !* Atomic polarisabilities (nAtom)
    real(dp), pointer :: rVdW2(:,:)   !* Van der Waals radii (nAtom)
    integer :: nAtom                  !* Nr. of atoms (without images)
    real(dp), pointer :: energies(:)  !* Energies
    real(dp), pointer :: gradients(:,:) !* Gradients (3, nAtom)
    logical :: tPeriodic              !* If system is periodic
    real(dp) :: rCutoff               !* Real space cutoff
    real(dp) :: gCutoff               !* Reciprocal space cutoff
    real(dp) :: dampCutoff            !* Cutoff, where damping function = 1
    real(dp) :: eta                   !* Periodic summation parameter
    real(dp) :: vol                   !* Volume of the unit cell
    logical :: coordsUpdated          !* If first coordinate update done
  end type OSlaterKirkwood


  !!* Consctructors
  interface create
    module procedure SlaterKirkwood_create
  end interface

  !!* Destructors
  interface destroy
    module procedure SlaterKirkwood_destroy
    module procedure SlaterKirkwoodInput_destroy
  end interface

  !!* Updates the coordinates of the atoms
  interface updateCoords
    module procedure SlaterKirkwood_updateCoords
  end interface

  !!* Gets the atomic resolved energies due to the dispersion
  interface getEnergies
    module procedure SlaterKirkwood_getEnergies
  end interface

  !!* Adds the atomic resoved gradients due to the dispersion
  interface addGradients
    module procedure SlaterKirkwood_addGradients
  end interface

  !!* Returns the real space cutoff for the dispersion interaction
  interface getRCutoff
    module procedure SlaterKirkwood_getRCutoff
  end interface

  !!* Returns the reciprocal space cutoff for the dispersion interaction
  interface getGCutoff
    module procedure SlaterKirkwood_getGCutoff
  end interface

  !! Some magic constants for the damping function (see paper)
  integer, parameter :: nn_ = 7         ! N
  integer, parameter :: mm_ = 4         ! M
  real(dp), parameter :: dd_ = 3.0_dp   ! d

  real(dp), pointer :: gLatPoint_(:,:)  !! Temporary dirty solution

  character(len=lc) :: error_string

  
contains

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!  Public routines
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  !!* Creates a SlaterKirkwood instance.
  !!* @param self Initialised instance at return.
  !!* @param inp Input parameters for Slater-Kirkwood.
  !!* @param latVecs Lattice vectors, if the system is periodic
  !!* @param recVecs Reciprocal vectors, if the system is periodic
  !!* @param vol Volume of the unit cell, if the system is periodic
  subroutine SlaterKirkwood_create(self, inp, latVecs, recVecs, vol)
    type(OSlaterKirkwood), pointer :: self
    type(TSlaterKirkwoodInput), intent(in) :: inp
    real(dp), intent(in), optional :: latVecs(:,:)
    real(dp), intent(in), optional :: recVecs(:,:)
    real(dp), intent(in), optional :: vol

    integer :: iAt1, iAt2
    real(dp) :: tol, rTmp, maxR

    ASSERT(size(inp%polar) > 0)
    ASSERT(size(inp%polar) == size(inp%rWaals))
    ASSERT(size(inp%polar) == size(inp%charges))
    ASSERT(all(inp%polar >= 0.0_dp))
    ASSERT(all(inp%rWaals >= 0.0_dp))
    ASSERT(present(latVecs) .eqv. present(recVecs))
    ASSERT(present(latVecs) .eqv. present(vol))
    ASSERT_ENV(if (present(latVecs)) then)
    ASSERT_ENV(ASSERT(all(shape(latVecs) == (/ 3, 3 /))))
    ASSERT_ENV(ASSERT(all(shape(recVecs) == (/ 3, 3 /))))
    ASSERT_ENV(ASSERT(vol > 0.0_dp))
    ASSERT_ENV(end if)


    INITALLOCATE_P(self)
    self%nAtom = size(inp%polar)
    INITALLOCATE_PARR(self%c6, (self%nAtom, self%nAtom))
    INITALLOCATE_PARR(self%rVdW2, (self%nAtom, self%nAtom))
    self%rCutoff = 0.0_dp
    self%c6 = 0.0_dp
    self%rVdW2 = 0.0_dp
    maxR = 0.0_dp
    tol = epsilon(1.0_dp)
    do iAt1 = 1, self%nAtom
      if (inp%polar(iAt1) < tol .or. inp%rWaals(iAt1) < tol) then
        cycle
      end if
      do iAt2 = 1, iAt1
        self%c6(iAt2, iAt1) = 1.5_dp * inp%polar(iAt1) * inp%polar(iAt2)&
            &/ (sqrt(inp%polar(iAt1)/inp%charges(iAt1)) &
            &+ sqrt(inp%polar(iAt2)/inp%charges(iAt2)))
        rTmp = (inp%rWaals(iAt1)**3 + inp%rWaals(iAt2)**3) &
            &/ (inp%rWaals(iAt1)**2 + inp%rWaals(iAt2)**2)
        self%rVdW2(iAt2, iAt1) = dd_ / rTmp**nn_
        maxR = max(maxR, rTmp)
        if (iAt1 /= iAt2) then
          self%c6(iAt1, iAt2) = self%c6(iAt2, iAt1)
          self%rVdW2(iAt1, iAt2) = self%rVdW2(iAt2, iAt1)
        end if
        self%rCutoff = max(self%rCutoff, &
            &(self%c6(iAt2, iAt1)/tolDispersion)**(1.0_dp/6.0_dp))
      end do
    end do

    self%tPeriodic = present(latVecs)
    if (self%tPeriodic) then
      self%vol = vol
      !! Scaling down optimal eta (as suggested in the literature) is purely
      !! empirical, it reduces the real space summation, and seems to yield
      !! shorter execution times. (It doesn't influence the result.)
      self%eta =  getOptimalEta_(latVecs, self%vol) / sqrt(2.0_dp)
      self%rCutoff = getMaxRDispersion_(self%eta, self%c6, self%vol, &
          &tolDispersion)
      !! Cutoff, beyond which dispersion is purely 1/r^6 without damping
      self%dampCutoff = getDampCutoff_(maxR, tolDispDamp)
      self%rCutoff = max(self%rCutoff, self%dampCutoff)
      self%gCutoff = getMaxGDispersion_(self%eta, self%c6, tolDispersion)
      INIT_PARR(gLatPoint_)
      call getLatticePoints(gLatPoint_, recVecs, latVecs/(2.0_dp*pi), &
          &self%gCutoff, onlyInside=.true., reduceByInversion=.true., &
          &withoutOrigin=.true.)
      gLatPoint_ = matmul(recVecs, gLatPoint_)
    end if

    INITALLOCATE_PARR(self%energies, (self%nAtom))
    INITALLOCATE_PARR(self%gradients, (3, self%nAtom))
    self%coordsUpdated = .false.

  end subroutine SlaterKirkwood_create



  !!* Destorys SlaterKirkwood instance.
  !!* @param self SK-instance.
  subroutine SlaterKirkwood_destroy(self)
    type(OSlaterKirkwood), pointer :: self

    if (.not. associated(self)) then
      return
    end if
    DEALLOCATE_PARR(self%c6)
    DEALLOCATE_PARR(self%rVdW2)
    DEALLOCATE_PARR(self%energies)
    DEALLOCATE_PARR(self%gradients)
    DEALLOCATE_P(self)

  end subroutine SlaterKirkwood_destroy



  !!* Notifies the objects about changed coordinates
  !!* @param self Object instance.
  !!* @param neigh Current neighbor list mapping
  !!* @param img2CentCell Current mapping of periodic images to the central cell
  !!* @param coords Current coordinates of the atoms
  subroutine SlaterKirkwood_updateCoords(self, neigh, img2CentCell, coords)
    type(OSlaterKirkwood), pointer :: self
    type(TNeighborList), intent(in) :: neigh
    integer, intent(in) :: img2CentCell(:)
    real(dp), intent(in) :: coords(:,:)

    integer, allocatable :: nNeighReal(:) ! Neighbors for real space summation
    integer, allocatable :: nNeighDamp(:) ! Nr. of neighbors with damping

    ALLOCATE_(nNeighReal, (self%nAtom))
    call getNrOfNeighborsForAll(nNeighReal, neigh, self%rCutoff)
    self%energies(:) = 0.0_dp
    self%gradients(:,:) = 0.0_dp
    if (self%tPeriodic) then
      !! Make Ewald summation for a pure 1/r^6 interaction
      call addDispEnergyAndGrad_periodic_(self%nAtom, coords, nNeighReal, &
          &neigh%iNeighbor, neigh%neighDist2, img2CentCell, self%c6, self%eta, &
          &self%vol, gLatPoint_, self%energies, self%gradients)
      !! Correct those terms, where damping is important
      ALLOCATE_(nNeighDamp, (self%nAtom))
      call getNrOfNeighborsForAll(nNeighDamp, neigh, self%dampCutoff)
      call addDispEnergyAndGrad_cluster_(self%nAtom, coords, nNeighDamp, &
          &neigh%iNeighbor, neigh%neighDist2, img2CentCell, self%c6, &
          &self%rVdW2, self%energies, self%gradients, dampCorrection=-1.0_dp)
    else
      call addDispEnergyAndGrad_cluster_(self%nAtom, coords, nNeighReal, &
          &neigh%iNeighbor, neigh%neighDist2, img2CentCell, self%c6, &
          &self%rVdW2, self%energies, self%gradients)
    end if
    self%coordsUpdated = .true.

    DEALLOCATE_(nNeighReal)
    DEALLOCATE_(nNeighDamp)
    
  end subroutine SlaterKirkwood_updateCoords
  
  

  !!* Returns the atomic resolved energies due to the dispersion.
  !!* @param self Object instance
  !!* @param energies Contains the atomic energy contributions on exit.
  subroutine SlaterKirkwood_getEnergies(self, energies)
    type(OSlaterKirkwood), pointer :: self
    real(dp), intent(out) :: energies(:)

    ASSERT(self%coordsUpdated)
    ASSERT(size(energies) == self%nAtom)
    
    energies(:) = self%energies(:)
    
  end subroutine SlaterKirkwood_getEnergies
  


  !!* Adds the atomic gradients to the provided vector.
  !!* @param self Object instance.
  !!* @param gradients The vector to increase by the gradients.
  subroutine SlaterKirkwood_addGradients(self, gradients)
    type(OSlaterKirkwood), pointer :: self
    real(dp), intent(inout) :: gradients(:,:)

    ASSERT(self%coordsUpdated)
    ASSERT(all(shape(gradients) == (/ 3, self%nAtom /)))
    
    gradients(:,:) = gradients(:,:) + self%gradients(:,:)
    
  end subroutine SlaterKirkwood_addGradients

  

  !!* Estimates the real space cutoff of the dispersion interaction.
  !!* @param self Object instance
  !!* @return Cutoff
  function SlaterKirkwood_getRCutoff(self) result(cutoff)
    type(OSlaterKirkwood), pointer :: self
    real(dp) :: cutoff

    cutoff = self%rCutoff

  end function SlaterKirkwood_getRCutoff


  
  !!* Estimates the reciprocal space cutoff of the dispersion interaction.
  !!* @param self Object instance
  !!* @return Cutoff
  function SlaterKirkwood_getGCutoff(self) result(cutoff)
    type(OSlaterKirkwood), pointer :: self
    real(dp) :: cutoff

    cutoff = self%gCutoff

  end function SlaterKirkwood_getGCutoff


  
  !!* Destroys the object instance.
  !!* @param self Object instance.
  subroutine SlaterKirkwoodInput_destroy(self)
    type(TSlaterKirkwoodInput), intent(inout) :: self

    DEALLOCATE_PARR(self%polar)
    DEALLOCATE_PARR(self%rWaals)
    DEALLOCATE_PARR(self%charges)
    
  end subroutine SlaterKirkwoodInput_destroy

  
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!  Private routines
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  
  !!* Adds the energy per atom and the gradients for the cluster case
  !!* @param nAtom Nr. of atoms (without periodic images)
  !!* @param coords Coordinates of the atoms (including images)
  !!* @param nNeighbors Nr. of neighbors for each atom
  !!* @param iNeighbor Neighborlist.
  !!* @param neighDist2 Square distances of the neighbours.
  !!* @param img2CentCell Mapping of periodic images onto the central cell
  !!* @param c6 Van der Waals coefficients (nAtom, nAtom)
  !!* @param rVdW2 Scaled inverse van der Waals radii (nAtom, nAtom)
  !!* @param energies Updated energy vector at return
  !!* @param gradients Updated gradient vector at return
  !!* @param dampCorrection Adds the provided value to the damping function
  !!*   (use -1.0 to sum up damped 1/r^6 terms and subtract pure 1/r^6 ones, in
  !!*   order to correct periodic Ewald sum for the short range damped terms.)
  subroutine addDispEnergyAndGrad_cluster_(nAtom, coords, nNeighbors, &
      &iNeighbor, neighDist2, img2CentCell, c6, rVdW2, energies, gradients, &
      &dampCorrection)
    integer, intent(in) :: nAtom
    real(dp), intent(in) :: coords(:,:)
    integer, intent(in) :: nNeighbors(:)
    integer, intent(in) :: iNeighbor(0:,:)
    real(dp), intent(in) :: neighDist2(0:,:)
    integer, intent(in) :: img2CentCell(:)
    real(dp), intent(in) :: c6(:,:)
    real(dp), intent(in) :: rVdW2(:,:)
    real(dp), intent(inout) :: energies(:)
    real(dp), intent(inout) :: gradients(:,:)
    real(dp), intent(in), optional :: dampCorrection

    integer :: iAt1, iNeigh, iAt2, iAt2f
    real(dp) :: dist2, dist, h0, h1, h2, rTmp
    real(dp) :: diff(3), gr(3)
    real(dp) :: corr

    if (present(dampCorrection)) then
      corr = dampCorrection
    else
      corr = 0.0_dp
    end if

    !! Cluster case => explicit sum of the contributions
    do iAt1 = 1, nAtom
      do iNeigh = 1, nNeighbors(iAt1)
        iAt2 = iNeighbor(iNeigh, iAt1)
        iAt2f = img2CentCell(iAt2)
        if (c6(iAt2f, iAt1) == 0.0_dp) then
          cycle
        end if
        dist2 = neighDist2(iNeigh, iAt1)
        if (dist2 > minNeighDist2) then
          dist = sqrt(dist2)
          h0 = rVdW2(iAt2f, iAt1)
          h1 = exp(-1.0_dp * h0 * dist**nn_)
          h2 = 1.0_dp - h1
          !! Energy
          rTmp = -0.5_dp * c6(iAt2f, iAt1) * (h2**mm_ + corr) / dist**6
          energies(iAt1) = energies(iAt1) + rTmp
          if (iAt1 /= iAt2f) then
            energies(iAt2f) = energies(iAt2f) + rTmp
          end if
          !! Gradients
          diff(:) = (coords(:,iAt1) - coords(:,iAt2))
          gr(:) = -c6(iAt2f, iAt1) * diff(:) &
              &* (mm_*h2**(mm_-1)*h1*h0*nn_*dist**(nn_-8) &
              &- 6.0_dp * (h2**mm_ + corr) * dist**(-8))
          gradients(:,iAt1) = gradients(:,iAt1) + gr(:)
          gradients(:,iAt2f) = gradients(:,iAt2f) - gr(:)
        end if
      end do
    end do
    
  end subroutine addDispEnergyAndGrad_cluster_
  


  !!* Adds the energy per atom and the gradients for the periodic case, but
  !!* without damping (pure 1/r^6 interaction)
  !!* @param nAtom Nr. of atoms (without periodic images)
  !!* @param coords Coordinates of the atoms (including images)
  !!* @param nNeighbors Nr. of neighbors for each atom
  !!* @param iNeighbor Neighborlist.
  !!* @param neighDist2 Square distances of the neighbours.
  !!* @param img2CentCell Mapping of periodic images onto the central cell
  !!* @param c6 Van der Waals coefficients (nAtom, nAtom)
  !!* @param eta Controling partitioning between real and reciprocal space sum
  !!* @param vol Volume of the unit cell
  !!* @param gLatVecs Set of reciprocal space vectors to include in the sum
  !!* @param energies Updated energy vector at return
  !!* @param gradients Updated gradient vector at return
  !!* @desc Fast converging Ewald like summation on 1/r^6 type interactions.
  !!*   For details see the references in the module description.
  subroutine addDispEnergyAndGrad_periodic_(nAtom, coords, nNeighbors, &
      &iNeighbor, neighDist2, img2CentCell, c6, eta, vol, gLatVecs, &
      &energies, gradients)
    integer, intent(in) :: nAtom
    real(dp), intent(in) :: coords(:,:)
    integer, intent(in) :: nNeighbors(:)
    integer, intent(in) :: iNeighbor(0:,:)
    real(dp), intent(in) :: neighDist2(0:,:)
    integer, intent(in) :: img2CentCell(:)
    real(dp), intent(in) :: c6(:,:)
    real(dp), intent(in) :: eta
    real(dp), intent(in) :: vol
    real(dp), intent(in) :: gLatVecs(:,:)
    real(dp), intent(inout) :: energies(:)
    real(dp), intent(inout) :: gradients(:,:)

    integer :: iAt1, iNeigh, iAt2, iAt2f, iG
    real(dp) :: rSum, rSum3(3), gSum, gSum3(3), gg(3), ggAbs
    real(dp) :: aam2, bb, bbm2, rTmp, rTmp2, rTmp3, rc, r3c, gc, g3c, ddp, etam3

    ASSERT(size(energies) == nAtom)
    ASSERT(all(shape(gradients) == (/ 3, nAtom /)))

    etam3 = eta**(-3)
    rc = 0.5_dp * etam3 * etam3
    r3c = 2.0_dp * rc / (eta * eta)
    gc = pi**1.5_dp / (12.0_dp * vol)
    g3c = pi**1.5_dp / (6.0_dp * vol)
    do iAt1 = 1, nAtom
      do iNeigh = 1, nNeighbors(iAt1)
        iAt2 = iNeighbor(iNeigh, iAt1)
        iAt2f = img2CentCell(iAt2)
        aam2 = (sqrt(neighDist2(iNeigh, iAt1))/eta)**(-2)
        rSum =  rc * c6(iAt2f, iAt1) &
            &* ((aam2 + 1.0_dp)*aam2 + 0.5_dp)*aam2 * exp(-1.0_dp / aam2)
        rSum3(:) = r3c * c6(iAt2f, iAt1) * exp(-1.0_dp / aam2) &
            &* (((6.0_dp*aam2 + 6.0_dp)*aam2 + 3.0_dp)*aam2 + 1.0_dp)*aam2 &
            & * (coords(:,iAt1) - coords(:,iAt2))
        energies(iAt1) = energies(iAt1) - rSum
        gradients(:,iAt1) = gradients(:,iAt1) + rSum3(:)
        if (iAt1 /= iAt2f) then
          energies(iAt2f) = energies(iAt2f) - rSum
          gradients(:,iAt2f) = gradients(:,iAt2f) - rSum3(:)
        end if
      end do
      
      gSum = 0.0_dp
      gSum3(:) = 0.0_dp
      do iG = 1, size(gLatVecs, dim=2)
        gg = gLatVecs(:, iG)
        ggAbs = sqrt(sum(gg**2))
        bb = 0.5_dp * ggAbs * eta
        bbm2 = bb**(-2)
        rTmp = 0.0_dp
        rTmp2 = 0.0_dp
        do iAt2 = 1, nAtom
          ddp = dot_product(gg, coords(:,iAt1) - coords(:,iAt2))
          rTmp = rTmp + cos(ddp) * c6(iAt1, iAt2)
          rTmp2 = rTmp2 + sin(ddp) * c6(iAt1, iAt2)
        end do
#ifdef EXTERNALERFC
        rTmp3 = sqrt(pi) * extErfc(bb) + (0.5_dp * bbm2 - 1.0_dp) / bb &
            &* exp(-1.0_dp / bbm2)
#else        
        rTmp3 = sqrt(pi) * erfc(bb) + (0.5_dp * bbm2 - 1.0_dp) / bb &
            &* exp(-1.0_dp / bbm2)
#endif        
        gSum = gSum + rTmp * ggAbs**3 * rTmp3
        gSum3(:) = gSum3(:) + gg(:) * rTmp2 * ggAbs**3 * rTmp3
      end do
      gSum = gSum * gc
      gSum3(:) = gSum3(:) * g3c
      energies(iAt1) = energies(iAt1) - gSum &
          &+ (c6(iAt1,iAt1)/12.0_dp * etam3 &
          &- pi**1.5 * sum(c6(:,iAt1))/(6.0_dp * vol)) * etam3 
      gradients(:,iAt1) =  gradients(:,iAt1) +  gSum3
    end do

  end subroutine addDispEnergyAndGrad_periodic_



  !!* Delivers the optimal paramater eta for the Ewald-like summation
  !!* @param latVecs Lattice vectors
  !!* @param vol Volume
  !!* @return Optimal parameter.
  function getOptimalEta_(latVecs, vol) result(eta)
    real(dp), intent(in) :: latVecs(:,:)
    real(dp), intent(in) :: vol
    real(dp) :: eta

    real(dp) :: vecLens(3), tmp(3)
    integer :: indx(3), i1, i2

    vecLens(:) = sqrt(sum(latVecs**2, dim=1))
    call index_heap_sort(indx, vecLens)
    !! *** Workaround for old compaq compilers ***
    i1 = indx(1)
    i2 = indx(2)
    call cross3(tmp, latVecs(:,i1), latVecs(:,i2))
    eta = sqrt(vecLens(indx(1)) * vol / (pi * sqrt(sum(tmp**2))))

  end function getOptimalEta_

    

  !!* Returns the longest real space vector needed to achive a given accuracy
  !!* in the Ewald summation for the dispersion.
  !!* @param eta Parameter of the ewald summation.
  !!* @param c6 Van der Waals coefficients (nAtom, nAtom)
  !!* @param vol Volume of the unit cell
  !!* @param minValue  Tolerance value.
  !!* @return Cutoff radius.
  function getMaxRDispersion_(eta, c6, vol, minValue) result(xx)
    real(dp), intent(in) :: eta
    real(dp), intent(in) :: c6(:,:)
    real(dp), intent(in) :: vol, minValue
    real(dp) :: xx

    real(dp), parameter :: rInit = 1.0e-8_dp
    real(dp) :: xLeft, xRight, yLeft, yRight, yy
    real(dp) :: c6sum
    integer  :: iError, iIter

    c6sum = sum(abs(c6))
    iError = 0
    xx = rInit
    yy = getDispRealError_(xx, c6sum, vol, eta)
    do while (yy > minValue .and. xx <= huge(1.0_dp))
      xx = 2.0_dp * xx
      yy = getDispRealError_(xx, c6sum, vol, eta)
    end do
    if (xx > huge(1.0_dp)) then
      iError = 1
    elseif (xx == rInit) then
      iError = 2
    end if

    if (iError == 0) then
      xLeft = 0.5_dp * xx
      yLeft = getDispRealError_(xLeft, c6sum, vol, eta)
      xRight = xx
      yRight = yy

      iIter = 1
      do while (yLeft - yRight > minValue .and. iIter <= nSearchIter)
        xx = 0.5_dp * (xLeft + xRight)
        yy = getDispRealError_(xx, c6sum, vol, eta)
        if (yy >= minValue) then
          xLeft = xx
          yLeft = yy
        else
          xRight = xx
          yRight = yy
        end if
        iIter = iIter + 1
      end do
      if (iIter > nSearchIter) then
        iError = 3
      end if
    end if

    if (iError /= 0) then
99020 format ('Failure in getMaxRDispersion_.', ' Error nr: ',I3)
      write(error_string, 99020) iError
      call error(error_string)
    end if

  end function getMaxRDispersion_



  !!* Returns the longest reciprocal space vector needed to achive a given 
  !!* accuracy in the Ewald summation for the dispersion.
  !!* @param eta Parameter of the ewald summation.
  !!* @param c6 Van der Waals coefficients (nAtom, nAtom)
  !!* @param minValue  Tolerance value.
  !!* @return Cutoff radius.
  function getMaxGDispersion_(eta, c6, minValue) result(xx)
    real(dp), intent(in) :: eta
    real(dp), intent(in) :: c6(:,:)
    real(dp), intent(in) :: minValue
    real(dp) :: xx

    real(dp), parameter :: gInit = 1.0e-8_dp
    real(dp) :: xLeft, xRight, yLeft, yRight, yy
    real(dp) :: c6sum
    integer :: iError, iIter

    c6sum = sum(abs(c6))
    iError = 0
    xx = gInit
    yy = getDispReciprocalError_(xx, c6sum, eta)
    do while (yy > minValue .and. xx <= huge(1.0_dp))
      xx = 2.0_dp * xx
      yy = getDispReciprocalError_(xx, c6sum, eta)
    end do
    if (xx > huge(1.0_dp)) then
      iError = 1
    elseif (xx == gInit) then
      iError = 2
    end if

    if (iError == 0) then
      xLeft = 0.5_dp * xx
      yLeft = getDispReciprocalError_(xLeft, c6sum, eta)
      xRight = xx
      yRight = yy

      iIter = 1
      do while (yLeft - yRight > minValue .and. iIter <= nSearchIter)
        xx = 0.5_dp * (xLeft + xRight)
        yy = getDispReciprocalError_(xx, c6sum, eta)
        if (yy >= minValue) then
          xLeft = xx
          yLeft = yy
        else
          xRight = xx
          yRight = yy
        end if
        iIter = iIter + 1
      end do
      if (iIter > nSearchIter) then
        iError = 3
      end if
    end if

    if (iError /= 0) then
99010 format ('Failure in getMaxGDispersion_.', ' Error nr: ',I3)
      write(error_string, 99010) iError
      call error(error_string)
    end if

  end function getMaxGDispersion_


  
  !!* Returns the distance, beyond that the damping function equals approx. 1.
  !!* @param r0 Length scaling parameter
  !!* @param tol  Tolerance value.
  !!* @return cutoff
  function getDampCutoff_(r0, tol) result(xx)
    real(dp), intent(in) :: r0, tol
    real(dp) :: xx

    !! solve: 1 - tol < (1-exp(-d*(r/r0)^N))^M  for r and hope that the 
    !! logarithm is not blowing up your computer.
    xx = r0 * (-1.0_dp/dd_ * log(1.0_dp &
        &- (1.0_dp - tol)**(1.0_dp/real(mm_,dp))))**(1.0_dp/real(nn_, dp))

  end function getDampCutoff_



  !!* Returns the error of the real space summation for a certain cutoff
  !!* @param rr Cutoff radius
  !!* @param c6sum VdW coefficients
  !!* @param vol Volume of the supercell
  !!* @param eta Summation parameter
  !!* @return Estimated error of the summation.
  function getDispRealError_(rr, c6sum, vol, eta) result(err)
    real(dp), intent(in) :: rr, c6sum, vol, eta
    real(dp) :: err

#ifdef EXTERNALERFC    
    err = (pi**1.5_dp * c6sum / vol) * eta * (1.0_dp/rr**4 &
        &+ 1.0_dp/(rr**2 * eta**2) + 1.0_dp / (2.0_dp * eta**4)) &
        &* extErfc(rr/eta)
#else
    err = (pi**1.5_dp * c6sum / vol) * eta * (1.0_dp/rr**4 &
        &+ 1.0_dp/(rr**2 * eta**2) + 1.0_dp / (2.0_dp * eta**4)) &
        &* erfc(rr/eta)
#endif    
    
  end function getDispRealError_


  
  !!* Returns the error of the reciprocal space summation for a certain cutoff
  !!* @param gg Cutoff radius
  !!* @param c6sum VdW coefficients
  !!* @param eta Summation parameter
  !!* @return Estimated error of the summation.
  function getDispReciprocalError_(gg, c6sum, eta) result(err)
    real(dp), intent(in) :: gg, c6sum, eta
    real(dp) :: err

#ifdef EXTERNALERFC
    err = c6sum/(6.0_dp * sqrt(pi)) * (1.0_dp / eta**6) &
        &*(gg * eta * exp(-1.0_dp * (0.5_dp*gg*eta)**2) &
        &+ sqrt(pi) * extErfc(0.5_dp * gg * eta))
#else
    err = c6sum/(6.0_dp * sqrt(pi)) * (1.0_dp / eta**6) &
        &*(gg * eta * exp(-1.0_dp * (0.5_dp*gg*eta)**2) &
        &+ sqrt(pi) * erfc(0.5_dp * gg * eta))
#endif
    
  end function getDispReciprocalError_



end module SlaterKirkwood
