!!* Module containing various routines for spinpolarised calculations
!!* Intended to be used with SCC switched on !
module Spin
#include "allocate.h"
#include "assert.h"
  use accuracy
  use message
  use CommonTypes
  implicit none
  private
  
  public :: buildHspin,getEnergySpin,project2direction, spinIndex
  public :: Spin_getOrbitalEquiv

  !!* Indexing array to map between Pauli spin indices and the order of storage
  !!* in sparse arrays
  integer, parameter :: spinIndex(2,2)=reshape((/1,4,3,2/),(/2,2/)) 

  interface buildHspin
    module procedure colinearHspin
    module procedure noncolinearHspin 
  end interface
  
  interface getEnergySpin
    module procedure getEnergySpin_total
    module procedure getEnergySpin_atom
  end interface getEnergySpin

  !!* Project spins back onto chosen direction(s)
  interface project2direction
    module procedure project2direction1_colinear
    module procedure project2direction2_colinear
    module procedure project2direction1_noncolinear
    module procedure project2direction2_noncolinear
    module procedure project2direction_DM
  end interface


contains
  
  !!* Construct the collinear Spin contribution to the Hamiltonian
  !!* @param Hspin The Spin Hamiltonian contribution
  !!* @param over The overlap
  !!* @param nNeigh number of surrounding neighbours for each atom
  !!* @param iNeigh list of surrounding neighbours for each atom
  !!* @param specie list of the specie for each atom
  !!* @param orb Orbital information
  !!* @param iPair indexing array for the Hamiltonian
  !!* @param img2CentCell Indexing array to fold from periodic image atom
  !!* numbers back to the central cell number
  !!* @param qOrbital Mulliken charges, both spin components are used !
  !!* @param ww array with W for all angular momentum combinations on all
  !!* species
  subroutine colinearHspin(Hspin, over, nNeigh, iNeigh, specie, orb, iPair, &
      &img2CentCell, qOrbital, ww)    
    real(dp), intent(out) :: Hspin(:)
    real(dp), intent(in) :: over(:)
    integer, intent(in) :: nNeigh(:)
    integer, intent(in) :: iNeigh(0:,:)
    integer, intent(in) :: specie(:)
    type(TOrbitals), intent(in) :: orb
    integer, intent(in) :: iPair(0:,:)
    integer, intent(in) :: img2CentCell(:)
    real(dp), intent(in) :: qOrbital(:,:,:)
    real(dp), intent(in) :: ww(:,:,:)

    real(dp) :: Htmp(orb%mOrb,orb%mOrb), Stmp(orb%mOrb,orb%mOrb)
    integer :: iAt1, iAt2f, iNeigh1, iSp1, iSp2, iSh1, iSh2
    integer :: iSh1l, iSh1r, iSh2l, iSh2r
    integer :: iOrig, nOrb1, nOrb2
    real(dp) :: deltaQ(size(qOrbital,dim=1),size(qOrbital,dim=2))
    real(dp) :: shiftPerL(orb%mShell,size(qOrbital,dim=2))
    real(dp) :: deltaQPerLShell(orb%mShell,size(qOrbital,dim=2))
    integer :: nAtom


    !ASSERT(size(qOrbital,dim=1) == (maxval(mAngSpecies)+1)**2)
    ASSERT(size(qOrbital,dim=3) == 2)
    ASSERT(size(nNeigh)==size(qOrbital,dim=2))
    ASSERT(size(over) == size(HSpin,dim=1))
    !ASSERT(all(shape(w)\
    !  == (/maxval(mAngSpecies)+1,maxval(mAngSpecies)+1,maxval(specie)/)))
    
    nAtom = size(qOrbital,dim=2)    
    Hspin(:) = 0.0_dp    

    deltaQ(:,:) = (qOrbital(:,:,1) - qOrbital(:,:,2))
    deltaQPerLShell(:,:) = 0.0_dp    
    do iAt1 = 1, nAtom
      iSp1 = specie(iAt1)
      do iSh1 = 1, orb%nShell(iSp1)
        deltaQPerLShell(iSh1,iAt1) = &
            &sum(deltaQ(orb%posShell(iSh1,iSp1):orb%posShell(iSh1+1,iSp1)-1,&
            &iAt1))
      end do
    end do
    
    shiftPerL(:,:) = 0.0_dp    
    do iAt1 = 1, nAtom
      iSp1 = specie(iAt1)
      do iSh1 = 1, orb%nShell(iSp1)
        do iSh2 = 1, orb%nShell(iSp1)
          shiftPerL(iSh1,iAt1) =  shiftPerL(iSh1,iAt1) + &
              & deltaQPerLShell(iSh2,iAt1) * ww(iSh1,iSh2,iSp1)
        end do
      end do
    end do

    do iAt1 = 1, nAtom
      iSp1 = specie(iAt1)
      nOrb1 = orb%nOrbSpecie(iSp1)
      do iNeigh1 = 0, nNeigh(iAt1)
        iAt2f = img2CentCell(iNeigh(iNeigh1,iAt1))
        iSp2 = specie(iAt2f)
        nOrb2 = orb%nOrbSpecie(iSp2)
        iOrig = iPair(iNeigh1,iAt1)

        Stmp(:,:) = 0.0_dp
        Stmp(1:nOrb2,1:nOrb1) = &
            & reshape(over(iOrig+1:iOrig+nOrb2*nOrb1), (/nOrb2,nOrb1/))
        Htmp(:,:) = 0.0_dp
        do iSh1 = 1, orb%nShell(iSp1)
          iSh1l = orb%posShell(iSh1,iSp1)
          iSh1r = orb%posShell(iSh1+1,iSp1) - 1
          do iSh2 = 1, orb%nShell(iSp2)
            iSh2l = orb%posShell(iSh2,iSp2)
            iSh2r = orb%posShell(iSh2+1,iSp2) - 1

            Htmp(iSh2l:iSh2r,iSh1l:iSh1r) = &
                &0.5_dp * Stmp(iSh2l:iSh2r,iSh1l:iSh1r) * &
                &(shiftPerL(iSh1,iAt1) + shiftPerL(iSh2,iAt2f))
          end do
        end do
        
        Hspin(iOrig+1:iOrig+nOrb2*nOrb1) = &
            & reshape(Htmp(1:nOrb2,1:nOrb1), (/nOrb2*nOrb1/))
      end do
    end do

  end subroutine colinearHspin
  

  
  !!* Construct the Spin contribution to the Hamiltonian
  !!* @param Hspin The Spin Hamiltonian contribution
  !!* @param over The overlap
  !!* @param nNeigh number of surrounding neighbours for each atom
  !!* @param iNeigh list of surrounding neighbours for each atom
  !!* @param specie list of the specie for each atom
  !!* @param mAngSpecies maximum angular momentum for each species
  !!* @param iPair indexing array for the Hamiltonian
  !!* @param img2CentCell Indexing array to fold from periodic image atom
  !!* numbers back to the central cell number
  !!* @param q Mulliken charges, all spin components are used!!
  !!* @param W array with W for all angular momentum combinations on all
  !!* species
  subroutine noncolinearHspin(Hspin,over,nNeigh,iNeigh,specie,mAngSpecies,&
      &iPair,img2CentCell,q,W)    
    complex(dp), intent(out) :: Hspin(:,:)
    real(dp), intent(in)     :: over(:)
    integer,  intent(in)     :: nNeigh(:)
    integer,  intent(in)     :: iNeigh(0:,:)
    integer,  intent(in)     :: specie(:)
    integer,  intent(in)     :: mAngSpecies(:)
    integer,  intent(in)     :: iPair(0:,:)
    integer,  intent(in)     :: img2CentCell(:)
    real(dp), intent(in)     :: q(:,:,0:) ! beware index from 0!
    real(dp), intent(in)     :: W(:,:,:)
    
    complex(dp) :: Htmp((maxval(mAngSpecies)+1)**2,(maxval(mAngSpecies)+1)**2)
    real(dp)    :: Stmp((maxval(mAngSpecies)+1)**2,(maxval(mAngSpecies)+1)**2)
    complex(dp) :: shiftPerL(maxval(mAngSpecies)+1,size(q,dim=2),4)
    real(dp)    :: PPerLShell(maxval(mAngSpecies)+1,size(q,dim=2),3)
    integer     :: ii, jj, kk, ll
    integer     :: iOrig, mAng12, mAng22
    integer     :: nAtom
    integer     :: iAlpha, iBeta, iIndex
    
    ASSERT(size(q,dim=1) == (maxval(mAngSpecies)+1)**2)
    ASSERT(size(q,dim=3) == 4)
    ASSERT(size(nNeigh)==size(q,dim=2))
    ASSERT(all(shape(Hspin) == (/size(over),4/)))
    ASSERT(all(shape(w)\
        == (/maxval(mAngSpecies)+1,maxval(mAngSpecies)+1,maxval(specie)/)))
        
    nAtom = size(q,dim=2)
    
    Hspin(:,:) = cmplx(0.0,0.0,dp)   
    
    PPerLShell(:,:,:) = 0.0_dp
    do kk=1,3
      do ii=1,nAtom
        do jj = 0, mAngSpecies(specie(ii))
          PPerLShell(jj+1,ii,kk) = sum(q(jj**2+1:(jj+1)**2 ,ii,kk))
        end do
      end do
    end do

    shiftPerL(:,:,:) = cmplx(0.0,0.0,dp)
    do ii= 1, nAtom
      do jj= 1, mAngSpecies(specie(ii)) + 1
        do kk= 1, mAngSpecies(specie(ii)) + 1
          shiftPerL(jj,ii,spinIndex(1,1)) = &
              & shiftPerL(jj,ii,spinIndex(1,1)) + &
              & cmplx(W(jj,kk,specie(ii))* &
              & PPerLShell(kk,ii,3),0.0_dp,dp)
          shiftPerL(jj,ii,spinIndex(2,2)) = &
              & shiftPerL(jj,ii,spinIndex(2,2)) - &
              & cmplx(W(jj,kk,specie(ii))* &
              & PPerLShell(kk,ii,3),0.0_dp,dp)
          shiftPerL(jj,ii,spinIndex(1,2)) = &
              & shiftPerL(jj,ii,spinIndex(1,2)) + &
              & cmplx(W(jj,kk,specie(ii)),0.0_dp,dp) * &
              & cmplx(PPerLShell(kk,ii,2),-PPerLShell(kk,ii,1),dp)
          shiftPerL(jj,ii,spinIndex(2,1)) = &
              & shiftPerL(jj,ii,spinIndex(2,1)) + &
              & cmplx(W(jj,kk,specie(ii)),0.0_dp,dp) * &
              & cmplx(PPerLShell(kk,ii,2),+PPerLShell(kk,ii,1),dp)
        end do
      end do
    end do

    do iAlpha =1,2
      do iBeta =1,2
        iIndex = spinIndex(iAlpha,iBeta)
        do ii=1,nAtom
          mAng12 = (mAngSpecies(specie(ii)) + 1)**2
          do jj = 0, nNeigh(ii)
            mAng22 = (mAngSpecies(specie(iNeigh(jj,ii))) + 1)**2
            iOrig = iPair(jj,ii)            
            Stmp(:,:) = 0.0_dp
            Stmp(1:mAng22,1:mAng12) = &
                & reshape(over(iOrig+1:iOrig+mAng22*mAng12),(/mAng22, mAng12/))
            Htmp(:,:) = cmplx(0.0_dp,0.0_dp,dp)
            do kk = 0, mAngSpecies(specie(ii))
              do ll = 0, mAngSpecies(specie(iNeigh(jj,ii)))
                Htmp((ll**2+1):(ll+1)**2,(kk**2+1):(kk+1)**2) = &
                    & 0.5_dp* &
                    & cmplx(Stmp((ll**2+1):(ll+1)**2,(kk**2+1):(kk+1)**2),0&
                    & ,dp) * &
                    & (shiftPerL(kk+1,ii,iIndex) + shiftPerL(ll+1, &
                    & img2CentCell(iNeigh(jj,ii)),iIndex))
              end do
            end do
            Hspin(iOrig + 1 : iOrig + mAng22*mAng12,iIndex) = &
                & reshape(Htmp(1:mAng22,1:mAng12),(/mAng22*mAng12/))
          end do
        end do
      end do
    end do
  end subroutine noncolinearHspin

  

  !!* Returns the total energy contribution of the spin polarisation
  !!* @param rslt The energy contribution on return.
  !!* @param orb Information about the orbitals and their angular momenta
  !!* @param specie Species of each atom
  !!* @param qOrbital population of the orbitals.
  !!* @param ww Spin coupling constants
  subroutine getEnergySpin_total(rslt, orb, specie, qOrbital, ww)    
    real(dp), intent(out) :: rslt
    type(TOrbitals), intent(in) :: orb
    integer, intent(in):: specie(:)
    real(dp), intent(in) :: qOrbital(:,:,:)
    real(dp), intent(in) :: ww(:,:,:)

    integer  :: nAtom
    real(dp), allocatable :: rslt_atom(:)
    
    ASSERT(size(qOrbital,dim=1) == orb%mOrb)
    ASSERT((size(qOrbital,dim=3) == 4) .or. (size(qOrbital,dim=3) == 2))
    ASSERT(all(shape(ww) == (/orb%mShell,orb%mShell,maxval(specie)/)))
    
    rslt = 0.0_dp
    nAtom = size(qOrbital,dim=2)

    ALLOCATE_(rslt_atom, (nAtom))

    call getEnergySpin_atom(rslt_atom, orb, specie, qOrbital, ww)
    
    rslt = sum(rslt_atom)
    DEALLOCATE_(rslt_atom)

  end subroutine getEnergySpin_total



  !!* Atom resolved part of the spin energy
  !!* @param rslt Contains the atomic contributions on exit
  !!* @param orb Information about the orbitals and their angular momenta
  !!* @param specie Species of each atom
  !!* @param qOrbital Orbital populations
  !!* @param ww Spin coupling constants.
  subroutine getEnergySpin_atom(rslt, orb, specie, qOrbital, ww)    
    real(dp), intent(out) :: rslt(:)
    type(TOrbitals), intent(in) :: orb
    integer,  intent(in) :: specie(:)
    real(dp), intent(in) :: qOrbital(:,:,:)
    real(dp), intent(in) :: ww(:,:,:)

    integer  :: iAt1, iSp1, iSh1, iSh2, iS, iSh1l, iSh1r, iSh2l, iSh2r
    real(dp),allocatable :: deltaQ(:,:)
    real(dp),allocatable :: deltaQPerLShell(:,:)    

    integer  :: nAtom

    ASSERT(size(qOrbital,dim=1) == orb%mOrb)
    ASSERT((size(qOrbital,dim=3) == 4) .or. (size(qOrbital,dim=3) == 2))
    ASSERT(all(shape(ww) == (/orb%mShell,orb%mShell,maxval(specie)/)))
    
    rslt = 0.0_dp
    nAtom = size(qOrbital,dim=2)

    rslt(:) = 0.0_dp

    if (size(qOrbital,dim=3) == 2) then
      ! collinear case, z-axis quantised
      ALLOCATE_(deltaQ, (size(qOrbital,dim=1),nAtom))
      ALLOCATE_(deltaQPerLShell, (orb%mShell,nAtom))
      
      deltaQ(:,:) = (qOrbital(:,:,1) - qOrbital(:,:,2))

      deltaQPerLShell(:,:) = 0.0_dp
      do iAt1 = 1, nAtom
        iSp1 = specie(iAt1)
        do iSh1 = 1, orb%nShell(iSp1)
          deltaQPerLShell(iSh1,iAt1) = &
              &sum(deltaQ(orb%posShell(iSh1,iSp1):orb%posShell(iSh1+1,iSp1)-1,&
              &iAt1))
        end do
      end do

      do iAt1 = 1, nAtom
        iSp1 = specie(iAt1)
        do iSh1 = 1, orb%nShell(iSp1)
          do iSh2 = 1, orb%nShell(iSp1)
            ! note sign change in gamma
            rslt(iAt1) = rslt(iAt1) + 0.5_dp * ww(iSh2,iSh1,iSp1) &
                &* deltaQPerLShell(iSh1,iAt1) * deltaQPerLShell(iSh2,iAt1)
          end do
        end do
      end do
      DEALLOCATE_(deltaQ)
      DEALLOCATE_(deltaQPerLShell)
    else
      !! non-collinear case
      do iS = 2, 4
        do iAt1 = 1, nAtom
          iSp1 = specie(iAt1)
          do iSh1 = 1, orb%nShell(iSp1)
            iSh1l = orb%posShell(iSh1, iSp1)
            iSh1r = orb%posShell(iSh1+1, iSp1) - 1
            do iSh2 = 1, orb%nShell(iSp1)
              iSh2l = orb%posShell(iSh2,iSp1)
              iSh2r = orb%posShell(iSh2+1,iSp1) - 1
              !! note sign change in gamma
              rslt(iAt1) = rslt(iAt1) &
                  &+ 0.5_dp * ww(iSh2,iSh1,iSp1) &
                  &* sum(qOrbital(iSh1l:iSh1r,iAt1,iS)) &
                  &* sum(qOrbital(iSh2l:iSh2r,iAt1,iS))
            end do
          end do
        end do
      end do
    end if
           
  end subroutine getEnergySpin_atom


  
  subroutine project2direction1_colinear(q,direction)
    real(dp), intent(inout) :: q(:,:)
    real(dp), intent(in)    :: direction(3)

    integer  :: ii
    real(dp) :: norm(3), mag, projection

    ASSERT(size(q,dim=2) == 4)

    norm(:) = direction(:) / sqrt(sum(direction(:)**2))

    do ii =1, size(q,dim=1)
      mag = sqrt(sum(q(ii,2:4)**2))
      projection = dot_product(q(ii,2:4),norm(:))
      if (projection < 0.0_dp) then
        q(ii,2:4) = -mag * norm(:)
      else
        q(ii,2:4) = mag * norm(:)
      end if
    end do

  end subroutine project2direction1_colinear



  subroutine project2direction2_colinear(q,direction)
    real(dp), intent(inout) :: q(:,:,:)
    real(dp), intent(in)    :: direction(3)

    integer  :: ii, jj
    real(dp) :: norm(3), mag, projection

    ASSERT(size(q,dim=3) == 4)

    norm(:) = direction(:) / sqrt(sum(direction(:)**2))

    do ii =1, size(q,dim=2)
      do jj = 1, size(q,dim=1)
        mag = sqrt(sum(q(jj,ii,2:4)**2))
        projection = dot_product(q(jj,ii,2:4),norm(:))
        if (projection < 0.0_dp) then
          q(jj,ii,2:4) = -mag * norm(:)
        else
          q(jj,ii,2:4) = mag * norm(:)
        end if
      end do
    end do

  end subroutine project2direction2_colinear



  subroutine project2direction1_noncolinear(q,direction)
    real(dp), intent(inout) :: q(:,:)
    real(dp), intent(in)    :: direction(:,:)

    integer  :: ii
    real(dp) :: norm(3), mag, projection

    ASSERT(size(q,dim=2) == 4)
    ASSERT(all(shape(q) == shape(direction)))

    do ii =1, size(q,dim=1)
      norm(:) = direction(ii,:) / sqrt(sum(direction(ii,:)**2))
      mag = sqrt(sum(q(ii,2:4)**2))
      projection = dot_product(q(ii,2:4),norm(:))
      if (projection < 0.0_dp) then
        q(ii,2:4) = -mag * norm(:)
      else
        q(ii,2:4) = mag * norm(:)
      end if
    end do

  end subroutine project2direction1_noncolinear



  subroutine project2direction2_noncolinear(q,direction)
    real(dp), intent(inout) :: q(:,:,:)
    real(dp), intent(in)    :: direction(:,:,:)

    integer  :: ii, jj
    real(dp) :: norm(3), mag, projection

    ASSERT(size(q,dim=3) == 4)
    ASSERT(all(shape(q) == shape(direction)))

    do ii = 1, size(q,dim=2)
      do jj = 1, size(q,dim=1)
        norm(:) = direction(jj,ii,:) / sqrt(sum(direction(jj,ii,:)**2))
        mag = sqrt(sum(q(jj,ii,2:4)**2))
        projection = dot_product(q(jj,ii,2:4),norm(:))
        if (projection < 0.0_dp) then
          q(jj,ii,2:4) = -mag * norm(:)
        else
          q(jj,ii,2:4) = mag * norm(:)
        end if
      end do
    end do

  end subroutine project2direction2_noncolinear



  subroutine project2direction_DM(DM,direction)
    complex(dp), intent(inout) :: DM(:,:)
    real(dp), intent(in)       :: direction(3)

    integer  :: ii
    real(dp) :: norm(3), projection
    real(dp) :: mag(2), vect(3)

    ASSERT(size(DM,dim=2) == 4)

    norm(:) = direction(:) / sqrt(sum(direction(:)**2))

    do ii =1, size(DM,dim=1)       
      mag(1) = real(DM(ii,spinIndex(1,1))+DM(ii,spinIndex(2,2)))
      vect(1) = aimag(DM(ii,spinIndex(2,1))-DM(ii,spinIndex(1,2)))
      vect(2) = real(DM(ii,spinIndex(2,1))+DM(ii,spinIndex(1,2)))
      vect(3) = real(DM(ii,spinIndex(1,1))-DM(ii,spinIndex(2,2)))
      mag(2) = sqrt(sum(vect(:)**2))              
      projection = dot_product(vect(:),norm(:))
      if (projection < 0.0_dp) then
        vect(:) = -mag(2) * norm(:)
      else
        vect(:) = mag(2) * norm(:)
      end if
      DM(ii,spinIndex(1,1)) = cmplx(0.5_dp*(mag(1)+vect(3)),0.0_dp,dp)
      DM(ii,spinIndex(2,2)) = cmplx(0.5_dp*(mag(1)-vect(3)),0.0_dp,dp)
      DM(ii,spinIndex(1,2)) = 0.5_dp*cmplx(vect(2),-vect(1),dp)
      DM(ii,spinIndex(2,1)) = 0.5_dp*cmplx(vect(2),vect(1),dp)
    end do

  end subroutine project2direction_DM


  
  !!* Returns the equivalence between the orbitals in the spin interaction.
  !!* @param orb  Information about the orbitals and their angular momenta
  !!* @param specie Species of each atom
  !!* @param equiv The equivalence vector on return.
  !!* @note The current version assumes, that no shells only the orbitals 
  !!* inside each shell are equivalent, which is in most cases true anyway.
  !!* @todo Proper analysis of the spin coupling constants to watch for
  !!*   eventual equivalence.
  subroutine Spin_getOrbitalEquiv(orb, specie, equiv)
    type(TOrbitals), intent(in) :: orb
    integer, intent(in) :: specie(:)
    integer, intent(out) :: equiv(:,:,:)

    integer :: nAtom, nSpin
    integer :: iAt, iOrb, iS, ind, iSp

    nAtom = size(equiv, dim=2)
    nSpin = size(equiv, dim=3)

    ASSERT(size(equiv, dim=1) == orb%mOrb)

    equiv(:,:,:) = 0
    ind = 1
    do iAt = 1, nAtom
      iSp = specie(iAt)
      do iOrb = 1, orb%nOrbSpecie(iSp)
        equiv(iOrb, iAt, 1) = ind + orb%iShellOrb(iOrb, iSp) - 1
      end do
      ind = ind + orb%nShell(iSp)
    end do
    do iS = 2, nSpin
      ind = maxval(equiv)
      where (equiv(:,:,1) /= 0)
        equiv(:,:,iS) = equiv(:,:,1) + ind
      end where
    end do

  end subroutine Spin_getOrbitalEquiv

    
end module Spin
