!!* Contains high level functions for converting the values in a XML/HSD 
!!* DOM-tree to Fortran intrinsic types.
!!* @todo Some more routines for complex numbers?
module HSDUtils
#include "assert.h"
#include "allocate.h"  
  use TokenReader
  use HSDParser
  use XMLUtils
  use CharManip
  use flib_dom
  use flib_wxml
  use m_strings
  use message
  use LinkedList
  use StringList
  use accuracy
  implicit none
  private

  public :: checkError, detailedError, detailedWarning
  public :: getFirstTextChild, getChildValue, setChildValue
  public :: writeChildValue, getAsString
  public :: getValueFromListOrRange, appendPathAndLine
  public :: getChild, getChildren, setChild
  public :: attrProcessed


  !!* Returns the value (the child) of a child node identified by its name.
  !!* @desc
  !!*   These routines investigate the provided node and look for a child with
  !!*   the supplied name. Is this child found, its child (which should be a
  !!*   single text node or a usual node if the value argument is of type node)
  !!*   is returned as value converted to the appropriate type. If the child is
  !!*   not found, an error is raised, unless a default value was specified.In
  !!*   that case, a child is created with the provided name and is appended to
  !!*   the node. Furthermore a text node containing the string converted
  !!*   default value is appended to the child node. If default value is
  !!*   provided, it must be also indicated, if the created child is only
  !!*   allowed to have one further child or not. (This corresponds to an
  !!*   assignment with '=' in the HSD input.) If the child (identified by the
  !!*   provided name) is allowed to have a modifier, an argument for the
  !!*   modifier must be provided to contain the the parsed value on return. If
  !!*   the argument for the modifier is missing, but a modifier is found, the
  !!*   program raises an error. The pointer to the found (or created) child can
  !!*   be queried through an appropriate argument. If the name of the child to
  !!*   look for is an empty string, the passed node itself is treated as if it
  !!*   would be the child, which had been found.
  interface getChildValue
    module procedure getChVal_logical
    module procedure getChVal_node
    module procedure getChVal_string
    module procedure getChVal_lString
    module procedure getChVal_lReal
    module procedure getChVal_lRealR1
    module procedure getChVal_lInt
    module procedure getChVal_lIntR1
    module procedure getChVal_real
    module procedure getChVal_realR1
    module procedure getChVal_realR2
    module procedure getChVal_int
    module procedure getChVal_intR1
    module procedure getChVal_intR2
    module procedure getChVal_lIntR1RealR1
    module procedure getChVal_lStringIntR1RealR1
  end interface


  !!* Sets the value (the child) of a child node identified by its name
  !!* @desc
  !!*   Those functions are the inverse of the getChildValue functions. They
  !!*   create a child with the provided name and append to that child a
  !!*   text node (or a normal node, if the provided value is of type node)
  !!*   containing the provided value. It must be indicated, if the
  !!*   created child is allowed to have only one single further child. If a
  !!*   child with the specified name already exists, the program raises an 
  !!*   error, unless replacement flag is set on .true.. In that case, the
  !!*   the existing child is replaced. If the name of the child is the empty
  !!*   string, the current node is treated as if it would be the child, which
  !!*   had been found.
  interface setChildValue
    module procedure setChVal_logical
    module procedure setChVal_node
    module procedure setChVal_char
    module procedure setChVal_charR1
    module procedure setChVal_real
    module procedure setChVal_realR1
    module procedure setChVal_realR2
    module procedure setChVal_int
    module procedure setChVal_intR1
    module procedure setChVal_intR2
    module procedure setChVal_intR2RealR2
    module procedure setChVal_charR1intR2RealR2
  end interface


  !!* Writes a child and its value to an xml-write stream
  interface writeChildValue
    module procedure writeChVal_logical
    module procedure writeChVal_real
    module procedure writeChVal_realR1
    module procedure writeChVal_realR2
    module procedure writeChVal_int
    module procedure writeChVal_intR1
    module procedure writeChVal_intR2
    module procedure writeChVal_intR2RealR2
    module procedure writeChVal_charR1
    module procedure writeChVal_charR1IntR2RealR2
  end interface


  !!* Returns a string representation of an object
  interface getAsString
    module procedure getAsString_logical
    module procedure getAsString_real
    module procedure getAsString_realR1
    module procedure getAsString_realR2
    module procedure getAsString_int
    module procedure getAsString_intR1
    module procedure getAsString_intR2
    module procedure getAsString_intR2RealR2
    module procedure getAsString_charR1
    module procedure getAsString_charR1IntR2RealR2
  end interface


  !! Error messages
  character(len=*), parameter :: MSG_MISSING_FIELD = "Missing child: "
  character(len=*), parameter :: MSG_EXISTING_CHILD = "Already existing child: "
  character(len=*), parameter :: MSG_NOMODIFIER = "Entity is not allowed to &
      &have a modifier"
  character(len=*), parameter :: MSG_MISSING_VALUES = "Not enough values &
      &provided."

  !!* Length of a line (for wrapping long lines when writing values)
  integer, parameter :: lineLength = 80

  !!* Maximal number of characters needed to represent an integer
  integer, parameter :: nCharInt = 50

  !!* Maximal number of characters needed to represent a real number
  integer, parameter :: nCharReal = 50

  !!* Attribute signalising that a tag was processed
  character(len=*), parameter :: attrProcessed = "proc"

  !!* Preallocateated size for temporary buffer strings
  integer, parameter :: preAllocSize = 1024


contains

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! getChildValue and getNode routines
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  
  !!* Returns the value (the child) of a child node as logical.
  !!* @param node     The node to investigate.
  !!* @param name     Name of the child to look for
  !!* @param value    Value on return
  !!* @param default  Default value for the child, if child is not found
  !!* @param modifier Modifier of the child on return
  !!* @param child    Pointer to the child node (with the spec. name) on return
  subroutine getChVal_logical(node, name, value, default, modifier, child)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    logical, intent(out) :: value
    logical, intent(in), optional :: default
    type(string), intent(inout), optional :: modifier
    type(fnode), pointer, optional :: child

    type(string) :: text, modif
    integer :: iStart, iErr
    type(fnode), pointer :: child2

    ASSERT(associated(node))

    child2 => getFirstChildByName(node, tolower(name))
    if (associated(child2)) then
      call getAttribute(child2, attrModifier, modif)
      if (present(modifier)) then
        modifier = modif
      elseif (len(modif) > 0) then
        call detailedError(child2, MSG_NOMODIFIER)
      end if
      call unstring(modif)
      iStart = 1
      call getFirstTextChild(child2, text)
      call getNextToken(char(text), value, iStart, iErr)
      call unstring(text)
      call checkError(child2, iErr, "Invalid logical value")
      call setAttribute(child2, attrProcessed, "")
    elseif (present(default)) then
      value = default
      if (present(modifier)) then
        modifier = ""
      end if
      call setChildValue(node, name, value, .false., child=child2)
    else
      call detailedError(node, MSG_MISSING_FIELD // name)
    end if
    if (present(child)) then
      child => child2
    end if
    
  end subroutine getChVal_logical
  


  !!* Returns the value (the child) of a child node as string.
  !!* @param node     The node to investigate.
  !!* @param name     Name of the child to look for
  !!* @param value    Value on return
  !!* @param default  Default value for the child, if child is not found
  !!* @param modifier Modifier of the child on return
  !!* @param child    Pointer to the child node (with the spec. name) on return
  subroutine getChVal_string(node, name, value, default, modifier, child)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    type(string), intent(inout) :: value
    character(len=*), intent(in), optional :: default
    type(string), intent(inout), optional :: modifier
    type(fnode), pointer, optional :: child

    type(string) :: text, modif
    integer :: iStart, iErr
    type(fnode), pointer :: child2

    ASSERT(associated(node))
    
    child2 => getFirstChildByName(node, tolower(name))
    if (associated(child2)) then
      call getAttribute(child2, attrModifier, modif)
      if (present(modifier)) then
        modifier = modif
      elseif (len(modif) > 0) then
        call detailedError(child2, MSG_NOMODIFIER)
      end if
      call unstring(modif)
      iStart = 1
      call getFirstTextChild(child2, text)
      call getNextToken(char(text), value, iStart, iErr)
      call unstring(text)
      call checkError(child2, iErr, "Invalid string value")
      call setAttribute(child2, attrProcessed, "")
    elseif (present(default)) then
      value = default
      if (present(modifier)) then
        modifier = ""
      end if
      call setChildValue(node, name, default, .false., child=child2)
    else
      call detailedError(node, MSG_MISSING_FIELD // name)
    end if
    if (present(child)) then
      child => child2
    end if

  end subroutine getChVal_string



  !!* Returns the value (the child) of a child node as real.
  !!* @param node     The node to investigate.
  !!* @param name     Name of the child to look for
  !!* @param value    Value on return
  !!* @param default  Default value for the child, if child is not found
  !!* @param modifier Modifier of the child on return
  !!* @param child    Pointer to the child node (with the spec. name) on return
  subroutine getChVal_real(node, name, value, default, modifier, child)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    real(dp), intent(out) :: value
    real(dp), intent(in), optional :: default
    type(string), intent(inout), optional :: modifier
    type(fnode), pointer, optional :: child

    type(string) :: text, modif
    integer :: iStart, iErr
    type(fnode), pointer :: child2

    ASSERT(associated(node))

    child2 => getFirstChildByName(node, tolower(name))
    if (associated(child2)) then
      call getAttribute(child2, attrModifier, modif)
      if (present(modifier)) then
        modifier = modif
      elseif (len(modif) > 0) then
        call detailedError(child2, MSG_NOMODIFIER)
      end if
      call unstring(modif)
      iStart = 1
      call getFirstTextChild(child2, text)
      call getNextToken(char(text), value, iStart, iErr)
      call unstring(text)
      call checkError(child2, iErr, "Invalid real value")
      call setAttribute(child2, attrProcessed, "")
    elseif (present(default)) then
      value = default
      if (present(modifier)) then
        modifier = ""
      end if
      call setChildValue(node, name, value, .false., child=child2)
    else
      call detailedError(node, MSG_MISSING_FIELD // name)
    end if
    call setAttribute(child2, attrProcessed, "")
    if (present(child)) then
      child => child2
    end if

  end subroutine getChVal_real



  !!* Returns the value (the child) of a child node as a rank one real array.
  !!* @param node     The node to investigate.
  !!* @param name     Name of the child to look for
  !!* @param value    Value on return
  !!* @param default  Default value for the child, if child is not found
  !!* @param nItem    Nr. of read items. If this argument is not passed, and
  !!*   the nr. of read items is less than the size of the array, the
  !!*   subroutine raises an error.
  !!* @param modifier Modifier of the child on return
  !!* @param child    Pointer to the child node (with the spec. name) on return
  subroutine getChVal_realR1(node, name, value, default, nItem, modifier, &
      &child)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    real(dp), intent(out) :: value(:)
    real(dp), intent(in), optional :: default(:)
    integer, intent(out), optional :: nItem
    type(string), intent(inout), optional :: modifier
    type(fnode), pointer, optional :: child

    type(string) :: text, modif
    integer :: iStart, iErr, nReadItem
    type(fnode), pointer :: child2

    ASSERT(associated(node))
    ASSERT_ENV(if (present(default)) then)
    ASSERT_ENV(ASSERT(all(shape(default) == shape(value))))
    ASSERT_ENV(end if)
    
    if (present(nItem)) then
      nItem = 0
    end if
    child2 => getFirstChildByName(node, tolower(name))
    if (associated(child2)) then
      call getAttribute(child2, attrModifier, modif)
      if (present(modifier)) then
        modifier = modif
      elseif (len(modif) > 0) then
        call detailedError(child2, MSG_NOMODIFIER)
      end if
      call unstring(modif)
      iStart = 1
      call getFirstTextChild(child2, text)
      call getNextToken(char(text), value, iStart, iErr, nReadItem)
      call unstring(text)
      call checkError(child2, iErr, "Invalid real value")
      if (present(nItem)) then
        nItem = nReadItem
      elseif (nReadItem /= size(value)) then
        call detailedError(node, MSG_MISSING_VALUES)
      end if
      call setAttribute(child2, attrProcessed, "")
    elseif (present(default)) then
      value = default
      if (present(modifier)) then
        modifier = ""
      end if
      call setChildValue(node, name, value, .false., child=child2)
    else
      call detailedError(node, MSG_MISSING_FIELD // name)
    end if
    call setAttribute(child2, attrProcessed, "")
    if (present(child)) then
      child => child2
    end if

  end subroutine getChVal_realR1



  !!* Returns the value (the child) of a child node as a rank two real array.
  !!* @param node     The node to investigate.
  !!* @param name     Name of the child to look for
  !!* @param value    Value on return
  !!* @param default  Default value for the child, if child is not found
  !!* @param nItem    Nr. of read items. If this argument is not passed, and
  !!*   the nr. of read items is less than the size of the array, the
  !!*   subroutine raises an error.
  !!* @param modifier Modifier of the child on return
  !!* @param child    Pointer to the child node (with the spec. name) on return
  !!* @note This is just a wrapper around the rank one version, to make sure
  !!*   that two dimensional arrays are pretty printed. For higher ranked
  !!*   arrays the rank one version should be used with some reshaping after.
  subroutine getChVal_realR2(node, name, value, default, nItem, modifier, &
      &child)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    real(dp), intent(out) :: value(:,:)
    real(dp), intent(in), optional :: default(:,:)
    integer, intent(out), optional :: nItem
    type(string), intent(inout), optional :: modifier
    type(fnode), pointer, optional :: child

    real(dp) :: buffer(size(value))
    integer :: nReadItem
    type(string) :: modif
    type(fnode), pointer :: child2

    ASSERT(associated(node))
    ASSERT_ENV(if (present(default)) then)
    ASSERT_ENV(ASSERT(all(shape(default) == shape(value))))
    ASSERT_ENV(end if)

    nReadItem = 0
    value = 0.0_dp
    if (present(default)) then
      call getChildValue(node, name, buffer, reshape(default, shape(buffer)), &
          &nReadItem, modifier=modif, child=child2)
    else
      call getChildValue(node, name, buffer, nItem=nReadItem, modifier=modif, &
          &child=child2)
    end if
    if (present(nItem)) then
      nItem = nReadItem
    elseif (nReadItem /= size(value)) then
      call detailedError(node, MSG_MISSING_VALUES)
    end if
    if (present(modifier)) then
      modifier = modif
    elseif (len(modif) > 0) then
      call detailedError(child2, MSG_NOMODIFIER)
    end if
    call unstring(modif)
    value(:,:) = reshape(buffer, shape(value))
    if (present(child)) then
      child => child2
    end if

  end subroutine getChVal_realR2



  !!* Returns the value (the child) of a child node as integer.
  !!* @param node     The node to investigate.
  !!* @param name     Name of the child to look for
  !!* @param value    Value on return
  !!* @param default  Default value for the child, if child is not found
  !!* @param modifier Modifier of the child on return
  !!* @param child    Pointer to the child node (with the spec. name) on return
  subroutine getChVal_int(node, name, value, default, modifier, child)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    integer, intent(out) :: value
    integer, intent(in), optional :: default
    type(string), intent(inout), optional :: modifier
    type(fnode), pointer, optional :: child

    type(string) :: text, modif
    integer :: iStart, iErr
    type(fnode), pointer :: child2

    ASSERT(associated(node))

    child2 => getFirstChildByName(node, tolower(name))
    if (associated(child2)) then
      call getAttribute(child2, attrModifier, modif)
      if (present(modifier)) then
        modifier = modif
      elseif (len(modif) > 0) then
        call detailedError(child2, MSG_NOMODIFIER)
      end if
      call unstring(modif)
      iStart = 1
      call getFirstTextChild(child2, text)
      call getNextToken(char(text), value, iStart, iErr)
      call unstring(text)
      call checkError(child2, iErr, "Invalid integer value")
      call setAttribute(child2, attrProcessed, "")
    elseif (present(default)) then
      value = default
      if (present(modifier)) then
        modifier = ""
      end if
      call setChildValue(node, name, value, .false., child=child2)
    else
      call detailedError(node, MSG_MISSING_FIELD // name)
    end if
    if (present(child)) then
      child => child2
    end if

  end subroutine getChVal_int



  !!* Returns the value (the child) of a child node as a rank one integer array.
  !!* @param node     The node to investigate.
  !!* @param name     Name of the child to look for
  !!* @param value    Value on return
  !!* @param default  Default value for the child, if child is not found
  !!* @param nItem    Nr. of read items. If this argument is not passed, and
  !!*   the nr. of read items is less than the size of the array, the
  !!*   subroutine raises an error.
  !!* @param modifier Modifier of the child on return
  !!* @param child    Pointer to the child node (with the spec. name) on return
  subroutine getChVal_intR1(node, name, value, default, nItem, modifier, &
      &child)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    integer, intent(out) :: value(:)
    integer, intent(in), optional :: default(:)
    integer, intent(out), optional :: nItem
    type(string), intent(inout), optional :: modifier
    type(fnode), pointer, optional :: child

    type(string) :: text, modif
    integer :: iStart, iErr, nReadItem
    type(fnode), pointer :: child2

    ASSERT(associated(node))
    ASSERT_ENV(if (present(default)) then)
    ASSERT_ENV(ASSERT(all(shape(default) == shape(value))))
    ASSERT_ENV(end if)

    if (present(nItem)) then
      nItem = 0
    end if
    child2 => getFirstChildByName(node, tolower(name))
    if (associated(child2)) then
      call getAttribute(child2, attrModifier, modif)
      if (present(modifier)) then
        modifier = modif
      elseif (len(modif) > 0) then
        call detailedError(child2, MSG_NOMODIFIER)
      end if
      call unstring(modif)
      iStart = 1
      call getFirstTextChild(child2, text)
      call getNextToken(char(text), value, iStart, iErr, nReadItem)
      call unstring(text)
      call checkError(child2, iErr, "Invalid integer value")
      if (present(nItem)) then
        nItem = nReadItem
      elseif (nReadItem /= size(value)) then
        call detailedError(node, MSG_MISSING_VALUES)
      end if
      call setAttribute(child2, attrProcessed, "")
    elseif (present(default)) then
      value = default
      if (present(modifier)) then
        modifier = ""
      end if
      call setChildValue(node, name, value, .false., child=child2)
    else
      call detailedError(node, MSG_MISSING_FIELD // name)
    end if
    if (present(child)) then
      child => child2
    end if

  end subroutine getChVal_intR1


  
  !!* Returns the value (the child) of a child node as a rank two integer array.
  !!* @param node     The node to investigate.
  !!* @param name     Name of the child to look for
  !!* @param value    Value on return
  !!* @param default  Default value for the child, if child is not found
  !!* @param nItem    Nr. of read items. If this argument is not passed, and
  !!*   the nr. of read items is less than the size of the array, the
  !!*   subroutine raises an error.
  !!* @param modifier Modifier of the child on return
  !!* @param child    Pointer to the child node (with the spec. name) on return
  !!* @note This is just a wrapper around the rank one version, to make sure
  !!*   that two dimensional arrays are pretty printed. For higher ranked
  !!*   arrays the rank one version should be used with some reshaping after.
  subroutine getChVal_intR2(node, name, value, default, nItem, modifier, &
      &child)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    integer, intent(out) :: value(:,:)
    integer, intent(in), optional :: default(:,:)
    integer, intent(out), optional :: nItem
    type(string), intent(inout), optional :: modifier
    type(fnode), pointer, optional :: child

    integer :: buffer(size(value))
    integer :: nReadItem
    type(string) :: modif
    type(fnode), pointer :: child2

    ASSERT(associated(node))
    ASSERT_ENV(if (present(default)) then)
    ASSERT_ENV(ASSERT(all(shape(default) == shape(value))))
    ASSERT_ENV(end if)

    nReadItem = 0
    if (present(default)) then
      call getChildValue(node, name, buffer, reshape(default, shape(buffer)), &
          &nReadItem, modif, child=child2)
    else
      call getChildValue(node, name, buffer, nItem=nReadItem, modifier=modif, &
          &child=child2)
    end if
    if (present(nItem)) then
      nItem = nReadItem
    elseif (nReadItem /= size(value)) then
      call detailedError(node, MSG_MISSING_VALUES)
    end if
    if (present(modifier)) then
      modifier = modif
    elseif (len(modif) > 0) then
      call detailedError(child2, MSG_NOMODIFIER)
    end if
    call unstring(modif)
    value(:,:) = reshape(buffer, shape(value))
    if (present(child)) then
      child => child2
    end if

  end subroutine getChVal_intR2



  !!* Returns the value (the child) of a child node as a linked list of strings.
  !!* @param node     The node to investigate.
  !!* @param name     Name of the child to look for
  !!* @param value    Value on return
  !!* @param modifier Modifier of the child on return
  !!* @param child    Pointer to the child node (with the spec. name) on return
  !!* @note In order to prevent a double packaging (from array to linked list
  !!*   and then from linked list to array), the setting of defaults for list
  !!*   types is not allowed. The presence of the child must be explicitely
  !!*   queried in the caller routine and an eventual default setting must be
  !!*   set with an explicit setChildValue call.
  subroutine getChVal_lString(node, name, value, modifier, child)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    type(listString), intent(inout) :: value
    type(string), intent(inout), optional :: modifier
    type(fnode), pointer, optional :: child

    type(string) :: text, modif
    type(fnode), pointer :: child2

    ASSERT(associated(node))
    child2 => getFirstChildByName(node, tolower(name))
    if (associated(child2)) then
      call getAttribute(child2, attrModifier, modif)
      if (present(modifier)) then
        modifier = modif
      elseif (len(modif) > 0) then
        call detailedError(child2, MSG_NOMODIFIER)
      end if
      call unstring(modif)
      call getFirstTextChild(child2, text)
      call getChVal_lString_h(char(text), value, child2)
      call unstring(text)
      call setAttribute(child2, attrProcessed, "")
    else
      call detailedError(node, MSG_MISSING_FIELD // name)
    end if
    if (present(child)) then
      child => child2
    end if

  end subroutine getChVal_lString



  !!* Helper function for getChVal_lString to avoid string to character
  !!* conversion in the do-loop.
  !!* @param text  Text to parse
  !!* @param value Contains the value of the parsed text
  subroutine getChVal_lString_h(text, value, node)
    character(len=*), intent(in) :: text
    type(listString), intent(inout) :: value
    type(fnode), pointer :: node

    integer :: iStart, iErr
    type(string) :: token
    
    iStart = 1
    call getNextToken(text, token, iStart, iErr)
    do while (iErr == TOKEN_OK)
      call append(value, trim(unquote(char(token))))
      call getNextToken(text, token, iStart, iErr)
    end do
    call unstring(token)
    if (iErr == TOKEN_ERROR) then
      call detailedError(node, "Invalid string")
    end if

  end subroutine getChVal_lString_h



  !!* Returns the value (the child) of a child node as a linked list of reals.
  !!* @param node     The node to investigate.
  !!* @param name     Name of the child to look for
  !!* @param value    Value on return
  !!* @param modifier Modifier of the child on return
  !!* @param child    Pointer to the child node (with the spec. name) on return
  !!* @note In order to prevent a double packaging (from array to linked list
  !!*   and then from linked list to array), the setting of defaults for list
  !!*   types is not allowed. The presence of the child must be explicitely
  !!*   queried in the caller routine and an eventual default setting must be
  !!*   set with an explicit setChildValue call.
  subroutine getChVal_lReal(node, name, value, modifier, child)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    type(listReal), intent(inout) :: value
    type(string), intent(inout), optional :: modifier
    type(fnode), pointer, optional :: child

    type(string) :: text, modif
    type(fnode), pointer :: child2

    ASSERT(associated(node))

    child2 => getFirstChildByName(node, tolower(name))
    if (associated(child2)) then
      call getAttribute(child2, attrModifier, modif)
      if (present(modifier)) then
        modifier = modif
      elseif (len(modif) > 0) then
        call detailedError(child2, MSG_NOMODIFIER)
      end if
      call unstring(modif)
      call getFirstTextChild(child2, text)
      call getChVal_lReal_h(char(text), value, child2)
      call unstring(text)
      call setAttribute(child2, attrProcessed, "")
    else
      call detailedError(node, MSG_MISSING_FIELD // name)
    end if
    if (present(child)) then
      child => child2
    end if

  end subroutine getChVal_lReal



  !!* Helper function for getChVal_lReal to avoid string to character
  !!* conversion in the do-loop.
  !!* @param text  Text to parse
  !!* @param value Contains the value of the parsed text
  subroutine getChVal_lReal_h(text, value, node)
    character(len=*), intent(in) :: text
    type(listReal), intent(inout) :: value
    type(fnode), pointer :: node
    
    integer :: iStart, iErr
    real(dp) :: buffer
    
    iStart = 1
    call getNextToken(text, buffer, iStart, iErr)
    do while (iErr == TOKEN_OK)
      call append(value, buffer)
      call getNextToken(text, buffer, iStart, iErr)
    end do
    if (iErr == TOKEN_ERROR) then
      call detailedError(node, "Invalid real value")
    end if

  end subroutine getChVal_lReal_h



  !!* Returns the value (the child) of a child node as a linked list of rank one
  !!* real arrays.
  !!* @param node     The node to investigate.
  !!* @param name     Name of the child to look for
  !!* @param dim      Dimension of the arrays    
  !!* @param value    Value on return
  !!* @param modifier Modifier of the child on return
  !!* @param child    Pointer to the child node (with the spec. name) on return
  !!* @note In order to prevent a double packaging (from array to linked list
  !!*   and then from linked list to array), the setting of defaults for list
  !!*   types is not allowed. The presence of the child must be explicitely
  !!*   queried in the caller routine and an eventual default setting must be
  !!*   set with an explicit setChildValue call.
  subroutine getChVal_lRealR1(node, name, dim, value, modifier, child)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    integer, intent(in) :: dim
    type(listRealR1), intent(inout) :: value
    type(string), intent(inout), optional :: modifier
    type(fnode), pointer, optional :: child

    type(string) :: text, modif
    type(fnode), pointer :: child2

    ASSERT(associated(node))

    child2 => getFirstChildByName(node, tolower(name))
    if (associated(child2)) then
      call getAttribute(child2, attrModifier, modif)
      if (present(modifier)) then
        modifier = modif
      elseif (len(modif) > 0) then
        call detailedError(child2, MSG_NOMODIFIER)
      end if
      call unstring(modif)
      call getFirstTextChild(child2, text)
      call getChVal_lRealR1_h(char(text), dim, value, child2)
      call unstring(text)
      call setAttribute(child2, attrProcessed, "")
    else
      call detailedError(node, MSG_MISSING_FIELD // name)
    end if
    if (present(child)) then
      child => child2
    end if

  end subroutine getChVal_lRealR1



  !!* Helper function for getChVal_lReal to avoid string to character
  !!* conversion in the do-loop.
  !!* @param text  Text to parse
  !!* @param value Contains the value of the parsed text
  subroutine getChVal_lRealR1_h(text, dim, value, node)
    character(len=*), intent(in) :: text
    integer, intent(in) :: dim
    type(listRealR1), intent(inout) :: value
    type(fnode), pointer :: node
    
    integer :: iStart, iErr
    real(dp) :: buffer(dim)
    integer :: nItem
    
    iStart = 1
    call getNextToken(text, buffer, iStart, iErr, nItem)
    do while (iErr == TOKEN_OK)
      call append(value, buffer)
      call getNextToken(text, buffer, iStart, iErr, nItem)
    end do
    if (iErr == TOKEN_ERROR) then
      call detailedError(node, "Invalid real value")
    elseif (iErr == TOKEN_EOS .and. nItem /= 0) then
      call detailedError(node, "Unexpected end of data")
    end if

  end subroutine getChVal_lRealR1_h



  !!* Returns the value (the child) of a child node as linked list of integers.
  !!* @param node     The node to investigate.
  !!* @param name     Name of the child to look for
  !!* @param value    Value on return
  !!* @param modifier Modifier of the child on return
  !!* @param child    Pointer to the child node (with the spec. name) on return
  !!* @note In order to prevent a double packaging (from array to linked list
  !!*   and then from linked list to array), the setting of defaults for list
  !!*   types is not allowed. The presence of the child must be explicitely
  !!*   queried in the caller routine and an eventual default setting must be
  !!*   set with an explicit setChildValue call.
  subroutine getChVal_lInt(node, name, value, modifier, child)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    type(listInt), intent(inout) :: value
    type(string), intent(inout), optional :: modifier
    type(fnode), pointer, optional :: child

    type(string) :: text, modif
    type(fnode), pointer :: child2

    ASSERT(associated(node))

    child2 => getFirstChildByName(node, tolower(name))
    if (associated(child2)) then
      call getAttribute(child2, attrModifier, modif)
      if (present(modifier)) then
        modifier = modif
      elseif (len(modif) > 0) then
        call detailedError(child2, MSG_NOMODIFIER)
      end if
      call unstring(modif)
      call getFirstTextChild(child2, text)
      call getChVal_lInt_h(char(text), value, child2)
      call unstring(text)
      call setAttribute(child2, attrProcessed, "")
    else
      call detailedError(node, MSG_MISSING_FIELD // name)
    end if
    if (present(child)) then
      child => child2
    end if

  end subroutine getChVal_lInt



  !!* Helper function for getChVal_lReal to avoid string to character
  !!* conversion in the do-loop.
  !!* @param text  Text to parse
  !!* @param value Contains the value of the parsed text
  subroutine getChVal_lInt_h(text, value, node)
    character(len=*), intent(in) :: text
    type(listInt), intent(inout) :: value
    type(fnode), pointer :: node
    
    integer :: iStart, iErr
    integer :: buffer
    
    iStart = 1
    call getNextToken(text, buffer, iStart, iErr)
    do while (iErr == TOKEN_OK)
      call append(value, buffer)
      call getNextToken(text, buffer, iStart, iErr)
    end do
    if (iErr == TOKEN_ERROR) then
      call detailedError(node, "Invalid real value")
    end if

  end subroutine getChVal_lInt_h



  !!* Returns the value (the child) of a child node as linked list of rank one
  !!* integer arrays.
  !!* @param node     The node to investigate.
  !!* @param name     Name of the child to look for
  !!* @param value    Value on return
  !!* @param modifier Modifier of the child on return
  !!* @param child    Pointer to the child node (with the spec. name) on return
  !!* @note In order to prevent a double packaging (from array to linked list
  !!*   and then from linked list to array), the setting of defaults for list
  !!*   types is not allowed. The presence of the child must be explicitely
  !!*   queried in the caller routine and an eventual default setting must be
  !!*   set with an explicit setChildValue call.
  subroutine getChVal_lIntR1(node, name, dim, value, modifier, child)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    integer, intent(in) :: dim
    type(listIntR1), intent(inout) :: value
    type(string), intent(inout), optional :: modifier
    type(fnode), pointer, optional :: child

    type(string) :: text, modif
    type(fnode), pointer :: child2

    ASSERT(associated(node))

    child2 => getFirstChildByName(node, tolower(name))
    if (associated(child2)) then
      call getAttribute(child2, attrModifier, modif)
      if (present(modifier)) then
        modifier = modif
      elseif (len(modif) > 0) then
        call detailedError(child2, MSG_NOMODIFIER)
      end if
      call unstring(modif)
      call getFirstTextChild(child2, text)
      call getChVal_lIntR1_h(char(text), dim, value, child2)
      call unstring(text)
      call setAttribute(child2, attrProcessed, "")
    else
      call detailedError(node, MSG_MISSING_FIELD // name)
    end if
    if (present(child)) then
      child => child2
    end if

  end subroutine getChVal_lIntR1



  !!* Helper function for getChVal_lReal to avoid string to character
  !!* conversion in the do-loop.
  !!* @param text  Text to parse
  !!* @param value Contains the value of the parsed text
  subroutine getChVal_lIntR1_h(text, dim, value, node)
    character(len=*), intent(in) :: text
    integer, intent(in) :: dim
    type(listIntR1), intent(inout) :: value
    type(fnode), pointer :: node
    
    integer :: iStart, iErr
    integer :: buffer(dim)
    integer :: nItem
    
    iStart = 1
    call getNextToken(text, buffer, iStart, iErr, nItem)
    do while (iErr == TOKEN_OK)
      call append(value, buffer)
      call getNextToken(text, buffer, iStart, iErr, nItem)
    end do
    if (iErr == TOKEN_ERROR) then
      call detailedError(node, "Invalid real value")
    elseif (iErr == TOKEN_EOS .and. nItem /= 0) then
      call detailedError(node, "Unexpected end of data")
    end if

  end subroutine getChVal_lIntR1_h



  !!* Returns the value (the child) of a child node as a linked list rank one
  !!* integer and rank one real arrays.
  !!* @param node      The node to investigate.
  !!* @param name      Name of the child to look for
  !!* @param dimInt    Dimension of the integer arrays in the list
  !!* @param valueInt  List of integer arrays on return
  !!* @param dimReal   Dimensio of the real arrays in the list
  !!* @param valueReal List of real array on return
  !!* @param modifier  Modifier of the child on return
  !!* @param child     Pointer to the child on return
  !!* @note In order to prevent a double packaging (from array to linked list
  !!*   and then from linked list to array), the setting of defaults for list
  !!*   types is not allowed. The presence of the child must be explicitely
  !!*   queried in the caller routine and an eventual default setting must be
  !!*   set with an explicit setChildValue call.
  subroutine getChVal_lIntR1RealR1(node, name, dimInt, valueInt, &
      &dimReal, valueReal, modifier, child)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    integer, intent(in) :: dimInt
    integer, intent(in) :: dimReal
    type(listIntR1), intent(inout) :: valueInt
    type(listRealR1), intent(inout) :: valueReal
    type(string), intent(inout), optional :: modifier
    type(fnode), pointer, optional :: child

    type(string) :: text, modif
    type(fnode), pointer :: child2

    ASSERT(associated(node))
    ASSERT(dimInt > 0)
    ASSERT(dimReal > 0)

    child2 => getFirstChildByName(node, tolower(name))
    if (associated(child2)) then
      call getAttribute(child2, attrModifier, modif)
      if (present(modifier)) then
        modifier = modif
      elseif (len(modif) > 0) then
        call detailedError(child2, MSG_NOMODIFIER)
      end if
      call unstring(modif)
      call getFirstTextChild(child2, text)
      call getChVal_lIntR1RealR1_h(char(text), dimInt, valueInt, &
          &dimReal, valueReal, child2)
      call unstring(text)
      if (len(valueInt) /= len(valueReal)) then
        call detailedError(node, "Unexpected end of data")
      end if
      call setAttribute(child2, attrProcessed, "")
    else
      call detailedError(node, MSG_MISSING_FIELD // name)
    end if
    if (present(child)) then
      child => child2
    end if

  end subroutine getChVal_lIntR1RealR1



  !!* Helper function for getChVal_lIntR1RealR1 to avoid string to char
  !!* conversion in the do-loop.
  !!* @param text  Text to parse
  !!* @param value Contains the value of the parsed text
  subroutine getChVal_lIntR1RealR1_h(text, dimInt, valueInt, &
      &dimReal, valueReal, node)
    character(len=*), intent(in) :: text
    integer, intent(in) :: dimInt
    type(listIntR1), intent(inout) :: valueInt
    integer, intent(in) :: dimReal
    type(listRealR1), intent(inout) :: valueReal
    type(fnode), pointer :: node
    
    integer :: iStart, iErr
    real(dp) :: bufferReal(dimReal)
    integer :: bufferInt(dimInt)
    integer :: nItem

    iErr = TOKEN_OK
    iStart = 1
    do while (iErr == TOKEN_OK)
      call getNextToken(text, bufferInt, iStart, iErr, nItem)
      if (iErr == TOKEN_ERROR) then
        call detailedError(node, "Invalid integer")
      elseif (iErr == TOKEN_EOS .and. nItem /= 0) then
        call detailedError(node, "Unexpected end of data")
      end if
      if (iErr == TOKEN_OK) then
        call append(valueInt, bufferInt)
        call getNextToken(text, bufferReal, iStart, iErr, nItem)
        call checkError(node, iErr, "Invalid real")
        if (iErr == TOKEN_OK) then
          call append(valueReal, bufferReal)
        end if
      end if
    end do

  end subroutine getChVal_lIntR1RealR1_h



  !!* Returns the value (the child) of a child node as a linked list of string,
  !!* rank one integer and rank one real arrays.
  !!* @param node      The node to investigate.
  !!* @param name      Name of the child to look for
  !!* @param valueStr  List of strings on return.
  !!* @param dimInt    Dimension of the integer arrays in the list
  !!* @param valueInt  List of integer arrays on return
  !!* @param dimReal   Dimension of the real arrays in the list
  !!* @param valueReal List of real array on return
  !!* @param modifier  Modifier of the child on return
  !!* @param child     Pointer to the child on return
  !!* @note In order to prevent a double packaging (from array to linked list
  !!*   and then from linked list to array), the setting of defaults for list
  !!*   types is not allowed. The presence of the child must be explicitely
  !!*   queried in the caller routine and an eventual default setting must be
  !!*   set with an explicit setChildValue call.
  subroutine getChVal_lStringIntR1RealR1(node, name, valueStr, dimInt, &
      &valueInt, dimReal, valueReal, modifier, child)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    type(listString), intent(inout) :: valueStr
    integer, intent(in) :: dimInt
    type(listIntR1), intent(inout) :: valueInt
    integer, intent(in) :: dimReal
    type(listRealR1), intent(inout) :: valueReal
    type(string), intent(inout), optional :: modifier
    type(fnode), pointer, optional :: child

    type(string) :: text, modif
    type(fnode), pointer :: child2

    ASSERT(associated(node))
    ASSERT(dimInt > 0)
    ASSERT(dimReal > 0)

    child2 => getFirstChildByName(node, tolower(name))
    if (associated(child2)) then
      call getAttribute(child2, attrModifier, modif)
      if (present(modifier)) then
        modifier = modif
      elseif (len(modif) > 0) then
        call detailedError(child2, MSG_NOMODIFIER)
      end if
      call unstring(modif)
      call getFirstTextChild(child2, text)
      call getChVal_lStringIntR1RealR1_h(char(text), valueStr, &
          &dimInt, valueInt, dimReal, valueReal, child2)
      call unstring(text)
      if (len(valueStr) /= len(valueInt) &
          &.or. len(valueInt) /= len(valueReal)) then
        call detailedError(node, "Unexpected end of data")
      end if
      call setAttribute(child2, attrProcessed, "")
    else
      call detailedError(node, MSG_MISSING_FIELD // name)
    end if
    if (present(child)) then
      child => child2
    end if

  end subroutine getChVal_lStringIntR1RealR1



  !!* Helper function for getChVal_lIntR1RealR1 to avoid string to char
  !!* conversion in the do-loop.
  !!* @param text  Text to parse
  !!* @param value Contains the value of the parsed text
  subroutine getChVal_lStringIntR1RealR1_h(text, valueStr, dimInt, &
      &valueInt, dimReal, valueReal, node)
    character(len=*), intent(in) :: text
    type(listString), intent(inout) :: valueStr
    integer, intent(in) :: dimInt
    type(listIntR1), intent(inout) :: valueInt
    integer, intent(in) :: dimReal
    type(listRealR1), intent(inout) :: valueReal
    type(fnode), pointer :: node
    
    integer :: iStart, iErr
    real(dp) :: bufferReal(dimReal)
    integer :: bufferInt(dimInt)
    integer :: nItem
    type(string) :: bufferStr

    iErr = TOKEN_OK
    iStart = 1
    do while (iErr == TOKEN_OK)
      call getNextToken(text, bufferStr, iStart, iErr)
      if (iErr == TOKEN_ERROR) then
        call detailedError(node, "Invalid string")
      elseif (iErr == TOKEN_EOS) then
        exit
      end if
      call append(valueStr, char(bufferStr))
      
      call getNextToken(text, bufferInt, iStart, iErr, nItem)
      call checkError(node, iErr, "Invalid integer")
      call append(valueInt, bufferInt)
      
      call getNextToken(text, bufferReal, iStart, iErr, nItem)
      call checkError(node, iErr, "Invalid real")
      call append(valueReal, bufferReal)
    end do
    call unstring(bufferStr)

  end subroutine getChVal_lStringIntR1RealR1_h



  !!* Returns the value (the child) of a child node as a node.
  !!* @param node     The node to investigate.
  !!* @param name     Name of the child to look for
  !!* @param value    Value on return
  !!* @param default  Default value for the child, if child is not found. If
  !!*   the empty string is passed as default value, the child is created
  !!*   but no value is added to it. The returned value pointer will be
  !!*   unassociated. (allowEmptyValue must be explicitely set to .true.)
  !!* @param modifier Modifier of the child on return
  !!* @param child    Pointer to the child node (with the spec. name) on return
  !!* @param list     If the node created as default should be tagged as list.
  !!* @param allowEmptyValue If the child is allowed to have an empty value.
  !!* @param dummyValue If true, the value is not marked as processed.
  !!* @caveat If allowEmptyValue is set to .true. and the child has no subnodes
  !!*   (empty value) the returned value is an unassociated pointer!
  subroutine getChVal_node(node, name, value, default, modifier, child, &
      &list, allowEmptyValue, dummyValue)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    type(fnode), pointer :: value
    character(len=*), intent(in), optional :: default
    type(string), intent(inout), optional :: modifier
    type(fnode), pointer, optional :: child
    logical, intent(in), optional :: list
    logical, intent(in), optional :: allowEmptyValue
    logical, intent(in), optional :: dummyValue

    type(string) :: modif
    type(fnode), pointer :: child2
    logical :: tList, tAllowEmptyVal, tDummyValue

    ASSERT(associated(node))
    ASSERT_ENV(if (present(default)) then)
    ASSERT_ENV(  if (len(default) == 0) then)
    ASSERT_ENV(    ASSERT(present(allowEmptyValue)))
    ASSERT_ENV(    ASSERT(allowEmptyValue))
    ASSERT_ENV(  end if)
    ASSERT_ENV(end if)

    if (present(list)) then
      tList = list
    else
      tList = .false.
    end if
    if (present(allowEmptyValue)) then
      tAllowEmptyVal = allowEmptyValue
    else
      tAllowEmptyVal = .false.
    end if
    if (present(dummyValue)) then
      tDummyValue = dummyValue
    else
      tDummyValue = .false.
    end if

    child2 => getFirstChildByName(node, tolower(name))
    if (associated(child2)) then
      call getAttribute(child2, attrModifier, modif)
      if (present(modifier)) then
        modifier = modif
      elseif (len(modif) > 0) then
        call detailedError(child2, MSG_NOMODIFIER)
      end if
      call unstring(modif)
      value => getFirstChild(child2)
      if ((.not. associated(value)) .and. (.not. tAllowEmptyVal)) then
        call detailedError(child2, "Missing value")
      end if
      call setAttribute(child2, attrProcessed, "")
    elseif (present(default)) then
      if (present(modifier)) then
        modifier = ""
      end if
      if (len(default) > 0) then
        value => createElement(tolower(default))
        call setChildValue(node, name, value, .false., child=child2, list=tList)
        call setAttribute(value, attrName, default)
      else
        nullify(value)
        call setChild(node, name, child2, .false., list=tList)
      end if
    else
      call detailedError(node, MSG_MISSING_FIELD // name)
    end if
    if (associated(value) .and. .not. tDummyValue) then
      if (getNodeType(value) == ELEMENT_NODE) then
        call setAttribute(value, attrProcessed, "")
      end if
    end if
    if (present(child)) then
      child => child2
    end if
    
  end subroutine getChVal_node


  
  !!* Gets a list of integers from a listNode type node, or generates
  !!* (and substitutes) them, if the node type is a range
  !!* @param field  Node to investigate
  !!* @param value  Contains an allocated array of integers on return.
  !!* @param mValue Maximal value, which can occur. If it is specified,
  !!*   negative numbers are interpreted as maxval-abs(value)+1.
  subroutine getValueFromListOrRange(field, value, mValue)
    type(fnode), pointer :: field
    integer, pointer :: value(:)
    integer, intent(in), optional :: mValue
  
    type(fnode), pointer :: node
    integer :: boundaries(2)
    integer :: ii, ind
    type(listInt) :: intList
    type(string) :: buffer

    ASSERT(associated(field))

    node => getFirstChild(field)
    if (.not. associated(node)) then
      INITALLOCATE_PARR(value, (0))
      return
    end if

    call getNodeName(node, buffer)
    select case(char(buffer))
    case (textNodeName)
      !! Explicit list was provided
      call init(intList)
      call getChildValue(field, "", intList)
      if (len(intList) == 0) then
        call detailedError(node, "Missing value")
      end if
      INITALLOCATE_PARR(value, (len(intList)))
      call asArray(intList, value)
      call destroy(intList)
      if (present(mValue) .and. any(value < 0)) then
        where (value < 0)
          value = value + mValue + 1
        end where
      end if

    case ("range")
      !! Range was specified, generate numbers
      call getChildValue(node, "", boundaries)
      if (present(mValue) .and. any(boundaries < 0)) then
        where (boundaries < 0)
          boundaries = boundaries + mValue + 1
        end where
      end if
      if (boundaries(1) > boundaries(2)) then
        call detailedError(node, "Lower range boundary must not be greater &
            &than upper one")
      end if
      INITALLOCATE_PARR(value, (boundaries(2) - boundaries(1) + 1))
      ind = 1
      do ii = boundaries(1), boundaries(2)
        value(ind) = ii
        ind = ind + 1
      end do

    case default
      call detailedError(node, "Cannot convert '" // char(buffer) &
          &// "' to a list")
    end select

    if (present(mValue)) then
      if (any(value > mValue) .or. any(value < 1)) then
        call detailedError(node, "Index out of the allowed range (1-" &
            &// i2c(mValue) // ")")
      end if
    end if
    
    !call setChildValue(field, "", value, .true.)
    call unstring(buffer)
    
  end subroutine getValueFromListOrRange



  !!* Returns a child node with a specified name
  !!* @param node      Node to investigate
  !!* @param name      Name of the child node to look for
  !!* @param child     Contains a pointer to the child on return
  !!* @param requested If true and child not found, error is issued
  !!* @param modifier  Contains modifier on exit.
  subroutine getChild(node, name, child, requested, modifier)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    type(fnode), pointer :: child
    logical, intent(in), optional :: requested
    type(string), intent(inout), optional :: modifier

    logical :: tRequested
    type(string) :: modif

    ASSERT(associated(node))

    if (present(requested)) then
      tRequested = requested
    else
      tRequested = .true.
    end if

    child => getFirstChildByName(node, tolower(name))
    if (associated(child)) then
      call getAttribute(child, attrModifier, modif)
      if (present(modifier)) then
        modifier = modif
      elseif (len(modif) > 0) then
        call detailedError(child, MSG_NOMODIFIER)
      end if
      call unstring(modif)
      call setAttribute(child, attrProcessed, "")
    elseif (tRequested) then
      call detailedError(node, MSG_MISSING_FIELD // name)
    end if
    
  end subroutine getChild

  

  !!* Returns a list of children with the specified name.
  !!* @param node     Parent node to investigate
  !!* @param name     Name of the children to look for
  !!* @param children List of the children.
  subroutine getChildren(node, name, children)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    type(fnodeList), pointer :: children

    type(fnode), pointer :: child
    integer :: ii

    children => getChildrenByName(node, tolower(name))
    do ii = 1, getLength(children)
      call getItem1(children, ii, child)
      call setAttribute(child, attrProcessed, "")
    end do
    
  end subroutine getChildren



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! setChildValue and setNode routines
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  !!* Sets the value (child) of a child with given name.
  !!* @param node    The node to investigate
  !!* @param name    Name of the child to look for
  !!* @param value   Value to set
  !!* @param replace Replace if child with same name already exists
  !!* @param child   Pointer to the child node (with the provided name)
  !!* @param modifier Optional modifier for the child
  subroutine setChVal_logical(node, name, value, replace, child, modifier)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    logical, intent(in) :: value
    logical, intent(in), optional :: replace
    type(fnode), pointer, optional :: child
    character(len=*), optional, intent(in) :: modifier

    type(string) :: strBuffer
    type(fnode), pointer :: child2
    logical :: tReplace

    if (present(replace)) then
      tReplace = replace
    else
      tReplace = .false.
    end if
    
    call getAsString(value, strBuffer)
    call createChild_local(node, name, .false., tReplace, child2, &
        &value=char(strBuffer))
    call unstring(strBuffer)
    if (present(child)) then
      child => child2
    end if
    if (present(modifier)) then
      call setAttribute(child2, attrModifier, modifier)
    end if
    
  end subroutine setChVal_logical


  
  !!* Writes the text representation of a node and its value to an xmlwriter.
  !!* @param xf    Xmlwriter stream
  !!* @param name  Name of the node
  !!* @param value Value of the node
  subroutine writeChVal_logical(xf, name, value)
    type(xmlf_t), intent(inout) :: xf
    character(len=*), intent(in) :: name
    logical, intent(in) :: value
    
    type(string) :: strBuffer

    call getAsString(value, strBuffer)
    call writeChild_local(xf, name, char(strBuffer))
    call unstring(strBuffer)

  end subroutine writeChVal_logical

  

  !!* Returns the text representation of the passed object
  !!* @param value     Value to represent
  !!* @param strBuffer Text representation on exit
  subroutine getAsString_logical(value, strBuffer)
    logical, intent(in) :: value
    type(string), intent(inout) :: strBuffer

    if (value) then
      strBuffer = LOGICAL_TRUE
    else
      strBuffer = LOGICAL_FALSE
    end if

  end subroutine getAsString_logical


  
  !!* Sets the value (child) of a child with given name.
  !!* @param node    The node to investigate
  !!* @param name    Name of the child to look for
  !!* @param value   Value to set
  !!* @param replace Replace if child with same name already exists
  !!* @param child   Pointer to the child node (with the provided name)
  !!* @param modifier Optional modifier for the child
  !!* @caveat This subroutines assumes, that a real can be represented as text
  !!*   with less than nCharReal characters.
  subroutine setChVal_real(node, name, value, replace, child, modifier)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    real(dp), intent(in) :: value
    logical, intent(in), optional :: replace
    type(fnode), pointer, optional :: child
    character(len=*), optional, intent(in) :: modifier

    type(string) :: strBuffer
    type(fnode), pointer :: child2
    logical :: tReplace

    if (present(replace)) then
      tReplace = replace
    else
      tReplace = .false.
    end if

    call getAsString(value, strBuffer)
    call createChild_local(node, name, .false., tReplace, child2, &
        &value=char(strBuffer))
    call unstring(strBuffer)
    if (present(child)) then
      child => child2
    end if
    if (present(modifier)) then
      call setAttribute(child2, attrModifier, modifier)
    end if
    
  end subroutine setChVal_real



  !!* Writes the text representation of a node and its value to an xmlwriter.
  !!* @param xf    Xmlwriter stream
  !!* @param name  Name of the node
  !!* @param value Value of the node
  subroutine writeChVal_real(xf, name, value)
    type(xmlf_t), intent(inout) :: xf
    character(len=*), intent(in) :: name
    real(dp), intent(in) :: value
    
    type(string) :: strBuffer

    call getAsString(value, strBuffer)
    call writeChild_local(xf, name, char(strBuffer))
    call unstring(strBuffer)

  end subroutine writeChVal_real

  

  !!* Returns the text representation of the passed object
  !!* @param value     Value to represent
  !!* @param strBuffer Text representation on exit
  subroutine getAsString_real(value, strBuffer)
    real(dp), intent(in) :: value
    type(string), intent(inout) :: strBuffer

    character(len=nCharReal) :: buffer
    
    write (buffer, *) value
    strBuffer = trim(adjustl(buffer))

  end subroutine getAsString_real



  !!* Sets the value (child) of a child with given name.
  !!* @param node    The node to investigate
  !!* @param name    Name of the child to look for
  !!* @param value   Value to set
  !!* @param replace Replace if child with same name already exists
  !!* @param child   Pointer to the child node (with the provided name)
  !!* @param modifier Optional modifier for the child
  !!* @caveat This subroutines assumes, that a real can be represented as text
  !!*   with less than nCharReal characters.
  subroutine setChVal_realR1(node, name, value, replace, child, modifier)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    real(dp), intent(in) :: value(:)
    logical, intent(in), optional :: replace
    type(fnode), pointer, optional :: child
    character(len=*), optional, intent(in) :: modifier

    type(string) :: strBuffer
    type(fnode), pointer :: child2
    logical :: tReplace

    if (present(replace)) then
      tReplace = replace
    else
      tReplace = .false.
    end if
    call getAsString(value, strBuffer)
    call createChild_local(node, name, .true., tReplace, child2, &
        &value=char(strBuffer))
    call unstring(strBuffer)
    if (present(child)) then
      child => child2
    end if
    if (present(modifier)) then
      call setAttribute(child2, attrModifier, modifier)
    end if
    
  end subroutine setChVal_realR1


  
  !!* Writes the text representation of a node and its value to an xmlwriter.
  !!* @param xf    Xmlwriter stream
  !!* @param name  Name of the node
  !!* @param value Value of the node
  subroutine writeChVal_realR1(xf, name, value)
    type(xmlf_t), intent(inout) :: xf
    character(len=*), intent(in) :: name
    real(dp), intent(in) :: value(:)
    
    type(string) :: strBuffer

    call getAsString(value, strBuffer)
    call writeChild_local(xf, name, char(strBuffer))
    call unstring(strBuffer)

  end subroutine writeChVal_realR1

  

  !!* Returns the text representation of the passed object
  !!* @param value     Value to represent
  !!* @param strBuffer Text representation on exit
  subroutine getAsString_realR1(value, strBuffer)
    real(dp), intent(in) :: value(:)
    type(string), intent(inout) :: strBuffer

    character(len=nCharReal) :: buffer
    integer :: buffLen, len
    integer :: ii

    call resize_string(strBuffer, preAllocSize)
    len = 0
    do ii = 1, size(value)
      write (buffer, *) value(ii)
      buffer = adjustl(buffer)
      buffLen = len_trim(buffer)
      len = len + buffLen
      if (len > lineLength) then
        call append_to_string(strBuffer, newline // trim(buffer))
        len = buffLen
      else
        call append_to_string(strBuffer, space // trim(buffer))
      end if
    end do

  end subroutine getAsString_realR1



  !!* Sets the value (child) of a child with given name.
  !!* @param node    The node to investigate
  !!* @param name    Name of the child to look for
  !!* @param value   Value to set
  !!* @param replace Replace if child with same name already exists
  !!* @caveat This subroutines assumes, that a real can be represented as text
  !!*   with less than nCharReal characters.
  !!* @param child   Pointer to the child node (with the provided name)
  !!* @param modifier Optional modifier for the child
  !!* @note This is just a wrapper around the rank one version, to make sure
  !!*   that two dimensional arrays are pretty printed. For higher ranked
  !!*   arrays the rank one version should be used with some reshaping before.
  subroutine setChVal_realR2(node, name, value, replace, child, modifier)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    real(dp), intent(in) :: value(:,:)
    logical, intent(in), optional :: replace
    type(fnode), pointer, optional :: child
    character(len=*), intent(in), optional :: modifier

    type(fnode), pointer :: child2
    type(string) :: strBuffer
    logical :: tReplace

    if (present(replace)) then
      tReplace = replace
    else
      tReplace = .false.
    end if

    call getAsString(value, strBuffer)
    call createChild_local(node, name, .true., tReplace, child2, &
        &value=char(strBuffer))
    call unstring(strBuffer)
    if (present(child)) then
      child => child2
    end if
    if (present(modifier)) then
      call setAttribute(child2, attrModifier, modifier)
    end if
    
  end subroutine setChVal_realR2


  
  !!* Writes the text representation of a node and its value to an xmlwriter.
  !!* @param xf    Xmlwriter stream
  !!* @param name  Name of the node
  !!* @param value Value of the node
  subroutine writeChVal_realR2(xf, name, value)
    type(xmlf_t), intent(inout) :: xf
    character(len=*), intent(in) :: name
    real(dp), intent(in) :: value(:,:)
    
    type(string) :: strBuffer

    call getAsString(value, strBuffer)
    call writeChild_local(xf, name, char(strBuffer))
    call unstring(strBuffer)

  end subroutine writeChVal_realR2

  

  !!* Returns the text representation of the passed object
  !!* @param value     Value to represent
  !!* @param strBuffer Text representation on exit
  subroutine getAsString_realR2(value, strBuffer)
    real(dp), intent(in) :: value(:,:)
    type(string), intent(inout) :: strBuffer

    character(len=nCharReal) :: buffer
    integer :: ii, jj

    call resize_string(strBuffer, preAllocSize)
    do ii = 1, size(value, dim=2)
      do jj = 1, size(value, dim=1)
        write (buffer, *) value(jj, ii)
        buffer = adjustl(buffer)
        call append_to_string(strBuffer, space // trim(buffer))
      end do
      call append_to_string(strBuffer, newline)
    end do

  end subroutine getAsString_realR2



  !!* Sets the value (child) of a child with given name.
  !!* @param node    The node to investigate
  !!* @param name    Name of the child to look for
  !!* @param value   Value to set
  !!* @param replace Replace if child with same name already exists
  !!* @param child   Pointer to the child node (with the provided name)
  !!* @param modifier Optional modifier for the child
  !!* @caveat This subroutines assumes, that an integer can be represented
  !!*   as text with less than nCharInt characters.
  subroutine setChVal_int(node, name, value, replace, child, modifier)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    integer, intent(in) :: value
    logical, intent(in), optional :: replace
    type(fnode), pointer, optional :: child
    character(len=*), optional, intent(in) :: modifier

    type(fnode), pointer :: child2
    type(string) :: strBuffer
    logical :: tReplace

    if (present(replace)) then
      tReplace = replace
    else
      tReplace = .false.
    end if
    call getAsString(value, strBuffer)
    call createChild_local(node, name, .false., tReplace, child2, &
        &value=char(strBuffer))
    call unstring(strBuffer)
    if (present(child)) then
      child => child2
    end if
    if (present(modifier)) then
      call setAttribute(child2, attrModifier, modifier)
    end if

  end subroutine setChVal_int



  !!* Writes the text representation of a node and its value to an xmlwriter.
  !!* @param xf    Xmlwriter stream
  !!* @param name  Name of the node
  !!* @param value Value of the node
  subroutine writeChVal_int(xf, name, value)
    type(xmlf_t), intent(inout) :: xf
    character(len=*), intent(in) :: name
    integer, intent(in) :: value
    
    type(string) :: strBuffer

    call getAsString(value, strBuffer)
    call writeChild_local(xf, name, char(strBuffer))
    call unstring(strBuffer)

  end subroutine writeChVal_int

  

  !!* Returns the text representation of the passed object
  !!* @param value     Value to represent
  !!* @param strBuffer Text representation on exit
  subroutine getAsString_int(value, strBuffer)
    integer, intent(in) :: value
    type(string), intent(inout) :: strBuffer

    character(len=nCharInt) :: buffer

    write (buffer, *) value
    strBuffer = trim(adjustl(buffer))
    
  end subroutine getAsString_int



  !!* Sets the value (child) of a child with given name.
  !!* @param node    The node to investigate
  !!* @param name    Name of the child to look for
  !!* @param value   Value to set
  !!* @param replace Replace if child with same name already exists
  !!* @param modifier Optional modifier for the child
  !!* @caveat This subroutines assumes, that an integer can be represented
  !!*   as text with less than nCharInt characters.
  subroutine setChVal_intR1(node, name, value, replace, child, modifier)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    integer, intent(in) :: value(:)
    logical, intent(in), optional :: replace
    type(fnode), pointer, optional :: child
    character(len=*), optional, intent(in) :: modifier

    type(fnode), pointer :: child2
    type(string) :: strBuffer
    logical :: tReplace

    if (present(replace)) then
      tReplace = replace
    else
      tReplace = .false.
    end if
    call getAsString(value, strBuffer)
    call createChild_local(node, name, .true., tReplace, child2, &
        &value=char(strBuffer))
    call unstring(strBuffer)
    if (present(child)) then
      child => child2
    end if
    if (present(modifier)) then
      call setAttribute(child2, attrModifier, modifier)
    end if

  end subroutine setChVal_intR1



  !!* Writes the text representation of a node and its value to an xmlwriter.
  !!* @param xf    Xmlwriter stream
  !!* @param name  Name of the node
  !!* @param value Value of the node
  subroutine writeChVal_intR1(xf, name, value)
    type(xmlf_t), intent(inout) :: xf
    character(len=*), intent(in) :: name
    integer, intent(in) :: value(:)
    
    type(string) :: strBuffer

    call getAsString(value, strBuffer)
    call writeChild_local(xf, name, char(strBuffer))
    call unstring(strBuffer)

  end subroutine writeChVal_intR1

  

  !!* Returns the text representation of the passed object
  !!* @param value     Value to represent
  !!* @param strBuffer Text representation on exit
  subroutine getAsString_intR1(value, strBuffer)
    integer, intent(in) :: value(:)
    type(string), intent(inout) :: strBuffer

    character(len=nCharInt) :: buffer
    integer :: buffLen, len
    integer :: ii

    call resize_string(strBuffer, preAllocSize)
    len = 0
    do ii = 1, size(value)
      write (buffer, *) value(ii)
      buffer = adjustl(buffer)
      buffLen = len_trim(buffer)
      len = len + buffLen
      if (len > lineLength) then
        call append_to_string(strBuffer, newline // trim(buffer))
        len = buffLen
      else
        call append_to_string(strBuffer, space // trim(buffer))
      end if
    end do
    
  end subroutine getAsString_intR1



  !!* Sets the value (child) of a child with given name.
  !!* @param node    The node to investigate
  !!* @param name    Name of the child to look for
  !!* @param value   Value to set
  !!* @param replace Replace if child with same name already exists
  !!* @param child   Pointer to the child node (with the provided name)
  !!* @caveat This subroutines assumes, that an integer can be represented as
  !!*   text with less than nCharInt characters.
  !!* @param modifier Optional modifier for the child
  !!* @note This is just a wrapper around the rank one version, to make sure
  !!*   that two dimensional arrays are pretty printed. For higher ranked
  !!*   arrays the rank one version should be used with some reshaping before.
  subroutine setChVal_intR2(node, name, value, replace, child, modifier)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    integer, intent(in) :: value(:,:)
    logical, intent(in), optional :: replace
    type(fnode), pointer, optional :: child
    character(len=*), optional, intent(in) :: modifier

    type(fnode), pointer :: child2
    type(string) :: strBuffer
    logical :: tReplace

    if (present(replace)) then
      tReplace = replace
    else
      tReplace = .false.
    end if
    call getAsString(value, strBuffer)
    call createChild_local(node, name, .true., tReplace, child2, &
        &value=char(strBuffer))
    call unstring(strBuffer)
    if (present(child)) then
      child => child2
    end if
    if (present(modifier)) then
      call setAttribute(child2, attrModifier, modifier)
    end if
    
  end subroutine setChVal_intR2



  !!* Writes the text representation of a node and its value to an xmlwriter.
  !!* @param xf    Xmlwriter stream
  !!* @param name  Name of the node
  !!* @param value Value of the node
  subroutine writeChVal_intR2(xf, name, value)
    type(xmlf_t), intent(inout) :: xf
    character(len=*), intent(in) :: name
    integer, intent(in) :: value(:,:)
    
    type(string) :: strBuffer

    call getAsString(value, strBuffer)
    call writeChild_local(xf, name, char(strBuffer))
    call unstring(strBuffer)

  end subroutine writeChVal_intR2

  

  !!* Returns the text representation of the passed object
  !!* @param value     Value to represent
  !!* @param strBuffer Text representation on exit
  subroutine getAsString_intR2(value, strBuffer)
    integer, intent(in) :: value(:,:)
    type(string), intent(inout) :: strBuffer

    character(len=nCharInt) :: buffer
    integer :: ii, jj
    
    call resize_string(strBuffer, preAllocSize)
    do ii = 1, size(value, dim=2)
      do jj = 1, size(value, dim=1)
        write (buffer, *) value(jj, ii)
        buffer = adjustl(buffer)
        call append_to_string(strBuffer, space // trim(buffer))
      end do
      call append_to_string(strBuffer, newline)
    end do

  end subroutine getAsString_intR2



  !!* Sets the value (child) of a child with given name.
  !!* @param node    The node to investigate
  !!* @param name    Name of the child to look for
  !!* @param value   Value to set
  !!* @param replace Replace if child with same name already exists
  !!* @param child   Pointer to the child node (with the provided name)
  !!* @param omitQuotes If qoutes around the string should be omitted
  !!* @param modifier Optional modifier for the child
  subroutine setChVal_char(node, name, value, replace, child, omitQuotes, &
      &modifier)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    character(len=*), intent(in) :: value
    logical, intent(in), optional :: replace
    type(fnode), pointer, optional :: child
    logical, intent(in), optional :: omitQuotes
    character(len=*), optional, intent(in) :: modifier

    type(fnode), pointer :: child2
    logical :: tReplace, tQuotes

    if (present(replace)) then
      tReplace = replace
    else
      tReplace = .false.
    end if
    if (present(omitQuotes)) then
      tQuotes = .not. omitQuotes
    else
      tQuotes = .true.
    end if
    if (tQuotes) then
      call createChild_local(node, name, .false., tReplace, child2, &
          &value='"'//value//'"')
    else
      call createChild_local(node, name, .false., tReplace, child2, value=value)
    end if
    
    if (present(child)) then
      child => child2
    end if
    if (present(modifier)) then
      call setAttribute(child2, attrModifier, modifier)
    end if
    
  end subroutine setChVal_char



  !!* Writes the text representation of a node and its value to an xmlwriter.
  !!* @param xf    Xmlwriter stream
  !!* @param name  Name of the node
  !!* @param value Value of the node
  subroutine writeChVal_char(xf, name, value)
    type(xmlf_t), intent(inout) :: xf
    character(len=*), intent(in) :: name
    character(len=*), intent(in) :: value
    
    call writeChild_local(xf, name, value)

  end subroutine writeChVal_char

  

  !!* Sets the value (child) of a child with given name.
  !!* @param node    The node to investigate
  !!* @param name    Name of the child to look for
  !!* @param value   Value to set
  !!* @param replace Replace if child with same name already exists
  !!* @param child   Pointer to the child node (with the provided name)
  !!* @param modifier Optional modifier for the child
  subroutine setChVal_charR1(node, name, value, replace, child, modifier)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    character(len=*), intent(in) :: value(:)
    logical, intent(in), optional :: replace
    type(fnode), pointer, optional :: child
    character(len=*), optional, intent(in) :: modifier

    type(string) :: strBuffer
    type(fnode), pointer :: child2
    logical :: tReplace

    if (present(replace)) then
      tReplace = replace
    else
      tReplace = .false.
    end if
    call getAsString(value, strBuffer)
    call createChild_local(node, name, .true., tReplace, child2, &
        &value=char(strBuffer))
    call unstring(strBuffer)
    if (present(child)) then
      child => child2
    end if
    if (present(modifier)) then
      call setAttribute(child2, attrModifier, modifier)
    end if
    
  end subroutine setChVal_charR1



  !!* Writes the text representation of a node and its value to an xmlwriter.
  !!* @param xf    Xmlwriter stream
  !!* @param name  Name of the node
  !!* @param value Value of the node
  subroutine writeChVal_charR1(xf, name, value)
    type(xmlf_t), intent(inout) :: xf
    character(len=*), intent(in) :: name
    character(len=*), intent(in) :: value(:)
    
    type(string) :: strBuffer

    call getAsString(value, strBuffer)
    call writeChild_local(xf, name, char(strBuffer))
    call unstring(strBuffer)

  end subroutine writeChVal_charR1

  

  !!* Returns the text representation of the passed object
  !!* @param value     Value to represent
  !!* @param strBuffer Text representation on exit
  subroutine getAsString_charR1(value, strBuffer)
    character(len=*), intent(in) :: value(:)
    type(string), intent(inout) :: strBuffer
    
    integer :: buffLen, len
    integer :: ii

    call resize_string(strBuffer, preAllocSize)
    len = 0
    do ii = 1, size(value)
      buffLen = len_trim(value(ii))
      len = len + buffLen
      if (len > lineLength) then
        call append_to_string(strBuffer, newline // '"'//trim(value(ii))//'"')
        len = buffLen
      else
        call append_to_string(strBuffer, space // '"'//trim(value(ii))//'"')
      end if
    end do

  end subroutine getAsString_charR1



  !!* Sets the value (child) of a child with given name.
  !!* @param node      The node to investigate
  !!* @param name      Name of the child to look for
  !!* @param intValue  Value for the integers
  !!* @param realValue Value for the reals
  !!* @param replace Replace if child with same name already exists
  !!* @param child     Pointer to the child node (with the provided name)
  !!* @param modifier Optional modifier for the child
  subroutine setChVal_intR2RealR2(node, name, intValue, realValue, &
      &replace, child, modifier)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    integer, intent(in) :: intValue(:,:)
    real(dp), intent(in) :: realValue(:,:)
    logical, intent(in), optional :: replace
    type(fnode), pointer, optional :: child
    character(len=*), optional, intent(in) :: modifier

    type(fnode), pointer :: child2
    type(string) :: strBuffer
    logical :: tReplace
        
    if (present(replace)) then
      tReplace = replace
    else
      tReplace = .false.
    end if
    call getAsString(intValue, realValue, strBuffer)
    call createChild_local(node, name, .true., tReplace, child2, &
        &value=char(strBuffer))
    call unstring(strBuffer)
    if (present(child)) then
      child => child2
    end if
    if (present(modifier)) then
      call setAttribute(child2, attrModifier, modifier)
    end if
    
  end subroutine setChVal_intR2RealR2



  !!* Writes the text representation of a node and its value to an xmlwriter.
  !!* @param xf    Xmlwriter stream
  !!* @param name  Name of the node
  !!* @param value Value of the node
  subroutine writeChVal_intR2RealR2(xf, name, intValue, realValue)
    type(xmlf_t), intent(inout) :: xf
    character(len=*), intent(in) :: name
    integer, intent(in) :: intValue(:,:)
    real(dp), intent(in) :: realValue(:,:)
    
    type(string) :: strBuffer

    call getAsString(intValue, realValue, strBuffer)
    call writeChild_local(xf, name, char(strBuffer))
    call unstring(strBuffer)

  end subroutine writeChVal_intR2RealR2

  

  !!* Returns the text representation of the passed object
  !!* @param intValue
  !!* @param realValue
  !!* @param strBuffer Text representation on exit
  subroutine getAsString_intR2RealR2(intValue, realValue, strBuffer)
    integer, intent(in) :: intValue(:,:)
    real(dp), intent(in) :: realValue(:,:)
    type(string), intent(inout) :: strBuffer

    character(len=100) :: buffer
    integer :: nRow, nCol1, nCol2
    integer :: ii, jj
    
    nRow = size(intValue, dim=2)
    ASSERT(size(realValue, dim=2) == nRow)
    
    nCol1 = size(intValue, dim=1)
    nCol2 = size(realValue, dim=1)
    call resize_string(strBuffer, preAllocSize)
    do ii = 1, nRow
      do jj = 1, nCol1
        write (buffer, *) intValue(jj, ii)
        buffer = adjustl(buffer)
        call append_to_string(strBuffer, space // trim(buffer))
      end do
      do jj = 1, nCol2
        write (buffer, *) realValue(jj, ii)
        buffer = adjustl(buffer)
        call append_to_string(strBuffer, space // trim(buffer))
      end do
      call append_to_string(strBuffer, newline)
    end do

  end subroutine getAsString_intR2RealR2



  !!* Sets the value (child) of a child with given name.
  !!* @param node      The node to investigate
  !!* @param name      Name of the child to look for
  !!* @param charValue Value for the characters
  !!* @param intValue  Value for the integers
  !!* @param realValue Value for the reals
  !!* @param replace Replace if child with same name already exists
  !!* @param child     Pointer to the child node (with the provided name)
  !!* @param modifier Optional modifier for the child
  subroutine setChVal_charR1IntR2RealR2(node, name, charValue, intValue, &
      &realValue, replace, child, modifier)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    character(len=*), intent(in) :: charValue(:)
    integer, intent(in) :: intValue(:,:)
    real(dp), intent(in) :: realValue(:,:)
    logical, intent(in), optional :: replace
    type(fnode), pointer, optional :: child
    character(len=*), optional, intent(in) :: modifier

    type(fnode), pointer :: child2
    type(string) :: strBuffer
    logical :: tReplace
        
    if (present(replace)) then
      tReplace = replace
    else
      tReplace = .false.
    end if
    call getAsString(charValue, intValue, realValue, strBuffer)
    call createChild_local(node, name, .true., tReplace, child2, &
        &value=char(strBuffer))
    call unstring(strBuffer)
    if (present(child)) then
      child => child2
    end if
    if (present(modifier)) then
      call setAttribute(child2, attrModifier, modifier)
    end if
    
  end subroutine setChVal_charR1IntR2RealR2



  !!* Writes the text representation of a node and its value to an xmlwriter.
  !!* @param xf    Xmlwriter stream
  !!* @param name  Name of the node
  !!* @param value Value of the node
  subroutine writeChVal_charR1IntR2RealR2(xf, name, charValue, intValue, &
      &realValue)
    type(xmlf_t), intent(inout) :: xf
    character(len=*), intent(in) :: name
    character(len=*), intent(in) :: charValue(:)
    integer, intent(in) :: intValue(:,:)
    real(dp), intent(in) :: realValue(:,:)
    
    type(string) :: strBuffer

    call getAsString(charValue, intValue, realValue, strBuffer)
    call writeChild_local(xf, name, char(strBuffer))
    call unstring(strBuffer)
    
  end subroutine writeChVal_charR1IntR2RealR2

  

  !!* Returns the text representation of the passed object
  !!* @param charValue
  !!* @param intValue
  !!* @param realValue
  !!* @param strBuffer Text representation on exit
  subroutine getAsString_charR1IntR2RealR2(charValue, intValue, realValue, &
      &strBuffer)
    character(len=*), intent(in) :: charValue(:)
    integer, intent(in) :: intValue(:,:)
    real(dp), intent(in) :: realValue(:,:)
    type(string), intent(inout) :: strBuffer

    character(len=100) :: buffer
    integer :: nRow, nCol1, nCol2
    integer :: ii, jj
    
    nRow = size(charValue)
    ASSERT(size(intValue, dim=2) == nRow)
    ASSERT(size(realValue, dim=2) == nRow)
    
    nCol1 = size(intValue, dim=1)
    nCol2 = size(realValue, dim=1)
    call resize_string(strBuffer, preAllocSize)
    do ii = 1, nRow
      call append_to_string(strBuffer, charValue(ii))
      do jj = 1, nCol1
        write (buffer, *) intValue(jj, ii)
        buffer = adjustl(buffer)
        call append_to_string(strBuffer, space // trim(buffer))
      end do
      do jj = 1, nCol2
        write (buffer, *) realValue(jj, ii)
        buffer = adjustl(buffer)
        call append_to_string(strBuffer, space // trim(buffer))
      end do
      call append_to_string(strBuffer, newline)
    end do

  end subroutine getAsString_charR1IntR2RealR2



  !!* Sets the value (child) of a child with given name.
  !!* @param node    The node to investigate
  !!* @param name    Name of the child to look for
  !!* @param value   Value to set
  !!* @param replace Replace if child with same name already exists
  !!* @param child   Pointer to the child node (with the provided name)
  !!* @param modifier Optional modifier for the child
  !!* @param list    If created child should be marked as a list.
  subroutine setChVal_node(node, name, value, replace, child, modifier, &
      &list)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    type(fnode), pointer :: value
    logical, intent(in), optional :: replace
    type(fnode), pointer, optional :: child
    character(len=*), optional, intent(in) :: modifier
    logical, optional, intent(in) :: list

    type(fnode), pointer :: child2, dummy
    logical :: tReplace, tList

    if (present(replace)) then
      tReplace = replace
    else
      tReplace = .false.
    end if
    if (present(list)) then
      tList = list
    else
      tList = .false.
    end if
    call createChild_local(node, name, tList, tReplace, child2)
    if (associated(value)) then
      dummy => appendChild(child2, value)
    end if
    if (present(child)) then
      child => child2
    end if
    if (present(modifier)) then
      call setAttribute(child2, attrModifier, modifier)
    end if
    
  end subroutine setChVal_node



  !!* Working horse for the setChildValue routines
  !!* @param node    The node to investigate
  !!* @param name    Name of the child to create
  !!* @param list    True, if child should be signed as a list
  !!* @param replace Replace if child with same name already exists
  !!* @param child   Pointer to the created child on return
  !!* @param value   Value to set (if empty, no child is appended to the created
  !!*   child)
  !!* @note If the empty string is provided as child name, no child is created,
  !!*   and the current node is replace instead. The pointer "node" becames
  !!*   associated with the new node, since the old instance will be destroyed.
  subroutine createChild_local(node, name, list, replace, child, value)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    logical, intent(in) :: list
    logical, intent(in) :: replace
    type(fnode), pointer :: child
    character(len=*), intent(in), optional :: value
    
    type(fnode), pointer :: parent, oldChild, child2, text, dummy
    character(len=len(name)) :: loName
    type(string) :: newName

    if (replace) then
      if (len(name) == 0) then
        call getNodeHSDName(node, newName)
        parent => getParentNode(node)
        oldChild => node
        child2 => createElement(tolower(char(newName)))
        node => child2
      else
        newName = name
        parent => node
        loName = tolower(name)
        oldChild => getFirstChildByName(node, loName)
        child2 => createElement(loName)
      end if
    else
      newName = name
      parent => node
      oldChild => null()
      child2 => createElement(tolower(name))
    end if
    if (associated(oldChild)) then
      dummy => replaceChild(parent, child2, oldChild)
      call destroyNode(oldChild)
    else
      dummy => appendChild(parent, child2)
    end if

    if (len(newName) > 0) then
      call setAttribute(child2, attrName, char(newName))
    end if
    if (list) then
      call setAttribute(child2, attrList, "")
    end if
      
    child => child2
    call setAttribute(child, attrProcessed, "")
    if (present(value)) then
      text => createTextNode(value)
      dummy => appendChild(child, text)
    end if

    call unstring(newName)
    
  end subroutine createChild_local



  subroutine writeChild_local(xf, name, value)
    type(xmlf_t), intent(inout) :: xf
    character(len=*), intent(in) :: name
    character(len=*), intent(in) :: value

    call xml_NewElement(xf, name)
    call xml_AddPCData(xf, value)
    call xml_EndElement(xf, name)
    
  end subroutine writeChild_local



  !!* Creates a child with the given name
  !!* @param node    Node to append the child to
  !!* @param name    Name of the child node to append
  !!* @param child   Contains the pointer to the added child node on return
  !!* @param replace If an already existing child with the same name should be
  !!*   replaced
  !!* @param list    If child should be signed as a list tag
  !!* @param modifier Optional modifier for the child
  subroutine setChild(node, name, child, replace, list, modifier)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: name
    type(fnode), pointer :: child
    logical, intent(in), optional :: replace
    logical, intent(in), optional :: list
    character(len=*), optional, intent(in) :: modifier

    logical :: tReplace, tList
    type(fnode), pointer :: dummy
    
    if (present(replace)) then
      tReplace = replace
    else
      tReplace = .false.
    end if
    if (present(list)) then
      tList = list
    else
      tList = .false.
    end if

    child => getFirstChildByName(node, tolower(name))
    if (associated(child)) then
      if (tReplace) then
        dummy => removeChild(node, child)
        call destroyNode(child)
      else
        call detailedError(node, MSG_EXISTING_CHILD // name)
      end if
    end if
    child => createElement(tolower(name))
    dummy => appendChild(node, child)
    call setAttribute(child, attrName, name)
    call setAttribute(child, attrProcessed, "")
    if (tList) then
      call setAttribute(child, attrList, "")
    end if
    if (present(modifier)) then
      call setAttribute(child, attrModifier, modifier)
    end if
    
  end subroutine setChild


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! Miscancellous routines
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  !!* Returns the content of the first TEXT_NODE child of a given node or
  !!* empty string, if such node does not exist.
  !!* @param node The node to investigate.
  !!* @return String representation of the TEXT_NODE.
  !!* @note If the document tree is normalized, every node has only one
  !!*   TEXT_NODE child.
  subroutine getFirstTextChild(node, str)
    type(fnode), pointer :: node
    type(string), intent(out) :: str

    type(fnode), pointer :: child

    child => getFirstChild(node)
    if (.not. associated(child)) then
      str = ""
    elseif (getNodeType(child) /= TEXT_NODE) then
      call detailedError(child, "Invalid node type.")
    else
      call getNodeValue(child, str)
    end if

  end subroutine getFirstTextChild


  
  !!* Checks if error flag signalizes an error. If yes, raises error.
  !!* @param node Node which the error flag was set for
  !!* @param iErr Content of the error flag.
  !!* @param msg  Message to print, if error occured
  subroutine checkError(node, iErr, msg)
    type(fnode), pointer :: node
    integer, intent(in) :: iErr
    character(len=*), intent(in) :: msg

    if (iErr == TOKEN_ERROR) then
      call detailedError(node, msg)
    elseif (iErr == TOKEN_EOS) then
      call detailedError(node, "Unexpected end of data")
    end if

  end subroutine checkError

  

  !!* Prints detailed error, including line number and path
  !!* @param node Node where the error occured.
  !!* @param msg  Message to print
  subroutine detailedError(node, msg)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: msg

    type(string) :: str

    str = msg
    call appendPathAndLine(node, str)
    call error(char(str) // newline)
    call unstring(str)

  end subroutine detailedError



  !!* Prints detailed warning, including line number and path
  !!* @param node Node where the error occured.
  !!* @param msg  Message to print
  subroutine detailedWarning(node, msg)
    type(fnode), pointer :: node
    character(len=*), intent(in) :: msg

    type(string) :: str

    str = msg
    call appendPathAndLine(node, str)
    call warning(char(str) // newline)
    call unstring(str)

  end subroutine detailedWarning



  !!* Appends path and line information to a string.
  !!* @param node Node, for which path and line should be added
  !!* @param str  String prepending the path and line information
  subroutine appendPathAndLine(node, str)
    type(fnode), pointer :: node
    type(string), intent(inout) :: str

    type(string) :: str2, str3

    call append_to_string(str, newline // "Path: ")
    call getHSDPath(node, str2, excludeRoot=.true.)
    call append_to_string(str, str2)
    call getAttribute(node, attrStart, str2)
    call getAttribute(node, attrEnd, str3)
    if (len(str2) /= 0) then
      call append_to_string(str, newline // "Line: ")
      call append_to_string(str, str2)
      if (len(str3) /= 0) then
        call append_to_string(str, "-")
        call append_to_string(str, str3)
      end if
    end if
    call getAttribute(node, attrFile, str2)
    if (len(str2) /= 0) then
      call append_to_string(str, " (File: ")
      call append_to_string(str, str2)
      call append_to_string(str, ")")
    end if
    call unstring(str2)
    call unstring(str3)

  end subroutine appendPathAndLine






end module HSDUtils
