!!* Contains routines related to calculating the density matrix from inverse
!!* iteration for banded matrices.
!!* @version \$Id: inverseIt.F90,v 1.2 2005/03/31 16:36:39 aradi Exp $
module invIt
#include "allocate.h"
#include "assert.h"
  use accuracy, only : dp
  use periodic
  use message
  implicit none

  character(len=100) :: error_string !* Used to return runtime diagnostics
  
  private
  public :: invIt
  
  interface invIt
    module procedure invIt_DM_real
  end interface
  
contains
  
  !!* Banded matrix routines to calculate eigenvectors from inverse iteration
  !!* (using lapack eigenvalues), then adds onto a given density matrix the
  !!* contribution from them weighted by the filling  
  !!* @param rho Density matrix
  !!* @param over overlap in sparse form
  !!* @param ham Hamiltonian in sparse form
  !!* @param ei eigenvalues of interest
  !!* @param filling occupation number of the eigenvectors
  !!* @param iNeighbor Neighbor list for each atom (First index from 0!)
  !!* @param nNeighbor Nr. of neighbors for each atom (incl. itself).
  !!* @param mAngAtom Maximal angular momentum for each atom.
  !!* @param mmAng Maximum total angular momentum
  !!* @param iSquare Atom offset for the squared Hamiltonian
  !!* @param img2CentCell Atomic mapping indexes.
  !!* @param bandwidth banded matrix width
  !!* @todo eigenvalues degenerate to within 3 decimal places need to be
  !!* orthogonalised, also generalisation of routines for k-points
  !!* @author Ben Hourahine
  subroutine invIt_DM_real( rho,over,ham,ei,filling,iNeighbor,nNeighbor, &
      & mAngAtom,mmAng,iSquare,img2CentCell,bandwidth )
    
    real(dp), intent(inout) :: rho(:)
    real(dp), intent(in)    :: over(:)
    real(dp), intent(in)    :: ham(:)
    real(dp), intent(in)    :: ei(:)
    real(dp), intent(in)    :: filling(:)
    integer,  intent(in)    :: iNeighbor(0:,:)
    integer,  intent(in)    :: nNeighbor(:)
    integer,  intent(in)    :: mAngAtom(:)
    integer,  intent(in)    :: mmAng
    integer,  intent(in)    :: iSquare(:)
    integer,  intent(in)    :: img2CentCell(:)
    integer,  intent(in)    :: bandwidth

    integer                 :: ii, jj, kk
    integer                 :: iSeed(4)
    real(dp), allocatable   :: AlambdaBsparse(:)
    real(dp), allocatable   :: AlambdaB(:,:)
    real(dp), allocatable   :: BB(:,:), eigenvec(:,:)
    
    ASSERT(size(ei) == size(filling))
    ASSERT(size(ham) == size(over))
    ASSERT(size(rho) == size(ham))
    ASSERT(bandwidth > 0)
    
    ALLOCATE_(AlambdaBsparse,(size(ham)))
    ALLOCATE_(AlambdaB,(bandwidth,size(filling)))
    ALLOCATE_(BB,(bandwidth,(size(filling))))
    ALLOCATE_(eigenvec,(size(filling),2))
    
    do ii = 1, size(ei) - 1
      if (ei(ii) >= (ei(ii+1) - 0.001_dp) ) then ! eigenvalues too degenerate
        ! so would have to orthonormalise, stop for the moment :
        write(error_string,*)'Eigenvalues too degenerate for inverse iteration'&
            & ,ii,ii+1
        call error(error_string)
      end if        
    end do

    ! get overlap matrix into a banded form
    call unpackHS(BB,over,iNeighbor, nNeighbor, mAngAtom, mmAng, &
        & iAtomStart, img2CentCell, bandwidth)

    iSeed(1) = 124
    iSeed(2) = 101
    iSeed(3) = 1184
    iSeed(4) = 10    
    
    do ii = 1, size(ei)
      
      if (abs(filling(ii)) >= epsilon(0.0_dp)) then
        
        ! get the characteristic equation for the ii-th eigenvector
        AlambdaBsparse(:) = ham(:) - filling(ii)*over(:)
        call unpackHS(AlambdaB,AlambdaBsparse,iNeighbor, &
            & nNeighbor, mAngAtom, mmAng, iAtomStart, img2CentCell, bandwidth)

        ! random start vector
        initEigenVec(iSeed,eigenvec(:,2))
        ! scale by the overlap
        call sbmv(eigenvec(:,1),BB,eigenvec(:,2))
        eigenvec(:,2) = eigenvec(:,1)
        
        ! need  dpbsv.f for this now to solve (A - e_i B) x = B x
        ! (potentially as a loop)
        
      end if
    end do

    DEALLOCATE_(AlambdaB)
    DEALLOCATE_(AlambdaBsparse)
    DEALLOCATE_(BB)
    DEALLOCATE_(eigenvec)
    
  end subroutine invIt_DM_real
  
  subroutine initEigenVec(iSeed,eigenvec)
    use lapack, only : larnv
    integer, intent(inout)   :: iSeed(4)
    real(dp), intent(out)    :: eigenvec(:)
    
    eigenvec(:) = 0.0_dp
    call larnv(2,iSeed,eigenvec)
    eigenvec(:) = eigenvec(:) / sqrt(sum(eigenvec(:)**2)) ! normalise to |1|    
  end subroutine initEigenVec
  
end module invIt
