!!* Reverse Cuthill-McKee reordering
module rcm
#include "allocate.h"
#include "assert.h"
  use constants
  use periodic
  use accuracy
  use sorting
  implicit none
  private

  public :: reorderAtoms, max_bandwidth

  integer, allocatable :: iUniqNeigh(:,:) !* List of neighbours, using
  !* numbering as in only the central cell
  integer, allocatable :: nUniqNeigh(:) !* Number of neighbours numbered to
  !* include all surroundings but without including shift for image atom
  !*  number
  integer, private :: maxNoNeigh

contains
  
  subroutine reorderAtoms(coord0,specie0,iNeighbor,nNeighbor,img2CentCell)
    real(dp),intent(inout) :: coord0(:,:)
    integer, intent(inout) :: specie0(:)
    integer, intent(in) :: iNeighbor(0:,:)
    integer, intent(in) :: nNeighbor(:)
    integer, intent(in) :: img2CentCell(:)
    
    real(dp) :: tmpCoord0(3,size(coord0,dim=2))
    integer :: tmpSpecie0(size(coord0,dim=2))
    
    integer, allocatable :: newOrder(:)
    integer, allocatable :: invMapNewOrder(:)
    
    integer :: ii

    ALLOCATE_(newOrder,(size(nNeighbor)))
    ALLOCATE_(invMapNewOrder,(size(nNeighbor)))

    tmpCoord0(:,:) = coord0(:,:)
    tmpSpecie0(:) = specie0(:)
    
    call bandwidth_minimise(iNeighbor,nNeighbor,img2CentCell,newOrder&
        &,invMapNewOrder)

    write(*,*)'New atom ordering :'
    do ii = 1,size(nNeighbor)
      write(*,'(''New atom '',i4,'' was atom '',i4)')newOrder(ii),ii
    end do      
    do ii = 1,size(nNeighbor)
      tmpCoord0(:,ii) = coord0(:,newOrder(ii))
      tmpSpecie0(ii) = specie0(newOrder(ii))
    end do
    coord0(:,:) = tmpCoord0(:,:)
    specie0(:) = tmpSpecie0(:)

    DEALLOCATE_(newOrder)
    DEALLOCATE_(invMapNewOrder)

  end subroutine reorderAtoms

  subroutine bandwidth_minimise(iNeighbor,nNeighbor,img2CentCell,newOrder&
      &,invMapNewOrder)
    integer, intent(in) :: iNeighbor(0:,:)
    integer, intent(in) :: nNeighbor(:)
    integer, intent(in) :: img2CentCell(:)
    integer, intent(out) :: newOrder(:)
    integer, intent(out) :: invMapNewOrder(:)

    integer :: n
    integer, allocatable :: RCM_map(:)
    integer, allocatable :: invRCM_map(:)
    integer :: ii, jj, kk
    integer ::  turner
    integer, allocatable :: tmpNeigh(:)
    integer :: maxBandwidth, minBandwidth

    ASSERT(size(nNeighbor) == size(iNeighbor, dim=2))
    ASSERT(size(img2CentCell) >= size(nNeighbor))
    ASSERT(size(newOrder) == size(nNeighbor))
    ASSERT(size(invMapnewOrder) == size(nNeighbor))

    call topologyNeighbours(nNeighbor,iNeighbor,img2CentCell)

    n = size(nNeighbor)
    ALLOCATE_(RCM_map,(n))
    ALLOCATE_(invRCM_map,(n))
    ALLOCATE_(tmpNeigh,(maxNoNeigh))

    minBandwidth = n
    do kk = 1, n
      turner = kk
        call RCM_order(RCM_map,invRCM_map,nUniqNeigh,iUniqNeigh,turner)
        maxBandwidth = 0
        do ii=1,n
          tmpNeigh(:)=0
          tmpNeigh(1:maxNoNeigh)=iUniqNeigh(1:maxNoNeigh,RCM_map(ii))
          do jj=1,nUniqNeigh(RCM_map(ii))
            tmpNeigh(jj) = invRCM_map(tmpNeigh(jj))
          end do
          call heap_sort(tmpNeigh(1:maxNoNeigh))
          if (abs(tmpNeigh(1)-ii) > maxBandwidth) then
            maxBandwidth = abs(tmpNeigh(1)-ii)
          end if
          if (abs(tmpNeigh(maxNoNeigh)-ii) > maxBandwidth) then
            maxBandwidth = abs(tmpNeigh(maxNoNeigh)-ii)
          end if
        end do
          if (minBandwidth > maxBandwidth) then
            minBandwidth = maxBandwidth
          end if
        if (minBandwidth == maxBandwidth) then
          newOrder(:) = RCM_map(:)
          invMapNewOrder(:) = invRCM_map(:)
        end if
    end do

    DEALLOCATE_(tmpNeigh)
    DEALLOCATE_(RCM_map)
    DEALLOCATE_(invRCM_map)
    DEALLOCATE_(iUniqNeigh)
    DEALLOCATE_(nUniqNeigh)

  end subroutine bandwidth_minimise


  !!* Counts the number of unique atoms (from the original central cell
  !!* list) bonded to each atom, including cases where i < j
  !!* @param nUniqNeigh
  !!* @param nNeigh The number of bonded interactions, including image atoms
  !!* @param iNeigh List of neighbours (Note: First array index running from
  !!* 0!)
  !!* @param img2CentCell Returns for all atom the index of an atom in
  !!* the central cell which the atom is mapped on to.
  subroutine topologyNeighbours(nNeigh,iNeigh,img2CentCell)
    integer, intent(in) :: nNeigh(:)
    integer, intent(in) :: iNeigh(0:,:)
    integer, intent(in) :: img2CentCell(:)

    integer :: ii, ij
    integer :: n, maxNeigh

    ASSERT(size(nNeigh) == size(iNeigh, dim=2))
    ASSERT(size(img2CentCell) >= size(nNeigh))
    n = size(iNeigh, dim=2)
    maxNeigh = size(iNeigh, dim=1) - 1
    ALLOCATE_(iUniqNeigh,(2*maxNeigh,n))
    ALLOCATE_(nUniqNeigh,(n))
    nUniqNeigh(:) = 0
    iUniqNeigh(:,:) = 0
    nUniqNeigh(:) = 0
    do ii = 1, n
      do ij = 1, nNeigh(ii)
        iUniqNeigh(ij,ii) = img2CentCell(iNeigh(ij,ii))
      end do
      if (nNeigh(ii) > 0) then
        call heap_sort(iUniqNeigh(1:nNeigh(ii),ii))
        nUniqNeigh(ii) = unique(iUniqNeigh(1:nNeigh(ii),ii))
      end if
    end do
    do ii = 1, n - 1
      do ij = 1, nUniqNeigh(ii)
        nUniqNeigh(iUniqNeigh(ij,ii)) = nUniqNeigh(iUniqNeigh(ij,ii)) + 1
        iUniqNeigh(nUniqNeigh(iUniqNeigh(ij,ii)),iUniqNeigh(ij,ii)) = ii
      end do
    end do
    do ii = 1, n
      if (nUniqNeigh(ii) > 0) then
        call heap_sort(iUniqNeigh(1:nUniqNeigh(ii),ii))
        nUniqNeigh(ii) = unique(iUniqNeigh(1:nUniqNeigh(ii),ii))
      end if
    end do
    do ii = 1, n
      if (nUniqNeigh(ii)>1) then
        if (iUniqNeigh(1,ii) == ii) then
          iUniqNeigh(1,ii) = iUniqNeigh(2,ii)
        end if
      else
        if (iUniqNeigh(1,ii) == ii) then
          iUniqNeigh(1,ii) = 0
        end if
      end if
      do ij=2,nUniqNeigh(ii)
        if (iUniqNeigh(ij,ii) == ii) then
          iUniqNeigh(ij,ii) = iUniqNeigh(ij-1,ii)
          exit
        end if
      end do
    end do
    maxNoNeigh = 0
    do ii = 1, n
      if (nUniqNeigh(ii) > 0) then
        call heap_sort(iUniqNeigh(1:nUniqNeigh(ii),ii))
        nUniqNeigh(ii) = unique(iUniqNeigh(1:nUniqNeigh(ii),ii))
        if (maxNoNeigh < nUniqNeigh(ii)) then
          maxNoNeigh = nUniqNeigh(ii)
        end if
      end if
    end do
  end subroutine topologyNeighbours

  !!* @todo trap for cases where there are disconnected graphs in the input
  subroutine RCM_order(RCM_ordering,reverseRCM,inNUniqNeigh,inIUniqNeigh,&
      &startNode)
    integer, intent(out) :: RCM_ordering(:)
    integer, intent(out) :: reverseRCM(:)
    integer, intent(in) :: inNUniqNeigh(:)
    integer, intent(in) :: inIUniqNeigh(:,:)
    integer, intent(in) :: startNode

    integer, allocatable :: localIUniqNeigh(:,:)
    integer, allocatable :: localNUniqNeigh(:)
    integer, allocatable :: visitedYet(:)
    integer :: icount, ii, jj, tmpInt

    ASSERT(size(RCM_ordering)==size(reverseRCM))
    ASSERT(size(RCM_ordering)==size(inNUniqNeigh))
    ASSERT(startNode <= size(reverseRCM))
    ASSERT(size(inNUniqNeigh) > 1)
    ALLOCATE_(localNUniqNeigh,(size(inNUniqNeigh,dim=1)))
    ALLOCATE_(visitedYet,(size(inNUniqNeigh,dim=1)))
    ALLOCATE_(localIUniqNeigh,(size(inIUniqNeigh,dim=1), \
        size(inNUniqNeigh, dim=1)))
    localNUniqNeigh(:) = inNUniqNeigh(:)
    localIUniqNeigh(:,:) = inIUniqNeigh(:,:)
    visitedYet(:) = 0
    RCM_ordering(:) = 0
    icount = 2
    RCM_ordering(1) = startNode
    visitedYet(RCM_ordering(1))=1
    do ii = 2, size(inNUniqNeigh)
      do jj = 1, localNUniqNeigh(RCM_ordering(ii-1))
        if (visitedYet(localIUniqNeigh(jj,RCM_ordering(ii-1))) == 0) then
          RCM_ordering(icount) = localIUniqNeigh(jj,RCM_ordering(ii-1))
          icount = icount + 1
          visitedYet(localIUniqNeigh(jj,RCM_ordering(ii-1))) = 1
        end if
      end do
    end do
    DEALLOCATE_(localNUniqNeigh)
    DEALLOCATE_(localIUniqNeigh)
    DEALLOCATE_(visitedYet)
    do ii = 1, size(inNUniqNeigh)/2
      tmpInt = RCM_ordering(ii)
      RCM_ordering(ii) = RCM_ordering(size(inNUniqNeigh)-ii+1)
      RCM_ordering(size(inNUniqNeigh)-ii+1) = tmpInt
    end do
    do ii = 1, size(inNUniqNeigh)
      reverseRCM(RCM_ordering(ii))= ii
    end do
  end subroutine RCM_order

  subroutine retroMap(A,B)
    integer, intent(inout) :: A(0:,:)
    integer, intent(inout) :: B(:)

    integer :: ii, jj, kk
    integer, allocatable :: C(:,:)
    integer, allocatable :: D(:)
    integer :: matches

    ASSERT(size(A,dim=2)==size(B))
    ALLOCATE_(C,(0:size(A,dim=1)-1,size(A,dim=2)))
    ALLOCATE_(D,(size(B)))

    C(:,:) = 0
    D(:) = 0

    do ii = size(B), 1, -1
      C(0,ii) = ii
      do jj = 1, B(ii)
        if (A(jj,ii) >= ii) then
          C(D(ii)+1,ii) = A(jj,ii)
          D(ii) = D(ii) + 1
        else
          matches = 0
          do kk = 1, B(A(jj,ii))
            if (A(kk,A(jj,ii)) == ii) then
              matches = 1
              exit
            end if
          end do
          if (matches == 0) then
            D(A(jj,ii)) = D(A(jj,ii)) + 1
            C(D(A(jj,ii)),A(jj,ii)) = ii
          end if
        end if
      end do
      call heap_sort(C(1:D(ii),ii))
    end do

    A(:,:) = C(:,:)
    B(:) = D(:)

    DEALLOCATE_(C)
    DEALLOCATE_(D)

    write(*,*)'Neighbours'
    do ii = 1, size(B)
      write(*,*)ii,B(ii),':'
      write(*,*)A(0,ii)
      write(*,*)A(1:B(ii),ii)
    end do

  end subroutine retroMap

  function max_bandwidth(nNeighbor,iNeighbor,img2CentCell,iSquare&
      &,mAngAtom)
    integer :: max_bandwidth
    integer, intent(in) :: nNeighbor(:)
    integer, intent(in) :: iNeighbor(:,:)
    integer, intent(in) :: img2CentCell(:)
    integer, intent(in) :: iSquare(:)
    integer, intent(in) :: mAngAtom(:)

    integer ii, jj, mAng22

    max_bandwidth = 0

    do ii = 1,size(nNeighbor)
      do jj = 1,nNeighbor(ii)
        mAng22 = (mAngAtom(img2CentCell(iNeighbor(jj,ii)))+1)**2
        if ((iSquare(img2CentCell(iNeighbor(jj,ii))) - iSquare(ii) &
            & + mAng22-1) > max_bandwidth)&
            & then
          max_bandwidth = iSquare(img2CentCell(iNeighbor(jj,ii))) &
              & - iSquare(ii) + mAng22 -1
        end if
      end do
    end do
    max_bandwidth = max_bandwidth + 1
  end function max_bandwidth

end module rcm
