!!* Andersen thermostat
!!* Two versions of the Andersen thermostat are implemented, either the global
!!* re-select or per atom reselect of velocities from the Maxwell-Boltzmann
!!* distribution
!!* @ref Andersen J. Chem. Phys. 72. 2384 (1980)
module AndersenThermostat
#include "assert.h"
#include "allocate.h"  
  use Accuracy
  use MDCommon
  use Ranlux
  use TempProfile
  implicit none
  private

  public :: OAndersenThermostat
  public :: create, destroy, getInitVelocities, updateVelocities

  !!* Data for the Andersen thermostat
  type OAndersenThermostat
    private
    integer :: nAtom                    !* Nr. of atoms
    type(ORanlux), pointer :: pRanlux   !* Random number generator
    real(dp), pointer :: mass(:)        !* Mass of the atoms
    type(OTempProfile), pointer :: pTempProfile !* Temperature generator
    logical :: tRescaleIndiv            !* Rescale velocities individually?
    real(dp) :: wvScale                 !* Rescaling probability
    type(OMDCommon), pointer :: pMDFramework  !* MD framework
  end type OAndersenThermostat

  
  interface create
    module procedure AndersenThermostat_create
  end interface

  interface destroy
    module procedure AndersenThermostat_destroy
  end interface

  interface getInitVelocities
    module procedure AndersenThermostat_getInitVelos
  end interface

  interface updateVelocities
    module procedure AndersenThermostat_updateVelos
  end interface


contains

  !!* Creates an Andersen thermostat instance.
  !!* @param self Initialised instance on exit.
  !!* @param pRanlux Pointer to the random generator.
  !!* @param masses Masses of the atoms.
  !!* @param tempProfile Pointer to a temperature profile object.
  !!* @param rescaleIndiv If velocities should be rescaled per atom
  !!* @param wvScale Rescaling probability.
  subroutine AndersenThermostat_create(self, pRanlux, masses, tempProfile, &
      &rescaleIndiv, wvScale, pMDFramework)
    type(OAndersenThermostat), pointer :: self
    type(ORanlux), pointer :: pRanlux
    real(dp), intent(in) :: masses(:)
    type(OTempProfile), pointer :: tempProfile
    logical, intent(in) :: rescaleIndiv
    real(dp), intent(in) :: wvScale
    type(OMDCommon), pointer :: pMDFramework

    ASSERT(associated(pRanlux))
    
    INITALLOCATE_P(self)
    self%pRanlux => pRanlux
    self%nAtom = size(masses)
    INITALLOCATE_PARR(self%mass, (self%nAtom))
    self%mass(:) = masses(:)
    self%pTempProfile => tempProfile
    self%tRescaleIndiv = rescaleIndiv
    self%wvScale = wvScale
    self%pMDFramework => pMDFramework
    
  end subroutine AndersenThermostat_create



  !!* Destroys an Ansersen thermostat.
  !!* @param self AndersenThermostat instance.
  subroutine AndersenThermostat_destroy(self)
    type(OAndersenThermostat), pointer :: self

    if (.not. associated(self)) then
      return
    end if
    call destroy(self%pTempProfile)
    DEALLOCATE_PARR(self%mass)
    DEALLOCATE_P(self)
    
  end subroutine AndersenThermostat_destroy



  !!* Returns the initial velocities.
  !!* @param self AndersenThermostat instance.
  !!* @param velocities Contains the velocities on return.
  subroutine AndersenThermostat_getInitVelos(self, velocities)
    type(OAndersenThermostat), pointer :: self
    real(dp), intent(out) :: velocities(:,:)

    real(dp) :: kT
    integer :: ii

    ASSERT(associated(self))
    ASSERT(all(shape(velocities) <= (/ 3, self%nAtom /)))
    
    call getTemperature(self%pTempProfile, kT)
    do ii = 1, self%nAtom
      call MaxwellBoltzmann(velocities(:,ii), self%mass(ii), kT, self%pRanlux)
    end do
    call restFrame(self%pMDFramework, velocities, self%mass)
    call rescaleTokT(self%pMDFramework, velocities, self%mass, kT)

  end subroutine AndersenThermostat_getInitVelos

  

  !!* Updates the provided velocities according the current temperature.
  !!* @param self AndersenThermostat instance.
  !!* @param velocities Updated velocities on exit.
  subroutine AndersenThermostat_updateVelos(self, velocities)
    type(OAndersenThermostat), pointer :: self
    real(dp), intent(inout) :: velocities(:,:)

    real(dp) :: rescaleChance
    real(dp) :: kT
    integer :: ii

    ASSERT(associated(self))
    ASSERT(all(shape(velocities) <= (/ 3, self%nAtom /)))

    call getTemperature(self%pTempProfile, kT)
    if (self%tRescaleIndiv) then
      do ii = 1, self%nAtom
        call getRandom(self%pRanlux, rescaleChance)
        if (rescaleChance <= self%wvScale) then
          call MaxwellBoltzmann(velocities(:,ii), self%mass(ii), kT, &
              &self%pRanlux)
        end if
      end do
      call restFrame(self%pMDFramework, velocities, self%mass)
    else
      ! all atoms re-set at random
      call getRandom(self%pRanlux, rescaleChance)
      if (rescaleChance <= self%wvScale) then
        do ii = 1, self%nAtom
          call MaxwellBoltzmann(velocities(:,ii), self%mass(ii), kT, &
              &self%pRanlux)
        end do
        call restFrame(self%pMDFramework, velocities, self%mass)
        call rescaleTokT(self%pMDFramework, velocities, self%mass, kT)
      end if
    end if

  end subroutine AndersenThermostat_updateVelos
  
  
end module AndersenThermostat
