!!* Contains wrapper for all thermostats.
module Thermostat
#include "allocate.h"  
  use Accuracy
  use DummyThermostat
  use AndersenThermostat
  use BerendsenThermostat
  implicit none
  private

  public :: OThermostat
  public :: create, destroy, getInitVelocities, updateVelocities

  !!* Data for the termostat wrapper.
  type OThermostat
    private
    integer :: thermostat
    type(ODummyThermostat), pointer :: pDummy
    type(OAndersenThermostat), pointer :: pAndersen
    type(OBerendsenThermostat), pointer :: pBerendsen
  end type OThermostat

  interface create
    module procedure Thermostat_create_Dummy
    module procedure Thermostat_create_Andersen
    module procedure Thermostat_create_Berendsen
  end interface

  interface destroy
    module procedure Thermostat_destroy
  end interface

  interface getInitVelocities
    module procedure Thermostat_getInitVelocities
  end interface

  interface updateVelocities
    module procedure Thermostat_updateVelocities
  end interface

  !! Thermostat types
  integer, parameter :: dummy_ = 0
  integer, parameter :: andersen_ = 1
  integer, parameter :: berendsen_ = 2

contains


  !!* Creates a thermostat wrapper for a DummyThermostat.
  !!* @param self Wrapper instance on exit.
  !!* @param pThermostat Pointer to a DummyThermostat.
  subroutine Thermostat_create_Dummy(self, pThermostat)
    type(OThermostat), pointer :: self
    type(ODummyThermostat), pointer :: pThermostat

    INITALLOCATE_P(self)
    self%thermostat = dummy_
    self%pDummy => pThermostat
    
  end subroutine Thermostat_create_Dummy
  

  
  !!* Creates a thermostat wrapper for an AndersenThermostat.
  !!* @param self Wrapper instance on exit.
  !!* @param pThermostat Pointer to a AndersenThermostat.
  subroutine Thermostat_create_Andersen(self, pThermostat)
    type(OThermostat), pointer :: self
    type(OAndersenThermostat), pointer :: pThermostat

    INITALLOCATE_P(self)
    self%thermostat = andersen_
    self%pAndersen => pThermostat
    
  end subroutine Thermostat_create_Andersen
    
  !!* Creates a thermostat wrapper for a BerendsenThermostat.
  !!* @param self Wrapper instance on exit.
  !!* @param pThermostat Pointer to a BerendsenThermostat.
  subroutine Thermostat_create_Berendsen(self, pThermostat)
    type(OThermostat), pointer :: self
    type(OBerendsenThermostat), pointer :: pThermostat

    INITALLOCATE_P(self)
    self%thermostat = berendsen_
    self%pBerendsen => pThermostat
    
  end subroutine Thermostat_create_Berendsen
  
  
  !!* Destroys the thermostat wrapper.
  !!* @param self Wrapper instance.
  subroutine Thermostat_destroy(self)
    type(OThermostat), pointer :: self

    if (.not. associated(self)) then
      return
    end if
    select case (self%thermostat)
    case (dummy_)
      call destroy(self%pDummy)
    case (andersen_)
      call destroy(self%pAndersen)
    case (berendsen_)
      call destroy(self%pBerendsen) 
    end select
    DEALLOCATE_P(self)
    
  end subroutine Thermostat_destroy
  
  !!* Returns the initial velocities
  !!* @param self Wrapper instance.
  !!* @param velocities Velocities on exit.
  subroutine Thermostat_getInitVelocities(self, velocities)
    type(OThermostat), pointer :: self
    real(dp), intent(out) :: velocities(:,:)

    select case (self%thermostat)
    case (dummy_)
      call getInitVelocities(self%pDummy, velocities)
    case(andersen_)
      call getInitVelocities(self%pAndersen, velocities)
    case(berendsen_)
      call getInitVelocities(self%pBerendsen, velocities)  
    end select

  end subroutine Thermostat_getInitVelocities

  !!* Updates the velocities.
  !!* @param self Wrapper instance.
  !!* @param velocities Updated velocities on exit.
  !!* @note The DummyThermostat has no method to update the velocities,
  !!*   so the wrapper returns without touching the velocities.
  subroutine Thermostat_updateVelocities(self, velocities)
    type(OThermostat), pointer :: self
    real(dp), intent(inout) :: velocities(:,:)

    select case (self%thermostat)
    case (dummy_)
      continue
    case(andersen_)
      call updateVelocities(self%pAndersen, velocities)
    case(berendsen_)
      call updateVelocities(self%pBerendsen, velocities)
    end select

  end subroutine Thermostat_updateVelocities
  

end module Thermostat
