!!* Velocity Verlet intergrator.
module VelocityVerlet
#include "assert.h"
#include "allocate.h"
  use Accuracy
  use Thermostat
  use FileId
  use message
  implicit none
  private

  public :: OVelocityVerlet
  public :: create, destroy, next

  character(len=120) :: error_string !* Used to return runtime diagnostics  

  !!* Data for the integrator.
  type OVelocityVerlet
    private
    integer :: nAtom                     !* Nr. of atoms
    real(dp) :: deltaT                   !* time step for the integrator
    real(dp), pointer :: positions(:,: ) !* list of particle positions
    real(dp), pointer :: velocities(:,:) !* list of particle velocities
    type(OThermostat), pointer :: pThermostat !* Thermostat
    logical           :: vHalfPresent = .false. !* do we have the v(t-.5)
    !* internal velocity state?
  end type OVelocityVerlet

  interface create
    module procedure VelocityVerlet_themostats
    module procedure VelocityVerlet_velocities
  end interface

  interface destroy
    module procedure VelocityVerlet_destroy
  end interface

  interface next
    module procedure VelocityVerlet_next
  end interface
    
contains

  !!* Creates a VelocityVerlet object from the thermostat settings
  !!* @param self Pointer to the initialised object on exit.
  !!* @param deltaT Integration time step.
  !!* @param positions Position of the atoms.
  !!* @param pThermostat Pointer to a thermostat
  subroutine VelocityVerlet_themostats(self, deltaT, positions, pThermostat)
    type(OVelocityVerlet), pointer       :: self
    real(dp), intent(in)                 :: deltaT
    real(dp), intent(in)                 :: positions(:,:)
    type(OThermostat), pointer           :: pThermostat    

    ASSERT(size(positions, dim=1) == 3)
    
    INITALLOCATE_P(self)
    self%nAtom = size(positions, dim=2)
    INITALLOCATE_PARR(self%velocities, (3, self%nAtom))
    INITALLOCATE_PARR(self%positions, (3, self%nAtom))
    
    self%deltaT = deltaT
    self%positions(:,:) = positions(:,:)
    self%pThermostat => pThermostat

    call getInitVelocities(pThermostat, self%velocities)

    self%vHalfPresent = .true. ! yes we have the t-.5 velocities
    
  end subroutine VelocityVerlet_themostats


  !!* Creates a VelocityVerlet object from given external velocities for the
  !!* t-th time step, this means later we have to reconstruct the Vel. Verlet
  !!* t+.5 velocities
  !!* @param self Pointer to the initialised object on exit.
  !!* @param deltaT Integration time step.
  !!* @param positions Position of the atoms.
  !!* @param pThermostat Pointer to a thermostat.
  !!* @param velocities list of initial velocities
  subroutine VelocityVerlet_velocities(self, deltaT, positions, pThermostat, &
      & velocities)
    type(OVelocityVerlet), pointer       :: self
    real(dp), intent(in)                 :: deltaT
    real(dp), intent(in)                 :: positions(:,:)
    type(OThermostat), pointer           :: pThermostat    
    real(dp), intent(in)                 :: velocities(:,:)

    ASSERT(size(positions, dim=1) == 3)
    
    INITALLOCATE_P(self)
    self%nAtom = size(positions, dim=2)
    INITALLOCATE_PARR(self%velocities, (3, self%nAtom))
    INITALLOCATE_PARR(self%positions, (3, self%nAtom))
    
    self%deltaT = deltaT
    self%positions(:,:) = positions(:,:)
    self%pThermostat => pThermostat
    
    self%velocities(:,:) = velocities(:,:)

    self%vHalfPresent = .false. ! assumes the V read in corresponds to the
    ! current coordinates, so we should reconstruct the t+.5 velocities when
    ! possible once forces are available for the coordinates
    
  end subroutine VelocityVerlet_velocities

  
  !!* removes an integrator example
  !!* @param self the instance to deallocate
  subroutine VelocityVerlet_destroy(self)
    type(OVelocityVerlet), pointer :: self

    if (.not. associated(self)) then
      return
    end if
    DEALLOCATE_PARR(self%velocities)
    DEALLOCATE_PARR(self%positions)
    if (associated(self%pThermostat)) then
      call destroy(self%pThermostat)
    end if
    DEALLOCATE_P(self)

  end subroutine VelocityVerlet_destroy

  
  !!* Takes a timestep for the MD integrator, optionally with a thermostat.
  !!* @param self integrator propogate
  !!* @param accel Accelerations.
  !!* @param newCoord displaced coordinates
  !!* @param newVelocity velocity of displaced coords
  !!* @caveat Due to the way the velocity Verlet is split, the returned
  !!* velocity is for 1 complete MD step behind the returned positions 
  !!* so print positions, then call next and then print velocities
  !!* to get agreement between the positions and velocities.
  subroutine VelocityVerlet_next(self, accel, newCoord, newVelocity)
    type(OVelocityVerlet), pointer :: self
    real(dp),intent(in) :: accel(:,:)
    real(dp),intent(out) :: newCoord(:,:)
    real(dp),intent(out) :: newVelocity(:,:)
    
    ASSERT(associated(self))

    newCoord(:,:) = 0.0_dp
    newVelocity(:,:) = 0.0_dp
    if (associated(self%pThermostat)) then
      call updateVelocities(self%pThermostat, self%velocities)
    end if
    
    ! start from the usual ordering of velocity verlet method (two cycles
    ! shown):
    ! a.1 v(t+.5dt) = v(t)        + .5*a(t)*dt -- a(t) external
    ! a.2 r(t + dt) = r(t)        + v(t+.5dt)*dt
    ! a.3 v(t + dt) = v(t+.5dt)   + .5*a(t+dt)*dt -- a(t+dt) external call
    ! b.1 v(t+1.5dt) = v(t+dt)    + .5*a(t+dt)*dt -- a(t+dt) external
    ! b.2 r(t + 2dt) = r(t+dt)    + v(t+1.5dt)*dt 
    ! b.3 v(t + 2dt) = v(t+1.5dt) + .5*a(t+2dt)*dt -- a(t+2dt) external call
    !
    ! and cut out a.3 b.1 b.2 as the cycle :
    ! a.3 v(t)      = v(t-.5dt)+ .5*a(t)*dt -- a(t) input
    ! b.1 v(t+.5dt) = v(t)     + .5*a(t)*dt
    ! b.2 r(t+dt)   = r(t)     + v(t+.5dt)*dt
    !
    ! so :
    ! v_out   = v_store + .5*a_input*dt
    ! v_store = v_out   + .5*a_input*dt
    ! r_out   = r_store + v_store*dt
    ! r_store = r_out
    ! where v_out is one MD step behind the positions returned.

    if (self%vHalfPresent) then
      newVelocity(:,:) = self%velocities(:,:) &
          & + 0.5_dp * accel(:,:) * self%deltaT
    else
      newVelocity(:,:) = self%velocities(:,:)
      self%vHalfPresent=.true.
    end if
    
    self%velocities(:,:) = newVelocity(:,:) + 0.5_dp * accel(:,:) * self%deltaT
    newCoord(:,:) = self%positions(:,:) + self%velocities(:,:) * self%deltaT
    self%positions(:,:) = newCoord(:,:)

  end subroutine VelocityVerlet_next

end module VelocityVerlet
