!!* Routines to read/write a TGeometry type in HSD and XML format.
module TypeGeometryHSD
# include "allocate.h"  
  use TypeGeometry
  use HSDUtils
  use HSDUtils2
  use TokenReader
  use UnitConversion
  use LinkedList
  use StringList
  use CharManip
  use flib_dom
  use flib_wxml
  use LapackRoutines
  implicit none

  private

  !!* Writes the content of a geometry object to a dom tree or to an xml-writer
  interface writeTGeometryHSD
    module procedure writeTGeometryHSD_dom
    module procedure writeTGeometryHSD_xmlf
  end interface

  !! Types/subroutines from TypeGeometry
  public :: TGeometry, destruct

  !! Locally defined subroutines
  public :: writeTGeometryHSD, readTGeometryHSD, readTGeometryGen


contains

  !!* Write the geometry in HSD format to a specified node
  !!* @param node Node in the HSD-tree which should contain the geometry
  !!* @param geo  The geometry
  subroutine writeTGeometryHSD_dom(node, geo)
    type(fnode), pointer :: node
    type(TGeometry), intent(in) :: geo

    call setChildValue(node, "TypeNames", geo%specieNames, .false.)
    call setChildValue(node, "TypesAndCoordinates", &
        &reshape(geo%species, (/ 1, size(geo%species) /)), geo%coords, .false.)
    call setChildValue(node, "Periodic", geo%tPeriodic, .false.)
    if (geo%tPeriodic) then
      call setChildValue(node, "LatticeVectors", geo%latVecs, .false.)
    end if
    
  end subroutine writeTGeometryHSD_dom



  !!* Write the geometry in HSD format to an xml writer
  !!* @param node Node in the HSD-tree which should contain the geometry
  !!* @param geo  The geometry
  subroutine writeTGeometryHSD_xmlf(xf, geo)
    type(xmlf_t), intent(inout) :: xf
    type(TGeometry), intent(in) :: geo

    call writeChildValue(xf, "TypeNames", geo%specieNames)
    call writeChildValue(xf, "TypesAndCoordinates", &
        &reshape(geo%species, (/ 1, size(geo%species) /)), geo%coords)
    call writeChildValue(xf, "Periodic", geo%tPeriodic)
    if (geo%tPeriodic) then
      call writeChildValue(xf, "LatticeVectors", geo%latVecs)
    end if
    
  end subroutine writeTGeometryHSD_xmlf



  !!* Read the geometry from a node in a HSD tree.
  !!* @param node Node in the HSD tree containing the geomery
  !!* @param geo  Contains the geometry on exit
  subroutine readTGeometryHSD(node, geo)
    type(fnode), pointer :: node
    type(TGeometry), intent(out) :: geo

    type(string) :: modifier
    integer :: ind, iErr
    type(listString) :: stringBuffer
    type(listRealR1) :: realBuffer
    type(listIntR1) :: intBuffer
    type(fnode), pointer :: child, typesAndCoords
    integer, allocatable :: tmpInt(:,:)
    real(dp) :: latvec(9)

    call getChildValue(node, "Periodic", geo%tPeriodic, default=.false.)
    call init(stringBuffer)
    call getChildValue(node, "TypeNames", stringBuffer)
    geo%nSpecie = len(stringBuffer)
    if (geo%nSpecie == 0) then
      call detailedError(node, "Missing specie names.")
    end if
    INITALLOCATE_PARR(geo%specieNames, (geo%nSpecie))
    call asArray(stringBuffer, geo%specieNames)
    call destroy(stringBuffer)
    call init(intBuffer)
    call init(realBuffer)
    call getChildValue(node, "TypesAndCoordinates", 1, intBuffer, 3, &
        &realBuffer, modifier=modifier, child=typesAndCoords)
    geo%nAtom = len(intBuffer)
    if (geo%nAtom == 0) then
      call detailedError(typesAndCoords, "Missing coordinates")
    end if
    INITALLOCATE_PARR(geo%species, (geo%nAtom))
    INITALLOCATE_PARR(geo%coords, (3, geo%nAtom))
    ALLOCATE_(tmpInt, (1, geo%nAtom))
    call asArray(intBuffer, tmpInt)
    geo%species(:) = tmpInt(1,:)
    DEALLOCATE_(tmpInt)
    !! Check validity of species
    if (any(geo%species < 1 .or. geo%species > geo%nSpecie)) then
      call detailedError(typesAndCoords, "Type index must be between 1 and " &
          &// i2c(geo%nSpecie) // ".")
    end if
    call asArray(realBuffer, geo%coords)
    call destroy(intBuffer)
    call destroy(realBuffer)
    geo%tFracCoord = .false.
    if (len(modifier) > 0) then
      select case(tolower(char(modifier)))
      case ("relative")
        if (.not. geo%tPeriodic) then
          call detailedError(typesAndCoords, "Relative coordinates are only &
              &allowed for periodic systems")
        end if
        geo%tFracCoord = .true.
      case default
        ind = getModifierIndex(char(modifier), lengthUnits, typesAndCoords)
        geo%coords(:,:) = geo%coords(:,:) * lengthUnits(ind)%value
        call setChildValue(typesAndCoords, "", &
            &reshape(geo%species, (/ 1, size(geo%species) /)), geo%coords, &
            &replace=.true.)
      end select
    end if
    if (geo%tPeriodic) then
      INITALLOCATE_PARR(geo%latVecs, (3,3))
      call getChildValue(node, "LatticeVectors", latvec, modifier=modifier, &
          &child=child)
      geo%latVecs(:,:) = reshape(latvec, (/3, 3/))
      if (len(modifier) > 0) then
        ind = getModifierIndex(char(modifier), lengthUnits, child)
        geo%latVecs = geo%latVecs(:,:) * lengthUnits(ind)%value
        call setChildValue(child, "", geo%latVecs, .true.)
      end if
      if (geo%tFracCoord) then
        geo%coords = matmul(geo%latVecs, geo%coords)
        !call setChildValue(typesAndCoords, "", &
        !    &reshape(geo%species, (/ 1, size(geo%species) /)), &
        !    &geo%coords, replace=.true.)
      end if
      INITALLOCATE_PARR(geo%recVecs2p, (3, 3))
      geo%recVecs2p(:,:) = geo%latVecs(:,:)
      call matinv(geo%recVecs2p, iError=iErr)
      if (iErr /= 0) then
        call detailedError(child, "Dependent lattice vectors")
      end if
      geo%recVecs2p(:,:) = reshape(geo%recVecs2p, (/3, 3/), order=(/2, 1/))
    end if
    call unstring(modifier)
    
  end subroutine readTGeometryHSD



  !!* Reads the geometry from a node in a HSD tree in GEN format
  !!* @param node Node containing the geometry in Gen format
  !!* @param geo  Contains the geometry on exit
  subroutine readTGeometryGen(node, geo)
    type(fnode), pointer :: node
    type(TGeometry), intent(out) :: geo

    type(string) :: text

    call getFirstTextChild(node, text)
    call readTGeometryGen_help(node, geo, char(text))
    call unstring(text)

  end subroutine readTGeometryGen



  !!* Helping routine for reading geometry from a HSD tree in GEN format
  !!* @param node Node to parse (only needed to produce proper error messages)
  !!* @param geo  Contains the geometry on exit
  !!* @param text Text content of the node
  subroutine readTGeometryGen_help(node, geo, text)
    type(fnode), pointer :: node
    type(TGeometry), intent(out) :: geo
    character(len=*), intent(in) :: text

    type(string) :: txt
    integer :: iStart, iOldStart, iErr
    integer :: ii, iTmp
    real(dp) :: coords(3), rTmp
    type(listString) :: specieNames

    !! Read first line of the gen file: Number of atoms, boundary conditions
    iStart = 1
    call getNextToken(text, geo%nAtom, iStart, iErr)
    call checkError(node, iErr, "Bad number of atoms.")
    call getNextToken(text, txt, iStart, iErr)
    select case (char(txt))
    case("S","s")
      geo%tPeriodic = .true.
      geo%tFracCoord = .false.
    case("F","f")
      geo%tPeriodic = .true.
      geo%tFracCoord = .true.
    case("C", "c")
      geo%tPeriodic = .false.
      geo%tFracCoord = .false.
    case default
      call detailedError(node, "Unknown boundary condition type '" &
          &// char(txt) // "'")
    end select

    !! Reading the 2nd line of a gen file.
    !! Since we can't rely on line breaks, we try to read in integers. If it
    !! fails, we had a specie name, so it will be read as string.
    call init(specieNames)
    iErr = TOKEN_ERROR
    iOldStart = iStart
    call getNextToken(text, iTmp, iStart, iErr)
    do while (iErr /= TOKEN_OK)
      call getNextToken(text, txt, iStart, iErr)
      if (iErr /= TOKEN_OK) then
        call detailedError(node, "Unexpected end of geometry information.")
      end if
      call append(specieNames, char(txt))
      iOldStart = iStart
      call getNextToken(text, iTmp, iStart, iErr)
    end do
    geo%nSpecie = len(specieNames)
    if (geo%nSpecie == 0) then
      call detailedError(node, "Number of species equals zero.")
    end if
    INITALLOCATE_PARR(geo%specieNames, (geo%nSpecie))
    call asArray(specieNames, geo%specieNames)
    call destroy(specieNames)

    !! Read in sequential and specie indices.
    INITALLOCATE_PARR(geo%species, (geo%nAtom))
    INITALLOCATE_PARR(geo%coords, (3, geo%nAtom))
    iStart = iOldStart
    do ii = 1, geo%nAtom
      call getNextToken(text, iTmp, iStart, iErr)
      call checkError(node, iErr, "Bad sequentual number for an atom.")
      call getNextToken(text, geo%species(ii), iStart, iErr)
      call checkError(node, iErr, "Bad specie number for an atom.")
      call getNextToken(text, coords, iStart, iErr)
      call checkError(node, iErr, "Bad coordinates for an atom.")
      geo%coords(:, ii) = coords(:)
    end do
    geo%nSpecie = maxval(geo%species(:))
    if (geo%nSpecie /= maxval(geo%species) .or. minval(geo%species) /= 1) then
      call detailedError(node, &
          &"Nr. of species and nr. of specified element do not match.")
    end if

    !! Read in origin an lattice vectors, if the structure is periodic
    if (geo%tPeriodic) then
      INITALLOCATE_PARR(geo%origin, (3))
      INITALLOCATE_PARR(geo%latVecs, (3, 3))
      call getNextToken(text, geo%origin, iStart, iErr)
      call checkError(node, iErr, "Invalid origin.")
      do ii = 1, 3
        call getNextToken(text, geo%latVecs(:, ii), iStart, iErr)
        call checkError(node, iErr, "Invalid lattice vectors.")
      end do
      geo%origin = geo%origin * AA__Bohr
      geo%latVecs = geo%latVecs * AA__Bohr
      if (geo%tFracCoord) then
        if (any(abs(geo%coords) > 1.0_dp)) then
          call detailedWarning(node, &
              &"Fractional coordinates with absolute value greater than one.")
        end if
        geo%coords = matmul(geo%latVecs, geo%coords) 
      else
        geo%coords = geo%coords * AA__Bohr
      end if
      INITALLOCATE_PARR(geo%recVecs2p, (3, 3))
      geo%recVecs2p(:,:) = geo%latVecs(:,:)
      call matinv(geo%recVecs2p, iError=iErr)
      if (iErr /= 0) then
        call detailedError(node, "Dependent lattice vectors")
      end if
    else
      geo%coords = geo%coords * AA__Bohr
    end if
    call getNextToken(text, rTmp, iStart, iErr)
    if (iErr /= TOKEN_EOS) then
      call detailedError(node, "Superfluous data found. Check if specified &
          &number of atoms matches the number of actually entered positions.")
    end if
    call unstring(txt)
    
  end subroutine readTGeometryGen_help



end module TypeGeometryHSD
