!!* The main dftb+ program
!!* @todo Wrapping the entire stuff using well defined APIs for initialisation,
!!* getting energy, forces etc. (Scheduled for DFTB+ 1.1)
program dftb
# include "allocate.h"
  use constants, only : Hartree__eV, Bohr__AA
  use initprogram
  use inputdata_
  use nonscc
  use Eigenvects
  use repulsive
  use eTemp
  use populations
  use densityMatrix
  use forces
  use TaggedOutput
  use SCC
  use Mixer
  use GeoOpt
  use Spin
  use DFTBplsU
  use FileId
  use FormatOut
  use MDCommon
  use Energies
  use OrbitalEquiv
  use Parser
  use flib_wxml
  use periodic, only : unpackHS
  use blasRoutines, only : hemv
  use HSDUtils
  use Charmanip
  implicit none
    
  !! Revision control strings
  character(len=*), parameter :: revision = &
      &"$Revision: 3130 $"
  character (len=*), parameter :: headURL = &
      &"$HeadURL: svn+ssh://svn/noodle/trunk/src/prg_dftb/dftb+.F90 $"
  
  type(inputData)          :: input             ! Contains the parsed input

  integer                  :: nk, iEgy, nSpin2, nK2, iSpin2, iK2
  complex(dp), allocatable :: HSqrCplx(:,:,:,:), SSqrCplx(:,:)
  real(dp),    allocatable :: HSqrReal(:,:,:), SSqrReal(:,:)
  real(dp),    allocatable :: eigen(:,:,:)
  real(dp), allocatable    :: rhoPrim(:,:)
  real(dp), allocatable    :: inputRhoPrim(:,:)
  real(dp), allocatable    :: rhoDiff(:,:)
  real(dp), allocatable    :: ERhoPrim(:,:)

  real(dp), allocatable    :: filling(:,:,:)
  real(dp)                 :: Eband(2), TS(2), Ef(2), E0(2), Eold

  type(TEnergies)          :: energy
  
  real(dp), allocatable    :: derivs(:,:),repulsiveDerivs(:,:),totalDeriv(:,:)
  real(dp), allocatable    :: chrgForces(:,:)

  integer                  :: iSCCIter
  integer                  :: iSpin
  integer                  :: ii, jj, kk

  logical                  :: tConverged

  logical, parameter       :: tDensOn = .false.  ! O(N^2) density mtx creation
  logical, parameter       :: tAppendDetailedOut = .false.

  character(len=8), parameter :: formatEnergy = '(8f12.5)'
  character(len=8), parameter :: formatEigen = '(8f14.8)'
  character(len=*), parameter :: formatGeoOut = "(I5,F16.8,F16.8,F16.8)"
  character(len=*), parameter :: formatE1 = "(' ',A,':',T30,F20.6)"
  character(len=*), parameter :: formatE2 = &
      &"(' ',A,':',T30,F18.10,' H',T50,F16.6,' eV')"

  !! Variables for the geometry optimization
  integer :: iGeoStep                      !* Geometry steps so far
  logical :: tGeomEnd                      !* Do we have the final geometry?
  real(dp), allocatable :: coord0Fold(:,:) !* Folded initial coords (3, nAtom)
  real(dp), allocatable :: newCoord(:)     !* New coordinates returned by
  !* the optimizer
  real(dp), allocatable :: new3Coord(:,:)     !* New coordinates returned by  
  !* the MD routines
  real(dp), allocatable :: velocities(:,:) !* MD velocities
  real(dp), allocatable :: movedVelo(:,:)  !* MD velocities for moved atoms
  real(dp), allocatable :: movedAccel(:,:) !* MD acceleration for moved atoms
  real(dp), allocatable :: movedMass(:,:)  !* Mass of the moved atoms
  real(dp) :: KE                           !* MD Kinetic energy
  real(dp) :: kT                           !* MD instantaneous thermal energy

  real(dp) :: diffGeo                      !* Difference between last calculated
  !* and new geometry.

  !!* Loop variables
  integer :: iAt1, iOrb1


  integer :: fdTagged  !!* File descriptor for the tagged writer
  integer :: fdUser    !!* File descriptor for the human readable output
  integer :: fdBand    !!* File descriptor for the band structure output
  integer :: fdEigvec  !!* File descriptor for the eigenvector output
  integer :: fdResultsTag !!* File descriptor for detailed.tag

  !!* Name of the human readable file
  character(*), parameter :: taggedOut = "autotest.tag"
  character(*), parameter :: userOut = "detailed.out"
  character(*), parameter :: bandOut = "band.out"
  character(*), parameter :: eigvecOut = "eigenvec.out"
  character(*), parameter :: eigvecBin = "eigenvec.bin"
  character(*), parameter :: fDensMtx = "densmtx.bin"
  character(*), parameter :: resultsTag = "results.tag"

  real(dp) :: sccErrorQ           !* Charge error in the last iterations
  real(dp) :: sccErrorRho         !* Density mtx. error in the last iterations
  real(dp) :: rTmp
  real(dp), allocatable    :: rVecTemp(:)
  complex(dp), allocatable :: cVecTemp(:)
  character(lc) :: lcTmp

  real(dp) :: EbandTmp, EfTmp, TSTmp, E0Tmp

  character(lc) :: tmpStr  !!* temporary character variable



  logical :: tWriteRestart = .false. !* flag to write out geometries (and 
  !* charge data if scc) when moving atoms about - in the case of conjugate
  !* gradient/steepest descent the geometries are written anyway
  integer :: minSCCIter                   !* Minimal number of SCC iterations

  type(xmlf_t) :: xf
  real(dp), allocatable :: bufferRealR2(:,:)
  logical :: tStopSCC, tStopDriver   ! if scf/driver should be stopped
  integer :: ang, iSh1, iSp1

  call printDFTBHeader(revision, headURL, parserVersion)
  write (*,'(/A/A/A/)') repeat("*", 80), "** Parsing and initializing", &
      &repeat("*", 80)

  !! Parse input and set the variables in the local scope according the input.
  !! These variables are defined in the initprogram module.

  call parseHSDInput(input)
  write (*,"(/A)") "Starting initialization..."
  write (*,"(A80)") repeat("-", 80)
  call initProgramVariables(input)
  write (*,"(A80)") repeat("-", 80)
  call destroy(input)
  write (*,*)
  
  if (tWriteTagged) then
    !! Initialize tagged writer and human readable output
    call initTaggedWriter()
    fdTagged = getFileId()
    open(fdTagged, file=taggedOut, position="rewind", status="replace")
    close(fdTagged)
  end if
  if (tWriteBandDat) then
    fdBand = getFileId()
  end if
  fdEigvec = getFileId()

  ! initially empty file
  if (tWriteDetailedOut) then
    fdUser = getFileId()
    open(fdUser, file=userOut, position="rewind", status="replace")
  end if

  ALLOCATE_(rhoPrim, (0, nSpin))
  if (tForces) then
    ALLOCATE_(ERhoPrim, (0, nSpin))
  end if
  if (tDFTBU) then
    ALLOCATE_(inputRhoPrim, (0, nSpin))
    ALLOCATE_(rhoDiff, (0, nSpin))
  end if

  if (tForces) then
    ALLOCATE_(derivs,(3,nAtom))
    ALLOCATE_(repulsiveDerivs,(3,nAtom))
    ALLOCATE_(totalDeriv, (3,nAtom))
    if (tExtChrg) then
      ALLOCATE_(chrgForces, (3, nExtChrg))
    end if
  end if

  call create(energy,nAtom,nSpin)

  ALLOCATE_(eigen, (nOrb, nKPoint,nSpin))
  ALLOCATE_(filling,(nOrb,nKpoint,nSpin))

  ALLOCATE_(coord0Fold, (3, nAtom))

  if (tMD) then
    ALLOCATE_(new3Coord, (3, nMovedAtom))
  end if
  if (tGeoOpt) then
    ALLOCATE_(newCoord, (nMovedCoord))
  end if

  eigen(:,:,:) = 0.0_dp
  filling(:,:,:) = 0.0_dp
  TS(:) = 0.0_dp
  Eband(:) = 0.0_dp
  Ef(:) = 0.0_dp
  E0(:) = 0.0_dp

  if (tStoreEigvecs) then
    nSpin2 = 1
    nK2 = 1
  else
    nSpin2 = nSpin
    nK2 = nKPoint
  end if

  ! If only H/S should be printed, no allocation for square HS is needed
  if (.not. (tWriteRealHS .or. tWriteHS)) then
    if (tRealHS) then
      ALLOCATE_(HSqrReal, (nOrb, nOrb, nSpin2))
      ALLOCATE_(SSqrReal, (nOrb, nOrb))
    else
      ALLOCATE_(HSqrCplx, (nOrb, nOrb, nK2, nSpin2))
      ALLOCATE_(SSqrCplx, (nOrb, nOrb))
    end if
  end if

  if (tMD) then
    ALLOCATE_(velocities,(3,nAtom))
    ALLOCATE_(movedVelo, (3, nMovedAtom))
    ALLOCATE_(movedAccel, (3, nMovedAtom))
    ALLOCATE_(movedMass, (3, nMovedAtom))
    movedMass(:,:) = spread(mass(specie0(indMovedAtom)),1,3)
    velocities(:,:) = 0.0_dp
  end if

  !! Document initial variables in tagged form
  if (tWriteTagged) then
    open(fdTagged, file=taggedOut, position="append")
    call writeTagged(fdTagged, tag_SCC, tSCC)
    call writeTagged(fdTagged, tag_spin, nSpin)
    call writeTagged(fdTagged, tag_DFTBU, tDFTBU)    
    if (tDFTBU) then
      call writeTagged(fdTagged, tag_nDFTBU, nDFTBUfunc)
    end if
    call writeTagged(fdTagged, tag_dispersn, tDispersion)
    !! Only temporary hack, as long as autotest is not recalculated
    call writeTagged(fdTagged, tag_mAngSpecie, orb%nShell(:)-1)
    call writeTagged(fdTagged, tag_kPoint, kPoint)
    call writeTagged(fdTagged, tag_kWeight, kWeight)
    call writeTagged(fdTagged, tag_atomEigVal, atomEigVal)
    call writeTagged(fdTagged, tag_hubbU, hubbU)
    call writeTagged(fdTagged, tag_tempElec, tempElec)
    call writeTagged(fdTagged, tag_distribFn, iDistribFn)
    if (size(nEl) == 1) then
      call writeTagged(fdTagged, tag_nElUp, nEl(1))
      call writeTagged(fdTagged, tag_nElDown, nEl(1))
    else
      call writeTagged(fdTagged, tag_nElUp, nEl(1))
      call writeTagged(fdTagged, tag_nElDown, nEl(2))
    end if
    close(fdTagged)
  end if

  E0(:) = 0.0_dp


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !! Geometry loop
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  tGeomEnd = nGeoSteps == 0

  iGeoStep = 0
  tStopDriver = .false.
  lpGeomOpt: do while (iGeoStep <= nGeoSteps)

    if (restartFreq > 0 .and. tGeoOpt .or. tMD) then
      tWriteRestart = (iGeoStep == nGeoSteps .or. &
          & (mod(iGeoStep, restartFreq) == 0 ))
    else
      tWriteRestart = .false.
    end if
    
    !! Write out geometry information
    write (*,'(/A)') repeat("*", 80)
    write (*,'(A,I5)') "** Geometry step:", iGeoStep
    write (*,'(A/)') repeat("*", 80)
    if (tGeoOpt .or. tMD) then
      write (lcTmp, "(A,A)") trim(geoOutFile), ".gen"
      call clearFile(trim(lcTmp))
      if (tPeriodic) then
        call writeGenFormat(trim(lcTmp), coord0, specie0, specieName, latVec, &
            & tFracCoord)
      else
        call writeGenFormat(trim(lcTmp), coord0, specie0, specieName)
      end if
      write (lcTmp, "(A,A)") trim(geoOutFile), ".xyz"
      if (iGeoStep == 0 .or. .not. (tMD .or. tAppendGeo)) then
        call clearFile(trim(lcTmp))
      end if      
    end if
    
    !! Save old coordinates and fold coords to unit cell
    coord0Fold(:,:) = coord0(:,:)
    if (tPeriodic) then
      call foldCoordToUnitCell(coord0Fold, latVec, recVec2p)
    end if

    !! Initialize neighborlists
    call updateNeighborListAndSpecies(coord, specie, img2CentCell, iCellVec, &
        &neighborList, nAllAtom, coord0Fold, specie0, mCutoff, rCellVec)
    nAllOrb = sum(orb%nOrbSpecie(specie(1:nAllAtom)))

    !! Calculate neigbhorlist for SK and repulsive calculation
    call getNrOfNeighborsForAll(nNeighbor, neighborList, skRepCutoff)

    !! Reallocate Hamiltonian and overlap based on the new neighbor list
    call reallocateHS(ham, over, iPair, neighborList%iNeighbor, nNeighbor, &
        &orb, img2CentCell)
    if (tSCC) then
      call reallocateHS(h0, h1, iPair, neighborList%iNeighbor, nNeighbor, &
          &orb, img2CentCell)
      if (nSpin == 2) then
        call reallocateHS(hspin, iPair, neighborList%iNeighbor, nNeighbor, &
            &orb, img2CentCell)
      end if
    end if

    !! Reallocate density matrixes if necessary
    if (size(ham, dim=1) > size(rhoPrim, dim=1)) then
      DEALLOCATE_(rhoPrim)
      ALLOCATE_(rhoPrim,(size(ham,dim=1),nSpin))
      if (tForces) then
        DEALLOCATE_(ERhoPrim)
        ALLOCATE_(ERhoPrim,(size(ham,dim=1),nSpin))
      end if
      if (tDFTBU) then
        DEALLOCATE_(inputRhoPrim)
        ALLOCATE_(inputRhoPrim,(size(ham,dim=1),nSpin))
        DEALLOCATE_(rhoDiff)
        ALLOCATE_(rhoDiff,(size(ham,dim=1),nSpin))
      end if
    end if

    !! (Re)Initialize mixer
    if (tSCC) then
      if (tDFTBU) then
        call reset(pChrgMixer, size(rhoPrim))
      else
        call reset(pChrgMixer, nIneqOrb)
      end if
    end if

    if (tWriteTagged) then
      !! Write out initialization information
      if (iGeoStep == 0) then
        open(fdTagged, file=taggedOut, position="append")
        call writeTagged(fdTagged, tag_initCoord, coord(:,:nAtom))
        call writeTagged(fdTagged, tag_specie, specie(:nAtom))
        call writeTagged(fdTagged, tag_nNeighbor, neighborList%nNeighbor)
        call writeTagged(fdTagged, tag_iNeighbor, neighborList%iNeighbor)
        close(fdTagged)
      end if
    end if

    !! Notificate the SCC module about the current coordinates
    if (tSCC) then
      call updateCoords_SCC(coord, specie, neighborList, img2CentCell)
    end if
    if (tDispersion) then
      call updateCoords(pDispersion, neighborList, img2CentCell, coord)
    end if

    !! Initialize density matrix for LDA+U from external file
    if (tDFTBU .and. tReadChrg) then
      call readRhoFromFile(inputRhoPrim, fDensMtx, nNeighbor, &
          &neighborList%iNeighbor, img2CentCell, iPair, orb)
    end if


    !! Build non-scc Hamiltonian
    if (tSCC) then
      !! Non-scc Ham for an SCC calc is built in a temporary hamiltonian
      call buildH0S(H0,over, skHamCont, skOverCont, atomEigVal, coord, &
          &nNeighbor, neighborList%iNeighbor, specie, iPair, orb)
    else
      !! Non-scc Ham for non-SCC calc is built in the final hamiltonian
      call buildH0S(ham(:,1),over, skHamCont, skOverCont, atomEigVal, &
          & coord, nNeighbor, neighborList%iNeighbor, specie, iPair, orb)
    end if
    
    !! Adapt electron temperature to MD, if necessary
    if (tSetFillingTemp) then
      call getTemperature(pTempProfile, tempElec)
    end if


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !! SCC-loop
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    if (tSCC .and. (.not. tAppendDetailedOut)) then
      if (.not. tDFTBU) then
        write(*,"(' ',A5,A18,A18,A18)") "iSCC", " Total electronic ", &
            & "  Diff electronic ", "     SCC error    "
      else
        write(*,"(' ',A5,A18,A18,A18,A18)") "iSCC", " Total electronic ", &
            & "  Diff electronic ", "     SCC error    ", &
            & " DFTBU-SCC error  "
      end if
    end if

    tConverged = .false.
    if (tSCC) then
      if (tDFTBU) then
        minSCCIter = 3
      else
        if (nSpin == 1) then
          minSCCIter = 1
        else
          minSCCIter = 2
        end if
      end if
    else
      minSCCIter = 1
    end if    

    energy%ETotal = 0.0_dp
    energy%atomTotal(:) = 0.0_dp
    
    !! Calculate repulsive energy
    call getERep(energy%atomRep, coord, nNeighbor, neighborList%iNeighbor, &
        &specie, pRepCont, img2CentCell)
    energy%Erep = sum(energy%atomRep)

    if (tDispersion) then
      call getEnergies(pDispersion, energy%atomDisp)
      energy%eDisp = sum(energy%atomDisp)
    else
      energy%atomDisp(:) = 0.0_dp
    end if
    
    iSCCIter = 1
    tStopSCC = .false.
    lpSCC: do while (iSCCiter <= nSCCIter)
      
      rhoPrim(:,:) = 0.0_dp

      !! Build charge contribution to the Hamiltonian
      if (.not. tSCC) then
        tConverged = .true.
      else
        call updateCharges_SCC(qInput, q0, orb, specie, &
            &neighborList%iNeighbor, img2CentCell)
        call buildH1(H1, over, nNeighbor, neighborList%iNeighbor, specie, &
            &orb, iPair, img2CentCell)

        !! Build spin contribution (if necessary)
        if (nSpin == 2) then
          Hspin(:)=0.0_dp          
          call buildHspin(Hspin, over, nNeighbor, neighborList%iNeighbor, &
              &specie, orb, iPair, img2CentCell, qInput, W)
          do ii = 1, size(H0)
            ham(ii,1) = H0(ii) + H1(ii) + Hspin(ii)
            ham(ii,2) = H0(ii) + H1(ii) - Hspin(ii)
          end do
        else
          ham(:,1) = H0(:) + H1(:)
        end if

        !! Apply LDA+U correction (if necessary)
        if (tDFTBU) then
          if (iSCCiter > 1 .or. tReadChrg .or. iGeoStep > 0) then
            call add_VplsU(ham(:,:),inputRhoPrim(:,:), nNeighbor, &
                & neighborList%iNeighbor, specie, orb, iPair, img2CentCell, &
                & UJ, nU, nLinU, iLVals,nDFTBUfunc)
          end if
        end if
      end if

      ! Write out matrices if necessary and quit.
      call writeHS(tWriteHS, tWriteRealHS, ham, over, neighborList%iNeighbor, &
          &nNeighbor, iAtomStart, iPair, img2CentCell, kPoint, iCellVec, &
          &cellVec)
      if (tWriteRealHS .or. tWriteHS) then
        write (*, "(A)") "Hamilton/Overlap written, exiting program."
        stop
      end if
      
      
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      !! Spin loop
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


      lpSpin: do iSpin = 1, nSpin

        if (tStoreEigvecs) then
          iSpin2 = 1
          if (tRealHS) then
            call reset(storeEigvecsReal(iSpin), (/ nOrb, nOrb /))
          else
            call reset(storeEigvecsCplx(iSpin), (/ nOrb, nOrb /))
          end if
        else
          iSpin2 = iSpin          
        end if

        !! Solve eigenproblem for real or complex Hamiltionian
        if (tRealHS) then

          call diagonalize(HSqrReal(:,:,iSpin2), SSqrReal, eigen(:,1,iSpin), &
              &ham(:,iSpin), over, neighborList%iNeighbor, nNeighbor, &
              &iAtomStart, iPair, img2CentCell, solver, 'V')

          if (tStoreEigvecs) then
            call push(storeEigvecsReal(iSpin), HSqrReal(:,:,iSpin2))
          end if

          !! Print eigenvectors
          if (tEigenvecs) then            
            if (iSpin == 1) then
              ! write eigenvectors in binary format as well
              open(fdEigvec, file=eigvecBin, action="write", status="replace", &
                  &position="rewind", form="unformatted")
              write (fdEigVec) runId
              do ii = 1, size(HSqrReal, dim=2)                
                write (fdEigvec) HSqrReal(:,ii,iSpin2)
              end do
              close(fdEigVec)

              open(fdEigvec, file=eigvecOut, action="write", status="replace", &
                  &position="rewind");
              write (fdEigvec,"(A/)") "Coefficients and Mulliken populations&
                    & of the atomic orbitals"
            else
              ! write eigenvectors in binary format as well
              open(fdEigvec, file=eigvecBin, action="write", status="old", &
                  &position="append", form="unformatted")
              do ii = 1, size(HSqrReal, dim=2)
                write (fdEigvec) HSqrReal(:,ii,iSpin2)
              end do
              close(fdEigVec)
              
              open(fdEigvec, file=eigvecOut, action="write", status="old", &
                  &position="append");
            end if
            ALLOCATE_(rVecTemp,(size(HSqrReal, dim=1)))
            call unpackHS(SSqrReal,over,neighborList%iNeighbor, nNeighbor, &
                &iAtomStart, iPair, img2CentCell)      
            do ii = 1, nOrb
              call hemv(rVecTemp,SSqrReal,HSqrReal(:,ii,iSpin2))
              write(fdEigvec, "('Eigenvector:',I4,4X,'(Spin: ',I1,')'/)") ii, &
                  &iSpin
              jj = 0
              do iAt1 = 1, nAtom
                iSp1 = specie(iAt1)
                do iSh1 = 1, orb%nShell(iSp1)
                  ang = orb%angShell(iSh1, iSp1)
                  if (iSh1 == 0) then
                    write(tmpStr, "(I5,1X,A2,2X,A1)") iAt1, specieName(iSp1), &
                        &orbitalName(ang+1)
                  else
                    write(tmpStr, "(10X,A1)") orbitalName(ang+1)
                  end if
                  do iOrb1 = 1, 2 * ang + 1
                    jj = jj + 1
                    write(fdEigvec,"(A,I1,T15,F12.6,3X,F12.6)") trim(tmpStr), &
                        &iOrb1, HSqrReal(jj, ii, iSpin2), &
                        & HSqrReal(jj, ii, iSpin2) * rVecTemp(jj)
                  end do
                end do
                write (fdEigvec,*)
              end do
            end do
            close(fdEigvec)
            DEALLOCATE_(rVecTemp)
          end if

          !! Occupy eigenlevels including possible spin polarisation
          call Efilling(Eband(iSpin),Ef(iSpin),TS(iSpin),&
              & E0(iSpin),&
              & filling(:,:,iSpin:iSpin),&
              & eigen(:,:,iSpin:iSpin),&
              & nEl(iSpin), tempElec,kWeight,iDistribFn)

          !! prefactor 2.0_dp for spin averaged, 1.0_dp for spin polarised
          Eband(iSpin) = real(3-nSpin,dp)*Eband(iSpin)
          E0(iSpin) = real(3-nSpin,dp)*E0(iSpin)
          TS(iSpin) = real(3-nSpin,dp)*TS(iSpin)
          filling(:,:,iSpin:iSpin) = real(3-nSpin,dp)* &
              & filling(:,:,iSpin:iSpin)

          if (tDensON) then
            call makeDensityMatrix(SSqrReal, HSqrReal(:,:,iSpin2), &
                &filling(:,1,iSpin), neighborlist%iNeighbor, nNeighbor, orb, &
                &iAtomStart, img2CentCell)
          else
            call makeDensityMatrix(SSqrReal, HSqrReal(:,:,iSpin2), &
                &filling(:,1,iSpin))
          end if
          call packHS(rhoPrim(:,iSpin), SSqrReal, neighborlist%iNeighbor, &
              &nNeighbor, orb%mOrb, iAtomStart, iPair, img2CentCell)
        else
          !! k-point calculation
          !! Do unpacking/diagonalizing/density mtx creating in each K-point

          nkLoop: do nk = 1, nKPoint
            if (tStoreEigvecs) then              
              iK2 = 1
            else
              iK2 = nK              
            end if
            call diagonalize(HSqrCplx(:,:,iK2,iSpin2), SSqrCplx(:,:), &
                &eigen(:,nk,iSpin), ham(:,iSpin), over,kPoint(:,nk), &
                &neighborList%iNeighbor, nNeighbor, iCellVec, cellVec, &
                &iAtomStart, iPair, img2CentCell, solver, "V")
            if (tStoreEigvecs) then
              call push(storeEigvecsCplx(iSpin), HSqrCplx(:,:,iK2,iSpin2))
            end if
            
          end do nkLoop

          !! Fill up levels with electrons
          if (.not. tFillKSep) then
            call Efilling(Eband(iSpin),Ef(iSpin),TS(iSpin),&
                & E0(iSpin), &
                & filling(:,:,iSpin:iSpin),&
                & eigen(:,:,iSpin:iSpin),&
                & nEl(iSpin),tempElec,kWeight,iDistribFn)
          else
            !! Make aufbau principle in each K-point separately
            do nk = 1, nKPoint
              call Efilling(EbandTmp,EfTmp,TSTmp,&
                  & E0Tmp, &
                  & filling(:,nk:nk,iSpin:iSpin),&
                  & eigen(:,nk:nk,iSpin:iSpin),&
                  & nEl(iSpin),tempElec,kWeight(nk:nk),iDistribFn)
              Eband(iSpin) = eBand(iSpin) + eBandTmp
              Ef(iSpin) = Ef(iSpin) + EfTmp
              TS(iSpin) = TS(iSpin) + TSTmp
              E0(iSpin) = E0(iSpin) + E0Tmp
            end do
            !! Fermi energy meaningless in that case -> put average Ef
            Ef(iSpin) = Ef(iSpin) / real(nKPoint, dp)
          end if

          ! prefactor 2.0_dp for spin averaged, 1.0_dp for spin polarised
          Eband(iSpin) = real(3-nSpin,dp)*Eband(iSpin)
          E0(iSpin) = real(3-nSpin,dp)*E0(iSpin)
          TS(iSpin) = real(3-nSpin,dp)*TS(iSpin)
          filling(:,:,iSpin:iSpin) = real(3-nSpin,dp)* &
              & filling(:,:,iSpin:iSpin)

          nkLoop2: do nk = 1, nKPoint

            if (tStoreEigvecs) then
              iK2 = 1
              call get(storeEigvecsCplx(iSpin), HSqrCplx(:,:,iK2,iSpin2))
            else
              iK2 = nK
            end if

            if (tDensON) then
              call makeDensityMatrix(SSqrCplx, HSqrCplx(:,:,iK2,iSpin2), &
                &filling(:,nK,iSpin), neighborlist%iNeighbor, nNeighbor, orb, &
                &iAtomStart, img2CentCell)
            else
              call makeDensityMatrix(SSqrCplx, HSqrCplx(:,:,iK2,iSpin2), &
                  &filling(:,nK,iSpin))
            end if
            call packHS(rhoPrim(:,iSpin), SSqrCplx, kPoint(:,nK), &
                &kWeight(nk), neighborList%iNeighbor, nNeighbor, orb%mOrb, &
                &iCellVec, cellVec, iAtomStart, iPair, img2CentCell)

            !! Print eigenvectors
            if (tEigenvecs) then
              if (nk == 1 .and. iSpin == 1) then
                ! write eigenvectors in binary format as well
                open(fdEigvec, file=eigvecBin, action="write", &
                    &status="replace",  position="rewind", form="unformatted")
                write (fdEigVec) runId
                do ii = 1, size(HSqrCplx, dim=2)
                  write (fdEigvec) HSqrCplx(:,ii,iK2, iSpin2)
                end do
                close(fdEigVec)
                
                open(fdEigvec, file=eigvecOut, action="write", status="replace")
                write (fdEigvec,"(A/)") "Coefficients and Mulliken populations&
                    & of the atomic orbitals"
              else
                ! write eigenvectors in binary format as well
                open(fdEigvec, file=eigvecBin, action="write", &
                    &status="old",  position="append", form="unformatted")
                do ii = 1, size(HSqrCplx, dim=2)
                  write (fdEigvec) HSqrCplx(:,ii,iK2, iSpin2)
                end do
                close(fdEigVec)

                open(fdEigvec, file=eigvecOut, action="write", status="old", &
                    &position="append")
              end if
              ALLOCATE_(cVecTemp,(size(HSqrCplx, dim=1)))
              call unpackHS(SSqrCplx, over, kPoint(:,nk), &
                &neighborList%iNeighbor, nNeighbor, iCellVec, cellVec, &
                &iAtomStart, iPair, img2CentCell)
              do ii = 1, nOrb
                call hemv(cVecTemp,SSqrCplx,HSqrCplx(:, ii, iK2, iSpin2))
                write(fdEigvec, "(A,I4,4X,A,I4,4X,'(Spin: ',I1,')'/)") &
                    &"K-point: ", nk, "Eigenvector: ", ii, iSpin
                jj = 0
                do iAt1 = 1, nAtom
                  iSp1 = specie(iAt1)
                  do iSh1 = 1, orb%nShell(iSp1)
                    ang = orb%angShell(iSh1,iSp1)
                    if (iSh1 == 1) then
                      write(tmpStr, "(I5,1X,A2,2X,A1)") iAt1, &
                          &specieName(iSp1), orbitalName(ang+1)
                    else
                      write(tmpStr, "(10X,A1)") orbitalName(ang+1)
                    end if
                    do iOrb1 = 1, 2*ang+1
                      jj = jj + 1
                      write(fdEigvec,&
                          &"(A,I1,T15,'(',F12.6,',',F12.6,')',3X,F12.6)") &
                          &trim(tmpStr), iOrb1, &
                          &real(HSqrCplx(jj, ii, iK2, iSpin2)), &
                          &aimag(HSqrCplx(jj, ii, iK2, iSpin2)), &
                          & real( conjg(HSqrCplx(jj, ii, iK2, iSpin2)) &
                          & * cVecTemp(jj))
                    end do
                  end do
                  write (fdEigvec,*)
                end do
              end do
              close(fdEigvec)
              DEALLOCATE_(cVecTemp)
            end if
          end do nkLoop2
        end if

        if (tWriteBandDat) then
          if (iSpin == 1) then
            open(unit=fdBand, file=bandOut, action="write")
          end if
          do nk=1,nKPoint
            write(fdBand,*)'KPT ',nk,' SPIN ', iSpin
            do iEgy=1,nOrb
              write(fdBand, formatEnergy) Hartree__eV*eigen(iEgy,nk,iSpin),&
                  & filling(iEgy,nk,iSpin)
            end do
            write(fdBand,*)
          end do
          if (iSpin == nSpin) then
            close(fdBand)
          end if
        end if


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !! Mulliken analysis
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        if (tMulliken) then
          qOutput(:,:,iSpin) = 0.0_dp
          call mulliken(qOutput(:,:,iSpin), over, rhoPrim(:,iSpin), &
              &orb, neighborList%iNeighbor, nNeighbor, img2CentCell, iPair)
        end if
      end do lpSpin

      if (tSCC) then
        ! recalculate the SCC shifts for the output charge.
        !! SCC contribution is calculated with the output charges.
        call updateCharges_SCC(qOutput, q0, orb, specie, &
            &neighborList%iNeighbor, img2CentCell)
      end if

      !! Calculate energies
      energy%atomElec(:) = 0.0_dp
      if (tSCC) then
        do iSpin = 1, nSpin
          call mulliken(energy%atomNonSCC(:,iSpin), rhoPrim(:,iSpin), H0, orb, &
              &neighborList%iNeighbor, nNeighbor, img2CentCell, iPair)
          energy%EnonSCC(iSpin) = sum(energy%atomNonSCC(:,iSpin))
        end do
        call getEnergySCC(energy%atomSCC(:), specie, neighborList%iNeighbor, &
            &img2CentCell)
        energy%ESCC = sum(energy%atomSCC(:))
        if (nSpin > 1) then
          call getEnergySpin(energy%atomSpin(:),orb,specie,qOutput,W)
          energy%Espin = sum(energy%atomSpin(:))
        end if
        if (tDFTBU.and.(iSCCiter > 1 .or. tReadChrg .or. iGeoStep > 0)) then
          call E_DFTBplsU(energy%atomDftbu, RhoPrim, nNeighbor, &
              &neighborList%iNeighbor, specie, orb, iPair, img2CentCell, &
              &UJ, nU, nLinU, iLVals, nDFTBUfunc)
          energy%Edftbu = sum(energy%atomDftbu(:))
        end if
        energy%Eelec = sum(energy%EnonSCC(:)) + energy%ESCC + energy%Espin &
            & + energy%Edftbu
      else
        call mulliken(energy%atomNonSCC(:,1), rhoPrim(:,1), ham(:,1), orb, &
            &neighborList%iNeighbor, nNeighbor, img2CentCell, iPair)
        energy%Eelec = sum(energy%atomNonSCC(:,1))
      end if
      
      energy%atomElec(:) = sum(energy%atomNonSCC(:,:),dim=2) &
          & + energy%atomSCC(:) + energy%atomSpin(:) + energy%atomDftbu(:)

      energy%atomTotal(:) = energy%atomElec(:) + energy%atomRep(:) + &
          & energy%atomDisp(:) 
      energy%Etotal = energy%Eelec + energy%Erep + energy%eDisp
      
      !! Stop SCC if appropriate stop file is present (We need this query here
      !! since the following block contains a check iSCCIter /= nSCCIter)
      inquire(file=fStopSCC, exist=tStopSCC)
      if (tStopSCC) then
        write (*,*) "Stop file '" // fStopSCC // "' found."
        nSCCIter = iSCCIter
        write (*,*) "Setting max number of scc cycles to current cycle."
      end if

      !! Mix charges
      if (tSCC) then
        if (.not.tDFTBU) then
          call OrbitalEquiv_reduce(qOutput, iEqOrbitals, orb, qOutRed)
          qDiffRed(:) = qOutRed(:) - qInpRed(:)
          sccErrorQ = maxval(abs(qDiffRed))

          tConverged = (sccErrorQ < sccTol) .and. &
              & (iSCCiter >= minSCCIter .or. tReadChrg)
          if ((.not. tConverged) .and. iSCCiter /= nSCCiter) then
            !! Avoid mixing of spin unpolarised density for spin polarised
            !! cases, this is only a problem in iteration 1, as there is
            !! only the (spin unpolarised!) atomic input density at that
            !! point. (Unless charges had been initialized externally)
            if (iSCCIter == 1 .and. nSpin > 1 .and. .not. tReadChrg) then
              qInput(:,:,:) = qOutput(:,:,:)
              qInpRed(:) = qOutRed(:)
            else
              call mix(pChrgMixer, qInpRed, qDiffRed)
              call OrbitalEquiv_expand(qInpRed, iEqOrbitals, orb, qInput)
            end if
          end if
        else
          !! Avoid mixing in 1st iteration (unless initial density
          !! matrix provided externally)
          if (iSCCiter > 1 .or. tReadChrg .or. iGeoStep > 0) then
            rhoDiff(:,:) = rhoPrim(:,:) - inputRhoPrim(:,:)
            sccErrorRho = maxval(abs(rhoDiff))
            tConverged = sccErrorRho < sccTol &
                &.and.(iSCCiter >= minSCCIter .or. tReadChrg .or.&
                & iGeoStep > 0)
            if (.not. tConverged) then
              call mix(pChrgMixer, inputRhoPrim, rhoDiff)
              do iSpin=1,nSpin
                call mulliken(qInput(:,:,iSpin), over, inputRhoPrim(:,iSpin), &
                    &orb, neighborList%iNeighbor, nNeighbor, img2CentCell, &
                    &iPair)
              end do
            end if
          else
            inputRhoPrim(:,:) = rhoPrim(:,:)
            qInput(:,:,:) = qOutput(:,:,:)
            sccErrorRho = 0.0_dp
            sccErrorQ = 0.0_dp
          end if
        end if

      end if

      !! Clear detailed.out if necessary
      if (tWriteDetailedOut .and. .not. tAppendDetailedOut) then
        close(fdUser)
        open(fdUser, file=userOut, position="rewind", status="replace")
        select case(iDistribFn)
        case(0)
          write(fdUser,*)'Fermi distribution function'
        case(1)
          write(fdUser,*)'Gaussian distribution function'
        case default
          write(fdUser,*)'Methfessel-Paxton distribution function order'&
              &,iDistribFn
        end select
      end if
      
      if (tSCC) then
        if (iSCCiter > 1) then
          rTmp = (energy%Eelec-Eold)
        else
          rTmp = 0.0_dp
        end if
        Eold = energy%Eelec

        if (tWriteDetailedOut) then
          write(fdUser,*)''
          write (fdUser,'(/A)') repeat("*", 80)
          if (tDFTBU) then          
            write(fdUser,"(' ',A5,A18,A18,A18,A18)") "iSCC", " Total electronic ", &
                & "  Diff electronic ", "     SCC error    ", &
                & " DFTBU-SCC error  "          
            write(fdUser,"(I5,E18.8,E18.8,E18.8,E18.8)") iSCCIter, energy%Eelec, rTmp, &
                &sccErrorQ, sccErrorRho              
          else
            write(fdUser,"(' ',A5,A18,A18,A18)") "iSCC", " Total electronic ", &
                & "  Diff electronic ", "     SCC error    "
            write(fdUser,"(I5,E18.8,E18.8,E18.8)") iSCCIter, energy%Eelec, rTmp, &
                &sccErrorQ
          end if
          write (fdUser,'(A)') repeat("*", 80)
        end if

        if (tDFTBU) then
          write(*,"(I5,E18.8,E18.8,E18.8,E18.8)") iSCCIter, energy%Eelec, rTmp, &
              &sccErrorQ, sccErrorRho
        else
          write(*,"(I5,E18.8,E18.8,E18.8)") iSCCIter, energy%Eelec, rTmp, sccErrorQ
        end if

      end if

      !! Not writing any restarting info if not converged and minimal number of
      !! SCC iterations not done.
      if (restartFreq > 0 .and. .not.(tMD.or.tGeoOpt)) then
        if (tConverged .or. ((iSCCIter >= minSCCIter &
            & .or. tReadChrg .or. iGeoStep > 0) &
            &.and. (iSCCIter == nSCCIter .or. mod(iSCCIter, restartFreq) ==&
            & 0))) then
          if (tMulliken) then
            call writeQToFile(qInput, fChargeIn, orb)
            print "('>> Charges saved for restart in ',A)", fChargeIn
          end if
          if (tDFTBU) then
            call writeRhoToFile(inputRhoPrim, fDensMtx, nNeighbor, &
                &neighborList%iNeighbor, img2CentCell, iPair, specie, orb)
            print "('>> Density matrix saved for restart in ',A)", fDensMtx
          end if
        end if
      end if
      if (tWriteDetailedOut) then
        if (nMovedAtom > 0) then
          write (fdUser,*) "Coordinates of moved atoms:"
          do ii = 1, nMovedAtom
            write(fdUser,formatGeoOut)indMovedAtom(ii), &
                & coord0(:, indMovedAtom(ii))
          end do
        end if

        !! Write out atomic charges
        if (tMulliken) then
          write (fdUser, "(/,A)") " Net atomic charges"
          write (fdUser, "(1X,A5,1X,A16)")" Atom", " Net charge"
          do ii = 1, nAtom
            write (fdUser, "(1X,I5,1X,F16.8)") ii, &
                &sum(q0(:, ii, :) - qOutput(:, ii, :))
          end do
        end if

        lpSpinPrint: do iSpin = 1, nSpin
          write(fdUser,*) ' '
          write(fdUser,*) 'SPIN = ',iSpin
          write(fdUser,*) ' '
          write(fdUser,*)'Eigenvalues /H'
          do iEgy=1,nOrb
            write(fdUser, formatEigen) (eigen(iEgy,ii,iSpin),ii=1,nKPoint)
          end do
          write (fdUser,*)
          write(fdUser,*)'Eigenvalues /eV'
          do iEgy=1,nOrb
            write(fdUser, formatEigen) &
                &(Hartree__eV*eigen(iEgy,ii,iSpin),ii=1,nKPoint)
          end do
          write (fdUser,*) ''
          write(fdUser,*)'Fillings'
          do iEgy=1,nOrb
            write(fdUser, formatEnergy) (filling(iEgy,ii,iSpin),ii=1,nKPoint)
          end do

          if (tMulliken) then
            write (fdUser,*) ''
            write (fdUser,"(' Nr. of electrons (spin: ',I1,'):',F16.8)") iSpin,&
                &sum(qOutput(:, :,iSpin))
            write (fdUser, "(' Atom populations (spin: ',I1,')')") iSpin
            write (fdUser, "(1X,A5,1X,A16)")" Atom", " Population"

            do ii = 1, nAtom
              write (fdUser, "(1X,I5,1X,F16.8)") ii, sum(qOutput(:, ii, iSpin))
            end do
            write (fdUser,*) ''
            write (fdUser, "(' l-shell populations (spin: ',I1,')')") iSpin
            write (fdUser, "(1X,A5,1X,A3,1X,A3,1X,A16)")" Atom", "Sh.", &
                &"  l", " Population"
            do ii = 1, nAtom
              iSp1 = specie(ii)
              do iSh1 = 1, orb%nShell(iSp1)
                write (fdUser, "(1X,I5,1X,I3,1X,I3,1X,F16.8)") ii, iSh1, &
                    &orb%angShell(iSh1,iSp1), &
                    &sum(qOutput(orb%posShell(iSh1,iSp1)&
                    &:orb%posShell(iSh1+1,iSp1)-1, ii, iSpin))
              end do
            end do
            write (fdUser,*) ''
            write (fdUser, "(' Orbital populations (spin: ',I1,')')") iSpin
            write (fdUser, "(1X,A5,1X,A3,1X,A3,1X,A3,1X,A16)") " Atom", "Sh.",&
                &"  l","  m", " Population"
            do ii = 1, nAtom
              iSp1 = specie(ii)
              do iSh1 = 1, orb%nShell(iSp1)
                ang = orb%angShell(iSh1, iSp1)
                do kk = 0, 2 * ang
                  write (fdUser, "(' ',I5,1X,I3,1X,I3,1X,I3,1X,F16.8)") &
                      &ii, iSh1, ang, kk - ang, &
                      &qOutput(orb%posShell(iSh1,iSp1)+kk, ii, iSpin)
                end do
              end do
            end do
          end if

          write(fdUser,*) ''
          write(fdUser,formatE2) 'Fermi energy', Ef(iSpin), &
              & Hartree__eV*Ef(iSpin)
          write(fdUser,formatE2) 'Band energy', Eband(iSpin), &
              &Hartree__eV*Eband(iSpin)
          write(fdUser,formatE2)'TS', TS(iSpin), Hartree__eV*TS(iSpin)
          write(fdUser,formatE2) 'Band free energy (E-TS)',&
              &Eband(iSpin)-TS(iSpin),&
              &Hartree__eV*(Eband(iSpin)-TS(iSpin))
          write(fdUser,formatE2)'Extrapolated E(0K)',E0(iSpin),&
              &Hartree__eV*(E0(iSpin))
          write (fdUser,*) ''
          if (tMulliken) then
            write(fdUser,"(' Input/Output charge (spin: ',I1,'):',F16.8,F16.8)") &
                &iSpin, sum(qInput(:, :, iSpin)), sum(qOutput(:, :,iSpin))
          end if
        end do lpSpinPrint

        write (fdUser,*) ''
        if (nSpin == 2) then
          write(fdUser,formatE2) 'Energy H0  up', energy%EnonSCC(1), &
              & energy%EnonSCC(1)*Hartree__eV
          write(fdUser,formatE2) 'Energy H0  down', energy%EnonSCC(2), &
              & energy%EnonSCC(2)*Hartree__eV
        else
          write(fdUser,formatE2) 'Energy H0', energy%EnonSCC(1), energy%EnonSCC(1)*Hartree__eV
        end if
        if (tSCC) then
          write (fdUser,formatE2) 'Energy SCC', energy%ESCC, energy%ESCC*Hartree__eV
          if (nSpin == 2) then
            write (fdUser,formatE2) 'Energy SPIN', energy%Espin, &
                & energy%Espin*Hartree__eV
          end if
          if (tDFTBU) then
            write (fdUser,formatE2) 'Energy DFTB+U', energy%Edftbu, &
                &energy%Edftbu*Hartree__eV
          end if
        end if
        write(fdUser,formatE2) 'Total Electronic energy', energy%Eelec, &
            & energy%Eelec*Hartree__eV

        !! Write out repulsive related data

        write(fdUser,formatE2) 'Repulsive energy', energy%Erep, energy%Erep*Hartree__eV
        write(fdUser,formatE2) 'Total energy', energy%Etotal, &
            &(energy%Etotal) * Hartree__eV
        write(fdUser,formatE2) 'Total Mermin free energy', energy%Etotal -&
            & sum(TS(:)), (energy%Etotal - sum(TS(:))) * Hartree__eV

        if (tAtomicEnergy) then
          write (fdUser,*) ''
          write (fdUser, "(' Atom resolved electronic energies ')")
          do ii = 1, nAtom
            write (fdUser, "(' Atom ',I5,F16.8,' H',F16.6,' eV')") ii, &
                & energy%atomElec(ii), Hartree__eV*energy%atomElec(ii)
          end do
          write (fdUser,*) ''
        end if

        if (tAtomicEnergy) then
          write (fdUser,*) ''
          write (fdUser, "(' Atom resolved repulsive energies ')")
          do ii = 1, nAtom
            write (fdUser, "(' Atom ',I5,F16.8,' H',F16.6,' eV')") ii, &
                & energy%atomRep(ii), Hartree__eV*energy%atomRep(ii)
          end do
          write (fdUser,*) ''
          write (fdUser, "(' Atom resolved total energies ')")
          do ii = 1, nAtom
            write (fdUser, "(' Atom ',I5,F16.8,' H',F16.6,' eV')") ii, &
                & energy%atomTotal(ii), Hartree__eV*energy%atomTotal(ii)
          end do
          write (fdUser,*) ''
        end if
        
      end if

      if (tConverged) then
        exit lpSCC
      end if

      iSCCIter = iSCCIter + 1
      
    end do lpSCC

    

    write (*,*)
    write (*, formatE1) "Total Energy", energy%Etotal
    write (*, formatE1) "Total Mermin free energy", &
        &energy%Etotal - sum(TS(:))


    !! Calculate energy weighted density matrix
    if (tForces) then

      !! Calculate the energy weighted density matrix
      ERhoPrim(:,:) = 0.0_dp
      if (tRealHS) then
        do iSpin = 1, nSpin
          
          if (tStoreEigvecs) then        
            iSpin2 = 1
            call get(storeEigvecsReal(iSpin), HSqrReal(:,:,iSpin2))
          else
            iSpin2 = iSpin
          end if

          ! Build energy weighted density matrix
          if (tDensON) then
            call makeDensityMatrix(SSqrReal, HSqrReal(:,:,iSpin2), &
                &filling(:,1,iSpin), eigen(:,1,iSpin), neighborlist%iNeighbor, &
                &nNeighbor, orb, iAtomStart, img2CentCell)
          else
            call makeDensityMatrix(SSqrReal, HSqrReal(:,:,iSpin2), &
                &filling(:,1,iSpin), eigen(:,1,iSpin))
          end if
          call packHS(ERhoPrim(:,iSpin), SSqrReal, neighborList%iNeighbor, &
              &nNeighbor, orb%mOrb, iAtomStart, iPair, img2CentCell)
        end do
      else
        do iSpin = 1, nSpin
          if (tStoreEigvecs) then           
            iSpin2 = 1
          else
            iSpin2 = iSpin
          end if

          do nK = 1, nKPoint
            !! Calculate eigenvectors, if necessary. Eigenvectors for the last
            !! spin in the last k-points are still there, so use those directly
            if (tStoreEigvecs) then
              iK2 = 1
              call get(storeEigvecsCplx(iSpin), HSqrCplx(:,:,iK2, iSpin2))
            else
              iK2 = nK
            end if

            if (tDensON) then
              call makeDensityMatrix(SSqrCplx, HSqrCplx(:,:,iK2,iSpin2), &
                  &filling(:,nK,iSpin), eigen(:,nK, iSpin), &
                  &neighborlist%iNeighbor, nNeighbor, orb, iAtomStart,&
                  &img2CentCell)
            else
              call makeDensityMatrix(SSqrCplx, HSqrCplx(:,:,iK2,iSpin2), &
                  &filling(:,nK,iSpin), eigen(:,nK, iSpin))
            end if
            call packHS(ERhoPrim(:,iSpin), SSqrCplx, kPoint(:,nk), &
                &kWeight(nk), neighborList%iNeighbor, nNeighbor, orb%mOrb, &
                &iCellVec, cellVec, iAtomStart, iPair, img2CentCell)
          end do
        end do
      end if

    end if

    if (tForces) then

      derivs(:,:) = 0.0_dp
      if (tExtChrg) then
        chrgForces(:,:) = 0.0_dp
      end if

      if (.not. tSCC) then
        call derivative_nonscc(derivs,rhoPrim(:,1), ERhoPrim(:,1), skHamCont, &
            &skOverCont, coord, specie, neighborList%iNeighbor, nNeighbor, &
            &img2CentCell, iPair, orb)
      else
        !! SCC contribution calculated with input charge density used to build
        !! up the hamiltonian and the shift vectors. This is only consistent
        !! with rhoPrim, if the charges are converged.
        if (nSpin == 1) then
          call derivative_scc(derivs, rhoPrim(:,1), ERhoPrim(:,1), skHamCont, &
              &skOverCont, coord, specie, neighborList%iNeighbor, nNeighbor, &
              &img2CentCell, iPair, orb)
        else
          call derivative_spin(derivs, rhoPrim(:,:), ERhoPrim(:,:), skHamCont, &
              &skOverCont, coord, specie, neighborList%iNeighbor, nNeighbor, &
              &img2CentCell, iPair, W, qInput, orb)
        end if

        ! add double counting term in force :
        if (tExtChrg) then
          call addForceDCSCC(derivs, specie, neighborList%iNeighbor, &
              &img2CentCell, coord, chrgForces)

        else
          call addForceDCSCC(derivs, specie, neighborList%iNeighbor, &
              &img2CentCell, coord)
        end if
      end if
      
      call getERepDeriv(repulsiveDerivs, coord, nNeighbor, &
          &neighborList%iNeighbor, specie,pRepCont, img2CentCell)
      totalDeriv(:,:) = repulsiveDerivs(:,:) + derivs(:,:)

      if (tDispersion) then
        call addGradients(pDispersion, totalDeriv)
      end if
      
    end if

    !! Write out information at the end of the SCC loop

    if (tWriteDetailedOut) then
      write(fdUser,*)''
      if (nGeoSteps > 0) then
        write(fdUser,*) "Geometry optimization step:", iGeoStep
      else
        write(fdUser,*) "Calculation with static geometry"
      end if
      write(fdUser,*)''

      if (tSCC) then
        if (tConverged) then
          write (fdUser,*) "SCC converged"
        else
          call warning("SCC is NOT converged, maximal iterations exceeded")
          write (fdUser,*)"SCC is NOT converged, maximal iterations exceeded"
        end if
      else
        write (fdUser,*) "Non-SCC calculation"
      end if

      if (tGeoOpt .or. tMD) then
        write(fdUser,*)''
        write(fdUser,*) "Full geometry written in ",trim(geoOutFile), &
            & ".{xyz|gen}"
        write(fdUser,*)
      end if
        
      !! Write out forces
      if (tForces) then

        write(fdUser,*) ''
        write(fdUser,*)'Electronic Forces'
        do ii = 1, nAtom
          write(fdUser,*)-derivs(:,ii)
        end do
        write(fdUser,*)''

        write(fdUser,*)'Repulsive Forces'
        do ii = 1, nAtom
          write(fdUser,*)-repulsiveDerivs(:,ii)
        end do
        write(fdUser,*)''

        write(fdUser,*)'Total Forces'
        do ii = 1, nAtom
          write(fdUser,*) -totalDeriv(:,ii)
        end do
        write(fdUser,*)''

        write(fdUser,*) "Maximal force component:", maxval(abs(totalderiv))
        if (nMovedAtom > 0) then
          write (fdUser,*) "Maximal force component for moved atoms:", &
              &maxval(abs(totalDeriv(:,indMovedAtom)))
        end if
        write(fdUser,*)''

        if (tExtChrg) then
          write (fdUser,*) "Forces on external charges"
          do ii = 1, nExtChrg
            write (fdUser, *) -chrgForces(:,ii)
          end do
        end if
        write(fdUser,*)''

      end if
    end if

    if (tForces) then
      !! Set force components along constraint vectors zero
      do ii = 1, nGeoConstr
        iAt1 = conAtom(ii)
        totalDeriv(:,iAt1) = totalDeriv(:,iAt1) &
            &- conVec(:,ii) * dot_product(conVec(:,ii), totalDeriv(:,iAt1))
      end do

      if (tGeoOpt .and. nMovedAtom > 0) then
        write (*,formatE1) "Maximal force component", &
            &maxval(abs(totalDeriv(:,indMovedAtom)))
      end if

      !! If geometry minimizer finished and the last calculated geometry is the
      !! minimal one (not necessary the case, depends on the optimizer!)
      !!-> we are finished.
      !! Otherwise we have to recalc everything in the converged geometry.
      if (tGeomEnd) then
        exit lpGeomOpt
      else
        if (tWriteRestart.and.(tGeoOpt.or.tMD)) then
          if (tMulliken) then
            call writeQToFile(qInput, fChargeIn, orb)
            print "('>> Charges saved for restart in ',A)", fChargeIn
          end if
          if (tDFTBU) then
            call writeRhoToFile(inputRhoPrim, fDensMtx, nNeighbor, &
                &neighborList%iNeighbor, img2CentCell, iPair, specie, orb)
            print "('>> Density matrix saved for restart in ',A)", fDensMtx
          end if
        end if
        if (tGeoOpt) then
          if (tMulliken) then
            call writeXYZFormat(trim(lcTmp), coord0, specie0, specieName,&
                &charges=sum(sum(qOutput,dim=3),dim=1))
          else
            call writeXYZFormat(trim(lcTmp), coord0, specie0, specieName)
          end if
          call next(pGeoOpt, energy%Etotal - sum(TS), &
              &reshape(totalDeriv(:,indMovedAtom),(/ nMovedCoord /)),newCoord,&
              &tGeomEnd)          
        elseif(tMD) then
          movedAccel(:,:) = -totalDeriv(:,indMovedAtom) / movedMass
          call next(pMDIntegrator, movedAccel ,new3Coord, movedVelo)
          if (tWriteRestart) then
            if (tMulliken) then            
              call writeXYZFormat(trim(lcTmp), coord0, specie0, specieName, &
                  &charges=sum(sum(qOutput,dim=3),dim=1),&
                  &velocities=velocities, comment="MD iter "//i2c(iGeoStep))
            else
              call writeXYZFormat(trim(lcTmp), coord0, specie0, specieName,&
                  & velocities=velocities, comment="MD iter "//i2c(iGeoStep))
            end if
          end if
          
          if (associated(pTempProfile)) then
            call next(pTempProfile)
          end if
          call evalKE(KE, movedVelo, movedMass(1,:))
          call evalkT(pMDFrame, kT, movedVelo, movedMass(1,:))
          velocities(:, indMovedAtom) = movedVelo(:,:)          
          if (tSetFillingTemp) then
            write (*, formatE1) "Electronic Temperature", tempElec
          end if
          write(*,formatE1) "MD Temperature", kT
          write(*,formatE1) "MD Kinetic Energy", KE
          write(*,formatE1) "Total MD Energy", KE + energy%Etotal - sum(TS)

          if (tWriteDetailedOut) then
            if (tSetFillingTemp) then
              write (fdUser, formatE2) "Electronic Temperature", tempElec, &
                  &tempElec * Hartree__eV
            end if
            write(fdUser,formatE2)'MD Kinetic Energy',KE,KE*Hartree__eV
            write(fdUser,formatE2)'Total MD Energy', &
                & KE+energy%Etotal - sum(TS), &
                & (KE+energy%Etotal - sum(TS))*Hartree__eV
            write(fdUser,formatE1)'MD temperature',kT
          end if
        end if
        
        if (tGeomEnd) then
          diffGeo = maxval(abs(reshape(coord0(:,indMovedAtom), (/nMovedCoord/))&
              &- newCoord))
          if (diffGeo < tolSameDist) then
            tGeomEnd = .true.
            
            exit lpGeomOpt
          end if
        end if
        if (.not. tGeomEnd) then
          if (tGeoOpt) then
            coord0(:,indMovedAtom) = reshape(newCoord(:), (/3, nMovedAtom/))
          else if (tMD) then
            coord0(:,indMovedAtom) = new3Coord(:,:)
          end if
        end if

      end if
    end if
    
    !! Stop reading initial density matrixes
    tReadChrg = .false.

    !! Stop SCC if appropriate stop file is present
    if (.not. tStopSCC) then
      inquire(file=fStopDriver, exist=tStopDriver)
      if (tStopDriver) then
        write (*,*) "Stop file '" // fStopDriver // "' found."
      end if
    end if
    if (tStopSCC .or. tStopDriver) then
      nGeoSteps = iGeoStep
      write (*,*) "Setting max number of geometry steps to current step number."
    end if
    iGeoStep = iGeoStep + 1

  end do lpGeomOpt

  tGeomEnd = tMD .or. tGeomEnd

  if (tWriteDetailedOut .and. (tGeoOpt .or. tMD)) then
    if (tGeomEnd) then
      write (fdUser,*) "Geometry converged"
    else
      write (fdUser, *) "!!! Geometry did NOT converge!"
    end if
    close(fdUser)
  end if

  if (tGeoOpt .or. tMD) then
    if (tGeomEnd) then
      write (*,*)
      write (*,*) "Geometry converged"
    else
      call warning("!!! Geometry did NOT converge!")
      write (*, *) "!!! Geometry did NOT converge!"
    end if
  end if

  if (tWriteTagged) then
    !! Write out final results in tagged format
    open(fdTagged, file=taggedOut, position="append")
    call writeTagged(fdTagged, tag_eigenVal, eigen(:,:,:))
    call writeTagged(fdTagged, tag_filling, filling)
    call writeTagged(fdTagged, tag_egyBand, Eband(:nSpin))
    call writeTagged(fdTagged, tag_egyBandT0, E0(:nSpin))
    call writeTagged(fdTagged, tag_egyRep, energy%Erep)
    if (tAtomicEnergy) then
      call writeTagged(fdTagged, tag_egyRepAt, energy%atomRep(:))
    end if
    if (tMulliken) then
      call writeTagged(fdTagged, tag_qOutput, qOutput(:,:,1))
      call writeTagged(fdTagged, tag_qOutputAt, sum(qOutput, dim=1))
    end if
    if (tForces) then
      call writeTagged(fdTagged, tag_forceTot, &
          &-repulsiveDerivs(:,:)-derivs(:,:))
      call writeTagged(fdTagged, tag_forceBand, -derivs(:,:))
      call writeTagged(fdTagged, tag_forceRep, -repulsiveDerivs(:,:))
      if (tExtChrg) then
        call writeTagged(fdTagged, tag_chrgForces, -chrgForces(:,:))
      end if
    end if
    if (tSCC) then
      call writeTagged(fdTagged, tag_egySCC, energy%ESCC)
      if (tAtomicEnergy) then
        call writeTagged(fdTagged, tag_egyRepAt, energy%atomSCC(:))
      end if
    end if
    if (tDispersion) then
      call writeTagged(fdTagged, tag_egyDispersn ,energy%eDisp )
      if (tAtomicEnergy) then
        call writeTagged(fdTagged, tag_egyDispAt ,energy%atomDisp(:) )
      end if
    end if
    if (nSpin == 2) then
      call writeTagged(fdTagged, tag_egySpin, energy%Espin)
      if (tAtomicEnergy) then
        call writeTagged(fdTagged, tag_egyRepAt, energy%atomSpin(:))
      end if
    end if
    if (tDFTBU) then
      call writeTagged(fdTagged, tag_egyDFTBU, energy%Edftbu)
      if (tAtomicEnergy) then
        call writeTagged(fdTagged, tag_egyRepAt, energy%atomDftbu(:))
      end if
    end if
    call writeTagged(fdTagged, tag_egyTotElec, energy%Eelec)
    call writeTagged(fdTagged, tag_egyTotal, energy%Etotal)
    if (tAtomicEnergy) then
      call writeTagged(fdTagged, tag_egyTotElAt, energy%atomElec(:))
      call writeTagged(fdTagged, tag_egyTotalAt, energy%atomTotal(:))
    end if
    call writeTagged(fdTagged, tag_entropy, sum(TS(:)))
    call writeTagged(fdTagged, tag_freeEgy, energy%Etotal - sum(TS(:)))
    call writeTagged(fdTagged, tag_endCoord, coord0)
    close(fdTagged)
  end if

  
  if (tWriteResultsTag) then
    fdResultsTag = getFileId()
    call initTaggedWriter()
    open(fdResultsTag, file=resultsTag, position="rewind", status="replace")
    call writeTagged(fdResultsTag, tag_egyTotal, energy%Etotal)
    if (tAtomicEnergy) then
      call writeTagged(fdResultsTag, tag_egyTotalAt, energy%atomTotal(:))
    end if
    call writeTagged(fdResultsTag, tag_forces, tForces)
    if (tForces) then
      call writeTagged(fdResultsTag, tag_forceTot, &
          &-repulsiveDerivs(:,:)-derivs(:,:))
      if (tExtChrg) then
        call writeTagged(fdResultsTag, tag_chrgForces, -chrgForces(:,:))
      end if
    end if
    call writeTagged(fdResultsTag, tag_scc, tSCC)
    if (tSCC) then
      call writeTagged(fdResultsTag, tag_nSCC, iSCCIter)
      call writeTagged(fdResultsTag, tag_sccConv, tConverged)
    end if
    if (tMulliken) then
      call writeTagged(fdResultsTag, tag_qOutputAt, sum(qOutput, dim=1))
      call writeTagged(fdResultsTag, tag_qOutAtNet, &
          &sum(sum(q0 - qOutput, dim=3), dim=1))
    end if
    call writeTagged(fdResultsTag, tag_eigenVal, eigen)
    call writeTagged(fdResultsTag, tag_filling, filling)
    call writeTagged(fdResultsTag, tag_efermi, Ef)
    if (size(nEl) == 1) then
      call writeTagged(fdResultsTag, tag_nElUp, nEl(1))
      call writeTagged(fdResultsTag, tag_nElDown, nEl(1))
    else
      call writeTagged(fdResultsTag, tag_nElUp, nEl(1))
      call writeTagged(fdResultsTag, tag_nElDown, nEl(2))
    end if
    close(fdResultsTag)
  end if


  if (tWriteDetailedXML) then
    !! Ugly hack for printing out xml info, will be removed later
    call xml_OpenFile("detailed.xml", xf, indent=.true.)
    call xml_ADDXMLDeclaration(xf)
    call xml_NewElement(xf, "detailedout")
    call writeChildValue(xf, "identity", runId)
    call xml_NewElement(xf, "geometry")
    call writeChildValue(xf, "typenames", specieName)
    call writeChildValue(xf, "typesandcoordinates", &
        &reshape(specie0, (/ 1, size(specie0) /)), coord0)
    call writeChildValue(xf, "periodic", tPeriodic)
    if (tPeriodic) then
      call writeChildValue(xf, "latticevectors", latVec)
    end if
    call xml_EndElement(xf, "geometry")
    call writeChildValue(xf, "real", tRealHS)
    call writeChildValue(xf, "nrofkpoints", nKPoint)
    call writeChildValue(xf, "nrofspins", nSpin)
    call writeChildValue(xf, "nrofstates", size(eigen, dim=1))
    call writeChildValue(xf, "nroforbitals", nOrb)
    ALLOCATE_(bufferRealR2, (4, nKPoint))
    bufferRealR2(1:3, :) = kPoint(:,:)
    bufferRealR2(4, :) = kWeight(:)
    call writeChildValue(xf, "kpointsandweights", bufferRealR2)
    DEALLOCATE_(bufferRealR2)
    call xml_NewElement(xf, "occupations")
    do ii = 1, nSpin
      call xml_NewElement(xf, "spin" // i2c(ii))
      do jj = 1, nKpoint
        call writeChildValue(xf, "k" // i2c(jj), filling(:, jj, ii))
      end do
      call xml_EndElement(xf, "spin" // i2c(ii))
    end do
    call xml_EndElement(xf, "occupations")    
    call xml_EndElement(xf, "detailedout")
    call xml_Close(xf)
  end if

  DEALLOCATE_(new3Coord)
  DEALLOCATE_(newCoord)

  !! Deallocate arrays
  DEALLOCATE_(filling)

  if (tForces) then
    DEALLOCATE_(derivs)
    DEALLOCATE_(repulsiveDerivs)
    DEALLOCATE_(chrgForces)
  end if

  call destroy(energy)

  if (tMulliken) then
    DEALLOCATE_(qOutput)
    DEALLOCATE_(qInput)
    DEALLOCATE_(q0)
  end if
  DEALLOCATE_(rhoPrim)
  if (tDFTBU) then
    DEALLOCATE_(inputRhoPrim)
  end if
  if (tForces) then
    DEALLOCATE_(ERhoPrim)
  end if

  if (tMD) then
    DEALLOCATE_(velocities)
    DEALLOCATE_(movedVelo)
    DEALLOCATE_(movedAccel)
    DEALLOCATE_(movedMass)
  end if

  DEALLOCATE_(HSqrCplx)
  DEALLOCATE_(SSqrCplx)
  DEALLOCATE_(HSqrReal)
  DEALLOCATE_(SSqrReal)
  DEALLOCATE_(eigen)

  if (tSCC) then
    call destruct_SCC()
    call destroy(pChrgMixer)
  end if
  call destroyProgramVariables()


contains

  ! Invokes the writing routines for the Hamiltonian and overlap matrices.
  subroutine writeHS(tWriteHS, tWriteRealHS, ham, over, iNeighbor, &
          &nNeighbor, iAtomStart, iPair, img2CentCell, kPoint, iCellVec, &
          &cellVec)
    logical, intent(in) :: tWriteHS, tWriteRealHS
    real(dp), intent(in) :: ham(:,:), over(:)
    integer, intent(in) :: iNeighbor(0:,:), nNeighbor(:)
    integer, intent(in) :: iAtomStart(:), iPair(0:,:), img2CentCell(:)
    real(dp), intent(in) :: kPoint(:,:)
    integer, intent(in) :: iCellVec(:)
    real(dp), intent(in) :: cellVec(:,:)

    integer :: iS, nSpin

    nSpin = size(ham, dim=2)

    if (tWriteRealHS) then
      do iS = 1, nSpin
        call writeSparse("hamreal" // i2c(iS) // ".dat", ham(:,iS), iNeighbor, &
            &nNeighbor, iAtomStart, iPair, img2CentCell, iCellVec, cellVec)
      end do
      call writeSparse("overreal.dat", over, iNeighbor, &
          &nNeighbor, iAtomStart, iPair, img2CentCell, iCellVec, cellVec)
    end if
    if (tWriteHS) then
      if (tRealHS) then
        do iS = 1, nSpin
          call writeSparseAsSquare("hamsqr" // i2c(iS) // ".dat", ham(:,iS), &
              &iNeighbor, nNeighbor, iAtomStart, iPair, img2CentCell)
        end do
        call writeSparseAsSquare("oversqr.dat", over, iNeighbor, nNeighbor, &
            &iAtomStart, iPair, img2CentCell)
      else
        do iS = 1, nSpin
          call writeSparseAsSquare("hamsqr" // i2c(iS) // ".dat", ham(:,iS), &
              &kPoint, iNeighbor, nNeighbor, iAtomStart, iPair, img2CentCell, &
              &iCellVec, cellVec)
        end do
        call writeSparseAsSquare("oversqr.dat", over, kPoint, iNeighbor, &
            &nNeighbor, iAtomStart, iPair, img2CentCell, iCellVec, cellVec)
      end if
    end if

  end subroutine writeHS
    
  
end program dftb
