!!* Module to wrap around the process of converting from a Hamiltonian and
!!* overlap in sparse form into eigenvectors
module Eigenvects
# include "assert.h"
  use accuracy
  use eigensolver
  use periodic, only : unpackHS
  use message
  implicit none
  
  public :: diagonalize
  private

  interface diagonalize
    module procedure realH
    ! module procedure realH_kpts
    module procedure cmplxH
    ! module procedure cmplxHScmplxRho non-collinear spin etc.
    ! module procedure cmplxHSrealRho collinear spin-orbit
  end interface

contains

  !!* Diagonalizes a sparse represented Hamiltonian and overlap to give
  !!* the eigenvectors and values, as well as often the Cholesky factorized
  !!* overlap matrix (due to a side effect of lapack)
  !!* @param HSqrReal Large square matrix for the resulting eigenvectors
  !!* @param SSqrReal Large square matrix for the overlap workspace, often
  !!* overwritten with the  Cholesky factorized form.
  !!* @param eigen The eigenvalues of the matrices
  !!* @param ham The sparse represented Hamiltonian in real space
  !!* @param over The sparse represented overlap matrix in real space
  !!* @param iNeighbor List of atomic neighbors for each central cell atom
  !!* @param nNeighbor Number of atomic neighbors for each central cell atom
  !!* @param iAtomStart Indexing array for the large square matrices to relate
  !!* atom number to position in the matrix
  !!* @param img2CentCell Array to relate image atoms outside the central cell
  !!* to their real counterpart inside the cell
  !!* @param iSolver Choice of eigensolver, 4 lapack dense solvers currently
  !!* supported
  !!* @jobz type of eigen-problem, either 'V'/'v' vectors or 'N'/'n' 
  !!* eigenvalues only
  subroutine realH(HSqrReal, SSqrReal, eigen, ham, over, iNeighbor, nNeighbor, &
      &iAtomStart, iPair, img2CentCell, iSolver, jobz)
    real(dp), intent(out) :: HSqrReal(:,:)
    real(dp), intent(out) :: SSqrReal(:,:)
    real(dp), intent(out) :: eigen(:)
    real(dp), intent(in)  :: ham(:)
    real(dp), intent(in)  :: over(:)
    integer,  intent(in)  :: iNeighbor(0:,:)
    integer,  intent(in)  :: nNeighbor(:)
    integer,  intent(in)  :: iAtomStart(:)
    integer,  intent(in)  :: iPair(0:,:)
    integer,  intent(in)  :: img2CentCell(:)
    integer,  intent(in)  :: iSolver
    character, intent(in) :: jobz

    integer :: nOrb

    ASSERT(size(HSqrReal, dim=1) == size(HSqrReal, dim=2))
    ASSERT(all(shape(HSqrReal) == shape(SSqrReal)))
    ASSERT(size(HSqrReal, dim=1) == size(eigen))
    ASSERT(size(over) == size(ham))
    ASSERT(size(img2CentCell) >= maxval(iNeighbor))
    ASSERT(jobz == 'n' .or. jobz == 'N' .or. jobz == 'v' .or. jobz == 'V')
    
    call unpackHS(HSqrReal,ham,iNeighbor,nNeighbor,iAtomStart,iPair, &
        &img2CentCell)
    call unpackHS(SSqrReal,over,iNeighbor,nNeighbor,iAtomStart,iPair, &
        & img2CentCell)
    select case(iSolver)
    case(1)
      call hegv(HSqrReal,SSqrReal,eigen,'L',jobz)
    case(2)
      call hegvd(HSqrReal,SSqrReal,eigen,'L',jobz)
    case(3)
      ! use subspace for all eigenstates
      nOrb = size(eigen)
      call gvr(HSqrReal,SSqrReal,eigen,'L',jobz,ilIn=1,iuIn=nOrb)
    case default
      call error('Unknown eigensolver')
    end select

  end subroutine realH


  
  !!* Diagonalizes a sparse represented Hamiltonian and overlap to give
  !!* the eigenvectors and values, as well as often the Cholesky factorized
  !!* overlap matrix (due to a side effect of lapack)
  !!* @param HSqrCplx Large square matrix for the resulting eigenvectors
  !!* @param SSqrCplx Large square matrix for the overlap workspace, often
  !!* overwritten with the  Cholesky factorized form.
  !!* @param eigen The eigenvalues of the matrices
  !!* @param ham The sparse represented Hamiltonian in real space
  !!* @param over The sparse represented overlap matrix in real space
  !!* @param kpoint The k-point to evaluate the phase factors for
  !!* @param iNeighbor List of atomic neighbors for each central cell atom
  !!* @param nNeighbor Number of atomic neighbors for each central cell atom
  !!* @param iCellVec Index of the cell translation vector for each atom.
  !!& @param cellVec Relative coordinates of the cell translation vectors.
  !!* @param iAtomStart Indexing array for the large square matrices to relate
  !!* atom number to position in the matrix
  !!* @param img2CentCell Array to relate image atoms outside the central cell
  !!* to their real counterpart inside the cell
  !!* @param iSolver Choice of eigensolver, 4 lapack dense solvers currently
  !!* supported
  !!* @jobz type of eigen-problem, either 'V'/'v' vectors or 'N'/'n' 
  !!* eigenvalues only
  subroutine cmplxH(HSqrCplx, SSqrCplx, eigen, ham, over, kpoint, iNeighbor, &
      &nNeighbor, iCellVec, cellVec, iAtomStart, iPair, img2CentCell, &
      &iSolver, jobz)
    complex(dp), intent(out) :: HSqrCplx(:,:)
    complex(dp), intent(out) :: SSqrCplx(:,:)
    real(dp), intent(out)    :: eigen(:)
    real(dp), intent(in)     :: ham(:)
    real(dp), intent(in)     :: over(:)
    real(dp), intent(in)     :: kpoint(3)
    integer,  intent(in)     :: iNeighbor(0:,:)
    integer,  intent(in)     :: nNeighbor(:)
    integer,  intent(in)     :: iCellVec(:)
    real(dp), intent(in)     :: cellVec(:,:)
    integer,  intent(in)     :: iAtomStart(:)
    integer,  intent(in)     :: iPair(0:,:)
    integer,  intent(in)     :: img2CentCell(:)
    integer,  intent(in)     :: iSolver
    character, intent(in)    :: jobz

    integer :: nOrb

    ASSERT(size(HSqrCplx, dim=1) == size(HSqrCplx, dim=2))
    ASSERT(all(shape(HSqrCplx) == shape(SSqrCplx)))
    ASSERT(size(HSqrCplx, dim=1) == size(eigen))
    ASSERT(size(over) == size(ham))
    ASSERT(size(img2CentCell) >= maxval(iNeighbor))
    ASSERT(jobz == 'n' .or. jobz == 'N' .or. jobz == 'v' .or. jobz == 'V')
    
    call unpackHS(HSqrCplx, ham, kPoint, iNeighbor, nNeighbor, iCellVec, &
        &cellVec, iAtomStart, iPair, img2CentCell)
    call unpackHS(SSqrCplx, over, kPoint, iNeighbor, nNeighbor, iCellVec, &
        &cellVec, iAtomStart, iPair, img2CentCell)
    select case(iSolver)
    case(1)
      call hegv(HSqrCplx,SSqrCplx,eigen,'L',jobz)
    case(2)
      call hegvd(HSqrCplx,SSqrCplx,eigen,'L',jobz)
    case(3)
      ! use subspace for all eigenstates
      nOrb = size(eigen)
      call gvr(HSqrCplx,SSqrCplx,eigen,'L',jobz,ilIn=1,iuIn=nOrb)
    case default
      call error('Unknown eigensolver')
    end select

  end subroutine cmplxH

end module Eigenvects
