!!* Module to wrap around the different energy components in the DFTB 
!!* total energy expression
module Energies
# include "allocate.h"
# include "assert.h"
  use accuracy
  implicit none

  public :: TEnergies, create, destroy
  
  private
  
  !!* data type to store components of the energy as named variables instead of
  !!* in the old arrays - makes extending energy expression easier.
  type TEnergies
    real(dp) :: Erep                 !* repulsive energy
    real(dp) :: EnonSCC(2)           !* Non-SCC energy - possibly from 2 spins
    real(dp) :: ESCC                 !* SCC double-counting energy
    real(dp) :: Espin                !* spin energy
    real(dp) :: Edftbu               !* DFTB+U energy
    real(dp) :: Eelec                !* total electronic energy
    real(dp) :: eDisp                !* Dispersion energy
    real(dp) :: Etotal               !* total energy (Erep+Etotal)
    real(dp), pointer :: atomRep(:)      !* atom resolved repulsive
    real(dp), pointer :: atomNonSCC(:,:) !* atom resolved non-SCC
    real(dp), pointer :: atomSCC(:)      !* atom resolved SCC
    real(dp), pointer :: atomSpin(:)     !* atom resolved spin
    real(dp), pointer :: atomDftbu(:)    !* atom resolved DFTB+U
    real(dp), pointer :: atomElec(:)     !* atom resolved electronic total
    real(dp), pointer :: atomTotal(:)    !* atom resolved total
    real(dp), pointer :: atomDisp(:)     !* atom resolved dispersion
    logical :: tInitialised = .false.
  end type TEnergies

  interface create
    module procedure Energies_create
  end interface
  
  interface destroy
    module procedure Energies_destroy
  end interface

  
contains

  !!* Allocates storage for the energy components
  !!* @param estruct data structure to allocate
  !!* @param nAtom number of atoms needed for atom resolved arrays
  !!* @param nSpin number of spins
  subroutine Energies_create(self,nAtom,nSpin)
    type(TEnergies), intent(out) :: self
    integer, intent(in) :: nAtom
    integer, intent(in) :: nSpin

    ASSERT(.not. self%tInitialised)
    
    INITALLOCATE_PARR(self%atomRep,(nAtom))
    INITALLOCATE_PARR(self%atomNonSCC,(nAtom,nSpin))
    INITALLOCATE_PARR(self%atomSCC,(nAtom))
    INITALLOCATE_PARR(self%atomSpin,(nAtom))
    INITALLOCATE_PARR(self%atomDftbu,(nAtom))
    INITALLOCATE_PARR(self%atomElec,(nAtom))
    INITALLOCATE_PARR(self%atomDisp, (nAtom))
    INITALLOCATE_PARR(self%atomTotal,(nAtom))
    self%atomRep(:) = 0.0_dp
    self%atomNonSCC(:,:) = 0.0_dp
    self%atomSCC(:) = 0.0_dp
    self%atomSpin(:) = 0.0_dp
    self%atomDftbu(:) = 0.0_dp
    self%atomElec(:) = 0.0_dp
    self%atomDisp(:) = 0.0_dp
    self%atomTotal(:) = 0.0_dp
    self%tInitialised = .true.
    
  end subroutine Energies_create


  
  !!* De-allocates storage for the energy components
  subroutine Energies_destroy(self)
    type(TEnergies), intent(inout) :: self

    ASSERT(self%tInitialised)
    
    self%tInitialised = .false.    
    DEALLOCATE_PARR(self%atomRep)
    DEALLOCATE_PARR(self%atomNonSCC)
    DEALLOCATE_PARR(self%atomSCC)
    DEALLOCATE_PARR(self%atomSpin)
    DEALLOCATE_PARR(self%atomDftbu)
    DEALLOCATE_PARR(self%atomElec)
    DEALLOCATE_PARR(self%atomDisp)
    DEALLOCATE_PARR(self%atomTotal)
    
  end subroutine Energies_destroy

  
end module Energies
