!!* Contains data type representing the input data for DFTB
!!* @todo
!!*   Has to be replaced by a more general approach, which should be
!!*   a kind of database (quite similar to Thomas' global library, but !'
!!*   with a dynamical approach).
module inputdata_
#include "assert.h"
#include "allocate.h"  
  use accuracy
  use TypeGeometry
  use message
  use SlaterKirkwood, only: TSlaterKirkwoodInput, destroy
  use SlakoCont
  use CommonTypes
  use RepCont
  implicit none
  
  private
  save
  
  type control
    integer       :: iSeed       = 0
    real(dp)      :: maxForce    = 0.0_dp
    logical       :: tScc        = .false.
    real(dp)      :: sccTol      = 0.0_dp
    logical       :: tReadChrg   = .false.
    logical       :: tGeoOpt     = .false. ! should probably be a packaged
    logical       :: tAppendGeo  = .false.
    integer       :: iGeoOpt     = 0
    logical       :: tMulliken   = .false. !
    logical       :: tAtomicEnergy = .false. !
    logical       :: tEigenvecs  = .false. !
    logical       :: tForces     = .false. !
    logical       :: tMD         = .false. !
    
    real(dp)               :: nrChrg        = 0.0_dp
    real(dp)               :: nrSpinPol     = 0.0_dp
    logical                :: tSpin         = .false.
    real(dp), pointer      :: initialSpins(:) => null() ! initial spin pattern
    logical                :: tDFTBU        = .false.
    integer                :: iSolver       = 0
    integer                :: iMixSwitch    = 0
    integer                :: maxIter       = 0
    real(dp)               :: almix         = 0.0_dp
    integer                :: iGenerations  = 0
    logical                :: tFromStart    = .true.
    real(dp)               :: broydenOmega0 = 0.01_dp
    real(dp)               :: broydenMinWeight = 1.0_dp
    real(dp)               :: broydenMaxWeight = 1.0e5_dp
    real(dp)               :: broydenWeightFac = 1.0e-2_dp
    real(dp)               :: andersonInitMixing = 0.01_dp
    integer                :: andersonNrDynMix = 0
    real(dp), pointer      :: andersonDynMixParams(:,:) => null()
    real(dp)               :: andersonOmega0 = 1.0e-2_dp
    integer                :: nrMoved       = 0
    integer, pointer       :: indMovedAtom(:)  => null()
    integer                :: nrConstr      = 0
    integer, pointer       :: conAtom(:)    => null()
    real(dp), pointer      :: conVec(:,:)   => null()
    character(lc)          :: outFile       = ''
    logical                :: tReadMDVelocities = .false. ! do we have MD velocities
    real(dp), pointer      :: initialVelocities(:,:) => null() ! initial MD velocities
    real(dp)               :: deltaT        = 0.0_dp

    real(dp)               :: tempAtom      = 0.0_dp
    integer                :: iThermostat   = 0
    logical                :: tMDstill ! whether to shift to a co-moving
    ! frame for MD
    logical                :: tRescale = .false.
    integer, pointer       :: tempMethods(:) => null()
    integer, pointer       :: tempSteps(:)  => null()
    real(dp), pointer      :: tempValues(:) => null()
    logical                :: tSetFillingTemp = .false.
    
    real(dp)               :: tempElec      = 0.0_dp
    logical                :: tFillKSep     = .false.
    integer                :: iDistribFn    = 0
    real(dp)               :: wvScale       = 0.0_dp
    integer                :: maxRun        = -1

    integer                :: nKPoint       = 0
    real(dp), pointer      :: kPoint(:,:)   => null()
    real(dp), pointer      :: kWeight(:)    => null()
    real(dp), pointer      :: W(:,:,:)      => null()
    integer                :: DFTBUfunc     = 0 ! choice of the DFTB+U
    ! functional    
    real(dp), pointer      :: UJ(:,:)       => null() ! list of U-J for species
    integer, pointer       :: nU(:)         => null() ! number of U-J needed
    ! for each species
    integer, pointer       :: nLinU(:,:)    => null() ! number of angular
    ! momenta in DFTB+U blocks
    integer, pointer       :: iLVals(:,:,:) => null() ! list of angular momenta
    ! in each block

    !! External charges
    integer :: nExtChrg = 0
    real(dp), pointer :: extChrg(:,:) => null()
    real(dp), pointer :: extChrgBlurWidth(:) => null()
    
    !! H short range damping
    logical :: tDampH = .false.
    

    !! Various options
    logical :: tWriteTagged = .false.
    integer :: restartFreq  = 20 ! Nr. of SCC iterations without restart info
    logical :: tWriteDetailedXML = .false.
    logical :: tWriteResultsTag = .false.
    logical :: tWriteDetailedOut = .true.
    logical :: tWriteBandDat = .true.
    logical :: oldSKInter = .false.
    logical :: tWriteHS = .false.
    logical :: tWriteRealHS = .false.
    logical :: tMinMemory = .false.

    !! Dispersion related stuff
    type(TSlaterKirkwoodInput), pointer :: pSlaKirk => null()
    
  end type control
  
  
  type geometry
    integer           :: nrAtoms         = 0
    logical           :: tPeriodic       = .false.
    logical           :: tFracCoord      = .false.
    integer,  pointer :: types(:)        => null()
    real(dp), pointer :: coords(:, :)    => null()
    integer           :: nrTypes         = 0
    real(dp), pointer :: origo(:)        => null()
    real(dp), pointer :: latVecs(:, :)   => null()
    character(mc), pointer :: specieName(:) => null()
  end type geometry

  
  type slater
    real(dp), pointer :: skSelf(:, :)          => null()
    real(dp), pointer :: skHubbU(:, :)         => null()
    real(dp), pointer :: skOcc(:, :)           => null()
    real(dp), pointer :: mass(:)               => null()

    type(OSlakoCont), pointer :: skHamCont
    type(OSlakoCont), pointer :: skOverCont
    type(ORepCont), pointer :: repCont
    type(TOrbitals), pointer :: orb
  end type slater


  type inputData
    type(control)  :: ctrl
    type(TGeometry) :: geom
    type(slater)   :: slako
    logical        :: tInitialized = .false.
  end type inputData


  
  interface init
    module procedure InputData_init
  end interface
  
  
  interface destroy
    module procedure InputData_destroy
  end interface

  public :: control, TGeometry, slater, inputData
  public :: init, destroy

  
contains


  subroutine InputData_init(self)
    type(inputData), intent(out) :: self

    call initControl(self%ctrl)
    call initGeometry(self%geom)
    call initSlater(self%slako)
    self%tInitialized = .true.

  end subroutine InputData_init



  subroutine InputData_destroy(self)
    type(inputData), intent(inout) :: self
    
    ASSERT(self%tInitialized)
    
    call destroyControl(self%ctrl)
    call destruct(self%geom)
    call destroySlater(self%slako)
    self%tInitialized = .false.

  end subroutine InputData_destroy



  !!* Initializes control data.
  !!* @param ctrl Holds control data.
  subroutine initControl(ctrl)
    type(control), intent(inout) :: ctrl
    
    INIT_PARR(ctrl%conAtom)
    INIT_PARR(ctrl%conVec)
    INIT_PARR(ctrl%initialVelocities)
    INIT_PARR(ctrl%kPoint)
    INIT_PARR(ctrl%kWeight)    
    INIT_PARR(ctrl%indMovedAtom)
    
  end subroutine initControl
  
  
  
  !!* Destroys control data.
  !!* @param ctrl Holds control data.
  subroutine destroyControl(ctrl)
    type(control), intent(inout) :: ctrl

    DEALLOCATE_PARR(ctrl%initialSpins)
    DEALLOCATE_PARR(ctrl%andersonDynMixParams)
    DEALLOCATE_PARR(ctrl%indMovedAtom)
    DEALLOCATE_PARR(ctrl%conAtom)
    DEALLOCATE_PARR(ctrl%conVec)
    DEALLOCATE_PARR(ctrl%initialVelocities)
    DEALLOCATE_PARR(ctrl%tempMethods)
    DEALLOCATE_PARR(ctrl%tempSteps)
    DEALLOCATE_PARR(ctrl%tempValues)
    DEALLOCATE_PARR(ctrl%kPoint)
    DEALLOCATE_PARR(ctrl%kWeight)
    DEALLOCATE_PARR(ctrl%W)
    DEALLOCATE_PARR(ctrl%UJ)
    DEALLOCATE_PARR(ctrl%nU)
    DEALLOCATE_PARR(ctrl%nLinU)
    DEALLOCATE_PARR(ctrl%iLVals)
    DEALLOCATE_PARR(ctrl%extChrg)
    DEALLOCATE_PARR(ctrl%extChrgBlurWidth)
    
    if (associated(ctrl%pSlaKirk)) then
      call destroy(ctrl%pSlaKirk)
      DEALLOCATE_P(ctrl%pSlaKirk)
    end if

  end subroutine destroyControl



  !!* Initializes geometry data.
  !!* @param geom Holds geometry data.
  subroutine initGeometry(geom)
    type(TGeometry), intent(inout) :: geom

    !INIT_PARR(geom%types)
    !INIT_PARR(geom%coords)
    !INIT_PARR(geom%origo)
    !INIT_PARR(geom%latVecs)
    !INIT_PARR(geom%specieName)

  end subroutine initGeometry



  !!* Destroys geometry data.
  !!* @param geom Holds geometry data.
  subroutine destroyGeometry(geom)
    type(geometry), intent(inout) :: geom

    DEALLOCATE_PARR(geom%types)
    DEALLOCATE_PARR(geom%coords)
    DEALLOCATE_PARR(geom%origo)
    DEALLOCATE_PARR(geom%latVecs)
    DEALLOCATE_PARR(geom%specieName)

  end subroutine destroyGeometry



  !!* Initalizes Slater-Koster data.
  !!* @param slako Holds Slater-Koster data.
  subroutine initSlater(slako)
    type(slater), intent(inout) :: slako

    INIT_PARR(slako%skSelf)
    INIT_PARR(slako%skHubbU)
    INIT_PARR(slako%skOcc)
    INIT_PARR(slako%mass)


  end subroutine initSlater



  !!* Initalizes Slater-Koster data.
  !!* @param slako Holds Slater-Koster data.
  subroutine destroySlater(slako)
    type(slater), intent(inout) :: slako

    DEALLOCATE_PARR(slako%skSelf)
    DEALLOCATE_PARR(slako%skHubbU)
    DEALLOCATE_PARR(slako%skOcc)
    DEALLOCATE_PARR(slako%mass)

  end subroutine destroySlater

end module inputdata_
