!!* Fills the derived type with the input parameters from an HSD or an XML file.
module Parser
#include "assert.h"
#include "allocate.h"
  use Accuracy
  use Constants
  use Inputdata_
  use TypeGeometryHSD
  use HSDParser, only : dumpHSD, dumpHSDAsXML, getNodeHSDName
  use HSDUtils
  use HSDUtils2
  use flib_dom
  use CharManip
  use Message
  use LinkedList
  use StringList
  use FileId
  use UnitConversion
  use OldCompat
  use LapackRoutines, only : matinv
  use Periodic
  use SimpleAlgebra, only: determinant33
  use SlaterKirkwood, only : TSlaterKirkwoodInput
  use SlakoCont
  use SlakoEqGrid
  use RepCont
  use RepSpline
  use RepPoly
  use CommonTypes
  use OldSKData
  implicit None

  private

  public :: parseHSDInput, parserVersion

  !! Private constants
  character(len=*), parameter :: hsdInputName = "dftb_in.hsd"
  character(len=*), parameter :: xmlInputName = "dftb_in.xml"
  character(len=*), parameter :: hsdProcInputName = "dftb_pin.hsd"
  character(len=*), parameter :: xmlProcInputName = "dftb_pin.xml"
  character(len=*), parameter :: rootTag = "dftb_in"

  !! Version of the current parser
  integer, parameter :: parserVersion = 3

  !! Version of the oldest parser, for which compatibility is maintained
  integer, parameter :: minVersion = 1

  !! Container type for parser related flags.
  type TParserFlags
    logical :: tStop                        ! stop after parsing?
    logical :: tIgnoreUnprocessed           ! Continue despite unprocessed nodes
    logical :: tWriteXML, tWriteHSD         ! XML or HSD output?
  end type TParserFlags
  
  
contains

  !!* Parse input from an HSD/XML file
  !!* @param input   Initialised input variables on exit
  subroutine parseHSDInput(input)
    type(inputData), intent(out) :: input

    type(fnode), pointer :: hsdTree
    type(fnode), pointer :: root, tmp, child, dummy
    type(TParserflags) :: parserFlags
    logical :: tHSD, missing

    !! Read in the input
    call readHSDOrXML(hsdInputName, xmlInputName, rootTag, hsdTree, tHSD, &
        &missing)

    !! If input is missing return
    if (missing) then
      call error("No input file found.")
    end if
    
    if (tHSD) then
      write (*, "(A)") "Interpreting input file '" // hsdInputName // "'"
    else
      write (*, "(A)") "Interpreting input file '" // xmlInputName //  "'"
    end if
    write (*, "(A)") repeat("-", 80)

    !! Get the root of all evil ;-)
    call getChild(hsdTree, rootTag, root)

    !! Handle parser options
    call getChildValue(root, "ParserOptions", dummy, "", child=child, &
        &list=.true., allowEmptyValue=.true., dummyValue=.true.)
    call readParserOptions(child, root, parserFlags)
    
    !! Read in the different blocks
    call getChild(root, "Geometry", tmp)

    call readGeometry(tmp, input)

    call getChildValue(root, "Driver", tmp, "", child=child, &
        &allowEmptyValue=.true.)
    call readDriver(tmp, child, input%ctrl, input%geom%nAtom)
    call getChildValue(root, "Hamiltonian", tmp)
    call readHamiltonian(tmp, input%ctrl, input%geom, input%slako)
    call getChildValue(root, "Options", dummy, "", child=child, list=.true., &
        &allowEmptyValue=.true., dummyValue=.true.)
    call readOptions(child, input%ctrl)
    input%tInitialized = .true.

    !! Issue warning about unprocessed nodes
    call warnUnprocessedNodes(root, parserFlags%tIgnoreUnprocessed)

    !! Dump processed tree in HSD and XML format
    if (parserFlags%tWriteHSD) then
      call dumpHSD(hsdTree, hsdProcInputName)
      write (*, '(/,/,A)') "Processed input in HSD format written to '" &
          &// hsdProcInputName // "'"
    end if
    if (parserFlags%tWriteXML) then
      call dumpHSDAsXML(hsdTree, xmlProcInputName)
      write (*, '(A,/)') "Processed input in XML format written to '" &
          &// xmlProcInputName // "'"
    end if
    
    !! Stop, if only parsing is required    
    if (parserFlags%tStop) then
      call error("Keyword 'StopAfterParsing' is set to Yes. Stopping.")
    end if
    
  end subroutine parseHSDInput



  !!* Read in parser options (options not passed to the main code)
  !!* @param node Node to get the information from
  !!* @param root Root of the entire tree (in the case it must be converted)
  !!* @param flags Contains parser flags on exit.
  subroutine readParserOptions(node, root, flags)
    type(fnode), pointer :: node
    type(fnode), pointer :: root
    type(TParserFlags), intent(out) :: flags

    integer :: inputVersion
    type(fnode), pointer :: child

    !! Check if input needs compatibility conversion.
    call getChildValue(node, "ParserVersion", inputVersion, parserVersion, &
        &child=child)
    if (inputVersion < 1 .or. inputVersion > parserVersion) then
      call detailedError(child, "Invalid parser version (" // i2c(inputVersion)&
          &// ")")
    elseif (inputVersion < minVersion) then
      call detailedError(child, &
          &"Sorry, no compatibility mode for parser version " &
          &// i2c(inputVersion) // " (too old)")
    elseif (inputVersion /= parserVersion) then
      write (*, "(A,I2,A,I2,A)") "***  Converting input from version ", &
          &inputVersion, " to version ", parserVersion, " ..."
      call convertOldHSD(root, inputVersion, parserVersion)
      write (*, "(A,/)") "***  Done."
    end if

    call getChildValue(node, "WriteHSDInput", flags%tWriteHSD, .true.)
    call getChildValue(node, "WriteXMLInput", flags%tWriteXML, .false.)
    if (.not. (flags%tWriteHSD .or. flags%tWriteXML)) then
      call detailedWarning(node, &
          &"WriteHSDInput and WriteXMLInput both turned off. You won't &
          &eventually be " &
          &// newline // &
          &" able to obtain the same results with a later version of the code!")
    end if
    call getChildValue(node, "StopAfterParsing", flags%tStop, .false.)

    call getChildValue(node, "IgnoreUnprocessedNodes", &
        &flags%tIgnoreUnprocessed, .false.)
    
  end subroutine readParserOptions

  

  !!* Read in Geometry
  !!* @param node  Node to get the information from
  !!* @param input Input structure to be filled
  subroutine readGeometry(node, input)
    type(fnode), pointer :: node
    type(inputData), intent(inout) :: input

    type(fnode), pointer :: value, child
    type(string) :: buffer

    call getChildValue(node, "", value, child=child)
    call getNodeName(value, buffer)
    select case (char(buffer))
    case ("genformat")
      call readTGeometryGen(value, input%geom)
    case default
      call setUnprocessed(value)
      call readTGeometryHSD(child, input%geom)
    end select

    call unstring(buffer)

  end subroutine readGeometry



  !!* Read in driver properties
  !!* @param node   Node to get the information from
  !!* @param parent Parent of node (for error messages)
  !!* @param ctrl   Control structure to be filled
  !!* @param nAtom  Nr. of atoms in the system
  subroutine readDriver(node, parent, ctrl, nAtom)
    type(fnode), pointer :: node
    type(fnode), pointer :: parent
    type(control), intent(inout) :: ctrl
    integer, intent(in) :: nAtom

    type(fnode), pointer :: child, child2, value, value2, field
    type(string) :: buffer, buffer2, modifier
    logical :: tChanged

    ctrl%tGeoOpt = .false.
    ctrl%iGeoOpt = 0
    ctrl%tMD = .false.
    ctrl%iThermostat = 0
    ctrl%tForces = .false.
    ctrl%tSetFillingTemp = .false.
    
    call getNodeName2(node, buffer)
    driver: select case (char(buffer))
    case ("")
      continue
    case ("steepestdescent")
      ctrl%tGeoOpt = .true.
      ctrl%iGeoOpt = 1
      ctrl%tForces = .true.
      ctrl%restartFreq = 1
      call getChild(node, "MovedAtoms", child, requested=.false.)
      if (.not. associated(child)) then
        call setChild(node, "MovedAtoms", child)
        call setChildValue(child, "Range", (/ 1, -1 /))
      end if
      call getValueFromListOrRange(child, ctrl%indMovedAtom, mvalue=nAtom)
      ctrl%nrMoved = size(ctrl%indMovedAtom)
      call getChildValue(node, "MaxForceComponent", ctrl%maxForce, 1e-4_dp, &
          &modifier=modifier, child=field)
      call convertByMul(char(modifier), forceUnits, field, ctrl%maxForce)
      call getChildValue(node, "MaxSteps", ctrl%maxRun, 200)
      call getChildValue(node, "StepSize", ctrl%deltaT, 100.0_dp, &
          &modifier=modifier, child=field)
      call convertByMul(char(modifier), timeUnits, field, ctrl%deltaT)
      call getChildValue(node, "OutputPrefix", buffer2, "geo_end")
      ctrl%outFile = unquote(char(buffer2))
      call getChildValue(node, "AppendGeometries", ctrl%tAppendGeo, .false.)
      call readGeoConstraints(node, ctrl, nAtom)

    case ("conjugategradient")
      ctrl%tGeoOpt = .true.
      ctrl%iGeoOpt = 2
      ctrl%tForces = .true.
      ctrl%restartFreq = 1
      call getChild(node, "MovedAtoms", child, requested=.false.)
      if (.not. associated(child)) then
        call setChild(node, "MovedAtoms", child)
        call setChildValue(child, "Range", (/ 1, -1 /))
      end if
      call getValueFromListOrRange(child, ctrl%indMovedAtom, mvalue=nAtom)
      ctrl%nrMoved = size(ctrl%indMovedAtom)
      call getChildValue(node, "MaxForceComponent", ctrl%maxForce, 1e-4_dp, &
          &modifier=modifier, child=field)
      call convertByMul(char(modifier), forceUnits, field, ctrl%maxForce)
      call getChildValue(node, "MaxSteps", ctrl%maxRun, 200)
      call getChildValue(node, "OutputPrefix", buffer2, "geo_end")
      ctrl%outFile = unquote(char(buffer2))
      call getChildValue(node, "AppendGeometries", ctrl%tAppendGeo, .false.)
      call readGeoConstraints(node, ctrl, nAtom)

    case ("velocityverlet")
      ctrl%tForces = .true.
      ctrl%tMD = .true.

      call getChildValue(node, "MDRestartFrequency", ctrl%restartFreq, 1)
      call getChild(node, "MovedAtoms", child, requested=.false.)
      if (.not. associated(child)) then
        call setChild(node, "MovedAtoms", child)
        call setChildValue(child, "Range", (/ 1, -1 /))
      end if
      call getValueFromListOrRange(child, ctrl%indMovedAtom, mvalue=nAtom)
      ctrl%nrMoved = size(ctrl%indMovedAtom)
      
      call readInitialVelocities(node, ctrl, nAtom)
      
      call getChildValue(node, "KeepStationary", ctrl%tMDstill,.true.)
      
      call getChildValue(node, "TimeStep", ctrl%deltaT, modifier=modifier, &
          &child=field)
      call convertByMul(char(modifier), timeUnits, field, ctrl%deltaT)
      
      call getChildValue(node, "Thermostat", value, child=child)
      call getNodeName(value, buffer2)

      thermostat: select case(char(buffer2))
      case ("berendsen")
        ctrl%iThermostat = 2
        !! Read temperature or temperatura profiles
        call getChildValue(value, "Temperature", value2, modifier=modifier, &
            &child=child2)
        call getNodeName(value2, buffer)
        
        select case(char(buffer))
        case (textNodeName)
          call readTemperature(child2, ctrl)
        case ("temperatureprofile")
          call readTemperatureProfile(value2, child2, char(modifier), ctrl)
        case default
          call detailedError(value2, "Invalid method name.")
        end select
        
        call getChildValue(value, "CouplingStrength", ctrl%wvScale)
        call getChildValue(value, "AdaptFillingTemp", ctrl%tSetFillingTemp, &
            &.false.)      

      case ("andersen")
        ctrl%iThermostat = 1
        !! Read temperature or temperature profiles
        call getChildValue(value, "Temperature", value2, modifier=modifier, &
            &child=child2)
        call getNodeName(value2, buffer)
        
        select case(char(buffer))
        case (textNodeName)
          call readTemperature(child2, ctrl)
        case ("temperatureprofile")
          call readTemperatureProfile(value2, child2, char(modifier), ctrl)
        case default
          call detailedError(value2, "Invalid method name.")
        end select
        
        call getChildValue(value, "ReselectProbability", ctrl%wvScale)
        call getChildValue(value, "ReselectIndividually", ctrl%tRescale)
        call getChildValue(value, "AdaptFillingTemp", ctrl%tSetFillingTemp, &
            &.false.)
        
      case ("none")
        ctrl%iThermostat = 0
        INITALLOCATE_PARR(ctrl%tempSteps, (1))
        INITALLOCATE_PARR(ctrl%tempValues, (1))
        
        if (ctrl%tReadMDVelocities) then
          ! without a thermostat, if we know the initial
          ! velocities, we don't need a temperature, so just set it to something
          ! 'safe'
          ctrl%tempAtom = minTemp
        else
          call getChildValue(value, "InitialTemperature", ctrl%tempAtom, &
              &modifier=modifier, child=field)
          if (ctrl%tempAtom < 0.0_dp) then
            call detailedError(field, "Negative temperature")
          end if
          tChanged = .false.
          call convertByMul(char(modifier), energyUnits, field, ctrl%tempAtom, &
              &.false., tChanged)
          if (ctrl%tempAtom < minTemp) then
            ctrl%tempAtom = minTemp
            tChanged = .true.
          end if
          if (tChanged) then
            call setChildValue(field, "", ctrl%tempAtom, .true.)
          end if
        end if
      case default
        call getNodeHSDName(value, buffer2)
        call detailedError(child, "Invalid thermostat '" // char(buffer) // "'")
      end select thermostat
      
      if (ctrl%maxRun < 0) then
        call getChildValue(node, "Steps", ctrl%maxRun)
      end if
      
      call getChildValue(node, "OutputPrefix", buffer2, "geo_end")
      ctrl%outFile = unquote(char(buffer2))
      
    case default
      call getNodeHSDName(node, buffer)
      call detailedError(parent, "Invalid driver '" // char(buffer) // "'")

    end select driver

    call unstring(buffer)
    call unstring(buffer2)
    call unstring(modifier)
    
  end subroutine readDriver

  

  !!* Reads geometry constraints.
  !!* @param node  Node to get the information from 
  !!* @param ctrl  Control structure to be filled
  !!* @param nAtom Nr. of atoms in the system
  subroutine readGeoConstraints(node, ctrl, nAtom)
    type(fnode), pointer :: node
    type(control), intent(inout) :: ctrl
    integer, intent(in) :: nAtom

    type(fnode), pointer :: value, child
    type(string) :: buffer
    type(listIntR1) :: intBuffer
    type(listRealR1) :: realBuffer

    call getChildValue(node, "Constraints", value, "", child=child, &
        &allowEmptyValue=.true.)
    call getNodeName2(value, buffer)
    if (char(buffer) == "") then
      ctrl%nrConstr = 0
    else
      call init(intBuffer)
      call init(realBuffer)
      call getChildValue(child, "", 1, intBuffer, 3, realBuffer)
      ctrl%nrConstr = len(intBuffer)
      INITALLOCATE_PARR(ctrl%conAtom, (ctrl%nrConstr))
      INITALLOCATE_PARR(ctrl%conVec, (3, ctrl%nrConstr))
      call asVector(intBuffer, ctrl%conAtom)
      if (.not.all(ctrl%conAtom<=nAtom)) then
        call detailedError(node,"Non-existent atom specified in constraint")
      end if
      call asArray(realBuffer, ctrl%conVec)
      call destroy(intBuffer)
      call destroy(realBuffer)
    end if
    call unstring(buffer)

  end subroutine readGeoConstraints


  
  !!* Reads MD velocities
  !!* @param node  Node to get the information from 
  !!* @param ctrl  Control structure to be filled
  !!* @param nAtom number of all atoms
  subroutine readInitialVelocities(node, ctrl, nAtom)
    type(fnode), pointer :: node
    type(control), intent(inout) :: ctrl
    integer, intent(in)  :: nAtom
    
    type(fnode), pointer :: value, child
    type(string) :: buffer, modifier
    type(listRealR1) :: realBuffer
    integer          :: nVelocities
    logical          :: tChanged
    real(dp), pointer :: tmpVelocities(:,:)
    
    call getChildValue(node, "Velocities", value, "", child=child, &
        & modifier=modifier, allowEmptyValue=.true.)
    call getNodeName2(value, buffer)
    if (char(buffer) == "") then
      ctrl%tReadMDVelocities = .false.
    else
      call init(realBuffer)
      call getChildValue(child, "", 3, realBuffer, modifier=modifier)
      nVelocities = len(realBuffer)
      if (nVelocities /= nAtom) then
        call detailedError(node, "Incorrect number of specified velocities: " &
            & // i2c(3*nVelocities) // " supplied, " &
            & // i2c(3*nAtom) // " required.")        
      end if
      INITALLOCATE_PARR(tmpVelocities, (3, nVelocities))      
      call asArray(realBuffer, tmpVelocities)
      if (len(modifier) > 0) then
        tChanged = .false.
        call convertByMul(char(modifier), VelocityUnits, child, &
            & tmpVelocities, .false., tChanged)
        if (tChanged) then
          call setChildValue(child, "", tmpVelocities, .true.)
        end if
      end if
      call destroy(realBuffer)
      INITALLOCATE_PARR(ctrl%initialVelocities, (3, ctrl%nrMoved))
      ctrl%initialVelocities(:,:) = tmpVelocities(:,ctrl%indMovedAtom(:))      
      DEALLOCATE_PARR(tmpVelocities)
      ctrl%tReadMDVelocities = .true.
    end if

    call unstring(buffer)
    call unstring(modifier)


  end subroutine readInitialVelocities

  
  
  !!* Reads Hamiltonian
  !!* @param node  Node to get the information from 
  !!* @param ctrl  Control structure to be filled
  !!* @param geo   Geometry structure to be filled
  !!* @param slako Slater-Koster structure to be filled
  subroutine readHamiltonian(node, ctrl, geo, slako)
    type(fnode), pointer :: node
    type(control), intent(inout) :: ctrl
    type(TGeometry), intent(in) :: geo
    type(slater), intent(inout) :: slako

    type(string) :: buffer

    call getNodeName(node, buffer)
    select case (char(buffer))
    case ("dftb")
      call readDFTBHam(node, ctrl, geo, slako)
    case default
      call detailedError(node, "Invalid Hamiltonian")
    end select
    
  end subroutine readHamiltonian



  !!* Reads DFTB-Hamiltonian
  !!* @param node  Node to get the information from 
  !!* @param ctrl  Control structure to be filled
  !!* @param geo   Geometry structure to be filled
  !!* @param slako Slater-Koster structure to be filled
  subroutine readDFTBHam(node, ctrl, geo, slako)
    type(fnode), pointer :: node
    type(control), intent(inout) :: ctrl
    type(TGeometry), intent(in) :: geo
    type(slater), intent(inout) :: slako

    type(fnode), pointer :: value, value2, child, child2, child3, field, dummy
    type(fnodeList), pointer :: children
    type(string) :: buffer, buffer2, modifier
    type(listIntR1) :: li1
    type(listRealR1) :: lr1
    type(listReal) :: lr
    type(listCharLc), allocatable :: skFiles(:,:)
    type(listString) :: lStr
    type(listIntR1), allocatable :: angShells(:)
    logical, allocatable :: repPoly(:,:)
    integer :: iSp1, iSp2, iSh1, iSh2, ii, jj, kk, ind, oldLen, iAt
    character(lc) :: prefix, suffix, separator, elem1, elem2, strTmp
    character(lc) :: errorStr
    logical :: tLower, tChanged, tExist, tOrbResolved
    real(dp), allocatable :: kpts(:,:), lconsts(:)
    integer, allocatable :: tmpI1(:)
    integer, pointer :: pTmpI1(:)
    real(dp) :: conv, rTmp, rTmp4(4), rTmp3(3)
    real(dp), allocatable :: rTmp41(:,:)
    real(dp) :: coeffsAndShifts(3, 4)
    integer :: iU, iLinU, nShell
    integer :: skInterMeth
    character(1) :: tmpCh
    logical :: tShellIncl(4), tFound
    integer :: angShell(maxL+1), angShellOrdered(maxL+1)

    tOrbResolved = .false.

    !! Read in maximal angular momenta or selected shells
    do ii = 1, maxL+1
      angShellOrdered(ii) = ii - 1
    end do
    call getChild(node, "MaxAngularMomentum", child)
    ALLOCATE_(angShells, (geo%nSpecie))
    do iSp1 = 1, geo%nSpecie
      call init(angShells(iSp1))
      call getChildValue(child, geo%specieNames(iSp1), value, child=child2)
      call getNodeName(value, buffer)
      select case(char(buffer))
      case("selectedshells")
        call init(lStr)
        call getChildValue(value, "", lStr)
        do ii = 1, len(lStr)
          call get(lStr, strTmp, ii)
          strTmp = tolower(unquote(trim(strTmp)))
          if (len_trim(strTmp) > 4 .or. len_trim(strTmp) < 1) then
            call detailedError(value, "Invalid shell selection '" &
                &// trim(strTmp) &
                &// "'. Nr. of selected shells must be between 1 and 4.")
          end if
          tShellIncl(:) = .false.
          nShell = len_trim(strTmp)
          do jj = 1, nShell
            tmpCh = strTmp(jj:jj)
            tFound = .false.
            do kk = 1, size(orbitalNames)
              if (tmpCh == trim(orbitalNames(kk))) then
                if (tShellIncl(kk)) then
                  call detailedError(value, "Double selection of the same shell '" &
                      & // tmpCh // "' in shell selection block '" &
                      &// trim(strTmp) // "'")
                end if
                tShellIncl(kk) = .true.
                angShell(jj) = kk - 1
                tFound = .true.
                exit
              end if
            end do
            if (.not. tFound) then
              call detailedError(value, "Invalid shell name '" // tmpCh // "'")
            end if
          end do
          call append(angShells(iSp1), angShell(1:nShell))
        end do
        call destroy(lStr)

      case(textNodeName)
        call getChildValue(child2, "", buffer)
        strTmp = unquote(char(buffer))
        do jj = 1, size(orbitalNames)
          if (trim(strTmp) == trim(orbitalNames(jj))) then
            call append(angShells(iSp1), angShellOrdered(:jj))
          end if
        end do
        if (len(angShells(iSp1)) < 1) then
          call detailedError(child2, "Invalid orbital name '" // &
              &trim(strTmp) // "'")
        end if

      case default
        call getNodeHSDName(value, buffer)
        call detailedError(child2, "Invalid shell specification method '" //&
            & char(buffer) // "'")
      end select
    end do
    
    !! Orbitals and angular momenta for the given shells (once the SK files
    !! will contain the full information about the basis, this will be moved
    !! to the SK reading routine).
    INITALLOCATE_P(slako%orb)
    INITALLOCATE_PARR(slako%orb%nShell, (geo%nSpecie))
    INITALLOCATE_PARR(slako%orb%nOrbSpecie, (geo%nSpecie))
    INITALLOCATE_PARR(slako%orb%nOrbAtom, (geo%nAtom))
    slako%orb%mOrb = 0
    slako%orb%mShell = 0
    do iSp1 = 1, geo%nSpecie
      slako%orb%nShell(iSp1) = 0
      slako%orb%nOrbSpecie(iSp1) = 0
      do ii = 1, len(angShells(iSp1))
        call intoArray(angShells(iSp1), angShell, nShell, ii)
        slako%orb%nShell(iSp1) = slako%orb%nShell(iSp1) + nShell
        do jj = 1, nShell
          slako%orb%nOrbSpecie(iSp1) = slako%orb%nOrbSpecie(iSp1) &
              &+ 2 * angShell(jj) + 1
        end do
      end do
    end do
    slako%orb%mShell = maxval(slako%orb%nShell)
    slako%orb%mOrb = maxval(slako%orb%nOrbSpecie)
    slako%orb%nOrbAtom(:) = slako%orb%nOrbSpecie(geo%species(:))
    INITALLOCATE_PARR(slako%orb%angShell, (slako%orb%mShell, geo%nSpecie))
    INITALLOCATE_PARR(slako%orb%iShellOrb, (slako%orb%mOrb, geo%nSpecie))
    INITALLOCATE_PARR(slako%orb%posShell, (slako%orb%mShell+1, geo%nSpecie))
    slako%orb%angShell(:,:) = 0
    do iSp1 = 1, geo%nSpecie
      ind = 1
      iSh1 = 1
      do ii = 1, len(angShells(iSp1))
        call intoArray(angShells(iSp1), angShell, nShell, ii)
        do jj = 1, nShell
          slako%orb%posShell(iSh1, iSp1) = ind
          slako%orb%angShell(iSh1, iSp1) = angShell(jj)
          slako%orb%iShellOrb(ind:ind+2*angShell(jj), iSp1) = iSh1
          ind = ind + 2 * angShell(jj) + 1
          iSh1 = iSh1 + 1
        end do
        slako%orb%posShell(iSh1, iSp1) = ind
      end do
    end do

    !! Slater-Koster files
    ALLOCATE_(skFiles, (geo%nSpecie, geo%nSpecie))
    do iSp1 = 1, geo%nSpecie
      do iSp2 = 1, geo%nSpecie
        call init(skFiles(iSp2, iSp1))
      end do
    end do
    call getChildValue(node, "SlaterKosterFiles", value, child=child)
    call getNodeName(value, buffer)
    select case(char(buffer))
    case ("type2filenames")
      call getChildValue(value, "Prefix", buffer2, "")
      prefix = unquote(char(buffer2))
      call getChildValue(value, "Suffix", buffer2, "")
      suffix = unquote(char(buffer2))
      call getChildValue(value, "Separator", buffer2, "")
      separator = unquote(char(buffer2))
      call getChildValue(value, "LowerCaseTypeName", tLower, .false.)
      do iSp1 = 1, geo%nSpecie
        if (tLower) then
          elem1 = tolower(geo%specieNames(iSp1))
        else
          elem1 = geo%specieNames(iSp1)
        end if
        do iSp2 = 1, geo%nSpecie
          if (tLower) then
            elem2 = tolower(geo%specieNames(iSp2))
          else
            elem2 = geo%specieNames(iSp2)
          end if
          strTmp = trim(prefix) // trim(elem1) // trim(separator) &
              &// trim(elem2) // trim(suffix)
          call append(skFiles(iSp2, iSp1), strTmp)
          inquire(file=strTmp, exist=tExist)
          if (.not. tExist) then
            call detailedError(value, "SK file with generated name '" &
                &// trim(strTmp) // "' does not exist.")
          end if
        end do
      end do
    case default
      call setUnprocessed(value)
      do iSp1 = 1, geo%nSpecie
        do iSp2 = 1, geo%nSpecie
          strTmp = trim(geo%specieNames(iSp1)) // "-" &
              &// trim(geo%specieNames(iSp2))
          call init(lStr)
          call getChildValue(child, trim(strTmp), lStr, child=child2)
          if (len(lStr) /= len(angShells(iSp1)) * len(angShells(iSp2))) then
            call detailedError(child2, "Incorrect number of Slater-Koster &
                &files")
          end if
          do ii = 1, len(lStr)
            call get(lStr, strTmp, ii)
            inquire(file=strTmp, exist=tExist)
            if (.not. tExist) then
              call detailedError(child2, "SK file '" // trim(strTmp) &
                  &// "' does not exist'")
            end if
            call append(skFiles(iSp2, iSp1), strTmp)
          end do
          call destroy(lStr)
        end do
      end do
    end select

    !! Which repulsive is defined by polynomial? (Default: None)
    ALLOCATE_(repPoly, (geo%nSpecie, geo%nSpecie))
    call getChildValue(node, "PolynomialRepulsive", value, "", child=child, &
        &list=.true., allowEmptyValue=.true., dummyValue=.true.)
    call getNodeName2(value, buffer)
    select case (char(buffer))
    case ("")
      repPoly(:,:) = .false.
    case("setforall")
      call getChildValue(value, "", repPoly(1,1))
      repPoly(:,:) = repPoly(1,1)
    case default
      do iSp1 = 1, geo%nSpecie
        do iSp2 = 1, geo%nSpecie
          strTmp = trim(geo%specieNames(iSp1)) // "-" &
              &// trim(geo%specieNames(iSp2))
          call getChildValue(child, trim(strTmp), repPoly(iSp2, iSp1), .false.)
        end do
      end do
      if (.not. all(repPoly .eqv. transpose(repPoly))) then
        call detailedError(value, "Assymetric definition (both A-B and B-A must&
            & be defined for using polynomial repulsive)")
      end if
    end select

    call getChildValue(node, "OrbitalResolvedSCC", tOrbResolved, .false.)
    call getChildValue(node, "OldSKInterpolation", ctrl%oldSKInter, .false.)
    if (ctrl%oldSKInter) then
      skInterMeth = skEqGridOld
    else
      skInterMeth = skEqGridNew
    end if
    call readSKFiles(skFiles, geo%nSpecie, slako, slako%orb, &
        &angShells, tOrbResolved, skInterMeth, repPoly)

    do iSp1 = 1, geo%nSpecie
      call destroy(angShells(iSp1))
      do iSp2 = 1, geo%nSpecie
        call destroy(skFiles(iSp2, iSp1))
      end do
    end do
    DEALLOCATE_(angShells)
    DEALLOCATE_(skFiles)
    DEALLOCATE_(repPoly)


    !! SCC parameters
    call getChildValue(node, "SCC", ctrl%tSCC, .false.)
    ifSCC: if (ctrl%tSCC) then
      call getChildValue(node, "SCCTolerance", ctrl%sccTol, 1.0e-5_dp)
      call getChildValue(node, "MaxSCCIterations", ctrl%maxIter, 100)
      call getChildValue(node, "Mixer", value, "Broyden", child=child)
      call getNodeName(value, buffer)
      select case(char(buffer))

      case ("broyden")
        ctrl%iMixSwitch = 3
        call getChildValue(value, "MixingParameter", ctrl%almix, 0.2_dp)
        call getChildValue(value, "CachedIterations", ctrl%iGenerations, -1, &
            &child=child)
        if (ctrl%iGenerations == -1) then
          ctrl%iGenerations = ctrl%maxIter
        elseif (ctrl%iGenerations < 1) then
          call detailedError(child, "Invalid nr. of cached iterations")
        end if
        call getChildValue(value, "InverseJacobiWeight", ctrl%broydenOmega0, &
            &0.01_dp)
        call getChildValue(value, "MinimalWeight", ctrl%broydenMinWeight, &
            &1.0_dp)
        call getChildValue(value, "MaximalWeight", ctrl%broydenMaxWeight, &
            &1.0e5_dp)
        call getChildValue(value, "WeightFactor", ctrl%broydenWeightFac, &
            &1.0e-2_dp)
        
      case ("anderson")
        ctrl%iMixSwitch = 2
        call getChildValue(value, "MixingParameter", ctrl%almix, 0.05_dp)
        call getChildValue(value, "Generations", ctrl%iGenerations, 4)
        call getChildValue(value, "InitMixingParameter", &
            &ctrl%andersonInitMixing, 0.01_dp)
        INIT_PARR(ctrl%andersonDynMixParams)
        call getChildValue(value, "DynMixingParameters", value2, "", &
            &child=child, allowEmptyValue=.true.)
        call getNodeName2(value2, buffer2)
        if (char(buffer2) == "") then
          ctrl%andersonNrDynMix = 0
        else
          call init(lr1)
          call getChildValue(child, "", 2, lr1, child=child2)
          if (len(lr1) < 1) then
            call detailedError(child2, "At least one dynamic mixing parameter&
                & must be defined.")
          end if
          ctrl%andersonNrDynMix = len(lr1)
          INITALLOCATE_PARR(ctrl%andersonDynMixParams, \
            (2, ctrl%andersonNrDynMix))
          call asArray(lr1, ctrl%andersonDynMixParams)
          call destroy(lr1)
        end if
        call getChildValue(value, "DiagonalRescaling", ctrl%andersonOmega0, &
            &1.0e-2_dp)
        
      case ("simple")
        ctrl%iMixSwitch = 1
        call getChildValue(value, "MixingParameter", ctrl%almix, 0.05_dp)

      case("diis")
        ctrl%iMixSwitch = 4
        call getChildValue(value, "InitMixingParameter", ctrl%almix, 0.2_dp)
        call getChildValue(value, "Generations", ctrl%iGenerations, 6)
        call getChildValue(value, "UseFromStart", ctrl%tFromStart, .true.)
      case default
        call getNodeHSDName(value, buffer)
        call detailedError(child, "Invalid mixer '" // char(buffer) // "'")
      end select

      !! External electric field
      call getChildValue(node, "ElectricField", value, "", child=child, &
          &allowEmptyValue=.true., dummyValue=.true., list=.true.)
      call getChildren(child, "PointCharges", children)
      INIT_PARR(ctrl%extChrg)
      INIT_PARR(ctrl%extChrgBlurWidth)
      if (getLength(children) > 0) then
        call init(lr1)
        if (.not. geo%tPeriodic) then
          call init(lr)
        end if
        do ii = 1, getLength(children)
          call getItem1(children, ii, child2)
          oldLen = len(lr1)
          call getChildValue(child2, "CoordsAndCharges", 4, lr1, &
              &modifier=modifier, child=child3)
          if (len(modifier) > 0) then
            ind = getModifierIndex(char(modifier), lengthUnits, child3)
            conv = lengthUnits(ind)%value
            ALLOCATE_(rTmp41, (4, len(lr1) - oldLen))
            do jj = oldLen + 1, len(lr1)
              call get(lr1, rTmp4, jj)
              rTmp41(1:3,jj - oldLen) = rTmp4(1:3) * conv
              rTmp41(4, jj - oldLen) = rTmp4(4)
              call set(lr1, rTmp41(:,jj - oldLen), jj)
            end do
            call setChildValue(child3, "", rTmp41, .true.)
            DEALLOCATE_(rTmp41)
          end if
          if (.not. geo%tPeriodic) then
            call getChildValue(child2, "GaussianBlurWidth", rTmp, 0.0_dp, &
                &modifier=modifier, child=child3)
            if (rTmp < 0.0_dp) then
              call detailedError(child3, "Gaussian blur width may not be &
                  &negative")
            end if
            call convertByMul(char(modifier), lengthUnits, child3, rTmp)
            do jj = oldLen + 1, len(lr1)
              call append(lr, rTmp )
            end do
          end if
        end do
        ctrl%nExtChrg = len(lr1)
        INITALLOCATE_PARR(ctrl%extChrg, (4, ctrl%nExtChrg))
        call asArray(lr1, ctrl%extChrg)
        call destroy(lr1)
        if (.not. geo%tPeriodic) then
          INITALLOCATE_PARR(ctrl%extChrgBlurWidth, (ctrl%nExtChrg))
          call asArray(lr, ctrl%extChrgBlurWidth)
          call destroy(lr)
        end if
      else
        ctrl%nExtChrg = 0
      end if
      call destroyNodeList(children)

      !! Misc.
      call getChildValue(node, "DampHShortRange", ctrl%tDampH, .false.)
    end if ifSCC

    !! Charge and spin
    call getChildValue(node, "Charge", ctrl%nrChrg, 0.0_dp)
    call getChildValue(node, "ReadInitialCharges", ctrl%tReadChrg, .false.)
    call getChildValue(node, "SpinPolarisation", value, "", child=child, &
        &allowEmptyValue=.true.)
    call getNodeName2(value, buffer)
    select case(char(buffer))
    case ("")
      ctrl%tSpin = .false.
      ctrl%nrSpinPol = 0.0_dp

    case ("colinear")
      ctrl%tSpin = .true.
      call getChildValue(value, "UnpairedElectrons", ctrl%nrSpinPol)
      if (.not. ctrl%tReadChrg) then
        call getChildValue(value, "InitialSpin", dummy, "", child=child, &
            &allowEmptyValue=.true., dummyValue=.true., list=.true.)
        call getChildren(child, "AtomSpin", children)
        INIT_PARR(ctrl%initialSpins)
        if (getLength(children) > 0) then
          INITALLOCATE_PARR(ctrl%initialSpins,(geo%nAtom))
          ctrl%initialSpins(:) = 0.0_dp
          do ii = 1, getLength(children)
            call getItem1(children, ii, child2)
            call getChild(child2, "Atoms", child3)
            call getValueFromListOrRange(child3, pTmpI1, mvalue=geo%nAtom)
            call getChildValue(child2, "SpinPerAtom", rTmp)
            do jj = 1, size(pTmpI1)
              iAt = pTmpI1(jj)
              if (ctrl%initialSpins(iAt) /= 0.0_dp) then
                call detailedWarning(child3, "Previoius setting for the spin &
                    &of atom" // i2c(iAt) // " overwritten")
              end if
              ctrl%initialSpins(iAt) = rTmp
            end do
            DEALLOCATE_PARR(pTmpI1)
          end do
        end if
      end if
      
    case default
      call getNodeHSDName(value, buffer)
      call detailedError(child, "Invalid spin polarisation type '" //&
          & char(buffer) // "'")
    end select
    if (ctrl%tSpin) then
      INITALLOCATE_PARR(ctrl%W, \
        (slako%orb%mShell, slako%orb%mShell, geo%nSpecie))
      call getChild(node, "SpinConstants", child)
      do iSp1 = 1, geo%nSpecie
        call getChildValue(child, geo%specieNames(iSp1), &
            &ctrl%W(:slako%orb%nShell(iSp1), :slako%orb%nShell(iSp1), &
            &iSp1))
      end do
    end if

    !! Solver
    call getChildValue(node, "Eigensolver", value, "DivideAndConquer")
    call getNodeName(value, buffer)
    select case(char(buffer))
    case ("standard")
      ctrl%iSolver = 1
    case ("divideandconquer")
      ctrl%iSolver = 2
    case ("relativelyrobust")
      ctrl%iSolver = 3
    end select

    !! Filling (temperature only read, if AdaptFillingTemp was not set
    !! for the selected MD thermostat.)
    call getChildValue(node, "Filling", value, "Fermi", child=child)
    call getNodeName(value, buffer)
    select case (char(buffer))
    case ("fermi")
      ctrl%iDistribFn = 0 ! Fermi function
      if (.not. ctrl%tSetFillingTemp) then
        call getChildValue(value, "Temperature", ctrl%tempElec, 0.0_dp, &
            &modifier=modifier, child=field)
        tChanged = .false.
        call convertByMul(char(modifier), energyUnits, field, ctrl%tempElec, &
            &.false., tChanged)
        if (ctrl%tempElec < minTemp) then
          ctrl%tempElec = minTemp
          tChanged = .true.
        end if
        if (tChanged) then
          call setChildValue(field, "", ctrl%tempElec, .true.)
        end if
      end if
    case ("methfesselpaxton")
      if (.not. ctrl%tSetFillingTemp) then
        call getChildValue(value, "Temperature", ctrl%tempElec, 0.0_dp, &
            &modifier=modifier, child=field)
        tChanged = .false.
        call convertByMul(char(modifier), energyUnits, field, ctrl%tempElec, &
            &.false., tChanged)
        if (ctrl%tempElec < minTemp) then
          ctrl%tempElec = minTemp
          tChanged = .true.
        end if
        if (tChanged) then
          call setChildValue(field, "", ctrl%tempElec, .true.)
        end if
      end if
      ! Set the order of the Methfessel-Paxton step function approximation,
      ! defaulting to 2
      call getChildValue(value, "Order", ctrl%iDistribFn, 2)
      if (ctrl%iDistribFn < 1) then
        call getNodeHSDName(value, buffer)
        write(errorStr, "(A,A,A,I4)")"Unsuported filling mode for '", &
            & char(buffer),"' :",ctrl%iDistribFn
        call detailedError(child, errorStr)!"Unsuported filling mode '" &
!            & //char(buffer)// "'")
      end if
    case default
      call getNodeHSDName(value, buffer)
      call detailedError(child, "Invalid filling method '" //char(buffer)// "'")
    end select
    if (geo%tPeriodic) then
      call getChildValue(value, "IndependentKFilling", ctrl%tFillKSep, .false.)
    end if
    

    !! K-Points
    if (geo%tPeriodic) then
      INIT_PARR(ctrl%kPoint)
      INIT_PARR(ctrl%kWeight)
      call getChildValue(node, "KPointsAndWeights", value, child=child, &
          &modifier=modifier)
      call getNodeName(value, buffer)
      select case(char(buffer))

      case ("supercellfolding")
        if (len(modifier) > 0) then
          call detailedError(child, "No modifier is allowed, if the &
              &SupercellFolding scheme is used.")
        end if
        call getChildValue(value, "", coeffsAndShifts)
        if (abs(determinant33(coeffsAndShifts(:,1:3))) - 1.0_dp < -1e-6_dp) then
          call detailedError(value, "Determinant of the supercell matrix must &
              &be greater than 1")
        end if
        if (any(abs(modulo(coeffsAndShifts(:,1:3) + 0.5_dp, 1.0_dp) - 0.5_dp) &
            &> 1e-6_dp)) then
          call detailedError(value, "The components of the supercell matrix &
              &must be integers.")
        end if
        call getSuperSampling(coeffsAndShifts(:,1:3), &
            &modulo(coeffsAndShifts(:,4), 1.0_dp), geo%latVecs, geo%recVecs2p, &
            &ctrl%kPoint, ctrl%kWeight, reduceByInversion=.true.)
        ctrl%nKPoint = size(ctrl%kPoint, dim=2)

      case ("klines")
        call init(li1)
        call init(lr1)
        call getChildValue(value, "", 1, li1, 3, lr1)
        if (len(li1) < 1) then
          call detailedError(value, "At least one line must be specified.")
        end if
        ALLOCATE_(tmpI1, (len(li1)))
        ALLOCATE_(kpts, (3, 0:len(lr1)))
        call asVector(li1, tmpI1)
        call asArray(lr1, kpts(:,1:len(lr1)))
        kpts(:,0) = (/ 0.0_dp, 0.0_dp, 0.0_dp /)
        call destroy(li1)
        call destroy(lr1)
        if (any(tmpI1 < 0)) then
          call detailedError(value, "Interval steps must be greater equal to &
              &zero.")
        end if
        ctrl%nKPoint = sum(tmpI1)
        if (ctrl%nKPoint < 1) then
          call detailedError(value, "Sum of the interval steps must be greater &
              &than zero.")
        end if
        ii = 1
        do while (tmpI1(ii) == 0)
          ii = ii + 1
        end do
        INITALLOCATE_PARR(ctrl%kPoint, (3, ctrl%nKPoint))
        INITALLOCATE_PARR(ctrl%kWeight, (ctrl%nKPoint))
        ind = 1
        do jj = ii, size(tmpI1)
          if (tmpI1(jj) == 0) then
            cycle
          end if
          rTmp3 = (kpts(:,jj) - kpts(:,jj-1)) / real(tmpI1(jj), dp)
          do kk = 1, tmpI1(jj)
            ctrl%kPoint(:,ind) = kpts(:,jj-1) + real(kk, dp) * rTmp3
            ind = ind + 1
          end do
        end do
        ctrl%kWeight(:) = 1.0_dp
        if (len(modifier) > 0) then
          select case (tolower(char(modifier)))
          case ("relative")
            call removeModifier(child)
          case ("absolute")
            ctrl%kPoint(:,:) =  matmul(transpose(geo%latVecs), ctrl%kPoint)
            kpts(:,:) = matmul(transpose(geo%latVecs), kpts)
            call removeModifier(child)
            call setChildValue(value, "", reshape(tmpI1, (/ 1,size(tmpI1) /)), &
                &kpts(:,1:), replace=.true.)
          case default
            call detailedError(child, "Invalid modifier: '" // char(modifier) &
                &// "'")
          end select
        end if
        DEALLOCATE_(tmpI1)
        DEALLOCATE_(kpts)
        if (ctrl%tSCC .and. ctrl%maxIter /= 1) then
          write(errorStr, "(A,I3)") "SCC cycle with k-lines probably will&
              & not converge, SCC iterations set to:", ctrl%maxIter
          call warning(errorStr)
        end if
        if (ctrl%tSCC .and. .not.ctrl%tReadChrg) then          
          call warning("It is strongly suggested you use the&
              & ReadInitialCharges option.")
        end if

      case (textNodeName)
        call init(lr1)
        call getChildValue(child, "", 4, lr1, modifier=modifier)
        if (len(lr1) < 1) then
          call detailedError(child, "At least one k-point must be defined.")
        end if
        ctrl%nKPoint = len(lr1)
        ALLOCATE_(kpts, (4, ctrl%nKPoint))
        call asArray(lr1, kpts)
        call destroy(lr1)
        if (len(modifier) > 0) then
          select case (tolower(char(modifier)))
          case ("relative")
            continue
          case ("absolute")
            kpts(1:3,:) =  matmul(transpose(geo%latVecs), kpts(1:3,:))
          case default
            call detailedError(child, "Invalid modifier: '" // char(modifier) &
                &// "'")
          end select
          call setChildValue(child, "", kpts, .true.)
        end if
        INITALLOCATE_PARR(ctrl%kPoint, (3, ctrl%nKPoint))
        INITALLOCATE_PARR(ctrl%kWeight, (ctrl%nKPoint))
        ctrl%kPoint(:,:) = kpts(1:3, :)
        ctrl%kWeight(:) = kpts(4, :)
        DEALLOCATE_(kpts)
      case default
        call detailedError(value, "Invalid K-point scheme")
      end select
    end if

    !! DFTB+U
    call getChildValue(node, "OrbitalPotential", value, "", child=child, &
        &allowEmptyValue=.true.)
    call getNodeName2(value, buffer)
    select case(char(buffer))
    case ("")
      ctrl%tDFTBU = .false.

    case ("fll")
      INITALLOCATE_PARR(ctrl%nU, (geo%nSpecie))
      INITALLOCATE_PARR(ctrl%nLinU, (slako%orb%mShell, geo%nSpecie))
      INITALLOCATE_PARR(ctrl%UJ, (slako%orb%mShell, geo%nSpecie))
      INITALLOCATE_PARR(ctrl%iLVals, \
          (slako%orb%mShell, slako%orb%mShell, geo%nSpecie))
      ctrl%tDFTBU = .true.
      ctrl%DFTBUfunc = 1
      ctrl%nU(:) = 0
      ctrl%UJ(:,:) = 0.0_dp
      ctrl%nLinU(:,:) = 0
      ctrl%iLVals(:,:,:) = 0
      call getChildValue(value, "Implementation", buffer2, child=child2)
      if (tolower(unquote(char(buffer2))) /= "on-site") then
        call detailedError(child2, "Unknown implementation '" // char(buffer2) &
            &// "'")
      end if
      call getChild(value, "LConstants", child2)
      do iSp1 = 1, geo%nSpecie
        ALLOCATE_(lconsts, (slako%orb%nShell(iSp1)))
        call getChildValue(child2, tolower(geo%specieNames(iSp1)), lconsts, &
            &default=spread(0.0_dp, 1, slako%orb%nShell(iSp1)), child=child3)
        iU = 0
        nShell = slako%orb%nShell(iSp1)
        do iSh1 = 1, nShell
          if (abs(lconsts(iSh1)) > 1e-10_dp) then
            iU = iU + 1
            ctrl%UJ(iU, iSp1) = lconsts(iSh1)
            iLinU = 1
            ctrl%iLVals(iLinU, iU, iSp1) = iSh1 - 1
            do iSh2 = iSh1 + 1, nShell
              if (abs(lconsts(iSh1) - lconsts(iSh2)) < 1e-10_dp) then
                iLinU = iLinU + 1
                ctrl%iLVals(iLinU, iU, iSp1) = iSh2 - 1
                lconsts(iSh2) = 0.0_dp
              end if
            end do
            ctrl%nLinU(iU, iSp1) = iLinU
          end if
        end do
        ctrl%nU(iSp1) = iU
        DEALLOCATE_(lconsts)
      end do
      
    case default
      call getNodeHSDName(value, buffer)
      call detailedError(child, "Unknown orbital potential '" // char(buffer) &
          &// "'")
    end select

    !! Dispersion
    call getChildValue(node, "Dispersion", value, "", child=child, &
        &allowEmptyValue=.true.)
    if (associated(value)) then
      INITALLOCATE_P(ctrl%pSlaKirk)
      call readDispersion(value, geo, ctrl%pSlaKirk)
    end if
    
  end subroutine readDFTBHam

  

  !!* Reads Slater-Koster files
  !!* @param skFiles List of SK file names to read in for every interaction
  !!* @param nrSpecie Nr. of species in the system
  !!* @param slako Data type for slako information
  !!* @param orb Information about the orbitals in the system
  !!* @param angShells For every specie a list of rank one arrays. Each array
  !!* contains angular momenta, to pick from the appropriate SK-files.
  !!* @param orbRes Hubbard Us different for each l-shell?
  !!* @param skInterMeth Method of the sk interpolation
  !!* @note Should be replaced with a more sophisticated one, once the new
  !!*   SK-format has been established
  subroutine readSKFiles(skFiles, nSpecie, slako, orb, angShells, orbRes, &
      &skInterMeth, repPoly)
    type(ListCharLc), intent(inout) :: skFiles(:,:)
    integer, intent(in)    :: nSpecie
    type(slater), intent(inout) :: slako
    type(TOrbitals), intent(in) :: orb
    type(listIntR1), intent(inout) :: angShells(:)
    logical, intent(in) :: orbRes
    integer, intent(in)  :: skInterMeth
    logical, intent(in) :: repPoly(:,:)
    
    integer :: iSp1, iSp2, nSK1, nSK2, iSK1, iSK2, ind, nInt, iSh1
    integer :: angShell(maxL+1), nShell
    logical :: readRep, readAtomic
    character(lc) :: fileName
    real(dp), allocatable :: skHam(:,:), skOver(:,:)
    real(dp) :: dist
    type(TOldSKData), allocatable :: skData12(:,:), skData21(:,:)
    type(OSlakoEqGrid), pointer :: pSlakoEqGrid1, pSlakoEqGrid2
    type(TRepSplineIn) :: repSplineIn1, repSplineIn2
    type(TRepPolyIn) :: repPolyIn1, repPolyIn2
    type(ORepSpline), pointer :: pRepSpline
    type(ORepPoly), pointer :: pRepPoly

    ASSERT(size(skFiles, dim=1) == size(skFiles, dim=2))
    ASSERT((size(skFiles, dim=1) > 0) .and. (size(skFiles, dim=1) == nSpecie))
    ASSERT(all(shape(repPoly) == shape(skFiles)))

    INITALLOCATE_PARR(slako%skSelf, (orb%mShell, nSpecie))
    INITALLOCATE_PARR(slako%skHubbU, (orb%mShell, nSpecie))
    INITALLOCATE_PARR(slako%skOcc, (orb%mShell, nSpecie))
    INITALLOCATE_PARR(slako%mass, (nSpecie))
    slako%skSelf(:,:) = 0.0_dp
    slako%skHubbU(:,:) = 0.0_dp
    slako%skOcc(:,:) = 0.0_dp

    INITALLOCATE_P(slako%skHamCont)
    call init(slako%skHamCont, nSpecie)
    INITALLOCATE_P(slako%skOverCont)
    call init(slako%skOverCont, nSpecie)
    INITALLOCATE_P(slako%repCont)
    call init(slako%repCont, nSpecie)

    lpSp1: do iSp1 = 1, nSpecie
      nSK1 = len(angShells(iSp1))
      lpSp2: do iSp2 = 1, nSpecie
        nSK2 = len(angShells(iSp2))
        ALLOCATE_(skData12, (nSK2, nSK1))
        ALLOCATE_(skData21, (nSK1, nSK2))
        ind = 1
        do iSK1 = 1, nSK1
          do iSK2 = 1, nSK2
            readRep = (iSK1 == 1 .and. iSK2 == 1)
            readAtomic = (iSp1 == iSp2 .and. iSK1 == iSK2)
            call get(skFiles(iSp2, iSp1), fileName, ind)
            if (readRep .and. repPoly(iSp2, iSp1)) then
              call readFromFile(skData12(iSK2,iSK1), fileName, readAtomic, &
                  &repPolyIn=repPolyIn1)
            elseif (readRep) then
              call readFromFile(skData12(iSK2,iSK1), fileName, readAtomic, &
                  &iSp1, iSp2, repSplineIn=repSplineIn1)
            else
              call readFromFile(skData12(iSK2,iSK1), fileName, readAtomic)
            end if
            ind = ind + 1
          end do
        end do
        if (iSp1 == iSp2) then
          skData21 = skData12
          if (repPoly(iSp1, iSp2)) then
            repPolyIn2 = repPolyIn1
          else
            repSplineIn2 = repSplineIn1
          end if
          ind = 1
          do iSK1 = 1, nSK1
            call intoArray(angShells(iSp1), angShell, nShell, iSK1)
            do iSh1 = 1, nShell
              slako%skSelf(ind, iSp1) = &
                  &skData12(iSK1,iSK1)%skSelf(angShell(iSh1)+1)
              slako%skOcc(ind:ind, iSp1) = &
                  &skData12(iSK1,iSK1)%skOcc(angShell(iSh1)+1)
              slako%skHubbU(ind, iSp1) = &
                  &skData12(iSK1,iSK1)%skHubbU(angShell(iSh1)+1)
              ind = ind + 1
            end do
          end do
          if (.not. orbRes) then
            slako%skHubbU(2:,iSp1) = slako%skHubbU(1,iSp1)
          end if
          slako%mass(iSp1) = skData12(1,1)%mass
        else
          ind = 1
          do iSK2 = 1, nSK2
            do iSK1 = 1, nSK1
              readRep = (iSK1 == 1 .and. iSK2 == 1)
              call get(sKFiles(iSp1, iSp2), fileName, ind)
              if (readRep .and. repPoly(iSp1, iSp2)) then
                call readFromFile(skData21(iSK1,iSK2), fileName, readAtomic, &
                    &repPolyIn=repPolyIn2)
              elseif (readRep) then
                call readFromFile(skData21(iSK1,iSK2), fileName, readAtomic, &
                    &iSp2, iSp1, repSplineIn=repSplineIn2)
              else
                call readFromFile(skData21(iSK1,iSK2), fileName, readAtomic)
              end if
              ind = ind + 1
            end do
          end do
        end if

        !! Check for SK and repulsive consistentcy
        !call checkSKCompElec(skData12, skData21, iSp1, iSp2)
        !if (repPoly(iSp1, iSp2)) then
        !  call checkSKCompRepPoly(repPolyIn1, repPolyIn2, iSp1, iSp2)
        !else
        !  call checkSKCompRepSpline(repSplineIn1, repSplineIn2, iSp1, iSp2)
        !end if

        !! Create full H/S table for all interactions of iSp1-iSp2
        nInt = getNSKIntegrals(iSp1, iSp2, orb)
        ALLOCATE_(skHam, (size(skData12(1,1)%skHam, dim=1), nInt))
        ALLOCATE_(skOver, (size(skData12(1,1)%skOver, dim=1), nInt))
        call getFullTable(skHam, skOver, skData12, skData21, angShells(iSp1), &
            &angShells(iSp2))

        !! Add H/S tables to the containers for iSp1-iSp2
        INITALLOCATE_P(pSlakoEqGrid1)
        INITALLOCATE_P(pSlakoEqGrid2)
        dist = skData12(1,1)%dist
        call init(pSlakoEqGrid1, dist, skHam, skInterMeth)
        call init(pSlakoEqGrid2, dist, skOver, skInterMeth)
        call addTable(slako%skHamCont, pSlakoEqGrid1, iSp1, iSp2)
        call addTable(slako%skOverCont, pSlakoEqGrid2, iSp1, iSp2)
        DEALLOCATE_(skHam)
        DEALLOCATE_(skOver)
        if (iSp1 /= iSp2) then
          !! Heteronuclear interactions: the same for the reverse interaction
          ALLOCATE_(skHam, (size(skData12(1,1)%skHam, dim=1), nInt))
          ALLOCATE_(skOver, (size(skData12(1,1)%skOver, dim=1), nInt))
          call getFullTable(skHam, skOver, skData21, skData12, angShells(iSp2),&
              &angShells(iSp1))
          INITALLOCATE_P(pSlakoEqGrid1)
          INITALLOCATE_P(pSlakoEqGrid2)
          call init(pSlakoEqGrid1, dist, skHam, skInterMeth)
          call init(pSlakoEqGrid2, dist, skOver, skInterMeth)
          call addTable(slako%skHamCont, pSlakoEqGrid1, iSp2, iSp1)
          call addTable(slako%skOverCont, pSlakoEqGrid2, iSp2, iSp1)
          DEALLOCATE_(skHam)
          DEALLOCATE_(skOver)
        end if
        DEALLOCATE_(skData12)
        DEALLOCATE_(skData21)

        !! Add repulsives to the containers.
        if (repPoly(iSp2, iSp1)) then
          INITALLOCATE_P(pRepPoly)
          call init(pRepPoly, repPolyIn1)
          call addRepulsive(slako%repCont, pRepPoly, iSp1, iSp2)
        else
          INITALLOCATE_P(pRepSpline)
          call init(pRepSpline, repSplineIn1)
          call addRepulsive(slako%repCont, pRepSpline, iSp1, iSp2)
          DEALLOCATE_PARR(repSplineIn1%xStart)
          DEALLOCATE_PARR(repSplineIn1%spCoeffs)
        end if
        if (iSp1 /= iSp2) then
          if (repPoly(iSp1, iSp2)) then
            INITALLOCATE_P(pRepPoly)
            call init(pRepPoly, repPolyIn2)
            call addRepulsive(slako%repCont, pRepPoly, iSp2, iSp1)
          else
            INITALLOCATE_P(pRepSpline)
            call init(pRepSpline, repSplineIn2)
            call addRepulsive(slako%repCont, pRepSpline, iSp2, iSp1)
            DEALLOCATE_PARR(repSplineIn2%xStart)
            DEALLOCATE_PARR(repSplineIn2%spCoeffs)
          end if
        end if
      end do lpSp2
    end do lpSp1

  end subroutine readSKFiles



  !!* Checks if the provided set of SK-tables for a the interactions A-B and
  !!* B-A is consistent.
  !!* @param skData12 Slater-Koster integral set for the interaction A-B
  !!* @param skData21 Slater-Koster integral set for the interaction B-A
  !!* @param repIn1 Repulsive for the interaction A-B
  !!* @param repIn2 Repulsive for the interaction B-A
  !!* @param sp1 Specie number for A (for error messages)
  !!* @param sp2 Specie number for B (for error messages)
  subroutine checkSKCompElec(skData12, skData21, sp1, sp2)
    type(TOldSKData), intent(in), target :: skData12(:,:), skData21(:,:)
    integer, intent(in) :: sp1, sp2
    
    integer :: iSK1, iSK2, nSK1, nSK2
    integer :: nGrid
    real(dp) :: dist
    type(TOldSKData), pointer :: pSK12, pSK21
    character(lc) :: errorStr

    nSK1 = size(skData12, dim=2)
    nSK2 = size(skData12, dim=1)

    ASSERT(size(skData21, dim=1) == nSK1)
    ASSERT(size(skData21, dim=2) == nSK2)

    nGrid = skData12(1,1)%nGrid
    dist = skData12(1,1)%dist

    !! All SK files should have the same grid separation and table length
    nGrid = skData12(1,1)%nGrid
    dist = skData12(1,1)%dist
    do iSK1 = 1, nSK1
      do iSK2 = 1, nSK2
        pSK12 => skData12(iSK2, iSK1)
        pSK21 => skData21(iSK1, iSK2)

        if (pSK12%dist /= dist .or. pSK21%dist /= dist) then
          write (errorStr, "(A,I2,A,I2)") "Incompatible SK grid separations &
              &for species ", sp1, ", ", sp2
          call error(errorStr)
        end if
        if (pSK12%nGrid /= nGrid .or. pSK21%nGrid /= nGrid) then
          write (errorStr, "(A,I2,A,I2)") "Incompatible SK grid lengths for &
              &species pair ", sp1, ", ", sp2
          call error(errorStr)
        end if
      end do
    end do

  end subroutine checkSKCompElec



  !!* Checks if the provided repulsive splines for A-B and B-A are compatible
  !!* @param repIn1 Repulsive spline for interaction A-B
  !!* @param repIn2 Repulsive spline for interaction B-A
  !!* @param sp1 Number of species A (for error messages only)
  !!* @param sp2 Number of species B (for error messages only)
  subroutine checkSKCompRepSpline(repIn1, repIn2, sp1, sp2)
    type(TRepSplineIn), intent(in) :: repIn1, repIn2
    integer, intent(in) :: sp1, sp2

    !! Tolerance for the agreement in the repulsive data
    real(dp), parameter :: tolRep = 1.0e-8_dp

    character(lc) :: errorStr

    !! Repulsives for A-B and B-A should be the same
    if (size(repIn1%xStart) /= size(repIn2%xStart)) then
      write(errorStr, "(A,I2,A,I2,A)") "Incompatible nr. of repulsive &
          &intervals for specie pair ", sp1, "-", sp2, "."
      call error(errorStr)
    end if
    if (maxval(abs(repIn1%xStart - repIn2%xStart)) > tolRep) then
      write(errorStr, "(A,I2,A,I2,A)") "Incompatible repulsive spline &
          &intervals for specie pair ", sp1, "-", sp2, "."
      call error(errorStr)
    end if
    if (maxval(abs(repIn1%spCoeffs - repIn2%spCoeffs)) > tolRep &
        &.or. maxval(abs(repIn1%spLastCoeffs - repIn2%spLastCoeffs)) &
        &> tolRep) then
      write(errorStr, "(A,I2,A,I2,A)") "Incompatible repulsive spline &
          &coefficients for specie pair ", sp1, "-", sp2, "."
      call error(errorStr)
    end if
    if (maxval(abs(repIn1%expCoeffs - repIn2%expCoeffs)) > tolRep) then
      write(errorStr, "(A,I2,A,I2,A)") "Incompatible repulsive spline &
          &exp. coefficients for specie pair ", sp1, "-", sp2, "."
      call error(errorStr)
    end if
    if (abs(repIn1%cutoff - repIn2%cutoff) > tolRep) then
      write(errorStr, "(A,I2,A,I2,A)") "Incompatible repulsive spline &
          &cutoffs for specie pair ", sp1, "-", sp2, "."
      call error(errorStr)
    end if

  end subroutine checkSKCompRepSpline

  

  !!* Checks if repulsive polynomials for A-B and B-A are compatible
  !!* @param repIn1 Repulsive polynomial for interaction A-B
  !!* @param repIn2 Repulsive polynomial for interaction B-A
  !!* @param sp1 Number of species A (for error messages only)
  !!* @param sp2 Number of species B (for error messages only)
  subroutine checkSKCompRepPoly(repIn1, repIn2, sp1, sp2)
    type(TRepPolyIn), intent(in) :: repIn1, repIn2
    integer, intent(in) :: sp1, sp2

    character(lc) :: errorStr

    if (any(repIn1%polyCoeffs /= repIn2%polyCoeffs)) then
      write(errorStr, "(A,I2,A,I2,A)") "Incompatible repulsive polynomial &
          &coefficients  for the species pair ", sp1, "-", sp2, "."
      call error(errorStr)
    end if
    if (repIn1%cutoff /= repIn2%cutoff) then
      write(errorStr, "(A,I2,A,I2,A)") "Incompatible repulsive cutoffs  &
          &for the species pair ", sp1, "-", sp2, "."
      call error(errorStr)
    end if

  end subroutine checkSKCompRepPoly



  !!* Returns the nr. of Slater-Koster integrals necessary to describe the
  !!* interactions between two species.
  !!* @param sp1 Index of the first species.
  !!* @param sp2 Index of the second species.
  !!* @param orb Information about the orbitals in the system.
  !!* @return Nr. of Slater-Koster interactions.
  pure function getNSKIntegrals(sp1, sp2, orb) result(nInt)
    integer, intent(in) :: sp1, sp2
    type(TOrbitals), intent(in) :: orb
    integer :: nInt

    integer :: iSh1, iSh2

    nInt = 0
    do iSh1 = 1, orb%nShell(sp1)
      do iSh2 = 1, orb%nShell(sp2)
        nInt = nInt + min(orb%angShell(iSh2, sp2), orb%angShell(iSh1, sp1)) + 1
      end do
    end do

  end function getNSKIntegrals

  

  !!* Creates from the columns of the Slater-Koster files for A-B and B-A
  !!* a full table for A-B, containing all integrals.
  !!* @param skHam Resulting table of H integrals
  !!* @param skOver Resulting table of S integrals
  !!* @param skData12 Contains all SK files describing interactions for A-B
  !!* @param skData21 Contains all SK files describing interactions for B-A
  !!* @param angShells1 Angular momenta to pick from the SK-files for specie A
  !!* @param angShells2 Angular momenta to pick from the SK-files for specie B
  subroutine getFullTable(skHam, skOver, skData12, skData21, angShells1, &
      &angShells2)
    real(dp), intent(out) :: skHam(:,:), skOver(:,:)
    type(TOldSKData), intent(in) :: skData12(:,:), skData21(:,:)
    type(listIntR1), intent(inout) :: angShells1, angShells2

    integer :: ind, iSK1, iSK2, iSh1, iSh2, nSh1, nSh2, l1, l2, lMin, lMax, mm
    integer :: angShell1(maxL+1), angShell2(maxL+1)
    real(dp), pointer :: pHam(:,:), pOver(:,:)

    !! Maps (mm, l1, l2 ) onto an element in the SK table.
    !! l2 >= l1 (l1 = 0, 1, ...; l2 = 0, 1, ...), m <= l1.
    integer, parameter :: skMap(0:maxL, 0:maxL, 0:maxL) &
        &= reshape((/&
        &20, 0,  0,  0,  19,  0,  0,  0,  18,  0,  0,  0,  17,  0,  0,  0,&
        & 0, 0,  0,  0,  15, 16,  0,  0,  13, 14,  0,  0,  11, 12,  0,  0,&
        & 0, 0,  0,  0,   0,  0,  0,  0,   8,  9, 10,  0,   5,  6,  7,  0,&
        & 0, 0,  0,  0,   0,  0,  0,  0,   0,  0,  0,  0,   1,  2,  3,  4/),&
        &(/maxL + 1, maxL + 1, maxL + 1/))

    ind = 1
    do iSK1 = 1, len(angShells1)
      call intoArray(angShells1, angShell1, nSh1, iSK1)
      do iSh1 = 1, nSh1
        l1 = angShell1(iSh1)
        do iSK2 = 1, len(angShells2)
          call intoArray(angShells2, angShell2, nSh2, iSK2)
          do iSh2 = 1, nSh2
            l2 = angShell2(iSh2)
            if (l1 <= l2) then
              pHam => skData12(iSK2,iSK1)%skHam
              pOver => skData12(iSK2,iSK1)%skOver
              lMin = l1
              lMax = l2
            else
              pHam => skData21(iSK1,iSK2)%skHam
              pOver => skData21(iSK1,iSK2)%skOver
              lMin = l2
              lMax = l1
            end if
            do mm = 0, lMin
              !! Safety check, if array size are appropriate
              ASSERT(all(shape(skHam) >= (/ size(pHam, dim=1), ind /)))
              ASSERT(all(shape(skOver) >= (/ size(pOver, dim=1), ind /)))
              ASSERT(size(pHam, dim=1) == size(pOver, dim=1))
              skHam(:,ind) = pHam(:,skMap(mm,lMax,lMin))
              skOver(:,ind) = pOver(:,skMap(mm,lMax,lMin))
              ind = ind + 1
            end do
          end do
        end do
      end do
    end do
    
  end subroutine getFullTable
  


  !!* Reads the option block
  !!* @param node Node to parse
  !!* @param ctrl Control structure to fill
  subroutine readOptions(node, ctrl)
    type(fnode), pointer :: node
    type(control), intent(inout) :: ctrl

    type(fnode), pointer :: child

    if (ctrl%tSCC) then
      ctrl%tMulliken = .true.
    else
      call getChildValue(node, "MullikenAnalysis", ctrl%tMulliken, .false.)
    end if
    if (ctrl%tMD .or. ctrl%tGeoOpt) then
      ctrl%tForces = .true.
    else
      call getChildValue(node, "CalculateForces", ctrl%tForces, .false.)
    end if
    call getChildValue(node, "WriteEigenvectors", ctrl%tEigenvecs, .false.)
    call getChildValue(node, "WriteAutotestTag", ctrl%tWriteTagged, .false.)
    call getChildValue(node, "WriteDetailedXML", ctrl%tWriteDetailedXML, &
        &.false.)
    call getChildValue(node, "WriteResultsTag", ctrl%tWriteResultsTag, &
        &.false.)
    call getChildValue(node, "WriteDetailedOut", ctrl%tWriteDetailedOut, &
        &.true.)
    call getChildValue(node, "WriteBandOut", ctrl%tWriteBandDat, .true.)
    call getChildValue(node, "AtomResolvedEnergies", ctrl%tAtomicEnergy, &
        &.false.)
    if (.not.(ctrl%tMD.or.ctrl%tGeoOpt).and.ctrl%tSCC) then
      call getChildValue(node, "RestartFrequency", ctrl%restartFreq, 20)      
    end if    
    call getChildValue(node, "RandomSeed", ctrl%iSeed, 0, child=child)
    if (ctrl%iSeed < 0) then
      call detailedError(child, "Random seed must be greater or equal zero")
    end if
    call getChildValue(node, "WriteHS", ctrl%tWriteHS, .false.)
    call getChildValue(node, "WriteRealHS", ctrl%tWriteRealHS, .false.)
    call getChildValue(node, "MinimiseMemoryUsage", ctrl%tMinMemory, .false.)
    
  end subroutine readOptions



  !!* Reads in the dispersion input data, works only for the Slater-Kirkwood
  !!* dispersion modell.
  !!* @param node Node to process.
  !!* @param geo Geometry of the current system.
  !!* @param input Contains the input for the dispersion module on exit.
  subroutine readDispersion(node, geo, input)
    type(fnode), pointer :: node
    type(TGeometry), intent(in) :: geo
    type(TSlaterKirkwoodInput), intent(out) :: input

    type(fnode), pointer :: value, value2, child, child2, child3
    type(string) :: buffer, buffer2, modifier, modifier2, modifiers(3)
    real(dp), allocatable :: tmpR2(:,:), tmp2R2(:,:), rCutoffs(:)
    real(dp) :: mCutoff, rTmp
    integer :: iAt1, iAt2f, iSp1, iSp2, iNeigh
    integer, allocatable :: nNeighs(:)
    logical :: tChanged
    real(dp), pointer :: cellVec(:,:), rCellVec(:,:), coords(:,:)
    integer, pointer :: img2CentCell(:), iCellVec(:)
    integer :: nAllAtom
    type(TNeighborList) :: neighs

    call getNodeName(node, buffer)
    select case (char(buffer))
    case ("slaterkirkwood")
      ALLOCATE_(tmpR2, (3, geo%nAtom))
      INITALLOCATE_PARR(input%polar, (geo%nAtom))
      INITALLOCATE_PARR(input%rWaals, (geo%nAtom))
      INITALLOCATE_PARR(input%charges, (geo%nAtom))
      call getChildValue(node, "PolarRadiusCharge", value, child=child, &
          &modifier=modifier)
      call getNodeName(value, buffer2)
      tChanged = .false.
      select case (char(buffer2))
      case (textNodeName)
        call getChildValue(child, "", tmpR2, modifier=modifier)
        if (len(modifier) > 0) then
          tChanged = .true.
          call splitModifier(char(modifier), child, modifiers)
          call convertByMul(char(modifiers(1)), volumeUnits, child, tmpR2(1,:),&
              &.false.)
          call convertByMul(char(modifiers(2)), lengthUnits, child, tmpR2(2,:),&
              &.false.)
          call convertByMul(char(modifiers(3)), chargeUnits, child, tmpR2(3,:),&
              &.false.)
        end if
        
      case ("hybriddependentpol")
        if (len(modifier) > 0) then
          call detailedError(child, "PolarRadiusCharge is not allowed to carry &
              &a modifier, if the HybridDependentPol method is used.")
        end if
        tChanged = .true.
        ALLOCATE_(rCutoffs, (geo%nSpecie))
        ALLOCATE_(tmp2R2, (13, geo%nSpecie))
        do iSp1 = 1, geo%nSpecie
          call getChildValue(value, geo%specieNames(iSp1), value2, &
              &child=child2, dummyValue=.true.)
          call getChildValue(child2, "CovalentRadius", rCutoffs(iSp1), &
              &modifier=modifier2, child=child3)
          call convertByMul(char(modifier2), lengthUnits, child3, &
              &rCutoffs(iSp1))
          call getChildValue(child2, "HybridPolarisations", tmp2R2(:, iSp1), &
              &modifier=modifier2, child=child3)
          if (len(modifier2) > 0) then
            call splitModifier(char(modifier2), child, modifiers)
            call convertByMul(char(modifiers(1)), volumeUnits, child, &
                &tmp2R2(1:6, iSp1), .false.)
            call convertByMul(char(modifiers(2)), lengthUnits, child, &
                &tmp2R2(7:12, iSp1), .false.)
            call convertByMul(char(modifiers(3)), chargeUnits, child, &
                &tmp2R2(13, iSp1), .false.)
          end if
        end do
        mCutoff = 2.0_dp * maxval(rCutoffs)
        if (geo%tPeriodic) then
          INIT_PARR(cellVec)
          INIT_PARR(rCellVec)
          call getCellTranslations(cellVec, rCellVec, geo%latVecs, &
              &geo%recVecs2p, mCutoff)
        else
          INITALLOCATE_PARR(cellVec, (3, 1))
          INITALLOCATE_PARR(rCellVec, (3, 1))
          cellVec(:, 1) = (/ 0.0_dp, 0.0_dp, 0.0_dp /)
          rCellVec(:, 1) = (/ 0.0_dp, 0.0_dp, 0.0_dp /)
        end if
        call init(neighs, geo%nAtom, 10)
        if (geo%tPeriodic) then
          !! Make some guess for the nr. of all interacting atoms
          nAllAtom = int((real(geo%nAtom, dp)**(1.0_dp/3.0_dp) + 3.0_dp)**3)
        else
          nAllAtom = geo%nAtom
        end if
        INITALLOCATE_PARR(coords, (3, nAllAtom))
        INITALLOCATE_PARR(img2CentCell, (nAllAtom))
        INITALLOCATE_PARR(iCellVec, (nAllAtom))
        call updateNeighborList(coords, img2CentCell, iCellVec, neighs, &
            &nAllAtom, geo%coords, mCutoff, rCellVec)
        ALLOCATE_(nNeighs, (geo%nAtom))
        nNeighs(:) = 0
        do iAt1 = 1, geo%nAtom
          iSp1 = geo%species(iAt1)
          do iNeigh = 1, neighs%nNeighbor(iAt1)
            iAt2f = img2CentCell(neighs%iNeighbor(iNeigh, iAt1))
            iSp2 = geo%species(iAt2f)
            rTmp = rCutoffs(iSp1) + rCutoffs(iSp2)
            if (neighs%neighDist2(iNeigh, iAt1) <= rTmp**2) then
              nNeighs(iAt1) = nNeighs(iAt1) + 1
              nNeighs(iAt2f) = nNeighs(iAt2f) + 1
            end if
          end do
        end do
        do iAt1 = 1, geo%nAtom
          iSp1 = geo%species(iAt1)
          if (nNeighs(iAt1) <= 4 ) then
            tmpR2(1, iAt1) = tmp2R2(1+nNeighs(iAt1), iSp1)
            tmpR2(2, iAt1) = tmp2R2(7+nNeighs(iAt1), iSp1)
          else
            tmpR2(1, iAt1) = tmp2R2(6, iSp1)
            tmpR2(2, iAt1) = tmp2R2(12, iSp1)
          end if
          tmpR2(3, iAt1) = tmp2R2(13, iSp1)
        end do
        DEALLOCATE_(nNeighs)
        DEALLOCATE_(tmp2R2)
        DEALLOCATE_PARR(cellVec)
        DEALLOCATE_PARR(rCellVec)
        DEALLOCATE_PARR(coords)
        DEALLOCATE_PARR(img2CentCell)
        DEALLOCATE_PARR(iCellVec)
        call destroy(neighs)
        DEALLOCATE_(rCutoffs)
        
      case default
        call detailedError(value, "Invalid method for PolarRadiusCharge.")
        
      end select

      if (tChanged) then
        call setChildValue(child, "", tmpR2, .true.)
      end if
      input%polar(:) = tmpR2(1,:)
      input%rWaals(:) = tmpR2(2,:)
      input%charges(:) = tmpR2(3,:)
      DEALLOCATE_(tmpR2)

    case default
      call detailedError(node, "Invalid dispersion model name.")
    end select

    call unstring(buffer)
    call unstring(buffer2)
    call unstring(modifier)
    call unstring(modifier2)
    call unstring(modifiers)
    
  end subroutine readDispersion



  subroutine readTemperature(node, ctrl)
    type(fnode), pointer :: node
    type(control), intent(inout) :: ctrl

    logical :: tChanged
    type(string) :: modifier
    
    INITALLOCATE_PARR(ctrl%tempSteps, (1))
    INITALLOCATE_PARR(ctrl%tempValues, (1))
    INITALLOCATE_PARR(ctrl%tempMethods, (1))
    ctrl%tempMethods(1) = 1
    ctrl%tempSteps(1) = 1
    call getChildValue(node, "", ctrl%tempValues(1), modifier=modifier)
    tChanged = .false.
    call convertByMul(char(modifier), energyUnits, node, ctrl%tempValues(1), &
        &.false., tChanged)
    if (ctrl%tempValues(1) < 0.0_dp) then
      call detailedError(node, "Negative temperature.")            
    end if
    if (ctrl%tempValues(1) < minTemp) then
      ctrl%tempValues(1) = minTemp
      tChanged = .true.
    end if
    call removeModifier(node)
    if (tChanged) then
      call setChildValue(node, "", ctrl%tempValues(1), .true.)
    end if

    call unstring(modifier)

  end subroutine readTemperature


  
  subroutine readTemperatureProfile(node, parent, modifier, ctrl)
    type(fnode), pointer :: node, parent
    character(len=*), intent(in) :: modifier
    type(control), intent(inout) :: ctrl

    character(len=*), parameter :: tempMethodNames(3) = (/ 'constant   ', &
        &'linear     ', 'exponential' /)

    type(listString) :: ls
    type(listIntR1) :: li1
    type(listRealR1) :: lr1
    character(len=20), allocatable :: tmpC1(:)
    logical :: tChanged
    integer :: ii, jj
    
    call init(ls)
    call init(li1)
    call init(lr1)
    call getChildValue(node, "", ls, 1, li1, 1, lr1)
    if (len(ls) < 1) then
      call detailedError(node, "At least one annealing step must be &
          &specified.")
    end if
    ALLOCATE_(tmpC1, (len(ls)))
    INITALLOCATE_PARR(ctrl%tempSteps, (len(li1)))
    INITALLOCATE_PARR(ctrl%tempValues, (len(lr1)))
    call asArray(ls, tmpC1)
    call asVector(li1, ctrl%tempSteps)
    call asVector(lr1, ctrl%tempValues)
    call destroy(ls)
    call destroy(li1)
    call destroy(lr1)
    INITALLOCATE_PARR(ctrl%tempMethods, (size(tmpC1)))
    lp2: do ii = 1, size(tmpC1)
      do jj = 1, size(tempMethodNames)
        if (trim(tmpC1(ii)) == tolower(trim(tempMethodNames(jj)))) then
          ctrl%tempMethods(ii) = jj
          cycle lp2
        end if
      end do
      call detailedError(node, "Invalid annealing method name '" &
          &// trim(tmpC1(ii)) // "'.")
    end do lp2

    if (any(ctrl%tempSteps < 0)) then
      call detailedError(node, "Step values must not be negative.")
    end if

    ii = sum(ctrl%tempSteps)
    if (ii < 1) then
      call detailedError(node, "Sum of steps in the profile must be &
          &greater than zero.")
    end if
    ctrl%maxRun = ii - 1

    if (any(ctrl%tempValues < 0.0_dp)) then
      call detailedError(node, "Negative temperature.")
    end if

    tChanged = .false.
    call convertByMul(modifier, energyUnits, node, ctrl%tempValues, .false., &
        &tChanged)
    call removeModifier(parent)
    if (any(ctrl%tempValues < minTemp)) then
      ctrl%tempValues = max(ctrl%tempValues, minTemp)
      tChanged = .true.
    end if
    if (tChanged) then
      call setChildValue(node, "", tmpC1, &
          &reshape(ctrl%tempSteps, (/ 1, size(ctrl%tempSteps) /)), &
          &reshape(ctrl%tempValues, (/ 1, size(ctrl%tempValues) /)), &
          &.true.)
    end if
    DEALLOCATE_(tmpC1)

  end subroutine readTemperatureProfile



end module Parser
